{"file":{},"class":{"name":"AmazonAS","file":"services\/as.class.php","url":"..\/amazonas\/index.html","description":"<p>This guide provides detailed information about Auto Scaling actions, data types, parameters,\nand errors. For detailed information about Auto Scaling features and their associated API\ncalls, go to the <a href=\n\"http:\/\/docs.amazonwebservices.com\/AutoScaling\/latest\/DeveloperGuide\/\">Auto Scaling Developer\nGuide<\/a>.<\/p>\n\n<p>Auto Scaling is a web service designed to automatically launch or terminate Amazon Elastic\nCompute Cloud (Amazon EC2) instances based on user-defined policies, schedules, and health\nchecks. This service is used in conjunction with Amazon CloudWatch and Elastic Load Balancing\nservices.<\/p>\n\n<p>This reference is based on the current WSDL, which is available at:<\/p>\n\n<p><a href=\n\"http:\/\/autoscaling.amazonaws.com\/doc\/2011-01-01\/AutoScaling.wsdl\">http:\/\/autoscaling.amazonaws.com\/doc\/2011-01-01\/AutoScaling.wsdl<\/a><\/p>\n\n<p><strong>Endpoints<\/strong><\/p>\n\n<p>For information about this product&#8217;s regions and endpoints, go to <a href=\n\"http:\/\/docs.amazonwebservices.com\/general\/latest\/gr\/index.html?rande.html\">Regions and\nEndpoints<\/a> in the Amazon Web Services General Reference.<\/p>","metadata":{"version":"2013.01.14","license":"See the included NOTICE.md file for complete information.","copyright":"See the included NOTICE.md file for complete information.","link":["http:\/\/aws.amazon.com\/as\/ Auto Scaling","http:\/\/aws.amazon.com\/as\/documentation\/ Auto Scaling documentation"]},"inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php","url":"..\/cfruntime\/index.html"}},"methods":{"@attributes":{"count":"64"},"method":[{"name":"__construct","modifier":"public","description":"<p>Constructs a new instance of <a href=\"..\/amazonas\/index.html\" rel=\"..\/..\/index.html#i=amazonas\"><code>AmazonAS<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"options","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>certificate_authority<\/code> - <code>boolean<\/code> - Optional - Determines which Cerificate Authority file to use. A value of boolean <code>false<\/code> will use the Certificate Authority file available on the system. A value of boolean <code>true<\/code> will use the Certificate Authority provided by the SDK. Passing a file system path to a Certificate Authority file (chmodded to <code>0755<\/code>) will use that. Leave this set to <code>false<\/code> if you&#8217;re not sure.<\/li><li><code>credentials<\/code> - <code>string<\/code> - Optional - The name of the credential set to use for authentication.<\/li><li><code>default_cache_config<\/code> - <code>string<\/code> - Optional - This option allows a preferred storage type to be configured for long-term caching. This can be changed later using the <a href=\"..\/amazonas\/set_cache_config.html\" rel=\"..\/..\/index.html#m=amazonas\/set_cache_config\"><code>set_cache_config()<\/code><\/a> method. Valid values are: <code>apc<\/code>, <code>xcache<\/code>, or a file system path such as <code>.\/cache<\/code> or <code>\/tmp\/cache\/<\/code>.<\/li><li><code>key<\/code> - <code>string<\/code> - Optional - Your AWS key, or a session key. If blank, the default credential set will be used.<\/li><li><code>secret<\/code> - <code>string<\/code> - Optional - Your AWS secret key, or a session secret key. If blank, the default credential set will be used.<\/li><li><code>token<\/code> - <code>string<\/code> - Optional - An AWS session token.<\/li><\/ul>"}},"returnvalue":{"@attributes":{"type":"void"}},"source":"public function __construct(array $options = array())\n{\n    $this-&gt;api_version = '2011-01-01';\n    $this-&gt;hostname = self::DEFAULT_URL;\n    $this-&gt;auth_class = 'AuthV4Query';\n\n    return parent::__construct($options);\n}"},{"name":"allow_hostname_override","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Disables any subsequent use of the <a href=\"..\/amazonas\/set_hostname.html\" rel=\"..\/..\/index.html#m=amazonas\/set_hostname\"><code>set_hostname()<\/code><\/a> method.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"override","type":"boolean","description":"<p>Whether or not subsequent calls to <a href=\"..\/amazonas\/set_hostname.html\" rel=\"..\/..\/index.html#m=amazonas\/set_hostname\"><code>set_hostname()<\/code><\/a> should be obeyed. A <code>false<\/code> value disables the further effectiveness of <a href=\"..\/amazonas\/set_hostname.html\" rel=\"..\/..\/index.html#m=amazonas\/set_hostname\"><code>set_hostname()<\/code><\/a>. Defaults to <code>true<\/code>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function allow_hostname_override($override = true)\n{\n    $this-&gt;override_hostname = $override;\n    return $this;\n}","examples":{"example":{"title":"<p>Configure hostname and resource prefix.<\/p>","code":"\/\/ Instantiate the class\n$ec2 = new AmazonEC2();\n\n\/\/ Point to the new endpoint\n$ec2-&gt;set_hostname('localhost', '3000');\n$ec2-&gt;set_resource_prefix('\/services\/mock');\n$ec2-&gt;allow_hostname_override(false);\n\n\/\/ Describe images\n$response = $ec2-&gt;describe_images();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"authenticate","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Default, shared method for authenticating a connection to AWS.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"operation","type":"string","description":"<p>Indicates the operation to perform.<\/p>"},{"@attributes":{"choice":"req"},"name":"payload","type":"array","description":"<p>An associative array of parameters for authenticating. See the individual methods for allowed keys.<\/p>"}]},"returnvalue":"<p>Object containing a parsed HTTP response.<\/p>","source":"public function authenticate($operation, $payload)\n{\n    $original_payload = $payload;\n    $method_arguments = func_get_args();\n    $curlopts = array();\n    $return_curl_handle = false;\n\n    if (substr($operation, 0, strlen($this-&gt;operation_prefix)) !== $this-&gt;operation_prefix)\n    {\n        $operation = $this-&gt;operation_prefix . $operation;\n    }\n\n    \/\/ Extract the custom CURLOPT settings from the payload\n    if (is_array($payload) &amp;&amp; isset($payload['curlopts']))\n    {\n        $curlopts = $payload['curlopts'];\n        unset($payload['curlopts']);\n    }\n\n    \/\/ Determine whether the response or curl handle should be returned\n    if (is_array($payload) &amp;&amp; isset($payload['returnCurlHandle']))\n    {\n        $return_curl_handle = isset($payload['returnCurlHandle']) ? $payload['returnCurlHandle'] : false;\n        unset($payload['returnCurlHandle']);\n    }\n\n    \/\/ Use the caching flow to determine if we need to do a round-trip to the server.\n    if ($this-&gt;use_cache_flow)\n    {\n        \/\/ Generate an identifier specific to this particular set of arguments.\n        $cache_id = $this-&gt;key . '_' . get_class($this) . '_' . $operation . '_' . sha1(serialize($method_arguments));\n\n        \/\/ Instantiate the appropriate caching object.\n        $this-&gt;cache_object = new $this-&gt;cache_class($cache_id, $this-&gt;cache_location, $this-&gt;cache_expires, $this-&gt;cache_compress);\n\n        if ($this-&gt;delete_cache)\n        {\n            $this-&gt;use_cache_flow = false;\n            $this-&gt;delete_cache = false;\n            return $this-&gt;cache_object-&gt;delete();\n        }\n\n        \/\/ Invoke the cache callback function to determine whether to pull data from the cache or make a fresh request.\n        $data = $this-&gt;cache_object-&gt;response_manager(array($this, 'cache_callback'), $method_arguments);\n\n        \/\/ Parse the XML body\n        $data = $this-&gt;parse_callback($data);\n\n        \/\/ End!\n        return $data;\n    }\n\n    \/*%******************************************************************************************%*\/\n\n    \/\/ Signer\n    $signer = new $this-&gt;auth_class($this-&gt;hostname, $operation, $payload, $this-&gt;credentials);\n    $signer-&gt;key = $this-&gt;key;\n    $signer-&gt;secret_key = $this-&gt;secret_key;\n    $signer-&gt;auth_token = $this-&gt;auth_token;\n    $signer-&gt;api_version = $this-&gt;api_version;\n    $signer-&gt;utilities_class = $this-&gt;utilities_class;\n    $signer-&gt;request_class = $this-&gt;request_class;\n    $signer-&gt;response_class = $this-&gt;response_class;\n    $signer-&gt;use_ssl = $this-&gt;use_ssl;\n    $signer-&gt;proxy = $this-&gt;proxy;\n    $signer-&gt;util = $this-&gt;util;\n    $signer-&gt;registered_streaming_read_callback = $this-&gt;registered_streaming_read_callback;\n    $signer-&gt;registered_streaming_write_callback = $this-&gt;registered_streaming_write_callback;\n    $request = $signer-&gt;authenticate();\n\n    \/\/ Update RequestCore settings\n    $request-&gt;request_class = $this-&gt;request_class;\n    $request-&gt;response_class = $this-&gt;response_class;\n    $request-&gt;ssl_verification = $this-&gt;ssl_verification;\n\n    \/*%******************************************************************************************%*\/\n\n    \/\/ Debug mode\n    if ($this-&gt;debug_mode)\n    {\n        $request-&gt;debug_mode = $this-&gt;debug_mode;\n    }\n\n    \/\/ Set custom CURLOPT settings\n    if (count($curlopts))\n    {\n        $request-&gt;set_curlopts($curlopts);\n    }\n\n    \/\/ Manage the (newer) batch request API or the (older) returnCurlHandle setting.\n    if ($this-&gt;use_batch_flow)\n    {\n        $handle = $request-&gt;prep_request();\n        $this-&gt;batch_object-&gt;add($handle);\n        $this-&gt;use_batch_flow = false;\n\n        return $handle;\n    }\n    elseif ($return_curl_handle)\n    {\n        return $request-&gt;prep_request();\n    }\n\n    \/\/ Send!\n    $request-&gt;send_request();\n\n    \/\/ Prepare the response.\n    $headers = $request-&gt;get_response_header();\n    $headers['x-aws-stringtosign'] = $signer-&gt;string_to_sign;\n\n    if (isset($signer-&gt;canonical_request))\n    {\n        $headers['x-aws-canonicalrequest'] = $signer-&gt;canonical_request;\n    }\n\n    $headers['x-aws-request-headers'] = $request-&gt;request_headers;\n    $headers['x-aws-body'] = $signer-&gt;querystring;\n\n    $data = new $this-&gt;response_class($headers, ($this-&gt;parse_the_response === true) ? $this-&gt;parse_callback($request-&gt;get_response_body()) : $request-&gt;get_response_body(), $request-&gt;get_response_code());\n\n    $response_body = (string) $request-&gt;get_response_body();\n\n    \/\/ Was it Amazon's fault the request failed? Retry the request until we reach $max_retries.\n    if (\n        (integer) $request-&gt;get_response_code() === 500 || \/\/ Internal Error (presumably transient)\n        (integer) $request-&gt;get_response_code() === 503)   \/\/ Service Unavailable (presumably transient)\n    {\n        if ($this-&gt;redirects &lt;= $this-&gt;max_retries)\n        {\n            \/\/ Exponential backoff\n            $delay = (integer) (pow(4, $this-&gt;redirects) * 100000);\n            usleep($delay);\n            $this-&gt;redirects++;\n            $data = $this-&gt;authenticate($operation, $original_payload);\n        }\n    }\n\n    \/\/ DynamoDB has additional, custom logic for retrying requests\n    else\n    {\n        \/\/ If the request to DynamoDB was throttled, we need to retry\n        $need_to_retry_dynamodb_request = (\n            (integer) $request-&gt;get_response_code() === 400 &amp;&amp;\n            stripos($response_body, 'com.amazonaws.dynamodb.') !== false &amp;&amp;\n            stripos($response_body, 'ProvisionedThroughputExceededException') !== false\n        );\n\n        \/\/ If the CRC32 of the response does not match the expected value, we need to retry\n        $response_headers = $request-&gt;get_response_header();\n        if (!$need_to_retry_dynamodb_request &amp;&amp; isset($response_headers['x-amz-crc32']))\n        {\n            $crc32_expected = $response_headers['x-amz-crc32'];\n            $crc32_actual = hexdec(hash('crc32b', $response_body));\n            $need_to_retry_dynamodb_request = ($crc32_expected != $crc32_actual);\n        }\n\n        \/\/ Perform retry if necessary using a more aggressive exponential backoff\n        if ($need_to_retry_dynamodb_request)\n        {\n            if ($this-&gt;redirects === 0)\n            {\n                $this-&gt;redirects++;\n                $data = $this-&gt;authenticate($operation, $original_payload);\n            }\n            elseif ($this-&gt;redirects &lt;= max($this-&gt;max_retries, 10))\n            {\n                \/\/ Exponential backoff\n                $delay = (integer) (pow(2, ($this-&gt;redirects - 1)) * 50000);\n                usleep($delay);\n                $this-&gt;redirects++;\n                $data = $this-&gt;authenticate($operation, $original_payload);\n            }\n        }\n    }\n\n    $this-&gt;redirects = 0;\n    return $data;\n}"},{"name":"batch","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Specifies that the intended request should be queued for a later batch request.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"queue","type":"CFBatchRequest","description":"<p>The <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a> instance to use for managing batch requests. If not available, it generates a new instance of <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function batch(CFBatchRequest &amp;$queue = null)\n{\n    if ($queue)\n    {\n        $this-&gt;batch_object = $queue;\n    }\n    elseif ($this-&gt;internal_batch_object)\n    {\n        $this-&gt;batch_object = &amp;$this-&gt;internal_batch_object;\n    }\n    else\n    {\n        $this-&gt;internal_batch_object = new $this-&gt;batch_class();\n        $this-&gt;batch_object = &amp;$this-&gt;internal_batch_object;\n    }\n\n    $this-&gt;use_batch_flow = true;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Batch several requests together, and then cache the responses to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"},{"title":"<p>Batch several requests together, and then cache the responses to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"}]}},{"name":"cache","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Specifies that the resulting <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object should be cached according to the settings from\n<a href=\"..\/amazonas\/set_cache_config.html\" rel=\"..\/..\/index.html#m=amazonas\/set_cache_config\"><code>set_cache_config()<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"expires","type":"string|integer","description":"<p>The time the cache is to expire. Accepts a number of seconds as an integer, or an amount of time, as a string, that is understood by <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a> (e.g. &#8220;1 hour&#8221;).<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function cache($expires)\n{\n    \/\/ Die if they haven't used set_cache_config().\n    if (!$this-&gt;cache_class)\n    {\n        throw new CFRuntime_Exception('Must call set_cache_config() before using cache()');\n    }\n\n    if (is_string($expires))\n    {\n        $expires = strtotime($expires);\n        $this-&gt;cache_expires = $expires - time();\n    }\n    elseif (is_int($expires))\n    {\n        $this-&gt;cache_expires = $expires;\n    }\n\n    $this-&gt;use_cache_flow = true;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Fire a single request, and then cache the response to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Fire a single request, and then cache the response to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Fire a single request, and then cache the response to Memcache.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config(array(\n\tarray('host' =&gt; '127.0.0.1')\n));\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Batch several requests together, and then cache the responses to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"},{"title":"<p>Batch several requests together, and then cache the responses to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"}]}},{"name":"cache_instance_profile_credentials","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Fetches and caches EC2 instance profile credentials. This is meant to be used by the constructor, and is not to\nbe manually invoked.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"cache","type":"CacheCore","description":"<p>The a reference to the cache object that is being used to handle the caching.<\/p>"},{"@attributes":{"choice":"req"},"name":"options","type":"array","description":"<p>The options that were passed into the constructor.<\/p>"}]},"returnvalue":"<p>The data to be cached, or NULL.<\/p>","source":"public function cache_instance_profile_credentials($cache, $options)\n{\n    $instance_profile_url = 'http:\/\/169.254.169.254\/latest\/meta-data\/iam\/security-credentials\/';\n    $connect_timeout = isset($options['instance_profile_timeout']) ? $options['instance_profile_timeout'] : 2;\n\n    try\n    {\n        \/\/ Make a call to the EC2 Metadata Service to find the available instance profile\n        $request = new RequestCore($instance_profile_url);\n        $request-&gt;set_curlopts(array(CURLOPT_CONNECTTIMEOUT =&gt; $connect_timeout));\n        $response = $request-&gt;send_request(true);\n\n        if ($response-&gt;isOK())\n        {\n            \/\/ Get the instance profile name\n            $profile = (string) $response-&gt;body;\n\n            \/\/ Make a call to the EC2 Metadata Service to get the instance profile credentials\n            $request = new RequestCore($instance_profile_url . $profile);\n            $request-&gt;set_curlopts(array(CURLOPT_CONNECTTIMEOUT =&gt; $connect_timeout));\n            $response = $request-&gt;send_request(true);\n\n            if ($response-&gt;isOK())\n            {\n                \/\/ Get the credentials\n                $credentials = json_decode($response-&gt;body, true);\n\n                if ($credentials['Code'] === 'Success')\n                {\n                    \/\/ Determine the expiration time\n                    $expiration_time = strtotime((string) $credentials['Expiration']);\n                    $expiration_duration = round(($expiration_time - time()) * 0.85);\n                    $cache-&gt;expire_in($expiration_duration);\n\n                    \/\/ Return the credential information\n                    return array(\n                        'key'     =&gt; $credentials['AccessKeyId'],\n                        'secret'  =&gt; $credentials['SecretAccessKey'],\n                        'token'   =&gt; $credentials['Token'],\n                        'expires' =&gt; $credentials['Expiration'],\n                    );\n                }\n            }\n        }\n    }\n    catch (cURL_Exception $e)\n    {\n        \/\/ The EC2 Metadata Service does not exist or had timed out.\n        \/\/ An exception will be thrown on the next line.\n    }\n\n    \/\/ @codeCoverageIgnoreStart\n    throw new CFCredentials_Exception('No credentials were provided. The SDK attempted to retrieve Instance '\n        . 'Profile credentials from the EC2 Instance Metadata Service, but failed to do so. Instance profile '\n        . 'credentials are only accessible on EC2 instances configured with a specific IAM role.');\n    \/\/ @codeCoverageIgnoreEnd\n}"},{"name":"cache_sts_credentials","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Fetches and caches STS credentials. This is meant to be used by the constructor, and is not to be\nmanually invoked.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"cache","type":"CacheCore","description":"<p>The a reference to the cache object that is being used to handle the caching.<\/p>"},{"@attributes":{"choice":"req"},"name":"options","type":"array","description":"<p>The options that were passed into the constructor.<\/p>"}]},"returnvalue":"<p>The data to be cached, or NULL.<\/p>","source":"public function cache_sts_credentials($cache, $options)\n{\n    $token = new AmazonSTS($options);\n    $response = $token-&gt;get_session_token();\n\n    if ($response-&gt;isOK())\n    {\n        \/\/ Update the expiration\n        $expiration_time = strtotime((string) $response-&gt;body-&gt;GetSessionTokenResult-&gt;Credentials-&gt;Expiration);\n        $expiration_duration = round(($expiration_time - time()) * 0.85);\n        $cache-&gt;expire_in($expiration_duration);\n\n        \/\/ Return the important data\n        $credentials = $response-&gt;body-&gt;GetSessionTokenResult-&gt;Credentials;\n\n        return array(\n            'key'     =&gt; (string) $credentials-&gt;AccessKeyId,\n            'secret'  =&gt; (string) $credentials-&gt;SecretAccessKey,\n            'token'   =&gt; (string) $credentials-&gt;SessionToken,\n            'expires' =&gt; (string) $credentials-&gt;Expiration,\n        );\n    }\n\n    \/\/ @codeCoverageIgnoreStart\n    throw new STS_Exception('Temporary credentials from the AWS Security '\n        . 'Token Service could not be retrieved using the provided long '\n        . 'term credentials. It\\'s possible that the provided long term '\n        . 'credentials were invalid.');\n    \/\/ @codeCoverageIgnoreEnd\n}"},{"name":"create_auto_scaling_group","modifier":"public","description":"<p>Creates a new Auto Scaling group with the specified name and other attributes. When the\ncreation request is completed, the Auto Scaling group is ready to be used in other calls.<\/p>\n\n<p class=\"note\">\nThe Auto Scaling group name must be unique within the scope of your AWS account, and under the\nquota of Auto Scaling groups allowed for your account.\n<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name of the Auto Scaling group. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"launch_configuration_name","type":"string","description":"<p>The name of the launch configuration to use with the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"min_size","type":"integer","description":"<p>The minimum size of the Auto Scaling group.<\/p>"},{"@attributes":{"choice":"req"},"name":"max_size","type":"integer","description":"<p>The maximum size of the Auto Scaling group.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DesiredCapacity<\/code> - <code>integer<\/code> - Optional - The number of Amazon EC2 instances that should be running in the group.<\/li><li><code>DefaultCooldown<\/code> - <code>integer<\/code> - Optional - The amount of time, in seconds, after a scaling activity completes before any further trigger-related scaling activities can start.<\/li><li><code>AvailabilityZones<\/code> - <code>string|array<\/code> - Optional - A list of Availability Zones for the Auto Scaling group. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>LoadBalancerNames<\/code> - <code>string|array<\/code> - Optional - A list of load balancers to use. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>HealthCheckType<\/code> - <code>string<\/code> - Optional - The service you want the health status from, Amazon EC2 or Elastic Load Balancer. Valid values are <code>EC2<\/code> or <code>ELB<\/code>. [Constraints: The value must be between 1 and 32 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>HealthCheckGracePeriod<\/code> - <code>integer<\/code> - Optional - Length of time in seconds after a new Amazon EC2 instance comes into service that Auto Scaling starts checking its health.<\/li><li><code>PlacementGroup<\/code> - <code>string<\/code> - Optional - Physical location of your cluster placement group created in Amazon EC2. For more information about cluster placement group, see <a href=\"http:\/\/docs.amazonwebservices.com\/AWSEC2\/latest\/UserGuide\/using_cluster_computing.html\">Using Cluster Instances<\/a> [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>VPCZoneIdentifier<\/code> - <code>string<\/code> - Optional - A comma-separated list of subnet identifiers of Amazon Virtual Private Clouds (Amazon VPCs). When you specify subnets and Availability Zones with this call, ensure that the subnets&#8217; Availability Zones match the Availability Zones specified. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>TerminationPolicies<\/code> - <code>string|array<\/code> - Optional - A standalone termination policy or a list of termination policies used to select the instance to terminate. The policies are executed in the order that they are listed. For more information on configuring a termination policy for your Auto Scaling group, go to <a href=\"http:\/\/docs.amazonwebservices.com\/AutoScaling\/latest\/DeveloperGuide\/us-termination-policy.html\">Instance Termination Policy for Your Auto Scaling Group<\/a> in the the <em>Auto Scaling Developer Guide<\/em>. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>Tags<\/code> - <code>array<\/code> - Optional - The tag to be created or updated. Each tag should be defined by its resource type, resource ID, key, value, and a propagate flag. Valid values: key= <em>value<\/em>, value= <em>value<\/em>, propagate= <em>true<\/em> or <em>false<\/em>. Value and propagate are optional parameters. <ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>ResourceId<\/code> - <code>string<\/code> - Optional - The name of the Auto Scaling group. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>ResourceType<\/code> - <code>string<\/code> - Optional - The kind of resource to which the tag is applied. Currently, Auto Scaling supports the <code>auto-scaling-group<\/code> resource type. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>Key<\/code> - <code>string<\/code> - Required - The key of the tag. [Constraints: The value must be between 1 and 128 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>Value<\/code> - <code>string<\/code> - Optional - The value of the tag. [Constraints: The value must be between 0 and 256 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>PropagateAtLaunch<\/code> - <code>boolean<\/code> - Optional - Specifies whether the new tag will be applied to instances launched after the tag is created. The same behavior applies to updates: If you change a tag, the changed tag will be applied to all instances launched after you made the change.<\/li><\/ul><\/li><\/ul><\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_auto_scaling_group($auto_scaling_group_name, $launch_configuration_name, $min_size, $max_size, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    $opt['LaunchConfigurationName'] = $launch_configuration_name;\n    $opt['MinSize'] = $min_size;\n    $opt['MaxSize'] = $max_size;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['AvailabilityZones']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'AvailabilityZones' =&gt; (is_array($opt['AvailabilityZones']) ? $opt['AvailabilityZones'] : array($opt['AvailabilityZones']))\n        ), 'member'));\n        unset($opt['AvailabilityZones']);\n    }\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['LoadBalancerNames']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'LoadBalancerNames' =&gt; (is_array($opt['LoadBalancerNames']) ? $opt['LoadBalancerNames'] : array($opt['LoadBalancerNames']))\n        ), 'member'));\n        unset($opt['LoadBalancerNames']);\n    }\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['TerminationPolicies']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'TerminationPolicies' =&gt; (is_array($opt['TerminationPolicies']) ? $opt['TerminationPolicies'] : array($opt['TerminationPolicies']))\n        ), 'member'));\n        unset($opt['TerminationPolicies']);\n    }\n    \n    \/\/ Optional list + map\n    if (isset($opt['Tags']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'Tags' =&gt; $opt['Tags']\n        ), 'member'));\n        unset($opt['Tags']);\n    }\n\n    return $this-&gt;authenticate('CreateAutoScalingGroup', $opt);\n}","examples":{"example":{"title":"<p>Create a new auto scaling group.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;create_auto_scaling_group('my-scaling-group', 'my-launch-config', 1, 2, array(\n\t'DesiredCapacity' =&gt; 2, \/\/ instances\n\t'DefaultCooldown' =&gt; 30, \/\/ seconds\n\t'HealthCheckGracePeriod' =&gt; 30, \/\/ seconds\n\t'AvailabilityZones' =&gt; array('us-east-1c', 'us-east-1d'),\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/delete_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_auto_scaling_group\"><code>delete_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/describe_auto_scaling_groups.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_auto_scaling_groups\"><code>describe_auto_scaling_groups()<\/code><\/a>","<a href=\"..\/amazonas\/describe_auto_scaling_instances.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_auto_scaling_instances\"><code>describe_auto_scaling_instances()<\/code><\/a>","<a href=\"..\/amazonas\/terminate_instance_in_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/terminate_instance_in_auto_scaling_group\"><code>terminate_instance_in_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/update_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/update_auto_scaling_group\"><code>update_auto_scaling_group()<\/code><\/a>"]}},{"name":"create_launch_configuration","modifier":"public","description":"<p>Creates a new launch configuration. The launch configuration name must be unique within the\nscope of the client&#8217;s AWS account. The maximum limit of launch configurations, which by default\nis 100, must not yet have been met; otherwise, the call will fail. When created, the new launch\nconfiguration is available for immediate use.<\/p>\n\n<p>You can create a launch configuration with Amazon EC2 security groups or with Amazon VPC\nsecurity groups. However, you can&#8217;t use Amazon EC2 security groups together with Amazon VPC\nsecurity groups, or vice versa.<\/p>\n\n<p class=\"note\">\nAt this time, Auto Scaling launch configurations don&#8217;t support compressed (e.g. zipped) user\ndata files.\n<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"launch_configuration_name","type":"string","description":"<p>The name of the launch configuration to create. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"image_id","type":"string","description":"<p>Unique ID of the <em>Amazon Machine Image<\/em> (AMI) which was assigned during registration. For more information about Amazon EC2 images, please see <a href=\"http:\/\/aws.amazon.com\/ec2\/\">Amazon EC2 product documentation<\/a>. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"instance_type","type":"string","description":"<p>The instance type of the Amazon EC2 instance. For more information about Amazon EC2 instance types, please see <a href=\"http:\/\/aws.amazon.com\/ec2\/\">Amazon EC2 product documentation<\/a> [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>KeyName<\/code> - <code>string<\/code> - Optional - The name of the Amazon EC2 key pair. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>SecurityGroups<\/code> - <code>string|array<\/code> - Optional - The names of the security groups with which to associate Amazon EC2 or Amazon VPC instances. Specify Amazon EC2 security groups using security group names, such as <code>websrv<\/code>. Specify Amazon VPC security groups using security group IDs, such as <code>sg-12345678<\/code>. For more information about Amazon EC2 security groups, go to <a href=\"http:\/\/docs.amazonwebservices.com\/AWSEC2\/latest\/UserGuide\/index.html?using-network-security.html\">Using Security Groups<\/a> in the Amazon EC2 product documentation. For more information about Amazon VPC security groups, go to <a href=\"http:\/\/docs.amazonwebservices.com\/AmazonVPC\/latest\/UserGuide\/index.html?VPC_SecurityGroups.html\">Security Groups<\/a> in the Amazon VPC product documentation. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>UserData<\/code> - <code>string<\/code> - Optional - The user data available to the launched Amazon EC2 instances. For more information about Amazon EC2 user data, please see <a href=\"http:\/\/aws.amazon.com\/ec2\/\">Amazon EC2 product documentation<\/a>. [Constraints: The value must be between 0 and 21847 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>KernelId<\/code> - <code>string<\/code> - Optional - The ID of the kernel associated with the Amazon EC2 AMI. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>RamdiskId<\/code> - <code>string<\/code> - Optional - The ID of the RAM disk associated with the Amazon EC2 AMI. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>BlockDeviceMappings<\/code> - <code>array<\/code> - Optional - A list of mappings that specify how block devices are exposed to the instance. Each mapping is made up of a <em>VirtualName<\/em>, a <em>DeviceName<\/em>, and an <em>ebs<\/em> data structure that contains information about the associated Elastic Block Storage volume. For more information about Amazon EC2 BlockDeviceMappings, go to <a href=\"http:\/\/docs.amazonwebservices.com\/AWSEC2\/latest\/UserGuide\/index.html?block-device-mapping-concepts.html\">Block Device Mapping<\/a> in the Amazon EC2 product documentation. <ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>VirtualName<\/code> - <code>string<\/code> - Optional - The virtual name associated with the device. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>DeviceName<\/code> - <code>string<\/code> - Required - The name of the device within Amazon EC2. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>Ebs<\/code> - <code>array<\/code> - Optional - The Elastic Block Storage volume information. <ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>SnapshotId<\/code> - <code>string<\/code> - Optional - The snapshot ID. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>VolumeSize<\/code> - <code>integer<\/code> - Optional - The volume size, in gigabytes.<\/li><\/ul><\/li><\/ul><\/li><\/ul><\/li><\/ul><\/li><li><code>InstanceMonitoring<\/code> - <code>array<\/code> - Optional - Enables detailed monitoring, which is enabled by default. When detailed monitoring is enabled, CloudWatch will generate metrics every minute and your account will be charged a fee. When you disable detailed monitoring, by specifying <code>False<\/code>, Cloudwatch will generate metrics every 5 minutes. For information about monitoring, see the <a href=\"http:\/\/aws.amazon.com\/cloudwatch\/\">Amazon CloudWatch<\/a> product page. <ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>Enabled<\/code> - <code>boolean<\/code> - Optional - If <code>True<\/code>, instance monitoring is enabled.<\/li><\/ul><\/li><\/ul><\/li><li><code>SpotPrice<\/code> - <code>string<\/code> - Optional - The maximum hourly price to be paid for any Spot Instance launched to fulfill the request. Spot Instances are launched when the price you specify exceeds the current Spot market price. For more information on launching Spot Instances, go to <a href=\"http:\/\/docs.amazonwebservices.com\/AutoScaling\/latest\/DeveloperGuide\/US-SpotInstances.html\">Using Auto Scaling to Launch Spot Instances<\/a> in the <em>Auto Scaling Developer Guide<\/em>.<\/li><li><code>IamInstanceProfile<\/code> - <code>string<\/code> - Optional - The name or the Amazon Resource Name (ARN) of the instance profile associated with the IAM role for the instance. For information on launching EC2 instances with an IAM role, go to <a href=\"http:\/\/docs.amazonwebservices.com\/AutoScaling\/latest\/DeveloperGuide\/us-iam-role.html\">Launching Auto Scaling Instances With an IAM Role<\/a> in the <em>Auto Scaling Developer Guide<\/em>. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>EbsOptimized<\/code> - <code>boolean<\/code> - Optional - <\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_launch_configuration($launch_configuration_name, $image_id, $instance_type, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['LaunchConfigurationName'] = $launch_configuration_name;\n    $opt['ImageId'] = $image_id;\n    $opt['InstanceType'] = $instance_type;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['SecurityGroups']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'SecurityGroups' =&gt; (is_array($opt['SecurityGroups']) ? $opt['SecurityGroups'] : array($opt['SecurityGroups']))\n        ), 'member'));\n        unset($opt['SecurityGroups']);\n    }\n    \n    \/\/ Optional list + map\n    if (isset($opt['BlockDeviceMappings']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'BlockDeviceMappings' =&gt; $opt['BlockDeviceMappings']\n        ), 'member'));\n        unset($opt['BlockDeviceMappings']);\n    }\n    \n    \/\/ Optional map (non-list)\n    if (isset($opt['InstanceMonitoring']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'InstanceMonitoring' =&gt; $opt['InstanceMonitoring']\n        ), 'member'));\n        unset($opt['InstanceMonitoring']);\n    }\n\n    return $this-&gt;authenticate('CreateLaunchConfiguration', $opt);\n}","examples":{"example":{"title":"<p>Create a new auto scaling launch configuration.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n\/\/ Using a micro instance running the 64-bit, EBS-backed Amazon Linux AMI.\n$response = $as-&gt;create_launch_configuration('my-launch-config', 'ami-38c33651', 't1.micro', array(\n\t'KeyName' =&gt; 'my-keypair',\n\t'SecurityGroups' =&gt; 'default'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/delete_launch_configuration.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_launch_configuration\"><code>delete_launch_configuration()<\/code><\/a>","<a href=\"..\/amazonas\/describe_launch_configurations.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_launch_configurations\"><code>describe_launch_configurations()<\/code><\/a>"]}},{"name":"create_or_update_tags","modifier":"public","description":"<p>Creates new tags or updates existing tags for an Auto Scaling group.<\/p>\n\n<p class=\"note\">\nA tag&#8217;s definition is composed of a resource ID, resource type, key and value, and the\npropagate flag. Value and the propagate flag are optional parameters. See the Request\nParameters for more information.\n<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"tags","type":"array","description":"<p>The tag to be created or updated. Each tag should be defined by its resource type, resource ID, key, value, and a propagate flag. The resource type and resource ID identify the type and name of resource for which the tag is created. Currently, <code>auto-scaling-group<\/code> is the only supported resource type. The valid value for the resource ID is <em>groupname<\/em>. The <code>PropagateAtLaunch<\/code> flag defines whether the new tag will be applied to instances launched by the Auto Scaling group. Valid values are <code>true<\/code> or <code>false<\/code>. However, instances that are already running will not get the new or updated tag. Likewise, when you modify a tag, the updated version will be applied only to new instances launched by the Auto Scaling group after the change. Running instances that had the previous version of the tag will continue to have the older tag. When you create a tag and a tag of the same name already exists, the operation overwrites the previous tag definition, but you will not get an error message.<\/p>\n\n<ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>ResourceId<\/code> - <code>string<\/code> - Optional - The name of the Auto Scaling group. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>ResourceType<\/code> - <code>string<\/code> - Optional - The kind of resource to which the tag is applied. Currently, Auto Scaling supports the <code>auto-scaling-group<\/code> resource type. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>Key<\/code> - <code>string<\/code> - Required - The key of the tag. [Constraints: The value must be between 1 and 128 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>Value<\/code> - <code>string<\/code> - Optional - The value of the tag. [Constraints: The value must be between 0 and 256 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>PropagateAtLaunch<\/code> - <code>boolean<\/code> - Optional - Specifies whether the new tag will be applied to instances launched after the tag is created. The same behavior applies to updates: If you change a tag, the changed tag will be applied to all instances launched after you made the change.<\/li><\/ul><\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_or_update_tags($tags, $opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Required list + map\n    $opt = array_merge($opt, CFComplexType::map(array(\n        'Tags' =&gt; (is_array($tags) ? $tags : array($tags))\n    ), 'member'));\n\n    return $this-&gt;authenticate('CreateOrUpdateTags', $opt);\n}"},{"name":"delete_auto_scaling_group","modifier":"public","description":"<p>Deletes the specified Auto Scaling group if the group has no instances and no scaling\nactivities in progress.<\/p>\n\n<p class=\"note\">\nTo remove all instances before calling <code>DeleteAutoScalingGroup<\/code>, you can call\n<code>UpdateAutoScalingGroup<\/code> to set the minimum and maximum size of the AutoScalingGroup\nto zero.\n<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>ForceDelete<\/code> - <code>boolean<\/code> - Optional - Starting with API version 2011-01-01, specifies that the Auto Scaling group will be deleted along with all instances associated with the group, without waiting for all instances to be terminated.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_auto_scaling_group($auto_scaling_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    \n    return $this-&gt;authenticate('DeleteAutoScalingGroup', $opt);\n}","examples":{"example":{"title":"<p>Delete an auto scaling group.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;delete_auto_scaling_group('my-scaling-group');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/create_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/create_auto_scaling_group\"><code>create_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/describe_auto_scaling_groups.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_auto_scaling_groups\"><code>describe_auto_scaling_groups()<\/code><\/a>","<a href=\"..\/amazonas\/describe_auto_scaling_instances.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_auto_scaling_instances\"><code>describe_auto_scaling_instances()<\/code><\/a>","<a href=\"..\/amazonas\/terminate_instance_in_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/terminate_instance_in_auto_scaling_group\"><code>terminate_instance_in_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/update_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/update_auto_scaling_group\"><code>update_auto_scaling_group()<\/code><\/a>"]}},{"name":"delete_cache","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Deletes a cached <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object using the specified cache storage type.<\/p>","returnvalue":"<p>A value of <code>true<\/code> if cached object exists and is successfully deleted, otherwise <code>false<\/code>.<\/p>","source":"public function delete_cache()\n{\n    $this-&gt;use_cache_flow = true;\n    $this-&gt;delete_cache = true;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Delete a cached response from APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ Delete the data\n$response = $sdb-&gt;delete_cache()-&gt;list_domains();\nvar_dump($response);","result":"bool(false)"},{"title":"<p>Delete a cached response from the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ Delete the data\n$response = $sdb-&gt;delete_cache()-&gt;list_domains();\nvar_dump($response);","result":"bool(true)"},{"title":"<p>Delete a cached response from Memcache.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config(array(\n\tarray('host' =&gt; '127.0.0.1')\n));\n\n\/\/ Delete the data\n$response = $sdb-&gt;delete_cache()-&gt;list_domains();\nvar_dump($response);","result":"bool(true)"}]}},{"name":"delete_launch_configuration","modifier":"public","description":"<p>Deletes the specified <code>LaunchConfiguration<\/code>.<\/p>\n\n<p>The specified launch configuration must not be attached to an Auto Scaling group. When this\ncall completes, the launch configuration is no longer available for use.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"launch_configuration_name","type":"string","description":"<p>The name of the launch configuration. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_launch_configuration($launch_configuration_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['LaunchConfigurationName'] = $launch_configuration_name;\n    \n    return $this-&gt;authenticate('DeleteLaunchConfiguration', $opt);\n}","examples":{"example":{"title":"<p>Delete an auto scaling launch configuration.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;delete_launch_configuration('my-launch-config');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/create_launch_configuration.html\" rel=\"..\/..\/index.html#m=amazonas\/create_launch_configuration\"><code>create_launch_configuration()<\/code><\/a>","<a href=\"..\/amazonas\/describe_launch_configurations.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_launch_configurations\"><code>describe_launch_configurations()<\/code><\/a>"]}},{"name":"delete_notification_configuration","modifier":"public","description":"<p>Deletes notifications created by <code>PutNotificationConfiguration<\/code>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"topic_arn","type":"string","description":"<p>The Amazon Resource Name (ARN) of the Amazon Simple Notification Service (SNS) topic. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_notification_configuration($auto_scaling_group_name, $topic_arn, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    $opt['TopicARN'] = $topic_arn;\n    \n    return $this-&gt;authenticate('DeleteNotificationConfiguration', $opt);\n}"},{"name":"delete_policy","modifier":"public","description":"<p>Deletes a policy created by <code>PutScalingPolicy<\/code>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"policy_name","type":"string","description":"<p>The name or PolicyARN of the policy you want to delete. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>AutoScalingGroupName<\/code> - <code>string<\/code> - Optional - The name of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_policy($policy_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['PolicyName'] = $policy_name;\n    \n    return $this-&gt;authenticate('DeletePolicy', $opt);\n}","examples":{"example":{"title":"<p>Delete an autoscaling policy.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;delete_policy('my-scaling-policy', array(\n\t'AutoScalingGroupName' =&gt; 'my-scaling-group',\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/describe_policies.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_policies\"><code>describe_policies()<\/code><\/a>","<a href=\"..\/amazonas\/execute_policy.html\" rel=\"..\/..\/index.html#m=amazonas\/execute_policy\"><code>execute_policy()<\/code><\/a>","<a href=\"..\/amazonas\/put_scaling_policy.html\" rel=\"..\/..\/index.html#m=amazonas\/put_scaling_policy\"><code>put_scaling_policy()<\/code><\/a>"]}},{"name":"delete_scheduled_action","modifier":"public","description":"<p>Deletes a scheduled action previously created using the\n<code>PutScheduledUpdateGroupAction<\/code>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"scheduled_action_name","type":"string","description":"<p>The name of the action you want to delete. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>AutoScalingGroupName<\/code> - <code>string<\/code> - Optional - The name of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_scheduled_action($scheduled_action_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['ScheduledActionName'] = $scheduled_action_name;\n    \n    return $this-&gt;authenticate('DeleteScheduledAction', $opt);\n}","examples":{"example":{"title":"<p>Delete a scheduled scaling action for an auto scaling group.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;delete_scheduled_action('my-scheduled-action', array(\n\t'AutoScalingGroupName' =&gt; 'my-scaling-group',\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/describe_scheduled_actions.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_scheduled_actions\"><code>describe_scheduled_actions()<\/code><\/a>","<a href=\"..\/amazonas\/put_scheduled_update_group_action.html\" rel=\"..\/..\/index.html#m=amazonas\/put_scheduled_update_group_action\"><code>put_scheduled_update_group_action()<\/code><\/a>"]}},{"name":"delete_tags","modifier":"public","description":"<p>Removes the specified tags or a set of tags from a set of resources.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"tags","type":"array","description":"<p>Each tag should be defined by its resource type, resource ID, key, value, and a propagate flag. Valid values are: Resource type = <em>auto-scaling-group<\/em>, Resource ID = <em>AutoScalingGroupName<\/em>, key= <em>value<\/em>, value= <em>value<\/em>, propagate= <em>true<\/em> or <em>false<\/em>.<\/p>\n\n<ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>ResourceId<\/code> - <code>string<\/code> - Optional - The name of the Auto Scaling group. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>ResourceType<\/code> - <code>string<\/code> - Optional - The kind of resource to which the tag is applied. Currently, Auto Scaling supports the <code>auto-scaling-group<\/code> resource type. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>Key<\/code> - <code>string<\/code> - Required - The key of the tag. [Constraints: The value must be between 1 and 128 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>Value<\/code> - <code>string<\/code> - Optional - The value of the tag. [Constraints: The value must be between 0 and 256 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>PropagateAtLaunch<\/code> - <code>boolean<\/code> - Optional - Specifies whether the new tag will be applied to instances launched after the tag is created. The same behavior applies to updates: If you change a tag, the changed tag will be applied to all instances launched after you made the change.<\/li><\/ul><\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_tags($tags, $opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Required list + map\n    $opt = array_merge($opt, CFComplexType::map(array(\n        'Tags' =&gt; (is_array($tags) ? $tags : array($tags))\n    ), 'member'));\n\n    return $this-&gt;authenticate('DeleteTags', $opt);\n}"},{"name":"describe_adjustment_types","modifier":"public","description":"<p>Returns policy adjustment types for use in the <code>PutScalingPolicy<\/code> action.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_adjustment_types($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeAdjustmentTypes', $opt);\n}","examples":{"example":{"title":"<p>Describe all adjustment types.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;describe_adjustment_types();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/describe_metric_collection_types.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_metric_collection_types\"><code>describe_metric_collection_types()<\/code><\/a>","<a href=\"..\/amazonas\/describe_scaling_activities.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_scaling_activities\"><code>describe_scaling_activities()<\/code><\/a>","<a href=\"..\/amazonas\/describe_scaling_process_types.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_scaling_process_types\"><code>describe_scaling_process_types()<\/code><\/a>"]}},{"name":"describe_auto_scaling_groups","modifier":"public","description":"<p>Returns a full description of each Auto Scaling group in the given list. This includes all\nAmazon EC2 instances that are members of the group. If a list of names is not provided, the\nservice returns the full details of all Auto Scaling groups.<\/p>\n\n<p>This action supports pagination by returning a token if there are more pages to retrieve. To\nget the next page, call this action again with the returned token as the <code>NextToken<\/code>\nparameter.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>AutoScalingGroupNames<\/code> - <code>string|array<\/code> - Optional - A list of Auto Scaling group names. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>NextToken<\/code> - <code>string<\/code> - Optional - A string that marks the start of the next batch of returned results. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to return.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_auto_scaling_groups($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Optional list (non-map)\n    if (isset($opt['AutoScalingGroupNames']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'AutoScalingGroupNames' =&gt; (is_array($opt['AutoScalingGroupNames']) ? $opt['AutoScalingGroupNames'] : array($opt['AutoScalingGroupNames']))\n        ), 'member'));\n        unset($opt['AutoScalingGroupNames']);\n    }\n\n    return $this-&gt;authenticate('DescribeAutoScalingGroups', $opt);\n}","examples":{"example":[{"title":"<p>Describe all auto scaling groups.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;describe_auto_scaling_groups();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Describe specific scaling groups.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n\/\/ Fetch list of all LaunchConfigurationName that contain the word &quot;Test&quot;.\n$response = $as-&gt;describe_auto_scaling_groups();\n$list_of_group_names = $response-&gt;body-&gt;query('\/\/AutoScalingGroupName[contains(., &quot;Test&quot;)]')-&gt;map_string();\n\n\/\/ Pass the list (array) into a new request.\n$response = $as-&gt;describe_auto_scaling_groups(array(\n\t'AutoScalingGroupNames' =&gt; $list_of_group_names\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}]},"related":{"method":["<a href=\"..\/amazonas\/create_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/create_auto_scaling_group\"><code>create_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/delete_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_auto_scaling_group\"><code>delete_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/describe_auto_scaling_instances.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_auto_scaling_instances\"><code>describe_auto_scaling_instances()<\/code><\/a>","<a href=\"..\/amazonas\/terminate_instance_in_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/terminate_instance_in_auto_scaling_group\"><code>terminate_instance_in_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/update_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/update_auto_scaling_group\"><code>update_auto_scaling_group()<\/code><\/a>"]}},{"name":"describe_auto_scaling_instances","modifier":"public","description":"<p>Returns a description of each Auto Scaling instance in the <code>InstanceIds<\/code> list. If a\nlist is not provided, the service returns the full details of all instances up to a maximum of\n50. By default, the service returns a list of 20 items.<\/p>\n\n<p>This action supports pagination by returning a token if there are more pages to retrieve. To\nget the next page, call this action again with the returned token as the <code>NextToken<\/code>\nparameter.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>InstanceIds<\/code> - <code>string|array<\/code> - Optional - The list of Auto Scaling instances to describe. If this list is omitted, all auto scaling instances are described. The list of requested instances cannot contain more than 50 items. If unknown instances are requested, they are ignored with no error. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of Auto Scaling instances to be described with each call.<\/li><li><code>NextToken<\/code> - <code>string<\/code> - Optional - The token returned by a previous call to indicate that there is more data available. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_auto_scaling_instances($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Optional list (non-map)\n    if (isset($opt['InstanceIds']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'InstanceIds' =&gt; (is_array($opt['InstanceIds']) ? $opt['InstanceIds'] : array($opt['InstanceIds']))\n        ), 'member'));\n        unset($opt['InstanceIds']);\n    }\n\n    return $this-&gt;authenticate('DescribeAutoScalingInstances', $opt);\n}","examples":{"example":{"title":"<p>Describe all auto scaling instances.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;describe_auto_scaling_instances();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/create_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/create_auto_scaling_group\"><code>create_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/delete_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_auto_scaling_group\"><code>delete_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/describe_auto_scaling_groups.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_auto_scaling_groups\"><code>describe_auto_scaling_groups()<\/code><\/a>","<a href=\"..\/amazonas\/terminate_instance_in_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/terminate_instance_in_auto_scaling_group\"><code>terminate_instance_in_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/update_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/update_auto_scaling_group\"><code>update_auto_scaling_group()<\/code><\/a>"]}},{"name":"describe_auto_scaling_notification_types","modifier":"public","description":"<p>Returns a list of all notification types that are supported by Auto Scaling.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_auto_scaling_notification_types($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeAutoScalingNotificationTypes', $opt);\n}"},{"name":"describe_launch_configurations","modifier":"public","description":"<p>Returns a full description of the launch configurations, or the specified launch\nconfigurations, if they exist.<\/p>\n\n<p>If no name is specified, then the full details of all launch configurations are returned.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>LaunchConfigurationNames<\/code> - <code>string|array<\/code> - Optional - A list of launch configuration names. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>NextToken<\/code> - <code>string<\/code> - Optional - A string that marks the start of the next batch of returned results. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of launch configurations. The default is 100.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_launch_configurations($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Optional list (non-map)\n    if (isset($opt['LaunchConfigurationNames']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'LaunchConfigurationNames' =&gt; (is_array($opt['LaunchConfigurationNames']) ? $opt['LaunchConfigurationNames'] : array($opt['LaunchConfigurationNames']))\n        ), 'member'));\n        unset($opt['LaunchConfigurationNames']);\n    }\n\n    return $this-&gt;authenticate('DescribeLaunchConfigurations', $opt);\n}","examples":{"example":[{"title":"<p>Describe all launch configurations.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;describe_launch_configurations();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Describe specific launch configurations.<\/p>","description":"<p>This is a fairly contrived example where we look up a list of launch configuration names, then pass that array back into a new call.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n\/\/ Fetch list of all LaunchConfigurationName that contain the word &quot;Test&quot;.\n$response = $as-&gt;describe_launch_configurations();\n$list_of_launch_configurations = $response-&gt;body-&gt;query('\/\/LaunchConfigurationName[contains(., &quot;Test&quot;)]')-&gt;map_string();\n\n\/\/ Pass the list (array) into a new request.\n$response = $as-&gt;describe_launch_configurations(array(\n\t'LaunchConfigurationNames' =&gt; $list_of_launch_configurations\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}]},"related":{"method":["<a href=\"..\/amazonas\/create_launch_configuration.html\" rel=\"..\/..\/index.html#m=amazonas\/create_launch_configuration\"><code>create_launch_configuration()<\/code><\/a>","<a href=\"..\/amazonas\/delete_launch_configuration.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_launch_configuration\"><code>delete_launch_configuration()<\/code><\/a>"]}},{"name":"describe_metric_collection_types","modifier":"public","description":"<p>Returns a list of metrics and a corresponding list of granularities for each metric.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_metric_collection_types($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeMetricCollectionTypes', $opt);\n}","examples":{"example":{"title":"<p>Describe all metric collection types.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;describe_metric_collection_types();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/describe_adjustment_types.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_adjustment_types\"><code>describe_adjustment_types()<\/code><\/a>","<a href=\"..\/amazonas\/describe_scaling_activities.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_scaling_activities\"><code>describe_scaling_activities()<\/code><\/a>","<a href=\"..\/amazonas\/describe_scaling_process_types.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_scaling_process_types\"><code>describe_scaling_process_types()<\/code><\/a>"]}},{"name":"describe_notification_configurations","modifier":"public","description":"<p>Returns a list of notification actions associated with Auto Scaling groups for specified\nevents.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>AutoScalingGroupNames<\/code> - <code>string|array<\/code> - Optional - The name of the Auto Scaling group. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>NextToken<\/code> - <code>string<\/code> - Optional - A string that is used to mark the start of the next batch of returned results for pagination. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - Maximum number of records to be returned.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_notification_configurations($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Optional list (non-map)\n    if (isset($opt['AutoScalingGroupNames']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'AutoScalingGroupNames' =&gt; (is_array($opt['AutoScalingGroupNames']) ? $opt['AutoScalingGroupNames'] : array($opt['AutoScalingGroupNames']))\n        ), 'member'));\n        unset($opt['AutoScalingGroupNames']);\n    }\n\n    return $this-&gt;authenticate('DescribeNotificationConfigurations', $opt);\n}"},{"name":"describe_policies","modifier":"public","description":"<p>Returns descriptions of what each policy does. This action supports pagination. If the response\nincludes a token, there are more records available. To get the additional records, repeat the\nrequest with the response token as the <code>NextToken<\/code> parameter.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>AutoScalingGroupName<\/code> - <code>string<\/code> - Optional - The name of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>PolicyNames<\/code> - <code>string|array<\/code> - Optional - A list of policy names or policy ARNs to be described. If this list is omitted, all policy names are described. If an auto scaling group name is provided, the results are limited to that group. The list of requested policy names cannot contain more than 50 items. If unknown policy names are requested, they are ignored with no error. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>NextToken<\/code> - <code>string<\/code> - Optional - A string that is used to mark the start of the next batch of returned results for pagination. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of policies that will be described with each call.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_policies($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Optional list (non-map)\n    if (isset($opt['PolicyNames']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'PolicyNames' =&gt; (is_array($opt['PolicyNames']) ? $opt['PolicyNames'] : array($opt['PolicyNames']))\n        ), 'member'));\n        unset($opt['PolicyNames']);\n    }\n\n    return $this-&gt;authenticate('DescribePolicies', $opt);\n}","examples":{"example":{"title":"<p>Describe all policies.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;describe_policies();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/delete_policy.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_policy\"><code>delete_policy()<\/code><\/a>","<a href=\"..\/amazonas\/execute_policy.html\" rel=\"..\/..\/index.html#m=amazonas\/execute_policy\"><code>execute_policy()<\/code><\/a>","<a href=\"..\/amazonas\/put_scaling_policy.html\" rel=\"..\/..\/index.html#m=amazonas\/put_scaling_policy\"><code>put_scaling_policy()<\/code><\/a>"]}},{"name":"describe_scaling_activities","modifier":"public","description":"<p>Returns the scaling activities for the specified Auto Scaling group.<\/p>\n\n<p>If the specified <code>ActivityIds<\/code> list is empty, all the activities from the past six\nweeks are returned. Activities are sorted by completion time. Activities still in progress\nappear first on the list.<\/p>\n\n<p>This action supports pagination. If the response includes a token, there are more records\navailable. To get the additional records, repeat the request with the response token as the\n<code>NextToken<\/code> parameter.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>ActivityIds<\/code> - <code>string|array<\/code> - Optional - A list containing the activity IDs of the desired scaling activities. If this list is omitted, all activities are described. If an <code>AutoScalingGroupName<\/code> is provided, the results are limited to that group. The list of requested activities cannot contain more than 50 items. If unknown activities are requested, they are ignored with no error. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>AutoScalingGroupName<\/code> - <code>string<\/code> - Optional - The name of the <code>AutoScalingGroup<\/code>. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of scaling activities to return.<\/li><li><code>NextToken<\/code> - <code>string<\/code> - Optional - A string that marks the start of the next batch of returned results for pagination. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_scaling_activities($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Optional list (non-map)\n    if (isset($opt['ActivityIds']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'ActivityIds' =&gt; (is_array($opt['ActivityIds']) ? $opt['ActivityIds'] : array($opt['ActivityIds']))\n        ), 'member'));\n        unset($opt['ActivityIds']);\n    }\n\n    return $this-&gt;authenticate('DescribeScalingActivities', $opt);\n}","examples":{"example":{"title":"<p>Describe all scaling activities.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;describe_scaling_activities();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/describe_adjustment_types.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_adjustment_types\"><code>describe_adjustment_types()<\/code><\/a>","<a href=\"..\/amazonas\/describe_metric_collection_types.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_metric_collection_types\"><code>describe_metric_collection_types()<\/code><\/a>","<a href=\"..\/amazonas\/describe_scaling_process_types.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_scaling_process_types\"><code>describe_scaling_process_types()<\/code><\/a>"]}},{"name":"describe_scaling_process_types","modifier":"public","description":"<p>Returns scaling process types for use in the <code>ResumeProcesses<\/code> and\n<code>SuspendProcesses<\/code> actions.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_scaling_process_types($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeScalingProcessTypes', $opt);\n}","examples":{"example":{"title":"<p>Describe all scaling process types.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;describe_scaling_process_types();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/describe_adjustment_types.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_adjustment_types\"><code>describe_adjustment_types()<\/code><\/a>","<a href=\"..\/amazonas\/describe_metric_collection_types.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_metric_collection_types\"><code>describe_metric_collection_types()<\/code><\/a>","<a href=\"..\/amazonas\/describe_scaling_activities.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_scaling_activities\"><code>describe_scaling_activities()<\/code><\/a>"]}},{"name":"describe_scheduled_actions","modifier":"public","description":"<p>Lists all the actions scheduled for your Auto Scaling group that haven&#8217;t been executed. To see\na list of actions already executed, see the activity record returned in\n<code>DescribeScalingActivities<\/code>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>AutoScalingGroupName<\/code> - <code>string<\/code> - Optional - The name of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>ScheduledActionNames<\/code> - <code>string|array<\/code> - Optional - A list of scheduled actions to be described. If this list is omitted, all scheduled actions are described. The list of requested scheduled actions cannot contain more than 50 items. If an auto scaling group name is provided, the results are limited to that group. If unknown scheduled actions are requested, they are ignored with no error. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>StartTime<\/code> - <code>string<\/code> - Optional - The earliest scheduled start time to return. If scheduled action names are provided, this field will be ignored. May be passed as a number of seconds since UNIX Epoch, or any string compatible with <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a>.<\/li><li><code>EndTime<\/code> - <code>string<\/code> - Optional - The latest scheduled start time to return. If scheduled action names are provided, this field is ignored. May be passed as a number of seconds since UNIX Epoch, or any string compatible with <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a>.<\/li><li><code>NextToken<\/code> - <code>string<\/code> - Optional - A string that marks the start of the next batch of returned results. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of scheduled actions to return.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_scheduled_actions($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Optional DateTime\n    if (isset($opt['StartTime']))\n    {\n        $opt['StartTime'] = $this-&gt;util-&gt;convert_date_to_iso8601($opt['StartTime']);\n    }\n    \n    \/\/ Optional DateTime\n    if (isset($opt['EndTime']))\n    {\n        $opt['EndTime'] = $this-&gt;util-&gt;convert_date_to_iso8601($opt['EndTime']);\n    }\n\n    \/\/ Optional list (non-map)\n    if (isset($opt['ScheduledActionNames']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'ScheduledActionNames' =&gt; (is_array($opt['ScheduledActionNames']) ? $opt['ScheduledActionNames'] : array($opt['ScheduledActionNames']))\n        ), 'member'));\n        unset($opt['ScheduledActionNames']);\n    }\n\n    return $this-&gt;authenticate('DescribeScheduledActions', $opt);\n}","examples":{"example":{"title":"<p>Describe all scheduled actions.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;describe_scheduled_actions();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/delete_scheduled_action.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_scheduled_action\"><code>delete_scheduled_action()<\/code><\/a>","<a href=\"..\/amazonas\/put_scheduled_update_group_action.html\" rel=\"..\/..\/index.html#m=amazonas\/put_scheduled_update_group_action\"><code>put_scheduled_update_group_action()<\/code><\/a>"]}},{"name":"describe_tags","modifier":"public","description":"<p>Lists the Auto Scaling group tags.<\/p>\n\n<p>You can use filters to limit results when describing tags. For example, you can query for tags\nof a particular Auto Scaling group. You can specify multiple values for a filter. A tag must\nmatch at least one of the specified values for it to be included in the results.<\/p>\n\n<p>You can also specify multiple filters. The result includes information for a particular tag\nonly if it matches all your filters. If there&#8217;s no match, no special message is returned.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>Filters<\/code> - <code>array<\/code> - Optional - The value of the filter type used to identify the tags to be returned. For example, you can filter so that tags are returned according to Auto Scaling group, the key and value, or whether the new tag will be applied to instances launched after the tag is created (PropagateAtLaunch). <ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>Name<\/code> - <code>string<\/code> - Optional - The name of the filter. Valid Name values are: <code>\"auto-scaling-group\"<\/code>, <code>\"key\"<\/code>, <code>\"value\"<\/code>, and <code>\"propagate-at-launch\"<\/code>. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>Values<\/code> - <code>string|array<\/code> - Optional - The value of the filter. Pass a string for a single value, or an indexed array for multiple values.<\/li><\/ul><\/li><\/ul><\/li><li><code>NextToken<\/code> - <code>string<\/code> - Optional - A string that marks the start of the next batch of returned results. [Constraints: The value must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to return.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_tags($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Optional list + map\n    if (isset($opt['Filters']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'Filters' =&gt; $opt['Filters']\n        ), 'member'));\n        unset($opt['Filters']);\n    }\n\n    return $this-&gt;authenticate('DescribeTags', $opt);\n}"},{"name":"describe_termination_policy_types","modifier":"public","description":"<p>Returns a list of all termination policies supported by Auto Scaling.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_termination_policy_types($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeTerminationPolicyTypes', $opt);\n}"},{"name":"disable_metrics_collection","modifier":"public","description":"<p>Disables monitoring of group metrics for the Auto Scaling group specified in\n<code>AutoScalingGroupName<\/code>. You can specify the list of affected metrics with the\n<code>Metrics<\/code> parameter.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name or ARN of the Auto Scaling Group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>Metrics<\/code> - <code>string|array<\/code> - Optional - The list of metrics to disable. If no metrics are specified, all metrics are disabled. The following metrics are supported:<ul><li>GroupMinSize<\/li><li>GroupMaxSize<\/li><li>GroupDesiredCapacity<\/li><li>GroupInServiceInstances<\/li><li>GroupPendingInstances<\/li><li>GroupTerminatingInstances<\/li><li>GroupTotalInstances<\/li><\/ul> Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function disable_metrics_collection($auto_scaling_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['Metrics']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'Metrics' =&gt; (is_array($opt['Metrics']) ? $opt['Metrics'] : array($opt['Metrics']))\n        ), 'member'));\n        unset($opt['Metrics']);\n    }\n\n    return $this-&gt;authenticate('DisableMetricsCollection', $opt);\n}","examples":{"example":{"title":"<p>Disable the collection of metrics.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;disable_metrics_collection('my-scaling-group', array(\n\t'Metrics' =&gt; array(\n\t\t'GroupMinSize',\n\t\t'GroupMaxSize',\n\t\t'GroupDesiredCapacity',\n\t\t'GroupInServiceInstances',\n\t\t'GroupPendingInstances',\n\t\t'GroupTerminatingInstances',\n\t\t'GroupTotalInstances'\n\t)\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":"<a href=\"..\/amazonas\/enable_metrics_collection.html\" rel=\"..\/..\/index.html#m=amazonas\/enable_metrics_collection\"><code>enable_metrics_collection()<\/code><\/a>"}},{"name":"disable_ssl","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Disables SSL\/HTTPS connections for hosts that don&#8217;t support them. Some services, however, still\nrequire SSL support.<\/p>\n\n<p>This method will throw a user warning when invoked, which can be hidden by changing your\n<a href=\"http:\/\/php.net\/error_reporting\"><code>error_reporting()<\/code><\/a> settings.<\/p>","returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function disable_ssl()\n{\n    trigger_error('Disabling SSL connections is potentially unsafe and highly discouraged.', E_USER_WARNING);\n    $this-&gt;use_ssl = false;\n    return $this;\n}","examples":{"example":{"title":"<p>Disable SSL for the service.<\/p>","description":"<ul>\n<li><strong>Note:<\/strong> It is not recommended to disable SSL.<\/li>\n<li><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/li>\n<\/ul>","code":"\/\/ Disable SSL (Using @ to suppress the warning from the output)\n$sdb = new AmazonSDB();\n@$sdb-&gt;disable_ssl(true);\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;use_ssl);","result":"bool(false)"}}},{"name":"disable_ssl_verification","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Disables the verification of the SSL Certificate Authority. Doing so can enable an attacker to carry\nout a man-in-the-middle attack.<\/p>\n\n<p>https:\/\/secure.wikimedia.org\/wikipedia\/en\/wiki\/Man-in-the-middle_attack<\/p>\n\n<p>This method will throw a user warning when invoked, which can be hidden by changing your\n<a href=\"http:\/\/php.net\/error_reporting\"><code>error_reporting()<\/code><\/a> settings.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"ssl_verification"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function disable_ssl_verification($ssl_verification = false)\n{\n    trigger_error('Disabling the verification of SSL certificates can lead to man-in-the-middle attacks. It is potentially unsafe and highly discouraged.', E_USER_WARNING);\n    $this-&gt;ssl_verification = $ssl_verification;\n    return $this;\n}"},{"name":"enable_debug_mode","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Enables HTTP request\/response header logging to <code>STDERR<\/code>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"enabled","type":"boolean","description":"<p>Whether or not to enable debug mode. Defaults to <code>true<\/code>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function enable_debug_mode($enabled = true)\n{\n    $this-&gt;debug_mode = $enabled;\n    return $this;\n}","examples":{"example":{"title":"<p>Enable the logging of HTTP request\/response headers to <code>STDERR<\/code>.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Adjust offset\n$sdb = new AmazonSDB();\n$sdb-&gt;enable_debug_mode();\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;debug_mode);","result":"bool(true)"}}},{"name":"enable_metrics_collection","modifier":"public","description":"<p>Enables monitoring of group metrics for the Auto Scaling group specified in\n<code>AutoScalingGroupName<\/code>. You can specify the list of enabled metrics with the\n<code>Metrics<\/code> parameter.<\/p>\n\n<p>Auto scaling metrics collection can be turned on only if the <code>InstanceMonitoring<\/code>\nflag, in the Auto Scaling group&#8217;s launch configuration, is set to <code>True<\/code>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name or ARN of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"granularity","type":"string","description":"<p>The granularity to associate with the metrics to collect. Currently, the only legal granularity is &#8220;1Minute&#8221;. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>Metrics<\/code> - <code>string|array<\/code> - Optional - The list of metrics to collect. If no metrics are specified, all metrics are enabled. The following metrics are supported:<ul><li>GroupMinSize<\/li><li>GroupMaxSize<\/li><li>GroupDesiredCapacity<\/li><li>GroupInServiceInstances<\/li><li>GroupPendingInstances<\/li><li>GroupTerminatingInstances<\/li><li>GroupTotalInstances<\/li><\/ul> Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function enable_metrics_collection($auto_scaling_group_name, $granularity, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    $opt['Granularity'] = $granularity;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['Metrics']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'Metrics' =&gt; (is_array($opt['Metrics']) ? $opt['Metrics'] : array($opt['Metrics']))\n        ), 'member'));\n        unset($opt['Metrics']);\n    }\n\n    return $this-&gt;authenticate('EnableMetricsCollection', $opt);\n}","examples":{"example":{"title":"<p>Enable the collection of metrics.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;enable_metrics_collection('my-scaling-group', '1Minute', array(\n\t'Metrics' =&gt; array(\n\t\t'GroupMinSize',\n\t\t'GroupMaxSize',\n\t\t'GroupDesiredCapacity',\n\t\t'GroupInServiceInstances',\n\t\t'GroupPendingInstances',\n\t\t'GroupTerminatingInstances',\n\t\t'GroupTotalInstances'\n\t)\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":"<a href=\"..\/amazonas\/disable_metrics_collection.html\" rel=\"..\/..\/index.html#m=amazonas\/disable_metrics_collection\"><code>disable_metrics_collection()<\/code><\/a>"}},{"name":"execute_policy","modifier":"public","description":"<p>Runs the policy you create for your Auto Scaling group in <code>PutScalingPolicy<\/code>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"policy_name","type":"string","description":"<p>The name or PolicyARN of the policy you want to run. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>AutoScalingGroupName<\/code> - <code>string<\/code> - Optional - The name or ARN of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>HonorCooldown<\/code> - <code>boolean<\/code> - Optional - Set to <code>True<\/code> if you want Auto Scaling to reject this request when the Auto Scaling group is in cooldown.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function execute_policy($policy_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['PolicyName'] = $policy_name;\n    \n    return $this-&gt;authenticate('ExecutePolicy', $opt);\n}","examples":{"example":{"title":"<p>Execute an autoscaling policy.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;execute_policy('my-scaling-policy', array(\n\t'AutoScalingGroupName' =&gt; 'my-scaling-group',\n\t'HonorCooldown' =&gt; 'false'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/delete_policy.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_policy\"><code>delete_policy()<\/code><\/a>","<a href=\"..\/amazonas\/describe_policies.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_policies\"><code>describe_policies()<\/code><\/a>","<a href=\"..\/amazonas\/put_scaling_policy.html\" rel=\"..\/..\/index.html#m=amazonas\/put_scaling_policy\"><code>put_scaling_policy()<\/code><\/a>"]}},{"name":"factory","modifier":"public static","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Alternate approach to constructing a new instance. Supports chaining.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"options","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>certificate_authority<\/code> - <code>boolean<\/code> - Optional - Determines which Cerificate Authority file to use. A value of boolean <code>false<\/code> will use the Certificate Authority file available on the system. A value of boolean <code>true<\/code> will use the Certificate Authority provided by the SDK. Passing a file system path to a Certificate Authority file (chmodded to <code>0755<\/code>) will use that. Leave this set to <code>false<\/code> if you&#8217;re not sure.<\/li><li><code>credentials<\/code> - <code>string<\/code> - Optional - The name of the credential set to use for authentication.<\/li><li><code>default_cache_config<\/code> - <code>string<\/code> - Optional - This option allows a preferred storage type to be configured for long-term caching. This can be changed later using the <a href=\"..\/amazonas\/set_cache_config.html\" rel=\"..\/..\/index.html#m=amazonas\/set_cache_config\"><code>set_cache_config()<\/code><\/a> method. Valid values are: <code>apc<\/code>, <code>xcache<\/code>, or a file system path such as <code>.\/cache<\/code> or <code>\/tmp\/cache\/<\/code>.<\/li><li><code>key<\/code> - <code>string<\/code> - Optional - Your AWS key, or a session key. If blank, the default credential set will be used.<\/li><li><code>secret<\/code> - <code>string<\/code> - Optional - Your AWS secret key, or a session secret key. If blank, the default credential set will be used.<\/li><li><code>token<\/code> - <code>string<\/code> - Optional - An AWS session token.<\/li><\/ul>"}},"returnvalue":{"@attributes":{"type":"void"}},"source":"public static function factory(array $options = array())\n{\n    if (version_compare(PHP_VERSION, '5.3.0', '&lt;'))\n    {\n        throw new Exception('PHP 5.3 or newer is required to instantiate a new class with CLASS::factory().');\n    }\n\n    $self = get_called_class();\n    return new $self($options);\n}"},{"name":"put_notification_configuration","modifier":"public","description":"<p>Configures an Auto Scaling group to send notifications when specified events take place.\nSubscribers to this topic can have messages for events delivered to an endpoint such as a web\nserver or email address.<\/p>\n\n<p>A new <code>PutNotificationConfiguration<\/code> overwrites an existing configuration.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"topic_arn","type":"string","description":"<p>The Amazon Resource Name (ARN) of the Amazon Simple Notification Service (SNS) topic. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"notification_types","type":"string|array","description":"<p>The type of events that will trigger the notification. For more information, go to <code>DescribeAutoScalingNotificationTypes<\/code>. Pass a string for a single value, or an indexed array for multiple values.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function put_notification_configuration($auto_scaling_group_name, $topic_arn, $notification_types, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    $opt['TopicARN'] = $topic_arn;\n    \n    \/\/ Required list (non-map)\n    $opt = array_merge($opt, CFComplexType::map(array(\n        'NotificationTypes' =&gt; (is_array($notification_types) ? $notification_types : array($notification_types))\n    ), 'member'));\n\n    return $this-&gt;authenticate('PutNotificationConfiguration', $opt);\n}"},{"name":"put_scaling_policy","modifier":"public","description":"<p>Creates or updates a policy for an Auto Scaling group. To update an existing policy, use the\nexisting policy name and set the parameter(s) you want to change. Any existing parameter not\nchanged in an update to an existing policy is not changed in this update request.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name or ARN of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"policy_name","type":"string","description":"<p>The name of the policy you want to create or update. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"scaling_adjustment","type":"integer","description":"<p>The number of instances by which to scale. <code>AdjustmentType<\/code> determines the interpretation of this number (e.g., as an absolute number or as a percentage of the existing Auto Scaling group size). A positive increment adds to the current capacity and a negative value removes from the current capacity.<\/p>"},{"@attributes":{"choice":"req"},"name":"adjustment_type","type":"string","description":"<p>Specifies whether the <code>ScalingAdjustment<\/code> is an absolute number or a percentage of the current capacity. Valid values are <code>ChangeInCapacity<\/code>, <code>ExactCapacity<\/code>, and <code>PercentChangeInCapacity<\/code>. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>Cooldown<\/code> - <code>integer<\/code> - Optional - The amount of time, in seconds, after a scaling activity completes before any further trigger-related scaling activities can start.<\/li><li><code>MinAdjustmentStep<\/code> - <code>integer<\/code> - Optional - Used with <code>AdjustmentType<\/code> with the value <code>PercentChangeInCapacity<\/code>, the scaling policy changes the <code>DesiredCapacity<\/code> of the Auto Scaling group by at least the number of instances specified in the value. You will get a <code>ValidationError<\/code> if you use <code>MinAdjustmentStep<\/code> on a policy with an <code>AdjustmentType<\/code> other than <code>PercentChangeInCapacity<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function put_scaling_policy($auto_scaling_group_name, $policy_name, $scaling_adjustment, $adjustment_type, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    $opt['PolicyName'] = $policy_name;\n    $opt['ScalingAdjustment'] = $scaling_adjustment;\n    $opt['AdjustmentType'] = $adjustment_type;\n    \n    return $this-&gt;authenticate('PutScalingPolicy', $opt);\n}","examples":{"example":{"title":"<p>Create or update an autoscaling policy.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;put_scaling_policy('my-scaling-group', 'my-scaling-policy', 1, 'ChangeInCapacity', array(\n\t'Cooldown' =&gt; 30,\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/delete_policy.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_policy\"><code>delete_policy()<\/code><\/a>","<a href=\"..\/amazonas\/describe_policies.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_policies\"><code>describe_policies()<\/code><\/a>","<a href=\"..\/amazonas\/execute_policy.html\" rel=\"..\/..\/index.html#m=amazonas\/execute_policy\"><code>execute_policy()<\/code><\/a>"]}},{"name":"put_scheduled_update_group_action","modifier":"public","description":"<p>Creates a scheduled scaling action for an Auto Scaling group. If you leave a parameter\nunspecified, the corresponding value remains unchanged in the affected Auto Scaling group.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name or ARN of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"scheduled_action_name","type":"string","description":"<p>The name of this scaling action. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>Time<\/code> - <code>string<\/code> - Optional - <code>Time<\/code> is deprecated. The time for this action to start. <code>Time<\/code> is an alias for <code>StartTime<\/code> and can be specified instead of <code>StartTime<\/code>, or vice versa. If both <code>Time<\/code> and <code>StartTime<\/code> are specified, their values should be identical. Otherwise, <code>PutScheduledUpdateGroupAction<\/code> will return an error. May be passed as a number of seconds since UNIX Epoch, or any string compatible with <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a>.<\/li><li><code>StartTime<\/code> - <code>string<\/code> - Optional - The time for this action to start, as in <code>--start-time 2010-06-01T00:00:00Z<\/code>. When <code>StartTime<\/code> and <code>EndTime<\/code> are specified with <code>Recurrence<\/code>, they form the boundaries of when the recurring action will start and stop. May be passed as a number of seconds since UNIX Epoch, or any string compatible with <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a>.<\/li><li><code>EndTime<\/code> - <code>string<\/code> - Optional - The time for this action to end. May be passed as a number of seconds since UNIX Epoch, or any string compatible with <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a>.<\/li><li><code>Recurrence<\/code> - <code>string<\/code> - Optional - The time when recurring future actions will start. Start time is specified by the user following the Unix cron syntax format. For information about cron syntax, go to <a href=\"http:\/\/en.wikipedia.org\/wiki\/Cron\">Wikipedia, The Free Encyclopedia<\/a>. When <code>StartTime<\/code> and <code>EndTime<\/code> are specified with <code>Recurrence<\/code>, they form the boundaries of when the recurring action will start and stop. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>MinSize<\/code> - <code>integer<\/code> - Optional - The minimum size for the new Auto Scaling group.<\/li><li><code>MaxSize<\/code> - <code>integer<\/code> - Optional - The maximum size for the Auto Scaling group.<\/li><li><code>DesiredCapacity<\/code> - <code>integer<\/code> - Optional - The number of Amazon EC2 instances that should be running in the group.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function put_scheduled_update_group_action($auto_scaling_group_name, $scheduled_action_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    $opt['ScheduledActionName'] = $scheduled_action_name;\n    \n    \/\/ Optional DateTime\n    if (isset($opt['Time']))\n    {\n        $opt['Time'] = $this-&gt;util-&gt;convert_date_to_iso8601($opt['Time']);\n    }\n    \n    \/\/ Optional DateTime\n    if (isset($opt['StartTime']))\n    {\n        $opt['StartTime'] = $this-&gt;util-&gt;convert_date_to_iso8601($opt['StartTime']);\n    }\n    \n    \/\/ Optional DateTime\n    if (isset($opt['EndTime']))\n    {\n        $opt['EndTime'] = $this-&gt;util-&gt;convert_date_to_iso8601($opt['EndTime']);\n    }\n\n    return $this-&gt;authenticate('PutScheduledUpdateGroupAction', $opt);\n}","examples":{"example":{"title":"<p>Create a scheduled scaling action for an auto scaling group.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;put_scheduled_update_group_action('my-scaling-group', 'my-scheduled-action', array(\n\t'Time' =&gt; '+10 minutes',\n\t'EndTime' =&gt; '+20 minutes',\n\t'MinSize' =&gt; 1,\n\t'MaxSize' =&gt; 2,\n\t'DesiredCapacity' =&gt; 2\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/delete_scheduled_action.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_scheduled_action\"><code>delete_scheduled_action()<\/code><\/a>","<a href=\"..\/amazonas\/describe_scheduled_actions.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_scheduled_actions\"><code>describe_scheduled_actions()<\/code><\/a>"]},"changelog":{"change":"<p>The <a href=\"http:\/\/docs.aws.amazon.com\/AutoScaling\/latest\/APIReference\/API_PutScheduledUpdateGroupAction.html\">PutScheduledUpdateGroupAction<\/a> operation introduced backwards-incompatible changes in the July 2011 update. These changes were accounted for in this version of the SDK.<\/p>"}},{"name":"register_streaming_read_callback","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Register a callback function to execute whenever a data stream is read from using\n<a href=\"..\/cfrequest\/streaming_read_callback.html\" rel=\"..\/..\/index.html#m=cfrequest\/streaming_read_callback\"><code>CFRequest::streaming_read_callback()<\/code><\/a>.<\/p>\n\n<p>The user-defined callback function should accept three arguments:<\/p>\n\n<ul><li><code>$curl_handle<\/code> - <code>resource<\/code> - Required - The cURL handle resource that represents the in-progress transfer.<\/li><li><code>$file_handle<\/code> - <code>resource<\/code> - Required - The file handle resource that represents the file on the local file system.<\/li><li><code>$length<\/code> - <code>integer<\/code> - Required - The length in kilobytes of the data chunk that was transferred.<\/li><\/ul>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"callback","type":"string|array|function","description":"<p>The callback function is called by <a href=\"http:\/\/php.net\/call_user_func\"><code>call_user_func()<\/code><\/a>, so you can pass the following values:<\/p>\n\n<ul><li>The name of a global function to execute, passed as a string.<\/li><li>A method to execute, passed as <code>array('ClassName', 'MethodName')<\/code>.<\/li><li>An anonymous function (PHP 5.3+).<\/li><\/ul>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function register_streaming_read_callback($callback)\n{\n    $this-&gt;registered_streaming_read_callback = $callback;\n    return $this;\n}","examples":{"example":{"title":"<p>Register a callback function to execute when data is read from an open stream.<\/p>","code":"$s3 = new AmazonS3();\n\n\/\/ Define callback function\nfunction callback($curl_handle, $file_handle, $length)\n{\n\t$total = curl_getinfo($curl_handle, CURLINFO_CONTENT_LENGTH_DOWNLOAD);\n\t$transferred = curl_getinfo($curl_handle, CURLINFO_SIZE_DOWNLOAD);\n\t$percentage = floor(($transferred \/ $total) * 100);\n\n\techo 'Transferred ' . $transferred . ' KB out of ' . $total . ' KB (' . $percentage . '%)' . PHP_EOL;\n}\n\n\/\/ Register a callback function to execute when a stream is read locally.\n$s3-&gt;register_streaming_read_callback('callback');\n\n\/\/ Download a public object.\n$response = $s3-&gt;create_object('my-bucket', 'my-large-file.mp4', array(\n\t'fileUpload' =&gt; '.\/uploads\/my-large-file.mp4'\n));","result":"Transferred 7200 bytes out of 9000000 (0%)\n...\nTransferred 9000000 bytes out of 9000000 (100%)"}}},{"name":"register_streaming_write_callback","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Register a callback function to execute whenever a data stream is written to using\n<a href=\"..\/cfrequest\/streaming_write_callback.html\" rel=\"..\/..\/index.html#m=cfrequest\/streaming_write_callback\"><code>CFRequest::streaming_write_callback()<\/code><\/a>.<\/p>\n\n<p>The user-defined callback function should accept two arguments:<\/p>\n\n<ul><li><code>$curl_handle<\/code> - <code>resource<\/code> - Required - The cURL handle resource that represents the in-progress transfer.<\/li><li><code>$length<\/code> - <code>integer<\/code> - Required - The length in kilobytes of the data chunk that was transferred.<\/li><\/ul>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"callback","type":"string|array|function","description":"<p>The callback function is called by <a href=\"http:\/\/php.net\/call_user_func\"><code>call_user_func()<\/code><\/a>, so you can pass the following values:<\/p>\n\n<ul><li>The name of a global function to execute, passed as a string.<\/li><li>A method to execute, passed as <code>array('ClassName', 'MethodName')<\/code>.<\/li><li>An anonymous function (PHP 5.3+).<\/li><\/ul>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function register_streaming_write_callback($callback)\n{\n    $this-&gt;registered_streaming_write_callback = $callback;\n    return $this;\n}","examples":{"example":{"title":"<p>Register a callback function to execute when data is written to an open stream.<\/p>","code":"$s3 = new AmazonS3();\n\n\/\/ Define callback function\nfunction callback($curl_handle, $length)\n{\n\t$total = curl_getinfo($curl_handle, CURLINFO_CONTENT_LENGTH_DOWNLOAD);\n\t$transferred = curl_getinfo($curl_handle, CURLINFO_SIZE_DOWNLOAD);\n\t$percentage = floor(($transferred \/ $total) * 100);\n\n\techo 'Transferred ' . $transferred . ' KB out of ' . $total . ' KB (' . $percentage . '%)' . PHP_EOL;\n}\n\n\/\/ Register a callback function to execute when a stream is written locally.\n$s3-&gt;register_streaming_write_callback('callback');\n\n\/\/ Download a public object.\n$response = $s3-&gt;get_object('my-bucket', 'my-large-file.mp4', array(\n\t'fileDownload' =&gt; '.\/downloads\/my-large-file.mp4'\n));","result":"Transferred 7200 bytes out of 9000000 (0%)\n...\nTransferred 9000000 bytes out of 9000000 (100%)"}}},{"name":"resume_processes","modifier":"public","description":"<p>Resumes Auto Scaling processes for an Auto Scaling group. For more information, see\n<code>SuspendProcesses<\/code> and <code>ProcessType<\/code>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name or Amazon Resource Name (ARN) of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>ScalingProcesses<\/code> - <code>string|array<\/code> - Optional - The processes that you want to suspend or resume, which can include one or more of the following:<ul><li>Launch<\/li><li>Terminate<\/li><li>HealthCheck<\/li><li>ReplaceUnhealthy<\/li><li>AZRebalance<\/li><li>AlarmNotification<\/li><li>ScheduledActions<\/li><li>AddToLoadBalancer<\/li><\/ul>To suspend all process types, omit this parameter. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function resume_processes($auto_scaling_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['ScalingProcesses']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'ScalingProcesses' =&gt; (is_array($opt['ScalingProcesses']) ? $opt['ScalingProcesses'] : array($opt['ScalingProcesses']))\n        ), 'member'));\n        unset($opt['ScalingProcesses']);\n    }\n\n    return $this-&gt;authenticate('ResumeProcesses', $opt);\n}","examples":{"example":{"title":"<p>Resume autoscaling processes.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;resume_processes('my-scaling-group', array(\n\t'ScalingProcesses' =&gt; array(\n\t\t'ReplaceUnhealthy',\n\t\t'HealthCheck',\n\t\t'ScheduledActions'\n\t)\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":"<a href=\"..\/amazonas\/suspend_processes.html\" rel=\"..\/..\/index.html#m=amazonas\/suspend_processes\"><code>suspend_processes()<\/code><\/a>"}},{"name":"send","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Executes the batch request queue by sending all queued requests.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"clear_after_send","type":"boolean","description":"<p>Whether or not to clear the batch queue after sending a request. Defaults to <code>true<\/code>. Set this to <code>false<\/code> if you are caching batch responses and want to retrieve results later.<\/p>"}},"returnvalue":"<p>An array of <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> objects.<\/p>","source":"public function send($clear_after_send = true)\n{\n    if ($this-&gt;use_batch_flow)\n    {\n        \/\/ When we send the request, disable batch flow.\n        $this-&gt;use_batch_flow = false;\n\n        \/\/ If we're not caching, simply send the request.\n        if (!$this-&gt;use_cache_flow)\n        {\n            $response = $this-&gt;batch_object-&gt;send();\n            $parsed_data = array_map(array($this, 'parse_callback'), $response);\n            $parsed_data = new CFArray($parsed_data);\n\n            \/\/ Clear the queue\n            if ($clear_after_send)\n            {\n                $this-&gt;batch_object-&gt;queue = array();\n            }\n\n            return $parsed_data;\n        }\n\n        \/\/ Generate an identifier specific to this particular set of arguments.\n        $cache_id = $this-&gt;key . '_' . get_class($this) . '_' . sha1(serialize($this-&gt;batch_object));\n\n        \/\/ Instantiate the appropriate caching object.\n        $this-&gt;cache_object = new $this-&gt;cache_class($cache_id, $this-&gt;cache_location, $this-&gt;cache_expires, $this-&gt;cache_compress);\n\n        if ($this-&gt;delete_cache)\n        {\n            $this-&gt;use_cache_flow = false;\n            $this-&gt;delete_cache = false;\n            return $this-&gt;cache_object-&gt;delete();\n        }\n\n        \/\/ Invoke the cache callback function to determine whether to pull data from the cache or make a fresh request.\n        $data_set = $this-&gt;cache_object-&gt;response_manager(array($this, 'cache_callback_batch'), array($this-&gt;batch_object));\n        $parsed_data = array_map(array($this, 'parse_callback'), $data_set);\n        $parsed_data = new CFArray($parsed_data);\n\n        \/\/ Clear the queue\n        if ($clear_after_send)\n        {\n            $this-&gt;batch_object-&gt;queue = array();\n        }\n\n        \/\/ End!\n        return $parsed_data;\n    }\n\n    \/\/ Load the class\n    $null = new CFBatchRequest();\n    unset($null);\n\n    throw new CFBatchRequest_Exception('You must use $object-&gt;batch()-&gt;send()');\n}"},{"name":"set_batch_class","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_batch_class($class = 'CFBatchRequest')\n{\n    $this-&gt;batch_class = $class;\n    return $this;\n}"},{"name":"set_cache_config","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set the caching configuration to use for response caching.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"location","type":"string","description":"<p>The location to store the cache object in. This may vary by cache method.<\/p><ul><li>File - The local file system paths such as <code>.\/cache<\/code> (relative) or <code>\/tmp\/cache\/<\/code> (absolute). The location must be server-writable.<\/li><li>APC - Pass in <code>apc<\/code> to use this lightweight cache. You must have the <a href=\"http:\/\/php.net\/apc\">APC extension<\/a> installed.<\/li><li>XCache - Pass in <code>xcache<\/code> to use this lightweight cache. You must have the <a href=\"http:\/\/xcache.lighttpd.net\">XCache<\/a> extension installed.<\/li><li>Memcached - Pass in an indexed array of associative arrays. Each associative array should have a <code>host<\/code> and a <code>port<\/code> value representing a <a href=\"http:\/\/php.net\/memcached\">Memcached<\/a> server to connect to.<\/li><li>PDO - A URL-style string (e.g. <code>pdo.mysql:\/\/user:pass@localhost\/cache<\/code>) or a standard DSN-style string (e.g. <code>pdo.sqlite:\/sqlite\/cache.db<\/code>). MUST be prefixed with <code>pdo.<\/code>. See <code>CachePDO<\/code> and <a href=\"http:\/\/php.net\/pdo\">PDO<\/a> for more details.<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"gzip","type":"boolean","description":"<p>Whether or not data should be gzipped before being stored. A value of <code>true<\/code> will compress the contents before caching them. A value of <code>false<\/code> will leave the contents uncompressed. Defaults to <code>true<\/code>.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_cache_config($location, $gzip = true)\n{\n    \/\/ If location is empty, don't do anything.\n    if (empty($location))\n    {\n        return $this;\n    }\n\n    \/\/ If we have an array, we're probably passing in Memcached servers and ports.\n    if (is_array($location))\n    {\n        $this-&gt;cache_class = 'CacheMC';\n    }\n    else\n    {\n        \/\/ I would expect locations like `\/tmp\/cache`, `pdo.mysql:\/\/user:pass@hostname:port`, `pdo.sqlite:memory:`, and `apc`.\n        $type = strtolower(substr($location, 0, 3));\n        switch ($type)\n        {\n            case 'apc':\n                $this-&gt;cache_class = 'CacheAPC';\n                break;\n\n            case 'xca': \/\/ First three letters of `xcache`\n                $this-&gt;cache_class = 'CacheXCache';\n                break;\n\n            case 'pdo':\n                $this-&gt;cache_class = 'CachePDO';\n                $location = substr($location, 4);\n                break;\n\n            default:\n                $this-&gt;cache_class = 'CacheFile';\n                break;\n        }\n    }\n\n    \/\/ Set the remaining cache information.\n    $this-&gt;cache_location = $location;\n    $this-&gt;cache_compress = $gzip;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Fire a single request, and then cache the response to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Fire a single request, and then cache the response to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"}]}},{"name":"set_desired_capacity","modifier":"public","description":"<p>Adjusts the desired size of the <code>AutoScalingGroup<\/code> by initiating scaling activities.\nWhen reducing the size of the group, it is not possible to define which Amazon EC2 instances\nwill be terminated. This applies to any Auto Scaling decisions that might result in terminating\ninstances.<\/p>\n\n<p>There are two common use cases for <code>SetDesiredCapacity<\/code>: one for users of the Auto\nScaling triggering system, and another for developers who write their own triggering systems.\nBoth use cases relate to the concept of cooldown.<\/p>\n\n<p>In the first case, if you use the Auto Scaling triggering system,\n<code>SetDesiredCapacity<\/code> changes the size of your Auto Scaling group without regard to\nthe cooldown period. This could be useful, for example, if Auto Scaling did something\nunexpected for some reason. If your cooldown period is 10 minutes, Auto Scaling would normally\nreject requests to change the size of the group for that entire 10-minute period. The\n<code>SetDesiredCapacity<\/code> command allows you to circumvent this restriction and change\nthe size of the group before the end of the cooldown period.<\/p>\n\n<p>In the second case, if you write your own triggering system, you can use\n<code>SetDesiredCapacity<\/code> to control the size of your Auto Scaling group. If you want the\nsame cooldown functionality that Auto Scaling offers, you can configure\n<code>SetDesiredCapacity<\/code> to honor cooldown by setting the <code>HonorCooldown<\/code>\nparameter to <code>true<\/code>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"desired_capacity","type":"integer","description":"<p>The new capacity setting for the Auto Scaling group.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>HonorCooldown<\/code> - <code>boolean<\/code> - Optional - By default, <code>SetDesiredCapacity<\/code> overrides any cooldown period. Set to <code>True<\/code> if you want Auto Scaling to reject this request when the Auto Scaling group is in cooldown.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function set_desired_capacity($auto_scaling_group_name, $desired_capacity, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    $opt['DesiredCapacity'] = $desired_capacity;\n    \n    return $this-&gt;authenticate('SetDesiredCapacity', $opt);\n}","examples":{"example":{"title":"<p>Set the desired capacity for an autoscaling group.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;set_desired_capacity('my-scaling-group', 0, array(\n\t'HonorCooldown' =&gt; 'false'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":"<a href=\"..\/amazonas\/set_instance_health.html\" rel=\"..\/..\/index.html#m=amazonas\/set_instance_health\"><code>set_instance_health()<\/code><\/a>"}},{"name":"set_hostname","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set the hostname to connect to. This is useful for alternate services that are API-compatible with\nAWS, but run from a different hostname.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"hostname","type":"string","description":"<p>The alternate hostname to use in place of the default one. Useful for mock or test applications living on different hostnames.<\/p>"},{"@attributes":{"choice":"opt"},"name":"port_number","type":"integer","description":"<p>The alternate port number to use in place of the default one. Useful for mock or test applications living on different port numbers.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_hostname($hostname, $port_number = null)\n{\n    if ($this-&gt;override_hostname)\n    {\n        $this-&gt;hostname = $hostname;\n\n        if ($port_number)\n        {\n            $this-&gt;port_number = $port_number;\n            $this-&gt;hostname .= ':' . (string) $this-&gt;port_number;\n        }\n    }\n\n    return $this;\n}","examples":{"example":{"title":"<p>Configure hostname and resource prefix.<\/p>","code":"\/\/ Instantiate the class\n$ec2 = new AmazonEC2();\n\n\/\/ Point to the new endpoint\n$ec2-&gt;set_hostname('localhost', '3000');\n$ec2-&gt;set_resource_prefix('\/services\/mock');\n$ec2-&gt;allow_hostname_override(false);\n\n\/\/ Describe images\n$response = $ec2-&gt;describe_images();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"set_instance_health","modifier":"public","description":"<p>Sets the health status of an instance.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"instance_id","type":"string","description":"<p>The identifier of the Amazon EC2 instance. [Constraints: The value must be between 1 and 16 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"health_status","type":"string","description":"<p>The health status of the instance. &#8220;Healthy&#8221; means that the instance is healthy and should remain in service. &#8220;Unhealthy&#8221; means that the instance is unhealthy. Auto Scaling should terminate and replace it. [Constraints: The value must be between 1 and 32 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>ShouldRespectGracePeriod<\/code> - <code>boolean<\/code> - Optional - If True, this call should respect the grace period associated with the group.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function set_instance_health($instance_id, $health_status, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['InstanceId'] = $instance_id;\n    $opt['HealthStatus'] = $health_status;\n    \n    return $this-&gt;authenticate('SetInstanceHealth', $opt);\n}","examples":{"example":{"title":"<p>Set the instance health for an autoscaling group.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;set_instance_health('i-be038dd3', 'Unhealthy', array(\n\t'ShouldRespectGracePeriod' =&gt; 'false'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":"<a href=\"..\/amazonas\/set_desired_capacity.html\" rel=\"..\/..\/index.html#m=amazonas\/set_desired_capacity\"><code>set_desired_capacity()<\/code><\/a>"}},{"name":"set_max_retries","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Sets the maximum number of times to retry failed requests.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"retries","type":"integer","description":"<p>The maximum number of times to retry failed requests. Defaults to <code>3<\/code>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_max_retries($retries = 3)\n{\n    $this-&gt;max_retries = $retries;\n    return $this;\n}","examples":{"example":{"title":"<p>Sets the maximum number of times to retry failed requests.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Adjust offset\n$sdb = new AmazonSDB();\n$sdb-&gt;set_max_retries(3);\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;max_retries);","result":"int(3)"}}},{"name":"set_parser_class","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfsimplexml\/index.html\" rel=\"..\/..\/index.html#i=cfsimplexml\"><code>CFSimpleXML<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_parser_class($class = 'CFSimpleXML')\n{\n    $this-&gt;parser_class = $class;\n    return $this;\n}"},{"name":"set_proxy","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set the proxy settings to use.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"proxy","type":"string","description":"<p>Accepts proxy credentials in the following format: <code>proxy:\/\/user:pass@hostname:port<\/code><\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_proxy($proxy)\n{\n    $this-&gt;proxy = $proxy;\n    return $this;\n}","examples":{"example":{"title":"<p>Set the proxy server to use for requests.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Adjust offset\n$sdb = new AmazonSDB();\n$sdb-&gt;set_proxy('proxy:\/\/user:pass@hostname:port');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;proxy);","result":"string(31) \"proxy:\/\/user:pass@hostname:port\""}}},{"name":"set_region","modifier":"public","description":"<p>This allows you to explicitly sets the region for the service to use.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"region","type":"string","description":"<p>The region to explicitly set. Available options are <a href=\"..\/amazonas\/constants.html#region_us_e1\" rel=\"..\/..\/index.html#c=amazonas\/region_us_e1\"><code>REGION_US_E1<\/code><\/a>, <a href=\"..\/amazonas\/constants.html#region_us_w1\" rel=\"..\/..\/index.html#c=amazonas\/region_us_w1\"><code>REGION_US_W1<\/code><\/a>, <a href=\"..\/amazonas\/constants.html#region_us_w2\" rel=\"..\/..\/index.html#c=amazonas\/region_us_w2\"><code>REGION_US_W2<\/code><\/a>, <a href=\"..\/amazonas\/constants.html#region_eu_w1\" rel=\"..\/..\/index.html#c=amazonas\/region_eu_w1\"><code>REGION_EU_W1<\/code><\/a>, <a href=\"..\/amazonas\/constants.html#region_apac_se1\" rel=\"..\/..\/index.html#c=amazonas\/region_apac_se1\"><code>REGION_APAC_SE1<\/code><\/a>, <a href=\"..\/amazonas\/constants.html#region_apac_se2\" rel=\"..\/..\/index.html#c=amazonas\/region_apac_se2\"><code>REGION_APAC_SE2<\/code><\/a>, <a href=\"..\/amazonas\/constants.html#region_apac_ne1\" rel=\"..\/..\/index.html#c=amazonas\/region_apac_ne1\"><code>REGION_APAC_NE1<\/code><\/a>, <a href=\"..\/amazonas\/constants.html#region_sa_e1\" rel=\"..\/..\/index.html#c=amazonas\/region_sa_e1\"><code>REGION_SA_E1<\/code><\/a>, <a href=\"..\/amazonas\/constants.html#region_us_gov1\" rel=\"..\/..\/index.html#c=amazonas\/region_us_gov1\"><code>REGION_US_GOV1<\/code><\/a>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_region($region)\n{\n    \/\/ @codeCoverageIgnoreStart\n    $this-&gt;set_hostname($region);\n    return $this;\n    \/\/ @codeCoverageIgnoreEnd\n}","examples":{"example":{"title":"<p>Set the region to use.<\/p>","code":"$as = new AmazonAS();\n\n\/\/ Set the region to US-East (Northern Virginia)\n$as-&gt;set_region(AmazonAS::REGION_US_E1);\n\n\/\/ Describe database engine versions\n$response = $as-&gt;describe_auto_scaling_groups();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"set_request_class","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfrequest\/index.html\" rel=\"..\/..\/index.html#i=cfrequest\"><code>CFRequest<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":{},"source":"public function set_request_class($class = 'CFRequest')\n{\n    $this-&gt;request_class = $class;\n    return $this;\n}","examples":{"example":{"title":"<p>Set the request class to use in the toolkit.<\/p>","description":"<p>The new request class must extend from <code>CFRequest<\/code>.<\/p>\n\n<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Custom class to extend RequestCore\nclass TestRequestCore extends RequestCore\n{\n\tpublic function test_method()\n\t{\n\t\treturn true;\n\t}\n}\n\n\/\/ Instantiate class and set new class\n$sdb = new AmazonSDB();\n$sdb-&gt;set_request_class('TestRequestCore');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;request_class);","result":"string(15) \"TestRequestCore\""}}},{"name":"set_resource_prefix","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set the resource prefix to use. This method is useful for alternate services that are API-compatible\nwith AWS.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"prefix","type":"string","description":"<p>An alternate prefix to prepend to the resource path. Useful for mock or test applications.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_resource_prefix($prefix)\n{\n    $this-&gt;resource_prefix = $prefix;\n    return $this;\n}","examples":{"example":{"title":"<p>Configure hostname and resource prefix.<\/p>","code":"\/\/ Instantiate the class\n$ec2 = new AmazonEC2();\n\n\/\/ Point to the new endpoint\n$ec2-&gt;set_hostname('localhost', '3000');\n$ec2-&gt;set_resource_prefix('\/services\/mock');\n$ec2-&gt;allow_hostname_override(false);\n\n\/\/ Describe images\n$response = $ec2-&gt;describe_images();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"set_response_class","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_response_class($class = 'CFResponse')\n{\n    $this-&gt;response_class = $class;\n    return $this;\n}","examples":{"example":{"title":"<p>Set the response class to use in the toolkit.<\/p>","description":"<p>The new response class must extend from <code>CFResponse<\/code>.<\/p>\n\n<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Custom class to extend ResponseCore\nclass TestResponseCore extends ResponseCore\n{\n\tpublic function test_method()\n\t{\n\t\treturn true;\n\t}\n}\n\n\/\/ Instantiate class and set new class\n$sdb = new AmazonSDB();\n$sdb-&gt;set_response_class('TestResponseCore');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;response_class);","result":"string(16) \"TestResponseCore\""}}},{"name":"set_utilities_class","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfutilities\/index.html\" rel=\"..\/..\/index.html#i=cfutilities\"><code>CFUtilities<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_utilities_class($class = 'CFUtilities')\n{\n    $this-&gt;utilities_class = $class;\n    $this-&gt;util = new $this-&gt;utilities_class();\n    return $this;\n}","examples":{"example":{"title":"<p>Set the utilities class to use in the toolkit.<\/p>","description":"<p>The new utilities class must extend from <code>CFUtilities<\/code>.<\/p>\n\n<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Custom class to extend CFUtilities\nclass TestUtilities extends CFUtilities\n{\n\tpublic function test_method()\n\t{\n\t\treturn true;\n\t}\n}\n\n\/\/ Instantiate class and set new class\n$sdb = new AmazonSDB();\n$sdb-&gt;set_utilities_class('TestUtilities');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;util-&gt;test_method());","result":"bool(true)"}}},{"name":"suspend_processes","modifier":"public","description":"<p>Suspends Auto Scaling processes for an Auto Scaling group. To suspend specific process types,\nspecify them by name with the <code>ScalingProcesses.member.N<\/code> parameter. To suspend all\nprocess types, omit the <code>ScalingProcesses.member.N<\/code> parameter.<\/p>\n\n<p class=\"important\"><\/p>\n\n<p>Suspending either of the two primary process types, <code>Launch<\/code> or\n<code>Terminate<\/code>, can prevent other process types from functioning properly. For more\ninformation about processes and their dependencies, see <code>ProcessType<\/code>.<\/p>\n\n<p>To resume processes that have been suspended, use <code>ResumeProcesses<\/code>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name or Amazon Resource Name (ARN) of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>ScalingProcesses<\/code> - <code>string|array<\/code> - Optional - The processes that you want to suspend or resume, which can include one or more of the following:<ul><li>Launch<\/li><li>Terminate<\/li><li>HealthCheck<\/li><li>ReplaceUnhealthy<\/li><li>AZRebalance<\/li><li>AlarmNotification<\/li><li>ScheduledActions<\/li><li>AddToLoadBalancer<\/li><\/ul>To suspend all process types, omit this parameter. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function suspend_processes($auto_scaling_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['ScalingProcesses']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'ScalingProcesses' =&gt; (is_array($opt['ScalingProcesses']) ? $opt['ScalingProcesses'] : array($opt['ScalingProcesses']))\n        ), 'member'));\n        unset($opt['ScalingProcesses']);\n    }\n\n    return $this-&gt;authenticate('SuspendProcesses', $opt);\n}","examples":{"example":{"title":"<p>Suspend autoscaling processes.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;suspend_processes('my-scaling-group', array(\n\t'ScalingProcesses' =&gt; array(\n\t\t'ReplaceUnhealthy',\n\t\t'HealthCheck',\n\t\t'ScheduledActions'\n\t)\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":"<a href=\"..\/amazonas\/resume_processes.html\" rel=\"..\/..\/index.html#m=amazonas\/resume_processes\"><code>resume_processes()<\/code><\/a>"}},{"name":"terminate_instance_in_auto_scaling_group","modifier":"public","description":"<p>Terminates the specified instance. Optionally, the desired group size can be adjusted.<\/p>\n\n<p class=\"note\">\nThis call simply registers a termination request. The termination of the instance cannot happen\nimmediately.\n<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"instance_id","type":"string","description":"<p>The ID of the Amazon EC2 instance to be terminated. [Constraints: The value must be between 1 and 16 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"req"},"name":"should_decrement_desired_capacity","type":"boolean","description":"<p>Specifies whether (<em>true<\/em>) or not (<em>false<\/em>) terminating this instance should also decrement the size of the <code>AutoScalingGroup<\/code>.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function terminate_instance_in_auto_scaling_group($instance_id, $should_decrement_desired_capacity, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['InstanceId'] = $instance_id;\n    $opt['ShouldDecrementDesiredCapacity'] = $should_decrement_desired_capacity;\n    \n    return $this-&gt;authenticate('TerminateInstanceInAutoScalingGroup', $opt);\n}","examples":{"example":{"title":"<p>Terminates the specified instance. Optionally, the desired group size can be adjusted.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;terminate_instance_in_auto_scaling_group('i-be038dd3', 'false');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/create_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/create_auto_scaling_group\"><code>create_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/delete_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_auto_scaling_group\"><code>delete_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/describe_auto_scaling_groups.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_auto_scaling_groups\"><code>describe_auto_scaling_groups()<\/code><\/a>","<a href=\"..\/amazonas\/describe_auto_scaling_instances.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_auto_scaling_instances\"><code>describe_auto_scaling_instances()<\/code><\/a>","<a href=\"..\/amazonas\/update_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/update_auto_scaling_group\"><code>update_auto_scaling_group()<\/code><\/a>"]}},{"name":"update_auto_scaling_group","modifier":"public","description":"<p>Updates the configuration for the specified <code>AutoScalingGroup<\/code>.<\/p>\n\n<p class=\"note\"><\/p>\n\n<p>To update an Auto Scaling group with a launch configuration that has the\n<code>InstanceMonitoring<\/code> flag set to <code>False<\/code>, you must first ensure that\ncollection of group metrics is disabled. Otherwise, calls to\n<code>UpdateAutoScalingGroup<\/code> will fail. If you have previously enabled group metrics\ncollection, you can disable collection of all group metrics by calling\n<code>DisableMetricsCollection<\/code>.<\/p>\n\n<p>The new settings are registered upon the completion of this call. Any launch configuration\nsettings take effect on any triggers after this call returns. Triggers that are currently in\nprogress aren&#8217;t affected.<\/p>\n\n<p class=\"note\"><\/p>\n\n<ul><li>If a new value is specified for <em>MinSize<\/em> without specifying the value for\n<em>DesiredCapacity<\/em>, and if the new <em>MinSize<\/em> is larger than the current\nsize of the Auto Scaling Group, there will be an implicit call to\n<code>SetDesiredCapacity<\/code> to set the group to the new <em>MinSize<\/em>.<\/li><li>If a new value is specified for <em>MaxSize<\/em> without specifying the value for\n<em>DesiredCapacity<\/em>, and the new <em>MaxSize<\/em> is smaller than the current size\nof the Auto Scaling Group, there will be an implicit call to\n<code>SetDesiredCapacity<\/code> to set the group to the new <em>MaxSize<\/em>.<\/li><li>All other optional parameters are left unchanged if not passed in the request.<\/li><\/ul>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auto_scaling_group_name","type":"string","description":"<p>The name of the Auto Scaling group. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>LaunchConfigurationName<\/code> - <code>string<\/code> - Optional - The name of the launch configuration. [Constraints: The value must be between 1 and 1600 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>MinSize<\/code> - <code>integer<\/code> - Optional - The minimum size of the Auto Scaling group.<\/li><li><code>MaxSize<\/code> - <code>integer<\/code> - Optional - The maximum size of the Auto Scaling group.<\/li><li><code>DesiredCapacity<\/code> - <code>integer<\/code> - Optional - The desired capacity for the Auto Scaling group.<\/li><li><code>DefaultCooldown<\/code> - <code>integer<\/code> - Optional - The amount of time, in seconds, after a scaling activity completes before any further trigger-related scaling activities can start.<\/li><li><code>AvailabilityZones<\/code> - <code>string|array<\/code> - Optional - Availability Zones for the group. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>HealthCheckType<\/code> - <code>string<\/code> - Optional - The service of interest for the health status check, either &#8220;EC2&#8221; for Amazon EC2 or &#8220;ELB&#8221; for Elastic Load Balancing. [Constraints: The value must be between 1 and 32 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>HealthCheckGracePeriod<\/code> - <code>integer<\/code> - Optional - The length of time that Auto Scaling waits before checking an instance&#8217;s health status. The grace period begins when an instance comes into service.<\/li><li><code>PlacementGroup<\/code> - <code>string<\/code> - Optional - The name of the cluster placement group, if applicable. For more information, go to <a href=\"http:\/\/docs.amazonwebservices.com\/AWSEC2\/latest\/UserGuide\/using_cluster_computing.html\">Using Cluster Instances<\/a> in the Amazon EC2 User Guide. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>VPCZoneIdentifier<\/code> - <code>string<\/code> - Optional - The subnet identifier for the Amazon VPC connection, if applicable. You can specify several subnets in a comma-separated list. When you specify <code>VPCZoneIdentifier<\/code> with <code>AvailabilityZones<\/code>, ensure that the subnets&#8217; Availability Zones match the values you specify for <code>AvailabilityZones<\/code>. [Constraints: The value must be between 1 and 255 characters, and must match the following regular expression pattern: <code>[\\u0020-\\uD7FF\\uE000-\\uFFFD\\uD800\\uDC00-\\uDBFF\\uDFFF\\r\\n\\t]*<\/code>]<\/li><li><code>TerminationPolicies<\/code> - <code>string|array<\/code> - Optional - A standalone termination policy or a list of termination policies used to select the instance to terminate. The policies are executed in the order that they are listed. For more information on creating a termination policy for your Auto Scaling group, go to <a href=\"http:\/\/docs.amazonwebservices.com\/AutoScaling\/latest\/DeveloperGuide\/us-termination-policy.html\">Instance Termination Policy for Your Auto Scaling Group<\/a> in the the <em>Auto Scaling Developer Guide<\/em>. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function update_auto_scaling_group($auto_scaling_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['AutoScalingGroupName'] = $auto_scaling_group_name;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['AvailabilityZones']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'AvailabilityZones' =&gt; (is_array($opt['AvailabilityZones']) ? $opt['AvailabilityZones'] : array($opt['AvailabilityZones']))\n        ), 'member'));\n        unset($opt['AvailabilityZones']);\n    }\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['TerminationPolicies']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'TerminationPolicies' =&gt; (is_array($opt['TerminationPolicies']) ? $opt['TerminationPolicies'] : array($opt['TerminationPolicies']))\n        ), 'member'));\n        unset($opt['TerminationPolicies']);\n    }\n\n    return $this-&gt;authenticate('UpdateAutoScalingGroup', $opt);\n}","examples":{"example":{"title":"<p>Update an auto scaling group.<\/p>","code":"\/\/ Instantiate the class\n$as = new AmazonAS();\n\n$response = $as-&gt;update_auto_scaling_group('my-scaling-group', array(\n\t'LaunchConfigurationName' =&gt; 'my-launch-config',\n\t'MinSize' =&gt; 0,\n\t'MaxSize' =&gt; 2,\n\t'DesiredCapacity' =&gt; 2, \/\/ instances\n\t'DefaultCooldown' =&gt; 30, \/\/ seconds\n\t'HealthCheckGracePeriod' =&gt; 30, \/\/ seconds\n\t'AvailabilityZones' =&gt; array(\n\t\t'us-east-1a',\n\t\t'us-east-1b',\n\t\t'us-east-1c',\n\t\t'us-east-1d'\n\t),\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonas\/create_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/create_auto_scaling_group\"><code>create_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/delete_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/delete_auto_scaling_group\"><code>delete_auto_scaling_group()<\/code><\/a>","<a href=\"..\/amazonas\/describe_auto_scaling_groups.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_auto_scaling_groups\"><code>describe_auto_scaling_groups()<\/code><\/a>","<a href=\"..\/amazonas\/describe_auto_scaling_instances.html\" rel=\"..\/..\/index.html#m=amazonas\/describe_auto_scaling_instances\"><code>describe_auto_scaling_instances()<\/code><\/a>","<a href=\"..\/amazonas\/terminate_instance_in_auto_scaling_group.html\" rel=\"..\/..\/index.html#m=amazonas\/terminate_instance_in_auto_scaling_group\"><code>terminate_instance_in_auto_scaling_group()<\/code><\/a>"]}}]},"constants":{"@attributes":{"count":"22"},"constant":[{"name":"BUILD","value":"20130314130000"},{"name":"DEFAULT_URL","value":"autoscaling.us-east-1.amazonaws.com"},{"name":"NAME","value":"aws-sdk-php"},{"name":"REGION_APAC_NE1","value":"autoscaling.ap-northeast-1.amazonaws.com"},{"name":"REGION_APAC_SE1","value":"autoscaling.ap-southeast-1.amazonaws.com"},{"name":"REGION_APAC_SE2","value":"autoscaling.ap-southeast-2.amazonaws.com"},{"name":"REGION_CALIFORNIA","value":"autoscaling.us-west-1.amazonaws.com"},{"name":"REGION_EU_W1","value":"autoscaling.eu-west-1.amazonaws.com"},{"name":"REGION_IRELAND","value":"autoscaling.eu-west-1.amazonaws.com"},{"name":"REGION_OREGON","value":"autoscaling.us-west-2.amazonaws.com"},{"name":"REGION_SAO_PAULO","value":"autoscaling.sa-east-1.amazonaws.com"},{"name":"REGION_SA_E1","value":"autoscaling.sa-east-1.amazonaws.com"},{"name":"REGION_SINGAPORE","value":"autoscaling.ap-southeast-1.amazonaws.com"},{"name":"REGION_SYDNEY","value":"autoscaling.ap-southeast-2.amazonaws.com"},{"name":"REGION_TOKYO","value":"autoscaling.ap-northeast-1.amazonaws.com"},{"name":"REGION_US_E1","value":"autoscaling.us-east-1.amazonaws.com"},{"name":"REGION_US_GOV1","value":"autoscaling.us-gov-west-1.amazonaws.com"},{"name":"REGION_US_W1","value":"autoscaling.us-west-1.amazonaws.com"},{"name":"REGION_US_W2","value":"autoscaling.us-west-2.amazonaws.com"},{"name":"REGION_VIRGINIA","value":"autoscaling.us-east-1.amazonaws.com"},{"name":"USERAGENT","value":"aws-sdk-php\/1.6.2 PHP\/5.4.9 curl\/7.15.5 openssl\/0.9.8e-fips-rhel5"},{"name":"VERSION","value":"1.6.2"}]},"properties":{"@attributes":{"count":"40"},"property":[{"name":"api_version","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The supported API version.<\/p>"},{"name":"auth_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The authentication class to use.<\/p>"},{"name":"auth_token","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The Amazon Authentication Token.<\/p>"},{"name":"batch_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFBatchRequest"}},"description":"<p>The default class to use for handling batch requests (defaults to <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a>).<\/p>"},{"name":"batch_object","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The current instantiated batch request object.<\/p>"},{"name":"cache_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The caching class to use.<\/p>"},{"name":"cache_compress","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The state of cache compression.<\/p>"},{"name":"cache_expires","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>When the cache should be considered stale.<\/p>"},{"name":"cache_location","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The caching location to use.<\/p>"},{"name":"cache_object","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The current instantiated cache object.<\/p>"},{"name":"credentials","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The credentials to use for authentication.<\/p>"},{"name":"debug_mode","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The state of the debug mode setting.<\/p>"},{"name":"delete_cache","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The state of the cache deletion setting.<\/p>"},{"name":"hostname","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The alternate hostname to use, if any.<\/p>"},{"name":"internal_batch_object","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The internally instantiated batch request object.<\/p>"},{"name":"key","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The Amazon API Key.<\/p>"},{"name":"max_retries","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"integer","value":"3"}},"description":"<p>The number of times to retry failed requests.<\/p>"},{"name":"operation","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The operation to execute.<\/p>"},{"name":"operation_prefix","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The string prefix to prepend to the operation name.<\/p>"},{"name":"override_hostname","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of the capability to override the hostname with <a href=\"..\/amazonas\/set_hostname.html\" rel=\"..\/..\/index.html#m=amazonas\/set_hostname\"><code>set_hostname()<\/code><\/a>.<\/p>"},{"name":"parse_the_response","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of whether the response should be parsed or not.<\/p>"},{"name":"parser_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFSimpleXML"}},"description":"<p>The default class to use for parsing XML (defaults to <a href=\"..\/cfsimplexml\/index.html\" rel=\"..\/..\/index.html#i=cfsimplexml\"><code>CFSimpleXML<\/code><\/a>).<\/p>"},{"name":"payload","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The payload to send.<\/p>"},{"name":"port_number","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The alternate port number to use, if any.<\/p>"},{"name":"proxy","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The proxy to use for connecting.<\/p>"},{"name":"redirects","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The number of times a request has been retried.<\/p>"},{"name":"registered_streaming_read_callback","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The user-defined callback function to call when a stream is read from.<\/p>"},{"name":"registered_streaming_write_callback","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The user-defined callback function to call when a stream is written to.<\/p>"},{"name":"request_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFRequest"}},"description":"<p>The default class to use for HTTP requests (defaults to <a href=\"..\/cfrequest\/index.html\" rel=\"..\/..\/index.html#i=cfrequest\"><code>CFRequest<\/code><\/a>).<\/p>"},{"name":"resource_prefix","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The alternate resource prefix to use, if any.<\/p>"},{"name":"response_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFResponse"}},"description":"<p>The default class to use for HTTP responses (defaults to <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a>).<\/p>"},{"name":"secret_key","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The Amazon API Secret Key.<\/p>"},{"name":"service","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>An identifier for the current AWS service.<\/p>"},{"name":"ssl_verification","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of SSL certificate verification.<\/p>"},{"name":"use_aws_query","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of whether auth should be handled as AWS Query.<\/p>"},{"name":"use_batch_flow","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The state of batch flow usage.<\/p>"},{"name":"use_cache_flow","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The state of cache flow usage.<\/p>"},{"name":"use_ssl","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of SSL\/HTTPS use.<\/p>"},{"name":"util","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>Handle for the utility functions.<\/p>"},{"name":"utilities_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFUtilities"}},"description":"<p>The default class to use for utilities (defaults to <a href=\"..\/cfutilities\/index.html\" rel=\"..\/..\/index.html#i=cfutilities\"><code>CFUtilities<\/code><\/a>).<\/p>"}]}}}