{"file":{},"class":{"name":"AmazonRDS","file":"services\/rds.class.php","url":"..\/amazonrds\/index.html","description":"<p>Amazon Relational Database Service (Amazon RDS) is a web service that makes it easier to set\nup, operate, and scale a relational database in the cloud. It provides cost-efficient,\nresizable capacity for an industry-standard relational database and manages common database\nadministration tasks, freeing up developers to focus on what makes their applications and\nbusinesses unique.<\/p>\n\n<p>Amazon RDS gives you access to the capabilities of a familiar MySQL or Oracle database server.\nThis means the code, applications, and tools you already use today with your existing MySQL or\nOracle databases work with Amazon RDS without modification. Amazon RDS automatically backs up\nyour database and maintains the database software that powers your DB Instance. Amazon RDS is\nflexible: you can scale your database instance&#8217;s compute resources and storage capacity to meet\nyour application&#8217;s demand. As with all Amazon Web Services, there are no up-front investments,\nand you pay only for the resources you use.<\/p>\n\n<p>This is the <em>Amazon RDS API Reference<\/em>. It contains a comprehensive description of all\nAmazon RDS Query APIs and data types. Note that this API is asynchronous and some actions may\nrequire polling to determine when an action has been applied. See the parameter description to\ndetermine if a change is applied immediately or on the next instance reboot or during the\nmaintenance window. To get started with Amazon RDS, go to the <a href=\n\"http:\/\/docs.amazonwebservices.com\/AmazonRDS\/latest\/GettingStartedGuide\/\">Amazon RDS Getting\nStarted Guide<\/a>. For more information on Amazon RDS concepts and usage scenarios, go to the\n<a href=\"http:\/\/docs.amazonwebservices.com\/AmazonRDS\/latest\/UserGuide\/\">Amazon RDS User\nGuide<\/a>.<\/p>","metadata":{"version":"2013.03.14","license":"See the included NOTICE.md file for complete information.","copyright":"See the included NOTICE.md file for complete information.","link":["http:\/\/aws.amazon.com\/rds\/ Amazon Relational Database Service","http:\/\/aws.amazon.com\/rds\/documentation\/ Amazon Relational Database Service documentation"]},"inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php","url":"..\/cfruntime\/index.html"}},"methods":{"@attributes":{"count":"82"},"method":[{"name":"__construct","modifier":"public","description":"<p>Constructs a new instance of <a href=\"..\/amazonrds\/index.html\" rel=\"..\/..\/index.html#i=amazonrds\"><code>AmazonRDS<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"options","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>certificate_authority<\/code> - <code>boolean<\/code> - Optional - Determines which Cerificate Authority file to use. A value of boolean <code>false<\/code> will use the Certificate Authority file available on the system. A value of boolean <code>true<\/code> will use the Certificate Authority provided by the SDK. Passing a file system path to a Certificate Authority file (chmodded to <code>0755<\/code>) will use that. Leave this set to <code>false<\/code> if you&#8217;re not sure.<\/li><li><code>credentials<\/code> - <code>string<\/code> - Optional - The name of the credential set to use for authentication.<\/li><li><code>default_cache_config<\/code> - <code>string<\/code> - Optional - This option allows a preferred storage type to be configured for long-term caching. This can be changed later using the <a href=\"..\/amazonrds\/set_cache_config.html\" rel=\"..\/..\/index.html#m=amazonrds\/set_cache_config\"><code>set_cache_config()<\/code><\/a> method. Valid values are: <code>apc<\/code>, <code>xcache<\/code>, or a file system path such as <code>.\/cache<\/code> or <code>\/tmp\/cache\/<\/code>.<\/li><li><code>key<\/code> - <code>string<\/code> - Optional - Your AWS key, or a session key. If blank, the default credential set will be used.<\/li><li><code>secret<\/code> - <code>string<\/code> - Optional - Your AWS secret key, or a session secret key. If blank, the default credential set will be used.<\/li><li><code>token<\/code> - <code>string<\/code> - Optional - An AWS session token.<\/li><\/ul>"}},"returnvalue":{"@attributes":{"type":"void"}},"source":"public function __construct(array $options = array())\n{\n    $this-&gt;api_version = '2013-02-12';\n    $this-&gt;hostname = self::DEFAULT_URL;\n    $this-&gt;auth_class = 'AuthV4Query';\n\n    return parent::__construct($options);\n}"},{"name":"add_source_identifier_to_subscription","modifier":"public","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"subscription_name","type":"string","description":{}},{"@attributes":{"choice":"req"},"name":"source_identifier","type":"string","description":{}},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function add_source_identifier_to_subscription($subscription_name, $source_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['SubscriptionName'] = $subscription_name;\n    $opt['SourceIdentifier'] = $source_identifier;\n    \n    return $this-&gt;authenticate('AddSourceIdentifierToSubscription', $opt);\n}"},{"name":"add_tags_to_resource","modifier":"public","description":"<p>Adds metadata tags to a DB Instance. These tags can also be used with cost allocation reporting\nto track cost associated with a DB Instance.<\/p>\n\n<p>For an overview on tagging DB Instances, see <a href=\n\"http:\/\/docs.amazonwebservices.com\/AmazonRDS\/latest\/UserGuide\/Overview.Tagging.html\">DB\nInstance Tags.<\/a><\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"resource_name","type":"string","description":"<p>The DB Instance the tags will be added to.<\/p>"},{"@attributes":{"choice":"req"},"name":"tags","type":"array","description":"<p>The tags to be assigned to the DB Instance.<\/p>\n\n<ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>Key<\/code> - <code>string<\/code> - Optional - A key is the required name of the tag. The string value can be from 1 to 128 Unicode characters in length and cannot be prefixed with &#8220;aws:&#8221;. The string may only contain only the set of Unicode letters, digits, white-space, &#8216;_&#8217;, &#8216;.&#8217;, &#8216;\/&#8217;, &#8216;=&#8217;, &#8216;+&#8217;, &#8216;-&#8217; (Java regex: &#8220;^([&#92;p{L}&#92;p{Z}&#92;p{N}_.:\/=+&#92;-]*)$&#8221;).<\/li><li><code>Value<\/code> - <code>string<\/code> - Optional - A value is the optional value of the tag. The string value can be from 1 to 256 Unicode characters in length and cannot be prefixed with &#8220;aws:&#8221;. The string may only contain only the set of Unicode letters, digits, white-space, &#8216;_&#8217;, &#8216;.&#8217;, &#8216;\/&#8217;, &#8216;=&#8217;, &#8216;+&#8217;, &#8216;-&#8217; (Java regex: &#8220;^([&#92;p{L}&#92;p{Z}&#92;p{N}_.:\/=+&#92;-]*)$&#8221;).<\/li><\/ul><\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function add_tags_to_resource($resource_name, $tags, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['ResourceName'] = $resource_name;\n    \n    \/\/ Required list + map\n    $opt = array_merge($opt, CFComplexType::map(array(\n        'Tags' =&gt; (is_array($tags) ? $tags : array($tags))\n    ), 'member'));\n\n    return $this-&gt;authenticate('AddTagsToResource', $opt);\n}"},{"name":"allow_hostname_override","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Disables any subsequent use of the <a href=\"..\/amazonrds\/set_hostname.html\" rel=\"..\/..\/index.html#m=amazonrds\/set_hostname\"><code>set_hostname()<\/code><\/a> method.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"override","type":"boolean","description":"<p>Whether or not subsequent calls to <a href=\"..\/amazonrds\/set_hostname.html\" rel=\"..\/..\/index.html#m=amazonrds\/set_hostname\"><code>set_hostname()<\/code><\/a> should be obeyed. A <code>false<\/code> value disables the further effectiveness of <a href=\"..\/amazonrds\/set_hostname.html\" rel=\"..\/..\/index.html#m=amazonrds\/set_hostname\"><code>set_hostname()<\/code><\/a>. Defaults to <code>true<\/code>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function allow_hostname_override($override = true)\n{\n    $this-&gt;override_hostname = $override;\n    return $this;\n}","examples":{"example":{"title":"<p>Configure hostname and resource prefix.<\/p>","code":"\/\/ Instantiate the class\n$ec2 = new AmazonEC2();\n\n\/\/ Point to the new endpoint\n$ec2-&gt;set_hostname('localhost', '3000');\n$ec2-&gt;set_resource_prefix('\/services\/mock');\n$ec2-&gt;allow_hostname_override(false);\n\n\/\/ Describe images\n$response = $ec2-&gt;describe_images();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"authenticate","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Default, shared method for authenticating a connection to AWS.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"operation","type":"string","description":"<p>Indicates the operation to perform.<\/p>"},{"@attributes":{"choice":"req"},"name":"payload","type":"array","description":"<p>An associative array of parameters for authenticating. See the individual methods for allowed keys.<\/p>"}]},"returnvalue":"<p>Object containing a parsed HTTP response.<\/p>","source":"public function authenticate($operation, $payload)\n{\n    $original_payload = $payload;\n    $method_arguments = func_get_args();\n    $curlopts = array();\n    $return_curl_handle = false;\n\n    if (substr($operation, 0, strlen($this-&gt;operation_prefix)) !== $this-&gt;operation_prefix)\n    {\n        $operation = $this-&gt;operation_prefix . $operation;\n    }\n\n    \/\/ Extract the custom CURLOPT settings from the payload\n    if (is_array($payload) &amp;&amp; isset($payload['curlopts']))\n    {\n        $curlopts = $payload['curlopts'];\n        unset($payload['curlopts']);\n    }\n\n    \/\/ Determine whether the response or curl handle should be returned\n    if (is_array($payload) &amp;&amp; isset($payload['returnCurlHandle']))\n    {\n        $return_curl_handle = isset($payload['returnCurlHandle']) ? $payload['returnCurlHandle'] : false;\n        unset($payload['returnCurlHandle']);\n    }\n\n    \/\/ Use the caching flow to determine if we need to do a round-trip to the server.\n    if ($this-&gt;use_cache_flow)\n    {\n        \/\/ Generate an identifier specific to this particular set of arguments.\n        $cache_id = $this-&gt;key . '_' . get_class($this) . '_' . $operation . '_' . sha1(serialize($method_arguments));\n\n        \/\/ Instantiate the appropriate caching object.\n        $this-&gt;cache_object = new $this-&gt;cache_class($cache_id, $this-&gt;cache_location, $this-&gt;cache_expires, $this-&gt;cache_compress);\n\n        if ($this-&gt;delete_cache)\n        {\n            $this-&gt;use_cache_flow = false;\n            $this-&gt;delete_cache = false;\n            return $this-&gt;cache_object-&gt;delete();\n        }\n\n        \/\/ Invoke the cache callback function to determine whether to pull data from the cache or make a fresh request.\n        $data = $this-&gt;cache_object-&gt;response_manager(array($this, 'cache_callback'), $method_arguments);\n\n        \/\/ Parse the XML body\n        $data = $this-&gt;parse_callback($data);\n\n        \/\/ End!\n        return $data;\n    }\n\n    \/*%******************************************************************************************%*\/\n\n    \/\/ Signer\n    $signer = new $this-&gt;auth_class($this-&gt;hostname, $operation, $payload, $this-&gt;credentials);\n    $signer-&gt;key = $this-&gt;key;\n    $signer-&gt;secret_key = $this-&gt;secret_key;\n    $signer-&gt;auth_token = $this-&gt;auth_token;\n    $signer-&gt;api_version = $this-&gt;api_version;\n    $signer-&gt;utilities_class = $this-&gt;utilities_class;\n    $signer-&gt;request_class = $this-&gt;request_class;\n    $signer-&gt;response_class = $this-&gt;response_class;\n    $signer-&gt;use_ssl = $this-&gt;use_ssl;\n    $signer-&gt;proxy = $this-&gt;proxy;\n    $signer-&gt;util = $this-&gt;util;\n    $signer-&gt;registered_streaming_read_callback = $this-&gt;registered_streaming_read_callback;\n    $signer-&gt;registered_streaming_write_callback = $this-&gt;registered_streaming_write_callback;\n    $request = $signer-&gt;authenticate();\n\n    \/\/ Update RequestCore settings\n    $request-&gt;request_class = $this-&gt;request_class;\n    $request-&gt;response_class = $this-&gt;response_class;\n    $request-&gt;ssl_verification = $this-&gt;ssl_verification;\n\n    \/*%******************************************************************************************%*\/\n\n    \/\/ Debug mode\n    if ($this-&gt;debug_mode)\n    {\n        $request-&gt;debug_mode = $this-&gt;debug_mode;\n    }\n\n    \/\/ Set custom CURLOPT settings\n    if (count($curlopts))\n    {\n        $request-&gt;set_curlopts($curlopts);\n    }\n\n    \/\/ Manage the (newer) batch request API or the (older) returnCurlHandle setting.\n    if ($this-&gt;use_batch_flow)\n    {\n        $handle = $request-&gt;prep_request();\n        $this-&gt;batch_object-&gt;add($handle);\n        $this-&gt;use_batch_flow = false;\n\n        return $handle;\n    }\n    elseif ($return_curl_handle)\n    {\n        return $request-&gt;prep_request();\n    }\n\n    \/\/ Send!\n    $request-&gt;send_request();\n\n    \/\/ Prepare the response.\n    $headers = $request-&gt;get_response_header();\n    $headers['x-aws-stringtosign'] = $signer-&gt;string_to_sign;\n\n    if (isset($signer-&gt;canonical_request))\n    {\n        $headers['x-aws-canonicalrequest'] = $signer-&gt;canonical_request;\n    }\n\n    $headers['x-aws-request-headers'] = $request-&gt;request_headers;\n    $headers['x-aws-body'] = $signer-&gt;querystring;\n\n    $data = new $this-&gt;response_class($headers, ($this-&gt;parse_the_response === true) ? $this-&gt;parse_callback($request-&gt;get_response_body()) : $request-&gt;get_response_body(), $request-&gt;get_response_code());\n\n    $response_body = (string) $request-&gt;get_response_body();\n\n    \/\/ Was it Amazon's fault the request failed? Retry the request until we reach $max_retries.\n    if (\n        (integer) $request-&gt;get_response_code() === 500 || \/\/ Internal Error (presumably transient)\n        (integer) $request-&gt;get_response_code() === 503)   \/\/ Service Unavailable (presumably transient)\n    {\n        if ($this-&gt;redirects &lt;= $this-&gt;max_retries)\n        {\n            \/\/ Exponential backoff\n            $delay = (integer) (pow(4, $this-&gt;redirects) * 100000);\n            usleep($delay);\n            $this-&gt;redirects++;\n            $data = $this-&gt;authenticate($operation, $original_payload);\n        }\n    }\n\n    \/\/ DynamoDB has additional, custom logic for retrying requests\n    else\n    {\n        \/\/ If the request to DynamoDB was throttled, we need to retry\n        $need_to_retry_dynamodb_request = (\n            (integer) $request-&gt;get_response_code() === 400 &amp;&amp;\n            stripos($response_body, 'com.amazonaws.dynamodb.') !== false &amp;&amp;\n            stripos($response_body, 'ProvisionedThroughputExceededException') !== false\n        );\n\n        \/\/ If the CRC32 of the response does not match the expected value, we need to retry\n        $response_headers = $request-&gt;get_response_header();\n        if (!$need_to_retry_dynamodb_request &amp;&amp; isset($response_headers['x-amz-crc32']))\n        {\n            $crc32_expected = $response_headers['x-amz-crc32'];\n            $crc32_actual = hexdec(hash('crc32b', $response_body));\n            $need_to_retry_dynamodb_request = ($crc32_expected != $crc32_actual);\n        }\n\n        \/\/ Perform retry if necessary using a more aggressive exponential backoff\n        if ($need_to_retry_dynamodb_request)\n        {\n            if ($this-&gt;redirects === 0)\n            {\n                $this-&gt;redirects++;\n                $data = $this-&gt;authenticate($operation, $original_payload);\n            }\n            elseif ($this-&gt;redirects &lt;= max($this-&gt;max_retries, 10))\n            {\n                \/\/ Exponential backoff\n                $delay = (integer) (pow(2, ($this-&gt;redirects - 1)) * 50000);\n                usleep($delay);\n                $this-&gt;redirects++;\n                $data = $this-&gt;authenticate($operation, $original_payload);\n            }\n        }\n    }\n\n    $this-&gt;redirects = 0;\n    return $data;\n}"},{"name":"authorize_db_security_group_ingress","modifier":"public","description":"<p>Enables ingress to a DBSecurityGroup using one of two forms of authorization. First, EC2 or VPC\nSecurity Groups can be added to the DBSecurityGroup if the application using the database is\nrunning on EC2 or VPC instances. Second, IP ranges are available if the application accessing\nyour database is running on the Internet. Required parameters for this API are one of CIDR\nrange, EC2SecurityGroupId for VPC, or (EC2SecurityGroupOwnerId and either EC2SecurityGroupName\nor EC2SecurityGroupId for non-VPC).<\/p>\n\n<p class=\"note\">\nYou cannot authorize ingress from an EC2 security group in one Region to an Amazon RDS DB\nInstance in another. You cannot authorize ingress from a VPC security group in one VPC to an\nAmazon RDS DB Instance in another.\n<\/p>\n\n<p>For an overview of CIDR ranges, go to the <a href=\n\"http:\/\/en.wikipedia.org\/wiki\/Classless_Inter-Domain_Routing\">Wikipedia Tutorial<\/a>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_security_group_name","type":"string","description":"<p>The name of the DB Security Group to add authorization to.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>CIDRIP<\/code> - <code>string<\/code> - Optional - The IP range to authorize.<\/li><li><code>EC2SecurityGroupName<\/code> - <code>string<\/code> - Optional - Name of the EC2 Security Group to authorize. For VPC DB Security Groups, <code>EC2SecurityGroupId<\/code> must be provided. Otherwise, EC2SecurityGroupOwnerId and either <code>EC2SecurityGroupName<\/code> or <code>EC2SecurityGroupId<\/code> must be provided.<\/li><li><code>EC2SecurityGroupId<\/code> - <code>string<\/code> - Optional - Id of the EC2 Security Group to authorize. For VPC DB Security Groups, <code>EC2SecurityGroupId<\/code> must be provided. Otherwise, EC2SecurityGroupOwnerId and either <code>EC2SecurityGroupName<\/code> or <code>EC2SecurityGroupId<\/code> must be provided.<\/li><li><code>EC2SecurityGroupOwnerId<\/code> - <code>string<\/code> - Optional - AWS Account Number of the owner of the EC2 Security Group specified in the EC2SecurityGroupName parameter. The AWS Access Key ID is not an acceptable value. For VPC DB Security Groups, <code>EC2SecurityGroupId<\/code> must be provided. Otherwise, EC2SecurityGroupOwnerId and either <code>EC2SecurityGroupName<\/code> or <code>EC2SecurityGroupId<\/code> must be provided.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function authorize_db_security_group_ingress($db_security_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBSecurityGroupName'] = $db_security_group_name;\n    \n    return $this-&gt;authenticate('AuthorizeDBSecurityGroupIngress', $opt);\n}","examples":{"example":[{"title":"<p>Authorize database security group access for a given security group.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;authorize_db_security_group_ingress('mySecurityGroup', array(\n\t'EC2SecurityGroupName' =&gt; 'default',\n\t'EC2SecurityGroupOwnerId' =&gt; CFCredentials::get('@default')-&gt;account_id\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Authorize database security group access for a CIDR IP Range.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;authorize_db_security_group_ingress('mySecurityGroup', array(\n\t'CIDRIP' =&gt; '0.0.0.0\/0'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}]},"related":{"method":["<a href=\"..\/amazonrds\/create_db_security_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_security_group\"><code>create_db_security_group()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_security_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_security_group\"><code>delete_db_security_group()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_security_groups.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_security_groups\"><code>describe_db_security_groups()<\/code><\/a>","<a href=\"..\/amazonrds\/revoke_db_security_group_ingress.html\" rel=\"..\/..\/index.html#m=amazonrds\/revoke_db_security_group_ingress\"><code>revoke_db_security_group_ingress()<\/code><\/a>"]}},{"name":"batch","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Specifies that the intended request should be queued for a later batch request.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"queue","type":"CFBatchRequest","description":"<p>The <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a> instance to use for managing batch requests. If not available, it generates a new instance of <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function batch(CFBatchRequest &amp;$queue = null)\n{\n    if ($queue)\n    {\n        $this-&gt;batch_object = $queue;\n    }\n    elseif ($this-&gt;internal_batch_object)\n    {\n        $this-&gt;batch_object = &amp;$this-&gt;internal_batch_object;\n    }\n    else\n    {\n        $this-&gt;internal_batch_object = new $this-&gt;batch_class();\n        $this-&gt;batch_object = &amp;$this-&gt;internal_batch_object;\n    }\n\n    $this-&gt;use_batch_flow = true;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Batch several requests together, and then cache the responses to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"},{"title":"<p>Batch several requests together, and then cache the responses to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"}]}},{"name":"cache","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Specifies that the resulting <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object should be cached according to the settings from\n<a href=\"..\/amazonrds\/set_cache_config.html\" rel=\"..\/..\/index.html#m=amazonrds\/set_cache_config\"><code>set_cache_config()<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"expires","type":"string|integer","description":"<p>The time the cache is to expire. Accepts a number of seconds as an integer, or an amount of time, as a string, that is understood by <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a> (e.g. &#8220;1 hour&#8221;).<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function cache($expires)\n{\n    \/\/ Die if they haven't used set_cache_config().\n    if (!$this-&gt;cache_class)\n    {\n        throw new CFRuntime_Exception('Must call set_cache_config() before using cache()');\n    }\n\n    if (is_string($expires))\n    {\n        $expires = strtotime($expires);\n        $this-&gt;cache_expires = $expires - time();\n    }\n    elseif (is_int($expires))\n    {\n        $this-&gt;cache_expires = $expires;\n    }\n\n    $this-&gt;use_cache_flow = true;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Fire a single request, and then cache the response to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Fire a single request, and then cache the response to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Fire a single request, and then cache the response to Memcache.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config(array(\n\tarray('host' =&gt; '127.0.0.1')\n));\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Batch several requests together, and then cache the responses to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"},{"title":"<p>Batch several requests together, and then cache the responses to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"}]}},{"name":"cache_instance_profile_credentials","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Fetches and caches EC2 instance profile credentials. This is meant to be used by the constructor, and is not to\nbe manually invoked.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"cache","type":"CacheCore","description":"<p>The a reference to the cache object that is being used to handle the caching.<\/p>"},{"@attributes":{"choice":"req"},"name":"options","type":"array","description":"<p>The options that were passed into the constructor.<\/p>"}]},"returnvalue":"<p>The data to be cached, or NULL.<\/p>","source":"public function cache_instance_profile_credentials($cache, $options)\n{\n    $instance_profile_url = 'http:\/\/169.254.169.254\/latest\/meta-data\/iam\/security-credentials\/';\n    $connect_timeout = isset($options['instance_profile_timeout']) ? $options['instance_profile_timeout'] : 2;\n\n    try\n    {\n        \/\/ Make a call to the EC2 Metadata Service to find the available instance profile\n        $request = new RequestCore($instance_profile_url);\n        $request-&gt;set_curlopts(array(CURLOPT_CONNECTTIMEOUT =&gt; $connect_timeout));\n        $response = $request-&gt;send_request(true);\n\n        if ($response-&gt;isOK())\n        {\n            \/\/ Get the instance profile name\n            $profile = (string) $response-&gt;body;\n\n            \/\/ Make a call to the EC2 Metadata Service to get the instance profile credentials\n            $request = new RequestCore($instance_profile_url . $profile);\n            $request-&gt;set_curlopts(array(CURLOPT_CONNECTTIMEOUT =&gt; $connect_timeout));\n            $response = $request-&gt;send_request(true);\n\n            if ($response-&gt;isOK())\n            {\n                \/\/ Get the credentials\n                $credentials = json_decode($response-&gt;body, true);\n\n                if ($credentials['Code'] === 'Success')\n                {\n                    \/\/ Determine the expiration time\n                    $expiration_time = strtotime((string) $credentials['Expiration']);\n                    $expiration_duration = round(($expiration_time - time()) * 0.85);\n                    $cache-&gt;expire_in($expiration_duration);\n\n                    \/\/ Return the credential information\n                    return array(\n                        'key'     =&gt; $credentials['AccessKeyId'],\n                        'secret'  =&gt; $credentials['SecretAccessKey'],\n                        'token'   =&gt; $credentials['Token'],\n                        'expires' =&gt; $credentials['Expiration'],\n                    );\n                }\n            }\n        }\n    }\n    catch (cURL_Exception $e)\n    {\n        \/\/ The EC2 Metadata Service does not exist or had timed out.\n        \/\/ An exception will be thrown on the next line.\n    }\n\n    \/\/ @codeCoverageIgnoreStart\n    throw new CFCredentials_Exception('No credentials were provided. The SDK attempted to retrieve Instance '\n        . 'Profile credentials from the EC2 Instance Metadata Service, but failed to do so. Instance profile '\n        . 'credentials are only accessible on EC2 instances configured with a specific IAM role.');\n    \/\/ @codeCoverageIgnoreEnd\n}"},{"name":"cache_sts_credentials","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Fetches and caches STS credentials. This is meant to be used by the constructor, and is not to be\nmanually invoked.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"cache","type":"CacheCore","description":"<p>The a reference to the cache object that is being used to handle the caching.<\/p>"},{"@attributes":{"choice":"req"},"name":"options","type":"array","description":"<p>The options that were passed into the constructor.<\/p>"}]},"returnvalue":"<p>The data to be cached, or NULL.<\/p>","source":"public function cache_sts_credentials($cache, $options)\n{\n    $token = new AmazonSTS($options);\n    $response = $token-&gt;get_session_token();\n\n    if ($response-&gt;isOK())\n    {\n        \/\/ Update the expiration\n        $expiration_time = strtotime((string) $response-&gt;body-&gt;GetSessionTokenResult-&gt;Credentials-&gt;Expiration);\n        $expiration_duration = round(($expiration_time - time()) * 0.85);\n        $cache-&gt;expire_in($expiration_duration);\n\n        \/\/ Return the important data\n        $credentials = $response-&gt;body-&gt;GetSessionTokenResult-&gt;Credentials;\n\n        return array(\n            'key'     =&gt; (string) $credentials-&gt;AccessKeyId,\n            'secret'  =&gt; (string) $credentials-&gt;SecretAccessKey,\n            'token'   =&gt; (string) $credentials-&gt;SessionToken,\n            'expires' =&gt; (string) $credentials-&gt;Expiration,\n        );\n    }\n\n    \/\/ @codeCoverageIgnoreStart\n    throw new STS_Exception('Temporary credentials from the AWS Security '\n        . 'Token Service could not be retrieved using the provided long '\n        . 'term credentials. It\\'s possible that the provided long term '\n        . 'credentials were invalid.');\n    \/\/ @codeCoverageIgnoreEnd\n}"},{"name":"copy_db_snapshot","modifier":"public","description":"<p>Copies the specified DBSnapshot. The source DBSnapshot must be in the &#8220;available&#8221; state.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"source_db_snapshot_identifier","type":"string","description":"<p>The identifier for the source DB snapshot. Constraints:<\/p>\n\n<ul><li>Must be the identifier for a valid system snapshot in the &#8220;available&#8221; state.<\/li><\/ul>\n\n<p>Example: <code>rds:mydb-2012-04-02-00-01<\/code><\/p>"},{"@attributes":{"choice":"req"},"name":"target_db_snapshot_identifier","type":"string","description":"<p>The identifier for the copied snapshot. Constraints:<\/p>\n\n<ul><li>Cannot be null, empty, or blank<\/li><li>Must contain from 1 to 255 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>\n\n<p>Example: <code>my-db-snapshot<\/code><\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function copy_db_snapshot($source_db_snapshot_identifier, $target_db_snapshot_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['SourceDBSnapshotIdentifier'] = $source_db_snapshot_identifier;\n    $opt['TargetDBSnapshotIdentifier'] = $target_db_snapshot_identifier;\n    \n    return $this-&gt;authenticate('CopyDBSnapshot', $opt);\n}"},{"name":"create_db_instance","modifier":"public","description":"<p>Creates a new DB instance.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_instance_identifier","type":"string","description":"<p>The DB Instance identifier. This parameter is stored as a lowercase string. Constraints:<\/p>\n\n<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens (1 to 15 for SQL Server).<\/li><li>First character must be a letter.<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens.<\/li><\/ul>\n\n<p>Example: <code>mydbinstance<\/code><\/p>"},{"@attributes":{"choice":"req"},"name":"allocated_storage","type":"integer","description":"<p>The amount of storage (in gigabytes) to be initially allocated for the database instance. <strong>MySQL<\/strong> Constraints: Must be an integer from 5 to 1024. Type: Integer <strong>Oracle<\/strong> Constraints: Must be an integer from 10 to 1024. <strong>SQL Server<\/strong> Constraints: Must be an integer from 200 to 1024 (Standard Edition and Enterprise Edition) or from 30 to 1024 (Express Edition and Web Edition)<\/p>"},{"@attributes":{"choice":"req"},"name":"db_instance_class","type":"string","description":"<p>The compute and memory capacity of the DB Instance. Valid Values: <code>db.t1.micro | db.m1.small | db.m1.medium | db.m1.large | db.m1.xlarge | db.m2.xlarge |db.m2.2xlarge | db.m2.4xlarge<\/code><\/p>"},{"@attributes":{"choice":"req"},"name":"engine","type":"string","description":"<p>The name of the database engine to be used for this instance. Valid Values: <code>MySQL<\/code> | <code>oracle-se1<\/code> | <code>oracle-se<\/code> | <code>oracle-ee<\/code> | <code>sqlserver-ee<\/code> | <code>sqlserver-se<\/code> | <code>sqlserver-ex<\/code> | <code>sqlserver-web<\/code><\/p>"},{"@attributes":{"choice":"req"},"name":"master_username","type":"string","description":"<p>The name of master user for the client DB Instance. <strong>MySQL<\/strong> Constraints:<\/p>\n\n<ul><li>Must be 1 to 16 alphanumeric characters.<\/li><li>First character must be a letter.<\/li><li>Cannot be a reserved word for the chosen database engine.<\/li><\/ul>\n\n<p>Type: String <strong>Oracle<\/strong> Constraints:<\/p>\n\n<ul><li>Must be 1 to 30 alphanumeric characters.<\/li><li>First character must be a letter.<\/li><li>Cannot be a reserved word for the chosen database engine.<\/li><\/ul>\n\n<p><strong>SQL Server<\/strong> Constraints:<\/p>\n\n<ul><li>Must be 1 to 128 alphanumeric characters.<\/li><li>First character must be a letter.<\/li><li>Cannot be a reserved word for the chosen database engine.<\/li><\/ul>"},{"@attributes":{"choice":"req"},"name":"master_user_password","type":"string","description":"<p>The password for the master database user. Can be any printable ASCII character except &#8220;\/&#8221;, &#8220;&#34;, or &#8220;@&#8221;. Type: String <strong>MySQL<\/strong> Constraints: Must contain from 8 to 41 alphanumeric characters. <strong>Oracle<\/strong> Constraints: Must contain from 8 to 30 alphanumeric characters. <strong>SQL Server<\/strong> Constraints: Must contain from 8 to 128 alphanumeric characters.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBName<\/code> - <code>string<\/code> - Optional - The meaning of this parameter differs according to the database engine you use. <strong>MySQL<\/strong> The name of the database to create when the DB Instance is created. If this parameter is not specified, no database is created in the DB Instance. Constraints:<ul><li>Must contain 1 to 64 alphanumeric characters<\/li><li>Cannot be a word reserved by the specified database engine<\/li><\/ul>Type: String <strong>Oracle<\/strong> The Oracle System ID (SID) of the created DB Instance. Default: <code>ORCL<\/code> Constraints:<ul><li>Cannot be longer than 8 characters<\/li><\/ul> <strong>SQL Server<\/strong> Not applicable. Must be null.<\/li><li><code>DBSecurityGroups<\/code> - <code>string|array<\/code> - Optional - A list of DB Security Groups to associate with this DB Instance. Default: The default DB Security Group for the database engine. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>VpcSecurityGroupIds<\/code> - <code>string|array<\/code> - Optional - A list of EC2 VPC Security Groups to associate with this DB Instance. Default: The default EC2 VPC Security Group for the DB Subnet group&#8217;s VPC. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>AvailabilityZone<\/code> - <code>string<\/code> - Optional - The EC2 Availability Zone that the database instance will be created in. Default: A random, system-chosen Availability Zone in the endpoint&#8217;s region. Example: <code>us-east-1d<\/code> Constraint: The AvailabilityZone parameter cannot be specified if the MultiAZ parameter is set to <code>true<\/code>. The specified Availability Zone must be in the same region as the current endpoint.<\/li><li><code>DBSubnetGroupName<\/code> - <code>string<\/code> - Optional - A DB Subnet Group to associate with this DB Instance. If there is no DB Subnet Group, then it is a non-VPC DB instance.<\/li><li><code>PreferredMaintenanceWindow<\/code> - <code>string<\/code> - Optional - The weekly time range (in UTC) during which system maintenance can occur. Format: <code>ddd:hh24:mi-ddd:hh24:mi<\/code> Default: A 30-minute window selected at random from an 8-hour block of time per region, occurring on a random day of the week. The following list shows the time blocks for each region from which the default maintenance windows are assigned.<ul><li> <strong>US-East (Northern Virginia) Region:<\/strong> 03:00-11:00 UTC<\/li><li> <strong>US-West (Northern California) Region:<\/strong> 06:00-14:00 UTC<\/li><li> <strong>EU (Ireland) Region:<\/strong> 22:00-06:00 UTC<\/li><li> <strong>Asia Pacific (Singapore) Region:<\/strong> 14:00-22:00 UTC<\/li><li> <strong>Asia Pacific (Tokyo) Region:<\/strong> 17:00-03:00 UTC<\/li><\/ul>Valid Days: Mon, Tue, Wed, Thu, Fri, Sat, Sun Constraints: Minimum 30-minute window.<\/li><li><code>DBParameterGroupName<\/code> - <code>string<\/code> - Optional - The name of the DB Parameter Group to associate with this DB instance. If this argument is omitted, the default DBParameterGroup for the specified engine will be used. Constraints:<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul><\/li><li><code>BackupRetentionPeriod<\/code> - <code>integer<\/code> - Optional - The number of days for which automated backups are retained. Setting this parameter to a positive number enables backups. Setting this parameter to 0 disables automated backups. Default: 1 Constraints:<ul><li>Must be a value from 0 to 8<\/li><li>Cannot be set to 0 if the DB Instance is a master instance with read replicas<\/li><\/ul><\/li><li><code>PreferredBackupWindow<\/code> - <code>string<\/code> - Optional - The daily time range during which automated backups are created if automated backups are enabled, using the <code>BackupRetentionPeriod<\/code> parameter. Default: A 30-minute window selected at random from an 8-hour block of time per region. The following list shows the time blocks for each region from which the default backup windows are assigned.<ul><li> <strong>US-East (Northern Virginia) Region:<\/strong> 03:00-11:00 UTC<\/li><li> <strong>US-West (Northern California) Region:<\/strong> 06:00-14:00 UTC<\/li><li> <strong>EU (Ireland) Region:<\/strong> 22:00-06:00 UTC<\/li><li> <strong>Asia Pacific (Singapore) Region:<\/strong> 14:00-22:00 UTC<\/li><li> <strong>Asia Pacific (Tokyo) Region:<\/strong> 17:00-03:00 UTC<\/li><\/ul>Constraints: Must be in the format <code>hh24:mi-hh24:mi<\/code>. Times should be Universal Time Coordinated (UTC). Must not conflict with the preferred maintenance window. Must be at least 30 minutes.<\/li><li><code>Port<\/code> - <code>integer<\/code> - Optional - The port number on which the database accepts connections. <strong>MySQL<\/strong> Default: <code>3306<\/code> Valid Values: <code>1150-65535<\/code> Type: Integer <strong>Oracle<\/strong> Default: <code>1521<\/code> Valid Values: <code>1150-65535<\/code> <strong>SQL Server<\/strong> Default: <code>1433<\/code> Valid Values: <code>1150-65535<\/code> except for <code>1434<\/code> and <code>3389<\/code>.<\/li><li><code>MultiAZ<\/code> - <code>boolean<\/code> - Optional - Specifies if the DB Instance is a Multi-AZ deployment. You cannot set the AvailabilityZone parameter if the MultiAZ parameter is set to true.<\/li><li><code>EngineVersion<\/code> - <code>string<\/code> - Optional - The version number of the database engine to use. <strong>MySQL<\/strong> Example: <code>5.1.42<\/code> Type: String <strong>Oracle<\/strong> Example: <code>11.2.0.2.v2<\/code> Type: String <strong>SQL Server<\/strong> Example: <code>10.50.2789.0.v1<\/code><\/li><li><code>AutoMinorVersionUpgrade<\/code> - <code>boolean<\/code> - Optional - Indicates that minor engine upgrades will be applied automatically to the DB Instance during the maintenance window. Default: <code>true<\/code><\/li><li><code>LicenseModel<\/code> - <code>string<\/code> - Optional - License model information for this DB Instance. Valid values: <code>license-included<\/code> | <code>bring-your-own-license<\/code> | <code>general-public-license<\/code><\/li><li><code>Iops<\/code> - <code>integer<\/code> - Optional - The amount of Provisioned IOPS (input\/output operations per second) to be initially allocated for the DB Instance. Constraints: Must be an integer greater than 1000.<\/li><li><code>OptionGroupName<\/code> - <code>string<\/code> - Optional - Indicates that the DB Instance should be associated with the specified option group.<\/li><li><code>CharacterSetName<\/code> - <code>string<\/code> - Optional - For supported engines, indicates that the DB Instance should be associated with the specified CharacterSet.<\/li><li><code>PubliclyAccessible<\/code> - <code>boolean<\/code> - Optional - <\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_db_instance($db_instance_identifier, $allocated_storage, $db_instance_class, $engine, $master_username, $master_user_password, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBInstanceIdentifier'] = $db_instance_identifier;\n    $opt['AllocatedStorage'] = $allocated_storage;\n    $opt['DBInstanceClass'] = $db_instance_class;\n    $opt['Engine'] = $engine;\n    $opt['MasterUsername'] = $master_username;\n    $opt['MasterUserPassword'] = $master_user_password;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['DBSecurityGroups']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'DBSecurityGroups' =&gt; (is_array($opt['DBSecurityGroups']) ? $opt['DBSecurityGroups'] : array($opt['DBSecurityGroups']))\n        ), 'member'));\n        unset($opt['DBSecurityGroups']);\n    }\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['VpcSecurityGroupIds']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'VpcSecurityGroupIds' =&gt; (is_array($opt['VpcSecurityGroupIds']) ? $opt['VpcSecurityGroupIds'] : array($opt['VpcSecurityGroupIds']))\n        ), 'member'));\n        unset($opt['VpcSecurityGroupIds']);\n    }\n\n    return $this-&gt;authenticate('CreateDBInstance', $opt);\n}","examples":{"example":[{"title":"<p>Create a MySQL database instance.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;create_db_instance('myInstance', 5, 'db.m1.small', 'MySQL', 'dbUser', 'dbPassword', array(\n\t'DBName' =&gt; 'myDatabase',\n\t'DBSecurityGroups' =&gt; 'default',\n\t'PreferredMaintenanceWindow' =&gt; 'Sun:05:00-Sun:09:00',\n\t'BackupRetentionPeriod' =&gt; 1,\n\t'PreferredBackupWindow' =&gt; '03:00-05:00',\n\t'MultiAZ' =&gt; true,\n\t'AutoMinorVersionUpgrade' =&gt; true\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Create an Oracle database instance.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;create_db_instance('myInstance-oracle', 10, 'db.m1.small', 'oracle-se1', 'dbUser', 'dbPassword', array(\n\t'DBName' =&gt; 'oracledb',\n\t'DBSecurityGroups' =&gt; 'default',\n\t'BackupRetentionPeriod' =&gt; 1,\n\t'PreferredBackupWindow' =&gt; '03:00-05:00',\n\t'Port' =&gt; 1521,\n\t'LicenseModel' =&gt; 'bring-your-own-license'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}]},"related":{"method":["<a href=\"..\/amazonrds\/create_db_instance_read_replica.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_instance_read_replica\"><code>create_db_instance_read_replica()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_instance\"><code>delete_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_instances.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_instances\"><code>describe_db_instances()<\/code><\/a>","<a href=\"..\/amazonrds\/modify_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/modify_db_instance\"><code>modify_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/reboot_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/reboot_db_instance\"><code>reboot_db_instance()<\/code><\/a>"]}},{"name":"create_db_instance_read_replica","modifier":"public","description":"<p>Creates a DB Instance that acts as a Read Replica of a source DB Instance.<\/p>\n\n<p>All Read Replica DB Instances are created as Single-AZ deployments with backups disabled. All\nother DB Instance attributes (including DB Security Groups and DB Parameter Groups) are\ninherited from the source DB Instance, except as specified below.<\/p>\n\n<p class=\"important\"><\/p>\n\n<p>The source DB Instance must have backup retention enabled.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_instance_identifier","type":"string","description":"<p>The DB Instance identifier of the Read Replica. This is the unique key that identifies a DB Instance. This parameter is stored as a lowercase string.<\/p>"},{"@attributes":{"choice":"req"},"name":"source_db_instance_identifier","type":"string","description":"<p>The identifier of the DB Instance that will act as the source for the Read Replica. Each DB Instance can have up to five Read Replicas. Constraints: Must be the identifier of an existing DB Instance that is not already a Read Replica DB Instance.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBInstanceClass<\/code> - <code>string<\/code> - Optional - The compute and memory capacity of the Read Replica. Valid Values: <code>db.m1.small | db.m1.medium | db.m1.large | db.m1.xlarge | db.m2.xlarge |db.m2.2xlarge | db.m2.4xlarge<\/code> Default: Inherits from the source DB Instance.<\/li><li><code>AvailabilityZone<\/code> - <code>string<\/code> - Optional - The Amazon EC2 Availability Zone that the Read Replica will be created in. Default: A random, system-chosen Availability Zone in the endpoint&#8217;s region. Example: <code>us-east-1d<\/code><\/li><li><code>Port<\/code> - <code>integer<\/code> - Optional - The port number that the DB Instance uses for connections. Default: Inherits from the source DB Instance Valid Values: <code>1150-65535<\/code><\/li><li><code>AutoMinorVersionUpgrade<\/code> - <code>boolean<\/code> - Optional - Indicates that minor engine upgrades will be applied automatically to the Read Replica during the maintenance window. Default: Inherits from the source DB Instance<\/li><li><code>Iops<\/code> - <code>integer<\/code> - Optional - The amount of Provisioned IOPS (input\/output operations per second) to be initially allocated for the DB Instance.<\/li><li><code>OptionGroupName<\/code> - <code>string<\/code> - Optional - The option group the DB instance will be associated with. If omitted, the default Option Group for the engine specified will be used.<\/li><li><code>PubliclyAccessible<\/code> - <code>boolean<\/code> - Optional - <\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_db_instance_read_replica($db_instance_identifier, $source_db_instance_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBInstanceIdentifier'] = $db_instance_identifier;\n    $opt['SourceDBInstanceIdentifier'] = $source_db_instance_identifier;\n    \n    return $this-&gt;authenticate('CreateDBInstanceReadReplica', $opt);\n}","examples":{"example":{"title":"<p>Create a database read replica.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;create_db_instance_read_replica('myInstancereplica', 'myInstance', array(\n\t'DBInstanceClass' =&gt; 'db.m1.small',\n\t'AutoMinorVersionUpgrade' =&gt; true\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_instance\"><code>create_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_instance\"><code>delete_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_instances.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_instances\"><code>describe_db_instances()<\/code><\/a>","<a href=\"..\/amazonrds\/modify_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/modify_db_instance\"><code>modify_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/reboot_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/reboot_db_instance\"><code>reboot_db_instance()<\/code><\/a>"]}},{"name":"create_db_parameter_group","modifier":"public","description":"<p>Creates a new DB Parameter Group.<\/p>\n\n<p>A DB Parameter Group is initially created with the default parameters for the database engine\nused by the DB Instance. To provide custom values for any of the parameters, you must modify\nthe group after creating it using <em>ModifyDBParameterGroup<\/em>. Once you&#8217;ve created a DB\nParameter Group, you need to associate it with your DB Instance using\n<em>ModifyDBInstance<\/em>. When you associate a new DB Parameter Group with a running DB\nInstance, you need to reboot the DB Instance for the new DB Parameter Group and associated\nsettings to take effect.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_parameter_group_name","type":"string","description":"<p>The name of the DB Parameter Group. Constraints:<\/p>\n\n<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>\n\n<p class=\"note\">This value is stored as a lower-case string.<\/p>"},{"@attributes":{"choice":"req"},"name":"db_parameter_group_family","type":"string","description":"<p>The DB Parameter Group Family name. A DB Parameter Group can be associated with one and only one DB Parameter Group Family, and can be applied only to a DB Instance running a database engine and engine version compatible with that DB Parameter Group Family.<\/p>"},{"@attributes":{"choice":"req"},"name":"description","type":"string","description":"<p>The description for the DB Parameter Group.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_db_parameter_group($db_parameter_group_name, $db_parameter_group_family, $description, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBParameterGroupName'] = $db_parameter_group_name;\n    $opt['DBParameterGroupFamily'] = $db_parameter_group_family;\n    $opt['Description'] = $description;\n    \n    return $this-&gt;authenticate('CreateDBParameterGroup', $opt);\n}","examples":{"example":{"title":"<p>Create a database parameter group.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;create_db_parameter_group('myParameterGroup', 'MySQL5.1', 'This is a sample description.');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/delete_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_parameter_group\"><code>delete_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_parameter_groups.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_parameter_groups\"><code>describe_db_parameter_groups()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_parameters.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_parameters\"><code>describe_db_parameters()<\/code><\/a>","<a href=\"..\/amazonrds\/modify_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/modify_db_parameter_group\"><code>modify_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/reset_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/reset_db_parameter_group\"><code>reset_db_parameter_group()<\/code><\/a>"]}},{"name":"create_db_security_group","modifier":"public","description":"<p>Creates a new DB Security Group. DB Security Groups control access to a DB Instance.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_security_group_name","type":"string","description":"<p>The name for the DB Security Group. This value is stored as a lowercase string. Constraints: Must contain no more than 255 alphanumeric characters or hyphens. Must not be &#8220;Default&#8221;. Example: <code>mysecuritygroup<\/code><\/p>"},{"@attributes":{"choice":"req"},"name":"db_security_group_description","type":"string","description":"<p>The description for the DB Security Group.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_db_security_group($db_security_group_name, $db_security_group_description, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBSecurityGroupName'] = $db_security_group_name;\n    $opt['DBSecurityGroupDescription'] = $db_security_group_description;\n    \n    return $this-&gt;authenticate('CreateDBSecurityGroup', $opt);\n}","examples":{"example":{"title":"<p>Create a database security group.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;create_db_security_group('mySecurityGroup', 'This is a sample description.');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/authorize_db_security_group_ingress.html\" rel=\"..\/..\/index.html#m=amazonrds\/authorize_db_security_group_ingress\"><code>authorize_db_security_group_ingress()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_security_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_security_group\"><code>delete_db_security_group()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_security_groups.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_security_groups\"><code>describe_db_security_groups()<\/code><\/a>","<a href=\"..\/amazonrds\/revoke_db_security_group_ingress.html\" rel=\"..\/..\/index.html#m=amazonrds\/revoke_db_security_group_ingress\"><code>revoke_db_security_group_ingress()<\/code><\/a>"]}},{"name":"create_db_snapshot","modifier":"public","description":"<p>Creates a DBSnapshot. The source DBInstance must be in &#8220;available&#8221; state.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_snapshot_identifier","type":"string","description":"<p>The identifier for the DB Snapshot. Constraints:<\/p>\n\n<ul><li>Cannot be null, empty, or blank<\/li><li>Must contain from 1 to 255 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>\n\n<p>Example: <code>my-snapshot-id<\/code><\/p>"},{"@attributes":{"choice":"req"},"name":"db_instance_identifier","type":"string","description":"<p>The DB Instance identifier. This is the unique key that identifies a DB Instance. This parameter isn&#8217;t case sensitive. Constraints:<\/p>\n\n<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_db_snapshot($db_snapshot_identifier, $db_instance_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBSnapshotIdentifier'] = $db_snapshot_identifier;\n    $opt['DBInstanceIdentifier'] = $db_instance_identifier;\n    \n    return $this-&gt;authenticate('CreateDBSnapshot', $opt);\n}","examples":{"example":{"title":"<p>Create a database snapshot.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;create_db_snapshot('mySnapshot', 'myInstanceID');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/delete_db_snapshot.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_snapshot\"><code>delete_db_snapshot()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_snapshots.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_snapshots\"><code>describe_db_snapshots()<\/code><\/a>"]}},{"name":"create_db_subnet_group","modifier":"public","description":"<p>Creates a new DB subnet group. DB subnet groups must contain at least one subnet in at least\ntwo AZs in the region.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_subnet_group_name","type":"string","description":"<p>The name for the DB Subnet Group. This value is stored as a lowercase string. Constraints: Must contain no more than 255 alphanumeric characters or hyphens. Must not be &#8220;Default&#8221;. Example: <code>mySubnetgroup<\/code><\/p>"},{"@attributes":{"choice":"req"},"name":"db_subnet_group_description","type":"string","description":"<p>The description for the DB Subnet Group.<\/p>"},{"@attributes":{"choice":"req"},"name":"subnet_ids","type":"string|array","description":"<p>The EC2 Subnet IDs for the DB Subnet Group. Pass a string for a single value, or an indexed array for multiple values.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_db_subnet_group($db_subnet_group_name, $db_subnet_group_description, $subnet_ids, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBSubnetGroupName'] = $db_subnet_group_name;\n    $opt['DBSubnetGroupDescription'] = $db_subnet_group_description;\n    \n    \/\/ Required list (non-map)\n    $opt = array_merge($opt, CFComplexType::map(array(\n        'SubnetIds' =&gt; (is_array($subnet_ids) ? $subnet_ids : array($subnet_ids))\n    ), 'member'));\n\n    return $this-&gt;authenticate('CreateDBSubnetGroup', $opt);\n}"},{"name":"create_event_subscription","modifier":"public","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"subscription_name","type":"string","description":{}},{"@attributes":{"choice":"req"},"name":"sns_topic_arn","type":"string","description":{}},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>SourceType<\/code> - <code>string<\/code> - Optional - <\/li><li><code>EventCategories<\/code> - <code>string|array<\/code> - Optional -  Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>SourceIds<\/code> - <code>string|array<\/code> - Optional -  Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>Enabled<\/code> - <code>boolean<\/code> - Optional - <\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_event_subscription($subscription_name, $sns_topic_arn, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['SubscriptionName'] = $subscription_name;\n    $opt['SnsTopicArn'] = $sns_topic_arn;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['EventCategories']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'EventCategories' =&gt; (is_array($opt['EventCategories']) ? $opt['EventCategories'] : array($opt['EventCategories']))\n        ), 'member'));\n        unset($opt['EventCategories']);\n    }\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['SourceIds']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'SourceIds' =&gt; (is_array($opt['SourceIds']) ? $opt['SourceIds'] : array($opt['SourceIds']))\n        ), 'member'));\n        unset($opt['SourceIds']);\n    }\n\n    return $this-&gt;authenticate('CreateEventSubscription', $opt);\n}"},{"name":"create_option_group","modifier":"public","description":"<p>Creates a new Option Group.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"option_group_name","type":"string","description":"<p>Specifies the name of the option group to be created. Constraints:<\/p>\n\n<ul><li>Must be 1 to 255 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>\n\n<p>Example: <code>myOptiongroup<\/code><\/p>"},{"@attributes":{"choice":"req"},"name":"engine_name","type":"string","description":"<p>Specifies the name of the engine that this option group should be associated with.<\/p>"},{"@attributes":{"choice":"req"},"name":"major_engine_version","type":"string","description":"<p>Specifies the major version of the engine that this option group should be associated with.<\/p>"},{"@attributes":{"choice":"req"},"name":"option_group_description","type":"string","description":"<p>The description of the option group.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function create_option_group($option_group_name, $engine_name, $major_engine_version, $option_group_description, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['OptionGroupName'] = $option_group_name;\n    $opt['EngineName'] = $engine_name;\n    $opt['MajorEngineVersion'] = $major_engine_version;\n    $opt['OptionGroupDescription'] = $option_group_description;\n    \n    return $this-&gt;authenticate('CreateOptionGroup', $opt);\n}"},{"name":"delete_cache","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Deletes a cached <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object using the specified cache storage type.<\/p>","returnvalue":"<p>A value of <code>true<\/code> if cached object exists and is successfully deleted, otherwise <code>false<\/code>.<\/p>","source":"public function delete_cache()\n{\n    $this-&gt;use_cache_flow = true;\n    $this-&gt;delete_cache = true;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Delete a cached response from APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ Delete the data\n$response = $sdb-&gt;delete_cache()-&gt;list_domains();\nvar_dump($response);","result":"bool(false)"},{"title":"<p>Delete a cached response from the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ Delete the data\n$response = $sdb-&gt;delete_cache()-&gt;list_domains();\nvar_dump($response);","result":"bool(true)"},{"title":"<p>Delete a cached response from Memcache.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config(array(\n\tarray('host' =&gt; '127.0.0.1')\n));\n\n\/\/ Delete the data\n$response = $sdb-&gt;delete_cache()-&gt;list_domains();\nvar_dump($response);","result":"bool(true)"}]}},{"name":"delete_db_instance","modifier":"public","description":"<p>The DeleteDBInstance API deletes a previously provisioned RDS instance. A successful response\nfrom the web service indicates the request was received correctly. If a final DBSnapshot is\nrequested the status of the RDS instance will be &#8220;deleting&#8221; until the DBSnapshot is created.\nDescribeDBInstance is used to monitor the status of this operation. This cannot be canceled or\nreverted once submitted.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_instance_identifier","type":"string","description":"<p>The DB Instance identifier for the DB Instance to be deleted. This parameter isn&#8217;t case sensitive. Constraints:<\/p>\n\n<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>SkipFinalSnapshot<\/code> - <code>boolean<\/code> - Optional - Determines whether a final DB Snapshot is created before the DB Instance is deleted. If <code>true<\/code> is specified, no DBSnapshot is created. If false is specified, a DB Snapshot is created before the DB Instance is deleted. <p class=\"note\">The FinalDBSnapshotIdentifier parameter must be specified if SkipFinalSnapshot is <code>false<\/code>.<\/p> Default: <code>false<\/code><\/li><li><code>FinalDBSnapshotIdentifier<\/code> - <code>string<\/code> - Optional - The DBSnapshotIdentifier of the new DBSnapshot created when SkipFinalSnapshot is set to <code>false<\/code>. <p class=\"note\">Specifying this parameter and also setting the SkipFinalShapshot parameter to true results in an error.<\/p> Constraints:<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul><\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_db_instance($db_instance_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBInstanceIdentifier'] = $db_instance_identifier;\n    \n    return $this-&gt;authenticate('DeleteDBInstance', $opt);\n}","examples":{"example":{"title":"<p>Delete a database instance.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;delete_db_instance('myInstance', array(\n\t'SkipFinalSnapshot' =&gt; true\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_instance\"><code>create_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/create_db_instance_read_replica.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_instance_read_replica\"><code>create_db_instance_read_replica()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_instances.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_instances\"><code>describe_db_instances()<\/code><\/a>","<a href=\"..\/amazonrds\/modify_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/modify_db_instance\"><code>modify_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/reboot_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/reboot_db_instance\"><code>reboot_db_instance()<\/code><\/a>"]}},{"name":"delete_db_parameter_group","modifier":"public","description":"<p>Deletes a specified DBParameterGroup. The DBParameterGroup cannot be associated with any RDS\ninstances to be deleted.<\/p>\n\n<p class=\"note\">\nThe specified DB Parameter Group cannot be associated with any DB Instances.\n<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_parameter_group_name","type":"string","description":"<p>The name of the DB Parameter Group. Constraints:<\/p>\n\n<ul><li>Must be the name of an existing DB Parameter Group<\/li><li>You cannot delete a default DB Parameter Group<\/li><li>Cannot be associated with any DB Instances<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_db_parameter_group($db_parameter_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBParameterGroupName'] = $db_parameter_group_name;\n    \n    return $this-&gt;authenticate('DeleteDBParameterGroup', $opt);\n}","examples":{"example":{"title":"<p>Delete a parameter group.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;delete_db_parameter_group('myParamGroup');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_parameter_group\"><code>create_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_parameter_groups.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_parameter_groups\"><code>describe_db_parameter_groups()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_parameters.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_parameters\"><code>describe_db_parameters()<\/code><\/a>","<a href=\"..\/amazonrds\/modify_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/modify_db_parameter_group\"><code>modify_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/reset_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/reset_db_parameter_group\"><code>reset_db_parameter_group()<\/code><\/a>"]}},{"name":"delete_db_security_group","modifier":"public","description":"<p>Deletes a DB Security Group.<\/p>\n\n<p class=\"note\">\nThe specified DB Security Group must not be associated with any DB Instances.\n<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_security_group_name","type":"string","description":"<p>The name of the DB Security Group to delete.<\/p>\n\n<p class=\"note\">You cannot delete the default DB Security Group.<\/p>\n\n<p>Constraints:<\/p>\n\n<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_db_security_group($db_security_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBSecurityGroupName'] = $db_security_group_name;\n    \n    return $this-&gt;authenticate('DeleteDBSecurityGroup', $opt);\n}","examples":{"example":{"title":"<p>Delete a security group.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;delete_db_security_group('mySecurityGroup');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/authorize_db_security_group_ingress.html\" rel=\"..\/..\/index.html#m=amazonrds\/authorize_db_security_group_ingress\"><code>authorize_db_security_group_ingress()<\/code><\/a>","<a href=\"..\/amazonrds\/create_db_security_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_security_group\"><code>create_db_security_group()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_security_groups.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_security_groups\"><code>describe_db_security_groups()<\/code><\/a>","<a href=\"..\/amazonrds\/revoke_db_security_group_ingress.html\" rel=\"..\/..\/index.html#m=amazonrds\/revoke_db_security_group_ingress\"><code>revoke_db_security_group_ingress()<\/code><\/a>"]}},{"name":"delete_db_snapshot","modifier":"public","description":"<p>Deletes a DBSnapshot.<\/p>\n\n<p class=\"note\">\nThe DBSnapshot must be in the <code>available<\/code> state to be deleted.\n<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_snapshot_identifier","type":"string","description":"<p>The DBSnapshot identifier. Constraints: Must be the name of an existing DB Snapshot in the <code>available<\/code> state.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_db_snapshot($db_snapshot_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBSnapshotIdentifier'] = $db_snapshot_identifier;\n    \n    return $this-&gt;authenticate('DeleteDBSnapshot', $opt);\n}","examples":{"example":{"title":"<p>Delete a database snapshot.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;delete_db_snapshot('mySnapshot');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_snapshot.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_snapshot\"><code>create_db_snapshot()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_snapshots.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_snapshots\"><code>describe_db_snapshots()<\/code><\/a>"]}},{"name":"delete_db_subnet_group","modifier":"public","description":"<p>Deletes a DB subnet group.<\/p>\n\n<p class=\"note\">\nThe specified database subnet group must not be associated with any DB instances.\n<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_subnet_group_name","type":"string","description":"<p>The name of the database subnet group to delete.<\/p>\n\n<p class=\"note\">You cannot delete the default subnet group.<\/p>\n\n<p>Constraints:<\/p>\n\n<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_db_subnet_group($db_subnet_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBSubnetGroupName'] = $db_subnet_group_name;\n    \n    return $this-&gt;authenticate('DeleteDBSubnetGroup', $opt);\n}"},{"name":"delete_event_subscription","modifier":"public","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"subscription_name","type":"string","description":{}},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_event_subscription($subscription_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['SubscriptionName'] = $subscription_name;\n    \n    return $this-&gt;authenticate('DeleteEventSubscription', $opt);\n}"},{"name":"delete_option_group","modifier":"public","description":"<p>Deletes an existing Option Group.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"option_group_name","type":"string","description":"<p>The name of the option group to be deleted.<\/p>\n\n<p class=\"note\">You cannot delete default Option Groups.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function delete_option_group($option_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['OptionGroupName'] = $option_group_name;\n    \n    return $this-&gt;authenticate('DeleteOptionGroup', $opt);\n}"},{"name":"describe_db_engine_versions","modifier":"public","description":"<p>Returns a list of the available DB engines.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>Engine<\/code> - <code>string<\/code> - Optional - The database engine to return.<\/li><li><code>EngineVersion<\/code> - <code>string<\/code> - Optional - The database engine version to return. Example: <code>5.1.49<\/code><\/li><li><code>DBParameterGroupFamily<\/code> - <code>string<\/code> - Optional - The name of a specific DB Parameter Group family to return details for. Constraints:<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul><\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more than the <code>MaxRecords<\/code> value is available, a pagination token called a marker is included in the response so that the following results can be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>DefaultOnly<\/code> - <code>boolean<\/code> - Optional - Indicates that only the default version of the specified engine or engine and major version combination is returned.<\/li><li><code>ListSupportedCharacterSets<\/code> - <code>boolean<\/code> - Optional - If this parameter is specified, and if the requested engine supports the CharacterSetName parameter for CreateDBInstance, the response includes a list of supported character sets for each engine version.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_db_engine_versions($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeDBEngineVersions', $opt);\n}","examples":{"example":[{"title":"<p>Describe the MySQL database engine versions.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_db_engine_versions(array(\n\t'Engine' =&gt; 'MySQL',\n\t'DBParameterGroupFamily' =&gt; 'mysql5.1',\n\t'DefaultOnly' =&gt; true\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Describe the Oracle database engine versions.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_db_engine_versions(array(\n\t'Engine' =&gt; 'oracle-se'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Describe the Oracle database engine versions.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_db_engine_versions(array(\n\t'Engine' =&gt; 'oracle-se1'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Describe the Oracle database engine versions.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_db_engine_versions(array(\n\t'Engine' =&gt; 'oracle-ee'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}]},"related":{"method":"<a href=\"..\/amazonrds\/describe_engine_default_parameters.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_engine_default_parameters\"><code>describe_engine_default_parameters()<\/code><\/a>"}},{"name":"describe_db_instances","modifier":"public","description":"<p>Returns information about provisioned RDS instances. This API supports pagination.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBInstanceIdentifier<\/code> - <code>string<\/code> - Optional - The user-supplied instance identifier. If this parameter is specified, information from only the specific DB Instance is returned. This parameter isn&#8217;t case sensitive. Constraints:<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul><\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results may be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous DescribeDBInstances request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_db_instances($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeDBInstances', $opt);\n}","examples":{"example":{"title":"<p>Describe a single database instance.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_db_instances(array(\n\t'DBInstanceIdentifier' =&gt; 'myInstance'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_instance\"><code>create_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/create_db_instance_read_replica.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_instance_read_replica\"><code>create_db_instance_read_replica()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_instance\"><code>delete_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/modify_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/modify_db_instance\"><code>modify_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/reboot_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/reboot_db_instance\"><code>reboot_db_instance()<\/code><\/a>"]}},{"name":"describe_db_log_files","modifier":"public","description":"<p>Returns a list of DB log files for the DB instance.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBInstanceIdentifier<\/code> - <code>string<\/code> - Optional - The customer-assigned name of the DB Instance that contains the log files you want to list. Constraints:<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul><\/li><li><code>FilenameContains<\/code> - <code>string<\/code> - Optional - Filters the available log files for log file names that contain the specified string.<\/li><li><code>FileLastWritten<\/code> - <code>long<\/code> - Optional - Filters the available log files for files written since the specified date.<\/li><li><code>FileSize<\/code> - <code>long<\/code> - Optional - Filters the available log files for files larger than the specified size.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified MaxRecords value, a pagination token called a marker is included in the response so that the remaining results can be retrieved.<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - The pagination token provided in the previous request. If this parameter is specified the response includes only records beyond the marker, up to MaxRecords.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_db_log_files($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeDBLogFiles', $opt);\n}"},{"name":"describe_db_parameter_groups","modifier":"public","description":"<p>Returns a list of DBParameterGroup descriptions. If a DBParameterGroupName is specified, the\nlist will contain only the description of the specified DBParameterGroup.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBParameterGroupName<\/code> - <code>string<\/code> - Optional - The name of a specific DB Parameter Group to return details for. Constraints:<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul><\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results may be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous DescribeDBParameterGroups request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_db_parameter_groups($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeDBParameterGroups', $opt);\n}","examples":{"example":{"title":"<p>Describe the database parameter groups.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_db_parameter_groups(array(\n\t'DBParameterGroupName' =&gt; 'myParamGroup'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_parameter_group\"><code>create_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_parameter_group\"><code>delete_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_parameters.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_parameters\"><code>describe_db_parameters()<\/code><\/a>","<a href=\"..\/amazonrds\/modify_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/modify_db_parameter_group\"><code>modify_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/reset_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/reset_db_parameter_group\"><code>reset_db_parameter_group()<\/code><\/a>"]}},{"name":"describe_db_parameters","modifier":"public","description":"<p>Returns the detailed parameter list for a particular DBParameterGroup.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_parameter_group_name","type":"string","description":"<p>The name of a specific DB Parameter Group to return details for. Constraints:<\/p>\n\n<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>Source<\/code> - <code>string<\/code> - Optional - The parameter types to return. Default: All parameter types returned Valid Values: <code>user | system | engine-default<\/code><\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results may be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous DescribeDBParameters request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_db_parameters($db_parameter_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBParameterGroupName'] = $db_parameter_group_name;\n    \n    return $this-&gt;authenticate('DescribeDBParameters', $opt);\n}","examples":{"example":{"title":"<p>Describe the database parameters.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_db_parameters('myParamGroup', array(\n\t'Source' =&gt; 'engine-default'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_parameter_group\"><code>create_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_parameter_group\"><code>delete_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_parameter_groups.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_parameter_groups\"><code>describe_db_parameter_groups()<\/code><\/a>","<a href=\"..\/amazonrds\/modify_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/modify_db_parameter_group\"><code>modify_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/reset_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/reset_db_parameter_group\"><code>reset_db_parameter_group()<\/code><\/a>"]}},{"name":"describe_db_security_groups","modifier":"public","description":"<p>Returns a list of DBSecurityGroup descriptions. If a DBSecurityGroupName is specified, the list\nwill contain only the descriptions of the specified DBSecurityGroup.<\/p>\n\n<p>For an overview of CIDR ranges, go to the <a href=\n\"http:\/\/en.wikipedia.org\/wiki\/Classless_Inter-Domain_Routing\">Wikipedia Tutorial<\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBSecurityGroupName<\/code> - <code>string<\/code> - Optional - The name of the DB Security Group to return details for.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results may be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous DescribeDBSecurityGroups request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_db_security_groups($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeDBSecurityGroups', $opt);\n}","examples":{"example":{"title":"<p>Describe a database security group.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_db_security_groups(array(\n\t'DBSecurityGroupName' =&gt; 'mySecurityGroup'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/authorize_db_security_group_ingress.html\" rel=\"..\/..\/index.html#m=amazonrds\/authorize_db_security_group_ingress\"><code>authorize_db_security_group_ingress()<\/code><\/a>","<a href=\"..\/amazonrds\/create_db_security_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_security_group\"><code>create_db_security_group()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_security_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_security_group\"><code>delete_db_security_group()<\/code><\/a>","<a href=\"..\/amazonrds\/revoke_db_security_group_ingress.html\" rel=\"..\/..\/index.html#m=amazonrds\/revoke_db_security_group_ingress\"><code>revoke_db_security_group_ingress()<\/code><\/a>"]}},{"name":"describe_db_snapshots","modifier":"public","description":"<p>Returns information about DBSnapshots. This API supports pagination.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBInstanceIdentifier<\/code> - <code>string<\/code> - Optional - A DB Instance Identifier to retrieve the list of DB Snapshots for. Cannot be used in conjunction with DBSnapshotIdentifier. This parameter isn&#8217;t case sensitive. Constraints:<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul><\/li><li><code>DBSnapshotIdentifier<\/code> - <code>string<\/code> - Optional - A specific DB Snapshot Identifier to describe. Cannot be used in conjunction with DBInstanceIdentifier. This value is stored as a lowercase string. Constraints:<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><li>If this is the identifier of an automated snapshot, the <code>SnapshotType<\/code> parameter must also be specified.<\/li><\/ul><\/li><li><code>SnapshotType<\/code> - <code>string<\/code> - Optional - An optional snapshot type for which snapshots will be returned. If not specified, the returned results will include snapshots of all types.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results may be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous DescribeDBSnapshots request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_db_snapshots($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeDBSnapshots', $opt);\n}","examples":{"example":{"title":"<p>Describe a single database snapshot.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_db_snapshots();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_snapshot.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_snapshot\"><code>create_db_snapshot()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_snapshot.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_snapshot\"><code>delete_db_snapshot()<\/code><\/a>"]}},{"name":"describe_db_subnet_groups","modifier":"public","description":"<p>Returns a list of DBSubnetGroup descriptions. If a DBSubnetGroupName is specified, the list\nwill contain only the descriptions of the specified DBSubnetGroup.<\/p>\n\n<p>For an overview of CIDR ranges, go to the <a href=\n\"http:\/\/en.wikipedia.org\/wiki\/Classless_Inter-Domain_Routing\">Wikipedia Tutorial<\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBSubnetGroupName<\/code> - <code>string<\/code> - Optional - The name of the DB Subnet Group to return details for.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results may be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous DescribeDBSubnetGroups request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_db_subnet_groups($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeDBSubnetGroups', $opt);\n}"},{"name":"describe_engine_default_parameters","modifier":"public","description":"<p>Returns the default engine and system parameter information for the specified database engine.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_parameter_group_family","type":"string","description":"<p>The name of the DB Parameter Group Family.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results may be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous DescribeEngineDefaultParameters request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_engine_default_parameters($db_parameter_group_family, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBParameterGroupFamily'] = $db_parameter_group_family;\n    \n    return $this-&gt;authenticate('DescribeEngineDefaultParameters', $opt);\n}","examples":{"example":{"title":"<p>Describe the database engine default parameters.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_engine_default_parameters('MySQL5.1');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":"<a href=\"..\/amazonrds\/describe_db_engine_versions.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_engine_versions\"><code>describe_db_engine_versions()<\/code><\/a>"}},{"name":"describe_event_categories","modifier":"public","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>SourceType<\/code> - <code>string<\/code> - Optional - <\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_event_categories($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeEventCategories', $opt);\n}"},{"name":"describe_event_subscriptions","modifier":"public","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>SubscriptionName<\/code> - <code>string<\/code> - Optional - <\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - <\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - <\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_event_subscriptions($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeEventSubscriptions', $opt);\n}"},{"name":"describe_events","modifier":"public","description":"<p>Returns events related to DB Instances, DB Security Groups, DB Snapshots and DB Parameter\nGroups for the past 14 days. Events specific to a particular DB Instance, DB Security Group,\ndatabase snapshot or DB Parameter Group can be obtained by providing the name as a parameter.\nBy default, the past hour of events are returned.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>SourceIdentifier<\/code> - <code>string<\/code> - Optional - The identifier of the event source for which events will be returned. If not specified, then all sources are included in the response. Constraints:<ul><li>If SourceIdentifier is supplied, SourceType must also be provided.<\/li><li>If the source type is DBInstance, then a DBInstanceIdentifier must be supplied.<\/li><li>If the source type is DBSecurityGroup, a DBSecurityGroupName must be supplied.<\/li><li>If the source type is DBParameterGroup, a DBParameterGroupName must be supplied.<\/li><li>If the source type is DBSnapshot, a DBSnapshotIdentifier must be supplied.<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens.<\/li><\/ul><\/li><li><code>SourceType<\/code> - <code>string<\/code> - Optional - The event source to retrieve events for. If no value is specified, all events are returned. [Allowed values: <code>db-instance<\/code>, <code>db-parameter-group<\/code>, <code>db-security-group<\/code>, <code>db-snapshot<\/code>]<\/li><li><code>StartTime<\/code> - <code>string<\/code> - Optional - The beginning of the time interval to retrieve events for, specified in ISO 8601 format. For more information about ISO 8601, go to the <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a>ISO8601 Wikipedia page.<a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a> Example: 2009-07-08T18:00Z May be passed as a number of seconds since UNIX Epoch, or any string compatible with <php:strtotime()>.<\/li><li><code>EndTime<\/code> - <code>string<\/code> - Optional - The end of the time interval for which to retrieve events, specified in ISO 8601 format. For more information about ISO 8601, go to the <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a>ISO8601 Wikipedia page.<a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a> Example: 2009-07-08T18:00Z May be passed as a number of seconds since UNIX Epoch, or any string compatible with <php:strtotime()>.<\/li><li><code>Duration<\/code> - <code>integer<\/code> - Optional - The number of minutes to retrieve events for. Default: 60<\/li><li><code>EventCategories<\/code> - <code>string|array<\/code> - Optional -  Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results may be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous DescribeEvents request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_events($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    \/\/ Optional DateTime\n    if (isset($opt['StartTime']))\n    {\n        $opt['StartTime'] = $this-&gt;util-&gt;convert_date_to_iso8601($opt['StartTime']);\n    }\n    \n    \/\/ Optional DateTime\n    if (isset($opt['EndTime']))\n    {\n        $opt['EndTime'] = $this-&gt;util-&gt;convert_date_to_iso8601($opt['EndTime']);\n    }\n\n    \/\/ Optional list (non-map)\n    if (isset($opt['EventCategories']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'EventCategories' =&gt; (is_array($opt['EventCategories']) ? $opt['EventCategories'] : array($opt['EventCategories']))\n        ), 'member'));\n        unset($opt['EventCategories']);\n    }\n\n    return $this-&gt;authenticate('DescribeEvents', $opt);\n}","examples":{"example":{"title":"<p>Describe a single database snapshot.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_db_snapshots();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{}},{"name":"describe_option_group_options","modifier":"public","description":"<p>Describes all available options.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"engine_name","type":"string","description":"<p>A required parameter. Options available for the given Engine name will be described.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>MajorEngineVersion<\/code> - <code>string<\/code> - Optional - If specified, filters the results to include only options for the specified major engine version.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results can be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_option_group_options($engine_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['EngineName'] = $engine_name;\n    \n    return $this-&gt;authenticate('DescribeOptionGroupOptions', $opt);\n}"},{"name":"describe_option_groups","modifier":"public","description":"<p>Describes the available option groups.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>OptionGroupName<\/code> - <code>string<\/code> - Optional - The name of the option group to describe. Cannot be supplied together with EngineName or MajorEngineVersion.<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous DescribeOptionGroups request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results can be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>EngineName<\/code> - <code>string<\/code> - Optional - Filters the list of option groups to only include groups associated with a specific database engine.<\/li><li><code>MajorEngineVersion<\/code> - <code>string<\/code> - Optional - Filters the list of option groups to only include groups associated with a specific database engine version. If specified, then EngineName must also be specified.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_option_groups($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeOptionGroups', $opt);\n}"},{"name":"describe_orderable_db_instance_options","modifier":"public","description":"<p>Returns a list of orderable DB Instance options for the specified engine.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"engine","type":"string","description":"<p>The name of the engine to retrieve DB Instance options for.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>EngineVersion<\/code> - <code>string<\/code> - Optional - The engine version filter value. Specify this parameter to show only the available offerings matching the specified engine version.<\/li><li><code>DBInstanceClass<\/code> - <code>string<\/code> - Optional - The DB Instance class filter value. Specify this parameter to show only the available offerings matching the specified DB Instance class.<\/li><li><code>LicenseModel<\/code> - <code>string<\/code> - Optional - The license model filter value. Specify this parameter to show only the available offerings matching the specified license model.<\/li><li><code>Vpc<\/code> - <code>boolean<\/code> - Optional - The VPC filter value. Specify this parameter to show only the available VPC or non-VPC offerings.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more records exist than the specified <code>MaxRecords<\/code> value, a pagination token called a marker is included in the response so that the remaining results can be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous DescribeOrderableDBInstanceOptions request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_orderable_db_instance_options($engine, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['Engine'] = $engine;\n    \n    return $this-&gt;authenticate('DescribeOrderableDBInstanceOptions', $opt);\n}","examples":{"example":[{"title":"<p>Get a list of orderable DB Instance options for MySQL.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_orderable_db_instance_options('MySQL');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Get a list of orderable DB Instance options for Oracle SE.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_orderable_db_instance_options('oracle-se');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Get a list of orderable DB Instance options for Oracle SE One.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_orderable_db_instance_options('oracle-se1');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Get a list of orderable DB Instance options for Oracle EE.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_orderable_db_instance_options('oracle-ee');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}]}},{"name":"describe_reserved_db_instances","modifier":"public","description":"<p>Returns information about reserved DB Instances for this account, or about a specified reserved\nDB Instance.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>ReservedDBInstanceId<\/code> - <code>string<\/code> - Optional - The reserved DB Instance identifier filter value. Specify this parameter to show only the reservation that matches the specified reservation ID.<\/li><li><code>ReservedDBInstancesOfferingId<\/code> - <code>string<\/code> - Optional - The offering identifier filter value. Specify this parameter to show only purchased reservations matching the specified offering identifier.<\/li><li><code>DBInstanceClass<\/code> - <code>string<\/code> - Optional - The DB Instance class filter value. Specify this parameter to show only those reservations matching the specified DB Instances class.<\/li><li><code>Duration<\/code> - <code>string<\/code> - Optional - The duration filter value, specified in years or seconds. Specify this parameter to show only reservations for this duration. Valid Values: <code>1 | 3 | 31536000 | 94608000<\/code><\/li><li><code>ProductDescription<\/code> - <code>string<\/code> - Optional - The product description filter value. Specify this parameter to show only those reservations matching the specified product description.<\/li><li><code>OfferingType<\/code> - <code>string<\/code> - Optional - The offering type filter value. Specify this parameter to show only the available offerings matching the specified offering type. Valid Values: <code>\"Light Utilization\" | \"Medium Utilization\" | \"Heavy Utilization\"<\/code><\/li><li><code>MultiAZ<\/code> - <code>boolean<\/code> - Optional - The Multi-AZ filter value. Specify this parameter to show only those reservations matching the specified Multi-AZ parameter.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more than the <code>MaxRecords<\/code> value is available, a pagination token called a marker is included in the response so that the following results can be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_reserved_db_instances($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeReservedDBInstances', $opt);\n}","examples":{"example":{"title":"<p>Describe reserved database instances.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_reserved_db_instances();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/describe_reserved_db_instances_offerings.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_reserved_db_instances_offerings\"><code>describe_reserved_db_instances_offerings()<\/code><\/a>","<a href=\"..\/amazonrds\/purchase_reserved_db_instances_offering.html\" rel=\"..\/..\/index.html#m=amazonrds\/purchase_reserved_db_instances_offering\"><code>purchase_reserved_db_instances_offering()<\/code><\/a>"]}},{"name":"describe_reserved_db_instances_offerings","modifier":"public","description":"<p>Lists available reserved DB Instance offerings.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>ReservedDBInstancesOfferingId<\/code> - <code>string<\/code> - Optional - The offering identifier filter value. Specify this parameter to show only the available offering that matches the specified reservation identifier. Example: <code>438012d3-4052-4cc7-b2e3-8d3372e0e706<\/code><\/li><li><code>DBInstanceClass<\/code> - <code>string<\/code> - Optional - The DB Instance class filter value. Specify this parameter to show only the available offerings matching the specified DB Instance class.<\/li><li><code>Duration<\/code> - <code>string<\/code> - Optional - Duration filter value, specified in years or seconds. Specify this parameter to show only reservations for this duration. Valid Values: <code>1 | 3 | 31536000 | 94608000<\/code><\/li><li><code>ProductDescription<\/code> - <code>string<\/code> - Optional - Product description filter value. Specify this parameter to show only the available offerings matching the specified product description.<\/li><li><code>OfferingType<\/code> - <code>string<\/code> - Optional - The offering type filter value. Specify this parameter to show only the available offerings matching the specified offering type. Valid Values: <code>\"Light Utilization\" | \"Medium Utilization\" | \"Heavy Utilization\"<\/code><\/li><li><code>MultiAZ<\/code> - <code>boolean<\/code> - Optional - The Multi-AZ filter value. Specify this parameter to show only the available offerings matching the specified Multi-AZ parameter.<\/li><li><code>MaxRecords<\/code> - <code>integer<\/code> - Optional - The maximum number of records to include in the response. If more than the <code>MaxRecords<\/code> value is available, a pagination token called a marker is included in the response so that the following results can be retrieved. Default: 100 Constraints: minimum 20, maximum 100<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - An optional pagination token provided by a previous request. If this parameter is specified, the response includes only records beyond the marker, up to the value specified by <code>MaxRecords<\/code>.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function describe_reserved_db_instances_offerings($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DescribeReservedDBInstancesOfferings', $opt);\n}","examples":{"example":{"title":"<p>Describe reserved database instance offerings.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;describe_reserved_db_instances_offerings(array(\n\t'MultiAZ' =&gt; 'false'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/describe_reserved_db_instances.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_reserved_db_instances\"><code>describe_reserved_db_instances()<\/code><\/a>","<a href=\"..\/amazonrds\/purchase_reserved_db_instances_offering.html\" rel=\"..\/..\/index.html#m=amazonrds\/purchase_reserved_db_instances_offering\"><code>purchase_reserved_db_instances_offering()<\/code><\/a>"]}},{"name":"disable_ssl","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Disables SSL\/HTTPS connections for hosts that don&#8217;t support them. Some services, however, still\nrequire SSL support.<\/p>\n\n<p>This method will throw a user warning when invoked, which can be hidden by changing your\n<a href=\"http:\/\/php.net\/error_reporting\"><code>error_reporting()<\/code><\/a> settings.<\/p>","returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function disable_ssl()\n{\n    trigger_error('Disabling SSL connections is potentially unsafe and highly discouraged.', E_USER_WARNING);\n    $this-&gt;use_ssl = false;\n    return $this;\n}","examples":{"example":{"title":"<p>Disable SSL for the service.<\/p>","description":"<ul>\n<li><strong>Note:<\/strong> It is not recommended to disable SSL.<\/li>\n<li><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/li>\n<\/ul>","code":"\/\/ Disable SSL (Using @ to suppress the warning from the output)\n$sdb = new AmazonSDB();\n@$sdb-&gt;disable_ssl(true);\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;use_ssl);","result":"bool(false)"}}},{"name":"disable_ssl_verification","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Disables the verification of the SSL Certificate Authority. Doing so can enable an attacker to carry\nout a man-in-the-middle attack.<\/p>\n\n<p>https:\/\/secure.wikimedia.org\/wikipedia\/en\/wiki\/Man-in-the-middle_attack<\/p>\n\n<p>This method will throw a user warning when invoked, which can be hidden by changing your\n<a href=\"http:\/\/php.net\/error_reporting\"><code>error_reporting()<\/code><\/a> settings.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"ssl_verification"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function disable_ssl_verification($ssl_verification = false)\n{\n    trigger_error('Disabling the verification of SSL certificates can lead to man-in-the-middle attacks. It is potentially unsafe and highly discouraged.', E_USER_WARNING);\n    $this-&gt;ssl_verification = $ssl_verification;\n    return $this;\n}"},{"name":"download_db_log_file_portion","modifier":"public","description":"<p>Downloads the last line of the specified log file.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBInstanceIdentifier<\/code> - <code>string<\/code> - Optional - The customer-assigned name of the DB Instance that contains the log files you want to list. Constraints:<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul><\/li><li><code>LogFileName<\/code> - <code>string<\/code> - Optional - The name of the log file to be downloaded.<\/li><li><code>Marker<\/code> - <code>string<\/code> - Optional - The pagination token provided in the previous request. If this parameter is specified the response includes only records beyond the marker, up to MaxRecords.<\/li><li><code>NumberOfLines<\/code> - <code>integer<\/code> - Optional - The number of lines remaining to be downloaded.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function download_db_log_file_portion($opt = null)\n{\n    if (!$opt) $opt = array();\n            \n    return $this-&gt;authenticate('DownloadDBLogFilePortion', $opt);\n}"},{"name":"enable_debug_mode","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Enables HTTP request\/response header logging to <code>STDERR<\/code>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"enabled","type":"boolean","description":"<p>Whether or not to enable debug mode. Defaults to <code>true<\/code>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function enable_debug_mode($enabled = true)\n{\n    $this-&gt;debug_mode = $enabled;\n    return $this;\n}","examples":{"example":{"title":"<p>Enable the logging of HTTP request\/response headers to <code>STDERR<\/code>.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Adjust offset\n$sdb = new AmazonSDB();\n$sdb-&gt;enable_debug_mode();\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;debug_mode);","result":"bool(true)"}}},{"name":"factory","modifier":"public static","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Alternate approach to constructing a new instance. Supports chaining.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"options","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>certificate_authority<\/code> - <code>boolean<\/code> - Optional - Determines which Cerificate Authority file to use. A value of boolean <code>false<\/code> will use the Certificate Authority file available on the system. A value of boolean <code>true<\/code> will use the Certificate Authority provided by the SDK. Passing a file system path to a Certificate Authority file (chmodded to <code>0755<\/code>) will use that. Leave this set to <code>false<\/code> if you&#8217;re not sure.<\/li><li><code>credentials<\/code> - <code>string<\/code> - Optional - The name of the credential set to use for authentication.<\/li><li><code>default_cache_config<\/code> - <code>string<\/code> - Optional - This option allows a preferred storage type to be configured for long-term caching. This can be changed later using the <a href=\"..\/amazonrds\/set_cache_config.html\" rel=\"..\/..\/index.html#m=amazonrds\/set_cache_config\"><code>set_cache_config()<\/code><\/a> method. Valid values are: <code>apc<\/code>, <code>xcache<\/code>, or a file system path such as <code>.\/cache<\/code> or <code>\/tmp\/cache\/<\/code>.<\/li><li><code>key<\/code> - <code>string<\/code> - Optional - Your AWS key, or a session key. If blank, the default credential set will be used.<\/li><li><code>secret<\/code> - <code>string<\/code> - Optional - Your AWS secret key, or a session secret key. If blank, the default credential set will be used.<\/li><li><code>token<\/code> - <code>string<\/code> - Optional - An AWS session token.<\/li><\/ul>"}},"returnvalue":{"@attributes":{"type":"void"}},"source":"public static function factory(array $options = array())\n{\n    if (version_compare(PHP_VERSION, '5.3.0', '&lt;'))\n    {\n        throw new Exception('PHP 5.3 or newer is required to instantiate a new class with CLASS::factory().');\n    }\n\n    $self = get_called_class();\n    return new $self($options);\n}"},{"name":"list_tags_for_resource","modifier":"public","description":"<p>Lists all tags on a DB Instance.<\/p>\n\n<p>For an overview on tagging DB Instances, see <a href=\n\"http:\/\/docs.amazonwebservices.com\/AmazonRDS\/latest\/UserGuide\/Overview.Tagging.html\">DB\nInstance Tags.<\/a><\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"resource_name","type":"string","description":"<p>The DB Instance with tags to be listed.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function list_tags_for_resource($resource_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['ResourceName'] = $resource_name;\n    \n    return $this-&gt;authenticate('ListTagsForResource', $opt);\n}"},{"name":"modify_db_instance","modifier":"public","description":"<p>Modify settings for a DB Instance. You can change one or more database configuration parameters\nby specifying these parameters and the new values in the request.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_instance_identifier","type":"string","description":"<p>The DB Instance identifier. This value is stored as a lowercase string. Constraints:<\/p>\n\n<ul><li>Must be the identifier for an existing DB Instance<\/li><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>AllocatedStorage<\/code> - <code>integer<\/code> - Optional - The new storage capacity of the RDS instance. Changing this parameter does not result in an outage and the change is applied during the next maintenance window unless the <code>ApplyImmediately<\/code> parameter is set to <code>true<\/code> for this request. <strong>MySQL<\/strong> Default: Uses existing setting Valid Values: 5-1024 Constraints: Value supplied must be at least 10% greater than the current value. Values that are not at least 10% greater than the existing value are rounded up so that they are 10% greater than the current value. Type: Integer <strong>Oracle<\/strong> Default: Uses existing setting Valid Values: 10-1024 Constraints: Value supplied must be at least 10% greater than the current value. Values that are not at least 10% greater than the existing value are rounded up so that they are 10% greater than the current value. <strong>SQL Server<\/strong> Cannot be modified.<\/li><li><code>DBInstanceClass<\/code> - <code>string<\/code> - Optional - The new compute and memory capacity of the DB Instance. To determine the instance classes that are available for a particular DB engine, use the <code>DescribeOrderableDBInstanceOptions<\/code> action. Passing a value for this parameter causes an outage during the change and is applied during the next maintenance window, unless the <code>ApplyImmediately<\/code> parameter is specified as <code>true<\/code> for this request. Default: Uses existing setting Valid Values: <code>db.t1.micro | db.m1.small | db.m1.medium | db.m1.large | db.m1.xlarge | db.m2.xlarge | db.m2.2xlarge | db.m2.4xlarge<\/code><\/li><li><code>DBSecurityGroups<\/code> - <code>string|array<\/code> - Optional - A list of DB Security Groups to authorize on this DB Instance. Changing this parameter does not result in an outage and the change is asynchronously applied as soon as possible. Constraints:<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul> Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>VpcSecurityGroupIds<\/code> - <code>string|array<\/code> - Optional - A list of EC2 VPC Security Groups to authorize on this DB Instance. This change is asynchronously applied as soon as possible. Constraints:<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul> Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>ApplyImmediately<\/code> - <code>boolean<\/code> - Optional - Specifies whether or not the modifications in this request and any pending modifications are asynchronously applied as soon as possible, regardless of the <code>PreferredMaintenanceWindow<\/code> setting for the DB Instance. If this parameter is passed as <code>false<\/code>, changes to the DB Instance are applied on the next call to <code>RebootDBInstance<\/code>, the next maintenance reboot, or the next failure reboot, whichever occurs first. See each parameter to determine when a change is applied. Default: <code>false<\/code><\/li><li><code>MasterUserPassword<\/code> - <code>string<\/code> - Optional - The new password for the DB Instance master user. Can be any printable ASCII character except &#8220;\/&#8221;, &#8220;&#34;, or &#8220;@&#8221;. Changing this parameter does not result in an outage and the change is asynchronously applied as soon as possible. Between the time of the request and the completion of the request, the <code>MasterUserPassword<\/code> element exists in the <code>PendingModifiedValues<\/code> element of the operation response. Default: Uses existing setting Constraints: Must be 8 to 41 alphanumeric characters (MySQL), 8 to 30 alphanumeric characters (Oracle), or 8 to 128 alphanumeric characters (SQL Server). <p class=\"note\">Amazon RDS API actions never return the password, so this action provides a way to regain access to a master instance user if the password is lost.<\/p><\/li><li><code>DBParameterGroupName<\/code> - <code>string<\/code> - Optional - The name of the DB Parameter Group to apply to this DB Instance. Changing this parameter does not result in an outage and the change is applied during the next maintenance window unless the <code>ApplyImmediately<\/code> parameter is set to <code>true<\/code> for this request. Default: Uses existing setting Constraints: The DB Parameter Group must be in the same DB Parameter Group family as this DB Instance.<\/li><li><code>BackupRetentionPeriod<\/code> - <code>integer<\/code> - Optional - The number of days to retain automated backups. Setting this parameter to a positive number enables backups. Setting this parameter to 0 disables automated backups. Changing this parameter can result in an outage if you change from 0 to a non-zero value or from a non-zero value to 0. These changes are applied during the next maintenance window unless the <code>ApplyImmediately<\/code> parameter is set to <code>true<\/code> for this request. If you change the parameter from one non-zero value to another non-zero value, the change is asynchronously applied as soon as possible. Default: Uses existing setting Constraints:<ul><li>Must be a value from 0 to 8<\/li><li>Cannot be set to 0 if the DB Instance is a master instance with read replicas or if the DB Instance is a read replica<\/li><\/ul><\/li><li><code>PreferredBackupWindow<\/code> - <code>string<\/code> - Optional - The daily time range during which automated backups are created if automated backups are enabled, as determined by the <code>BackupRetentionPeriod<\/code>. Changing this parameter does not result in an outage and the change is asynchronously applied as soon as possible. Constraints:<ul><li>Must be in the format hh24:mi-hh24:mi<\/li><li>Times should be Universal Time Coordinated (UTC)<\/li><li>Must not conflict with the preferred maintenance window<\/li><li>Must be at least 30 minutes<\/li><\/ul><\/li><li><code>PreferredMaintenanceWindow<\/code> - <code>string<\/code> - Optional - The weekly time range (in UTC) during which system maintenance can occur, which may result in an outage. Changing this parameter does not result in an outage, except in the following situation, and the change is asynchronously applied as soon as possible. If there are pending actions that cause a reboot, and the maintenance window is changed to include the current time, then changing this parameter will cause a reboot of the DB Instance. If moving this window to the current time, there must be at least 30 minutes between the current time and end of the window to ensure pending changes are applied. Default: Uses existing setting Format: ddd:hh24:mi-ddd:hh24:mi Valid Days: Mon | Tue | Wed | Thu | Fri | Sat | Sun Constraints: Must be at least 30 minutes<\/li><li><code>MultiAZ<\/code> - <code>boolean<\/code> - Optional - Specifies if the DB Instance is a Multi-AZ deployment. Changing this parameter does not result in an outage and the change is applied during the next maintenance window unless the <code>ApplyImmediately<\/code> parameter is set to <code>true<\/code> for this request. Constraints: Cannot be specified if the DB Instance is a read replica.<\/li><li><code>EngineVersion<\/code> - <code>string<\/code> - Optional - The version number of the database engine to upgrade to. Changing this parameter results in an outage and the change is applied during the next maintenance window unless the <code>ApplyImmediately<\/code> parameter is set to <code>true<\/code> for this request. For major version upgrades, if a nondefault DB Parameter Group is currently in use, a new DB Parameter Group in the DB Parameter Group Family for the new engine version must be specified. The new DB Parameter Group can be the default for that DB Parameter Group Family. Example: <code>5.1.42<\/code><\/li><li><code>AllowMajorVersionUpgrade<\/code> - <code>boolean<\/code> - Optional - Indicates that major version upgrades are allowed. Changing this parameter does not result in an outage and the change is asynchronously applied as soon as possible. Constraints: This parameter must be set to true when specifying a value for the EngineVersion parameter that is a different major version than the DB Instance&#8217;s current version.<\/li><li><code>AutoMinorVersionUpgrade<\/code> - <code>boolean<\/code> - Optional - Indicates that minor version upgrades will be applied automatically to the DB Instance during the maintenance window. Changing this parameter does not result in an outage except in the following case and the change is asynchronously applied as soon as possible. An outage will result if this parameter is set to <code>true<\/code> during the maintenance window, and a newer minor version is available, and RDS has enabled auto patching for that engine version.<\/li><li><code>Iops<\/code> - <code>integer<\/code> - Optional - The new Provisioned IOPS (I\/O operations per second) value for the RDS instance. Changing this parameter does not result in an outage and the change is applied during the next maintenance window unless the <code>ApplyImmediately<\/code> parameter is set to <code>true<\/code> for this request. Default: Uses existing setting Constraints: Value supplied must be at least 10% greater than the current value. Values that are not at least 10% greater than the existing value are rounded up so that they are 10% greater than the current value. Type: Integer<\/li><li><code>OptionGroupName<\/code> - <code>string<\/code> - Optional - Indicates that the DB Instance should be associated with the specified option group. Changing this parameter does not result in an outage except in the following case and the change is applied during the next maintenance window unless the <code>ApplyImmediately<\/code> parameter is set to <code>true<\/code> for this request. If the parameter change results in an option group that enables OEM, this change can cause a brief (sub-second) period during which new connections are rejected but existing connections are not interrupted.<\/li><li><code>NewDBInstanceIdentifier<\/code> - <code>string<\/code> - Optional - The new DB Instance identifier for the DB Instance when renaming a DB Instance. This value is stored as a lowercase string. Constraints:<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul><\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function modify_db_instance($db_instance_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBInstanceIdentifier'] = $db_instance_identifier;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['DBSecurityGroups']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'DBSecurityGroups' =&gt; (is_array($opt['DBSecurityGroups']) ? $opt['DBSecurityGroups'] : array($opt['DBSecurityGroups']))\n        ), 'member'));\n        unset($opt['DBSecurityGroups']);\n    }\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['VpcSecurityGroupIds']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'VpcSecurityGroupIds' =&gt; (is_array($opt['VpcSecurityGroupIds']) ? $opt['VpcSecurityGroupIds'] : array($opt['VpcSecurityGroupIds']))\n        ), 'member'));\n        unset($opt['VpcSecurityGroupIds']);\n    }\n\n    return $this-&gt;authenticate('ModifyDBInstance', $opt);\n}","examples":{"example":{"title":"<p>Modify database instance.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;modify_db_instance('myInstance', array(\n\t'AllocatedStorage' =&gt; 10,\n\t'DBSecurityGroups' =&gt; 'default',\n\t'ApplyImmediately' =&gt; 'true',\n\t'DBParameterGroupName' =&gt; 'myParamGroup',\n\t'MultiAZ' =&gt; true,\n\t'AllowMajorVersionUpgrade' =&gt; true,\n\t'AutoMinorVersionUpgrade' =&gt; true,\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_instance\"><code>create_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/create_db_instance_read_replica.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_instance_read_replica\"><code>create_db_instance_read_replica()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_instance\"><code>delete_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_instances.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_instances\"><code>describe_db_instances()<\/code><\/a>","<a href=\"..\/amazonrds\/reboot_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/reboot_db_instance\"><code>reboot_db_instance()<\/code><\/a>"]}},{"name":"modify_db_parameter_group","modifier":"public","description":"<p>Modifies the parameters of a DBParameterGroup. To modify more than one parameter submit a list\nof the following: ParameterName, ParameterValue, and ApplyMethod. A maximum of 20 parameters\ncan be modified in a single request.<\/p>\n\n<p class=\"note\"><\/p>\n\n<p>The <code>apply-immediate<\/code> method can be used only for dynamic parameters; the\n<code>pending-reboot<\/code> method can be used with MySQL and Oracle DB Instances for either\ndynamic or static parameters. For Microsoft SQL Server DB Instances, the\n<code>pending-reboot<\/code> method can be used only for static parameters.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_parameter_group_name","type":"string","description":"<p>The name of the DB Parameter Group. Constraints:<\/p>\n\n<ul><li>Must be the name of an existing DB Parameter Group<\/li><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"req"},"name":"parameters","type":"array","description":"<p>An array of parameter names, values, and the apply method for the parameter update. At least one parameter name, value, and apply method must be supplied; subsequent arguments are optional. A maximum of 20 parameters may be modified in a single request. Valid Values (for the application method): <code>immediate | pending-reboot<\/code><\/p>\n\n<p class=\"note\">You can use the immediate value with dynamic parameters only. You can use the pending-reboot value for both dynamic and static parameters, and changes are applied when DB Instance reboots.<\/p>\n\n<ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>ParameterName<\/code> - <code>string<\/code> - Optional - Specifies the name of the parameter.<\/li><li><code>ParameterValue<\/code> - <code>string<\/code> - Optional - Specifies the value of the parameter.<\/li><li><code>Description<\/code> - <code>string<\/code> - Optional - Provides a description of the parameter.<\/li><li><code>Source<\/code> - <code>string<\/code> - Optional - Indicates the source of the parameter value.<\/li><li><code>ApplyType<\/code> - <code>string<\/code> - Optional - Specifies the engine specific parameters type.<\/li><li><code>DataType<\/code> - <code>string<\/code> - Optional - Specifies the valid data type for the parameter.<\/li><li><code>AllowedValues<\/code> - <code>string<\/code> - Optional - Specifies the valid range of values for the parameter.<\/li><li><code>IsModifiable<\/code> - <code>boolean<\/code> - Optional - Indicates whether (<code>true<\/code>) or not (<code>false<\/code>) the parameter can be modified. Some parameters have security or operational implications that prevent them from being changed.<\/li><li><code>MinimumEngineVersion<\/code> - <code>string<\/code> - Optional - The earliest engine version to which the parameter can apply.<\/li><li><code>ApplyMethod<\/code> - <code>string<\/code> - Optional - Indicates when to apply parameter updates. [Allowed values: <code>immediate<\/code>, <code>pending-reboot<\/code>]<\/li><\/ul><\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function modify_db_parameter_group($db_parameter_group_name, $parameters, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBParameterGroupName'] = $db_parameter_group_name;\n    \n    \/\/ Required list + map\n    $opt = array_merge($opt, CFComplexType::map(array(\n        'Parameters' =&gt; (is_array($parameters) ? $parameters : array($parameters))\n    ), 'member'));\n\n    return $this-&gt;authenticate('ModifyDBParameterGroup', $opt);\n}","examples":{"example":{"title":"<p>Modify database parameter group values.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;modify_db_parameter_group('myParamGroup', array(\n\tarray('ParameterName' =&gt; 'max_user_connections', 'ParameterValue' =&gt; '24',   'ApplyMethod' =&gt; 'pending-reboot'),\n\tarray('ParameterName' =&gt; 'max_allowed_packet',   'ParameterValue' =&gt; '1024', 'ApplyMethod' =&gt; 'immediate'),\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_parameter_group\"><code>create_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_parameter_group\"><code>delete_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_parameter_groups.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_parameter_groups\"><code>describe_db_parameter_groups()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_parameters.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_parameters\"><code>describe_db_parameters()<\/code><\/a>","<a href=\"..\/amazonrds\/reset_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/reset_db_parameter_group\"><code>reset_db_parameter_group()<\/code><\/a>"]}},{"name":"modify_db_subnet_group","modifier":"public","description":"<p>Modifies an existing DB subnet group. DB subnet groups must contain at least one subnet in at\nleast two AZs in the region.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_subnet_group_name","type":"string","description":"<p>The name for the DB Subnet Group. This value is stored as a lowercase string. Constraints: Must contain no more than 255 alphanumeric characters or hyphens. Must not be &#8220;Default&#8221;. Example: <code>mySubnetgroup<\/code><\/p>"},{"@attributes":{"choice":"req"},"name":"subnet_ids","type":"string|array","description":"<p>The EC2 Subnet IDs for the DB Subnet Group. Pass a string for a single value, or an indexed array for multiple values.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBSubnetGroupDescription<\/code> - <code>string<\/code> - Optional - The description for the DB Subnet Group.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function modify_db_subnet_group($db_subnet_group_name, $subnet_ids, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBSubnetGroupName'] = $db_subnet_group_name;\n    \n    \/\/ Required list (non-map)\n    $opt = array_merge($opt, CFComplexType::map(array(\n        'SubnetIds' =&gt; (is_array($subnet_ids) ? $subnet_ids : array($subnet_ids))\n    ), 'member'));\n\n    return $this-&gt;authenticate('ModifyDBSubnetGroup', $opt);\n}"},{"name":"modify_event_subscription","modifier":"public","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"subscription_name","type":"string","description":{}},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>SnsTopicArn<\/code> - <code>string<\/code> - Optional - <\/li><li><code>SourceType<\/code> - <code>string<\/code> - Optional - <\/li><li><code>EventCategories<\/code> - <code>string|array<\/code> - Optional -  Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>Enabled<\/code> - <code>boolean<\/code> - Optional - <\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function modify_event_subscription($subscription_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['SubscriptionName'] = $subscription_name;\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['EventCategories']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'EventCategories' =&gt; (is_array($opt['EventCategories']) ? $opt['EventCategories'] : array($opt['EventCategories']))\n        ), 'member'));\n        unset($opt['EventCategories']);\n    }\n\n    return $this-&gt;authenticate('ModifyEventSubscription', $opt);\n}"},{"name":"modify_option_group","modifier":"public","description":"<p>Modifies an existing Option Group.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"option_group_name","type":"string","description":"<p>The name of the option group to be modified.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>OptionsToInclude<\/code> - <code>array<\/code> - Optional - Options in this list are added to the Option Group or, if already present, the specified configuration is used to update the existing configuration. <ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>OptionName<\/code> - <code>string<\/code> - Required - The configuration of options to include in a group.<\/li><li><code>Port<\/code> - <code>integer<\/code> - Optional - The optional port for the option.<\/li><li><code>DBSecurityGroupMemberships<\/code> - <code>string|array<\/code> - Optional - A list of DBSecurityGroupMemebrship name strings used for this option. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>VpcSecurityGroupMemberships<\/code> - <code>string|array<\/code> - Optional - A list of VpcSecurityGroupMemebrship name strings used for this option. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>OptionSettings<\/code> - <code>array<\/code> - Optional - A list of option settings applied for this option. <ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>Name<\/code> - <code>string<\/code> - Optional - The name of the setting.<\/li><li><code>Value<\/code> - <code>string<\/code> - Optional - The value of this setting.<\/li><li><code>DefaultValue<\/code> - <code>string<\/code> - Optional - Default value for this setting.<\/li><li><code>Description<\/code> - <code>string<\/code> - Optional - The description of the setting.<\/li><li><code>ApplyType<\/code> - <code>string<\/code> - Optional - Specifies the apply type for this setting.<\/li><li><code>DataType<\/code> - <code>string<\/code> - Optional - Specifies the valid data type of this setting<\/li><li><code>AllowedValues<\/code> - <code>string<\/code> - Optional - Specifies a valid list\/range of values allowed for this setting.<\/li><li><code>IsModifiable<\/code> - <code>boolean<\/code> - Optional - Indicates if the setting is modifiable or not.<\/li><li><code>IsCollection<\/code> - <code>boolean<\/code> - Optional - Indicates if the value for the setting can be a list of values or a single value.<\/li><\/ul><\/li><\/ul><\/li><\/ul><\/li><\/ul><\/li><li><code>OptionsToRemove<\/code> - <code>string|array<\/code> - Optional - Options in this list are removed from the Option Group. Pass a string for a single value, or an indexed array for multiple values.<\/li><li><code>ApplyImmediately<\/code> - <code>boolean<\/code> - Optional - Indicates whether the changes should be applied immediately, or during the next maintenance window for each instance associated with the Option Group.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function modify_option_group($option_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['OptionGroupName'] = $option_group_name;\n    \n    \/\/ Optional list + map\n    if (isset($opt['OptionsToInclude']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'OptionsToInclude' =&gt; $opt['OptionsToInclude']\n        ), 'member'));\n        unset($opt['OptionsToInclude']);\n    }\n    \n    \/\/ Optional list (non-map)\n    if (isset($opt['OptionsToRemove']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'OptionsToRemove' =&gt; (is_array($opt['OptionsToRemove']) ? $opt['OptionsToRemove'] : array($opt['OptionsToRemove']))\n        ), 'member'));\n        unset($opt['OptionsToRemove']);\n    }\n\n    return $this-&gt;authenticate('ModifyOptionGroup', $opt);\n}"},{"name":"promote_read_replica","modifier":"public","description":"<p>Promotes a Read Replica DB Instance to a standalone DB Instance.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_instance_identifier","type":"string","description":"<p>The DB Instance identifier. This value is stored as a lowercase string. Constraints:<\/p>\n\n<ul><li>Must be the identifier for an existing Read Replica DB Instance<\/li><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>\n\n<p>Example:<copy>mydbinstance<\/copy><\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>BackupRetentionPeriod<\/code> - <code>integer<\/code> - Optional - The number of days to retain automated backups. Setting this parameter to a positive number enables backups. Setting this parameter to 0 disables automated backups. Default: 1 Constraints:<ul><li>Must be a value from 0 to 8<\/li><\/ul><\/li><li><code>PreferredBackupWindow<\/code> - <code>string<\/code> - Optional - The daily time range during which automated backups are created if automated backups are enabled, using the <code>BackupRetentionPeriod<\/code> parameter. Default: A 30-minute window selected at random from an 8-hour block of time per region. The following list shows the time blocks for each region from which the default backup windows are assigned.<ul><li> <strong>US-East (Northern Virginia) Region:<\/strong> 03:00-11:00 UTC<\/li><li> <strong>US-West (Northern California) Region:<\/strong> 06:00-14:00 UTC<\/li><li> <strong>EU (Ireland) Region:<\/strong> 22:00-06:00 UTC<\/li><li> <strong>Asia Pacific (Singapore) Region:<\/strong> 14:00-22:00 UTC<\/li><li> <strong>Asia Pacific (Tokyo) Region:<\/strong> 17:00-03:00 UTC<\/li><\/ul>Constraints: Must be in the format <code>hh24:mi-hh24:mi<\/code>. Times should be Universal Time Coordinated (UTC). Must not conflict with the preferred maintenance window. Must be at least 30 minutes.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function promote_read_replica($db_instance_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBInstanceIdentifier'] = $db_instance_identifier;\n    \n    return $this-&gt;authenticate('PromoteReadReplica', $opt);\n}"},{"name":"purchase_reserved_db_instances_offering","modifier":"public","description":"<p>Purchases a reserved DB Instance offering.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"reserved_db_instances_offering_id","type":"string","description":"<p>The ID of the Reserved DB Instance offering to purchase. Example: 438012d3-4052-4cc7-b2e3-8d3372e0e706<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>ReservedDBInstanceId<\/code> - <code>string<\/code> - Optional - Customer-specified identifier to track this reservation. Example: myreservationID<\/li><li><code>DBInstanceCount<\/code> - <code>integer<\/code> - Optional - The number of instances to reserve. Default: <code>1<\/code><\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function purchase_reserved_db_instances_offering($reserved_db_instances_offering_id, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['ReservedDBInstancesOfferingId'] = $reserved_db_instances_offering_id;\n    \n    return $this-&gt;authenticate('PurchaseReservedDBInstancesOffering', $opt);\n}","related":{"method":["<a href=\"..\/amazonrds\/describe_reserved_db_instances.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_reserved_db_instances\"><code>describe_reserved_db_instances()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_reserved_db_instances_offerings.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_reserved_db_instances_offerings\"><code>describe_reserved_db_instances_offerings()<\/code><\/a>"]}},{"name":"reboot_db_instance","modifier":"public","description":"<p>Reboots a previously provisioned RDS instance. This API results in the application of modified\nDBParameterGroup parameters with ApplyStatus of pending-reboot to the RDS instance. This action\nis taken as soon as possible, and results in a momentary outage to the RDS instance during\nwhich the RDS instance status is set to rebooting. If the RDS instance is configured for\nMultiAZ, it is possible that the reboot will be conducted through a failover. A DBInstance\nevent is created when the reboot is completed.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_instance_identifier","type":"string","description":"<p>The DB Instance identifier. This parameter is stored as a lowercase string. Constraints:<\/p>\n\n<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>ForceFailover<\/code> - <code>boolean<\/code> - Optional - When <code>true<\/code>, the reboot will be conducted through a MultiAZ failover. Constraint: You cannot specify <code>true<\/code> if the instance is not configured for MultiAZ.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function reboot_db_instance($db_instance_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBInstanceIdentifier'] = $db_instance_identifier;\n    \n    return $this-&gt;authenticate('RebootDBInstance', $opt);\n}","examples":{"example":{"title":"<p>Reboot a database instance.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;reboot_db_instance('myInstance');\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_instance\"><code>create_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/create_db_instance_read_replica.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_instance_read_replica\"><code>create_db_instance_read_replica()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_instance\"><code>delete_db_instance()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_instances.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_instances\"><code>describe_db_instances()<\/code><\/a>","<a href=\"..\/amazonrds\/modify_db_instance.html\" rel=\"..\/..\/index.html#m=amazonrds\/modify_db_instance\"><code>modify_db_instance()<\/code><\/a>"]}},{"name":"register_streaming_read_callback","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Register a callback function to execute whenever a data stream is read from using\n<a href=\"..\/cfrequest\/streaming_read_callback.html\" rel=\"..\/..\/index.html#m=cfrequest\/streaming_read_callback\"><code>CFRequest::streaming_read_callback()<\/code><\/a>.<\/p>\n\n<p>The user-defined callback function should accept three arguments:<\/p>\n\n<ul><li><code>$curl_handle<\/code> - <code>resource<\/code> - Required - The cURL handle resource that represents the in-progress transfer.<\/li><li><code>$file_handle<\/code> - <code>resource<\/code> - Required - The file handle resource that represents the file on the local file system.<\/li><li><code>$length<\/code> - <code>integer<\/code> - Required - The length in kilobytes of the data chunk that was transferred.<\/li><\/ul>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"callback","type":"string|array|function","description":"<p>The callback function is called by <a href=\"http:\/\/php.net\/call_user_func\"><code>call_user_func()<\/code><\/a>, so you can pass the following values:<\/p>\n\n<ul><li>The name of a global function to execute, passed as a string.<\/li><li>A method to execute, passed as <code>array('ClassName', 'MethodName')<\/code>.<\/li><li>An anonymous function (PHP 5.3+).<\/li><\/ul>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function register_streaming_read_callback($callback)\n{\n    $this-&gt;registered_streaming_read_callback = $callback;\n    return $this;\n}","examples":{"example":{"title":"<p>Register a callback function to execute when data is read from an open stream.<\/p>","code":"$s3 = new AmazonS3();\n\n\/\/ Define callback function\nfunction callback($curl_handle, $file_handle, $length)\n{\n\t$total = curl_getinfo($curl_handle, CURLINFO_CONTENT_LENGTH_DOWNLOAD);\n\t$transferred = curl_getinfo($curl_handle, CURLINFO_SIZE_DOWNLOAD);\n\t$percentage = floor(($transferred \/ $total) * 100);\n\n\techo 'Transferred ' . $transferred . ' KB out of ' . $total . ' KB (' . $percentage . '%)' . PHP_EOL;\n}\n\n\/\/ Register a callback function to execute when a stream is read locally.\n$s3-&gt;register_streaming_read_callback('callback');\n\n\/\/ Download a public object.\n$response = $s3-&gt;create_object('my-bucket', 'my-large-file.mp4', array(\n\t'fileUpload' =&gt; '.\/uploads\/my-large-file.mp4'\n));","result":"Transferred 7200 bytes out of 9000000 (0%)\n...\nTransferred 9000000 bytes out of 9000000 (100%)"}}},{"name":"register_streaming_write_callback","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Register a callback function to execute whenever a data stream is written to using\n<a href=\"..\/cfrequest\/streaming_write_callback.html\" rel=\"..\/..\/index.html#m=cfrequest\/streaming_write_callback\"><code>CFRequest::streaming_write_callback()<\/code><\/a>.<\/p>\n\n<p>The user-defined callback function should accept two arguments:<\/p>\n\n<ul><li><code>$curl_handle<\/code> - <code>resource<\/code> - Required - The cURL handle resource that represents the in-progress transfer.<\/li><li><code>$length<\/code> - <code>integer<\/code> - Required - The length in kilobytes of the data chunk that was transferred.<\/li><\/ul>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"callback","type":"string|array|function","description":"<p>The callback function is called by <a href=\"http:\/\/php.net\/call_user_func\"><code>call_user_func()<\/code><\/a>, so you can pass the following values:<\/p>\n\n<ul><li>The name of a global function to execute, passed as a string.<\/li><li>A method to execute, passed as <code>array('ClassName', 'MethodName')<\/code>.<\/li><li>An anonymous function (PHP 5.3+).<\/li><\/ul>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function register_streaming_write_callback($callback)\n{\n    $this-&gt;registered_streaming_write_callback = $callback;\n    return $this;\n}","examples":{"example":{"title":"<p>Register a callback function to execute when data is written to an open stream.<\/p>","code":"$s3 = new AmazonS3();\n\n\/\/ Define callback function\nfunction callback($curl_handle, $length)\n{\n\t$total = curl_getinfo($curl_handle, CURLINFO_CONTENT_LENGTH_DOWNLOAD);\n\t$transferred = curl_getinfo($curl_handle, CURLINFO_SIZE_DOWNLOAD);\n\t$percentage = floor(($transferred \/ $total) * 100);\n\n\techo 'Transferred ' . $transferred . ' KB out of ' . $total . ' KB (' . $percentage . '%)' . PHP_EOL;\n}\n\n\/\/ Register a callback function to execute when a stream is written locally.\n$s3-&gt;register_streaming_write_callback('callback');\n\n\/\/ Download a public object.\n$response = $s3-&gt;get_object('my-bucket', 'my-large-file.mp4', array(\n\t'fileDownload' =&gt; '.\/downloads\/my-large-file.mp4'\n));","result":"Transferred 7200 bytes out of 9000000 (0%)\n...\nTransferred 9000000 bytes out of 9000000 (100%)"}}},{"name":"remove_source_identifier_from_subscription","modifier":"public","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"subscription_name","type":"string","description":{}},{"@attributes":{"choice":"req"},"name":"source_identifier","type":"string","description":{}},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function remove_source_identifier_from_subscription($subscription_name, $source_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['SubscriptionName'] = $subscription_name;\n    $opt['SourceIdentifier'] = $source_identifier;\n    \n    return $this-&gt;authenticate('RemoveSourceIdentifierFromSubscription', $opt);\n}"},{"name":"remove_tags_from_resource","modifier":"public","description":"<p>Removes metadata tags from a DB Instance.<\/p>\n\n<p>For an overview on tagging DB Instances, see <a href=\n\"http:\/\/docs.amazonwebservices.com\/AmazonRDS\/latest\/UserGuide\/Overview.Tagging.html\">DB\nInstance Tags.<\/a><\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"resource_name","type":"string","description":"<p>The DB Instance the tags will be removed from.<\/p>"},{"@attributes":{"choice":"req"},"name":"tag_keys","type":"string|array","description":"<p>The tag key (name) of the tag to be removed. Pass a string for a single value, or an indexed array for multiple values.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function remove_tags_from_resource($resource_name, $tag_keys, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['ResourceName'] = $resource_name;\n    \n    \/\/ Required list (non-map)\n    $opt = array_merge($opt, CFComplexType::map(array(\n        'TagKeys' =&gt; (is_array($tag_keys) ? $tag_keys : array($tag_keys))\n    ), 'member'));\n\n    return $this-&gt;authenticate('RemoveTagsFromResource', $opt);\n}"},{"name":"reset_db_parameter_group","modifier":"public","description":"<p>Modifies the parameters of a DBParameterGroup to the engine\/system default value. To reset\nspecific parameters submit a list of the following: ParameterName and ApplyMethod. To reset the\nentire DBParameterGroup specify the DBParameterGroup name and ResetAllParameters parameters.\nWhen resetting the entire group, dynamic parameters are updated immediately and static\nparameters are set to pending-reboot to take effect on the next DB instance restart or\nRebootDBInstance request.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_parameter_group_name","type":"string","description":"<p>The name of the DB Parameter Group. Constraints:<\/p>\n\n<ul><li>Must be 1 to 255 alphanumeric characters<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>ResetAllParameters<\/code> - <code>boolean<\/code> - Optional - Specifies whether (<code>true<\/code>) or not (<code>false<\/code>) to reset all parameters in the DB Parameter Group to default values. Default: <code>true<\/code><\/li><li><code>Parameters<\/code> - <code>array<\/code> - Optional - An array of parameter names, values, and the apply method for the parameter update. At least one parameter name, value, and apply method must be supplied; subsequent arguments are optional. A maximum of 20 parameters may be modified in a single request. <strong>MySQL<\/strong> Valid Values (for Apply method): <code>immediate<\/code> | <code>pending-reboot<\/code> You can use the immediate value with dynamic parameters only. You can use the <code>pending-reboot<\/code> value for both dynamic and static parameters, and changes are applied when DB Instance reboots. <strong>Oracle<\/strong> Valid Values (for Apply method): <code>pending-reboot<\/code> <ul><li><code>x<\/code> - <code>array<\/code> - Optional - This represents a simple array index. <ul><li><code>ParameterName<\/code> - <code>string<\/code> - Optional - Specifies the name of the parameter.<\/li><li><code>ParameterValue<\/code> - <code>string<\/code> - Optional - Specifies the value of the parameter.<\/li><li><code>Description<\/code> - <code>string<\/code> - Optional - Provides a description of the parameter.<\/li><li><code>Source<\/code> - <code>string<\/code> - Optional - Indicates the source of the parameter value.<\/li><li><code>ApplyType<\/code> - <code>string<\/code> - Optional - Specifies the engine specific parameters type.<\/li><li><code>DataType<\/code> - <code>string<\/code> - Optional - Specifies the valid data type for the parameter.<\/li><li><code>AllowedValues<\/code> - <code>string<\/code> - Optional - Specifies the valid range of values for the parameter.<\/li><li><code>IsModifiable<\/code> - <code>boolean<\/code> - Optional - Indicates whether (<code>true<\/code>) or not (<code>false<\/code>) the parameter can be modified. Some parameters have security or operational implications that prevent them from being changed.<\/li><li><code>MinimumEngineVersion<\/code> - <code>string<\/code> - Optional - The earliest engine version to which the parameter can apply.<\/li><li><code>ApplyMethod<\/code> - <code>string<\/code> - Optional - Indicates when to apply parameter updates. [Allowed values: <code>immediate<\/code>, <code>pending-reboot<\/code>]<\/li><\/ul><\/li><\/ul><\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function reset_db_parameter_group($db_parameter_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBParameterGroupName'] = $db_parameter_group_name;\n    \n    \/\/ Optional list + map\n    if (isset($opt['Parameters']))\n    {\n        $opt = array_merge($opt, CFComplexType::map(array(\n            'Parameters' =&gt; $opt['Parameters']\n        ), 'member'));\n        unset($opt['Parameters']);\n    }\n\n    return $this-&gt;authenticate('ResetDBParameterGroup', $opt);\n}","examples":{"example":{"title":"<p>Reset database parameter group values.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;reset_db_parameter_group('myParamGroup', array(\n\t'Parameters' =&gt; array(\n\t\tarray('ParameterName' =&gt; 'max_user_connections', 'ApplyMethod' =&gt; 'pending-reboot'),\n\t\tarray('ParameterName' =&gt; 'max_allowed_packet',   'ApplyMethod' =&gt; 'immediate'),\n\t)\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}},"related":{"method":["<a href=\"..\/amazonrds\/create_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_parameter_group\"><code>create_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_parameter_group\"><code>delete_db_parameter_group()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_parameter_groups.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_parameter_groups\"><code>describe_db_parameter_groups()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_parameters.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_parameters\"><code>describe_db_parameters()<\/code><\/a>","<a href=\"..\/amazonrds\/modify_db_parameter_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/modify_db_parameter_group\"><code>modify_db_parameter_group()<\/code><\/a>"]}},{"name":"restore_db_instance_from_db_snapshot","modifier":"public","description":"<p>Creates a new DB Instance from a DB snapshot. The target database is created from the source\ndatabase restore point with the same configuration as the original source database, except that\nthe new RDS instance is created with the default security group.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_instance_identifier","type":"string","description":"<p>The identifier for the DB Snapshot to restore from. Constraints:<\/p>\n\n<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"req"},"name":"db_snapshot_identifier","type":"string","description":"<p>Name of the DB Instance to create from the DB Snapshot. This parameter isn&#8217;t case sensitive. Constraints:<\/p>\n\n<ul><li>Must contain from 1 to 255 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>\n\n<p>Example: <code>my-snapshot-id<\/code><\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>DBInstanceClass<\/code> - <code>string<\/code> - Optional - The compute and memory capacity of the Amazon RDS DB instance. Valid Values: <code>db.t1.micro | db.m1.small | db.m1.medium | db.m1.large | db.m1.xlarge | db.m2.2xlarge | db.m2.4xlarge<\/code><\/li><li><code>Port<\/code> - <code>integer<\/code> - Optional - The port number on which the database accepts connections. Default: The same port as the original DB Instance Constraints: Value must be <code>1150-65535<\/code><\/li><li><code>AvailabilityZone<\/code> - <code>string<\/code> - Optional - The EC2 Availability Zone that the database instance will be created in. Default: A random, system-chosen Availability Zone. Constraint: You cannot specify the AvailabilityZone parameter if the MultiAZ parameter is set to <code>true<\/code>. Example: <code>us-east-1a<\/code><\/li><li><code>DBSubnetGroupName<\/code> - <code>string<\/code> - Optional - The DB Subnet Group name to use for the new instance.<\/li><li><code>MultiAZ<\/code> - <code>boolean<\/code> - Optional - Specifies if the DB Instance is a Multi-AZ deployment. Constraint: You cannot specify the AvailabilityZone parameter if the MultiAZ parameter is set to <code>true<\/code>.<\/li><li><code>PubliclyAccessible<\/code> - <code>boolean<\/code> - Optional - <\/li><li><code>AutoMinorVersionUpgrade<\/code> - <code>boolean<\/code> - Optional - Indicates that minor version upgrades will be applied automatically to the DB Instance during the maintenance window.<\/li><li><code>LicenseModel<\/code> - <code>string<\/code> - Optional - License model information for the restored DB Instance. Default: Same as source. Valid values: <code>license-included<\/code> | <code>bring-your-own-license<\/code> | <code>general-public-license<\/code><\/li><li><code>DBName<\/code> - <code>string<\/code> - Optional - The database name for the restored DB Instance. <p class=\"note\">This parameter doesn&#8217;t apply to the MySQL engine.<\/p><\/li><li><code>Engine<\/code> - <code>string<\/code> - Optional - The database engine to use for the new instance. Default: The same as source Constraint: Must be compatible with the engine of the source Example: <code>oracle-ee<\/code><\/li><li><code>Iops<\/code> - <code>integer<\/code> - Optional - The amount of Provisioned IOPS (input\/output operations per second) to be initially allocated for the DB Instance. Constraints: Must be an integer greater than 1000.<\/li><li><code>OptionGroupName<\/code> - <code>string<\/code> - Optional - <\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function restore_db_instance_from_db_snapshot($db_instance_identifier, $db_snapshot_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBInstanceIdentifier'] = $db_instance_identifier;\n    $opt['DBSnapshotIdentifier'] = $db_snapshot_identifier;\n    \n    return $this-&gt;authenticate('RestoreDBInstanceFromDBSnapshot', $opt);\n}","examples":{"example":{"title":"<p>Restore a database instance from a database snapshot.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;restore_db_instance_from_db_snapshot('mySnapshot', 'mySnapshot', array(\n\t'DBInstanceClass' =&gt; 'db.m1.small',\n\t'AutoMinorVersionUpgrade' =&gt; 'true'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());"}},"related":{"method":"<a href=\"..\/amazonrds\/restore_db_instance_to_point_in_time.html\" rel=\"..\/..\/index.html#m=amazonrds\/restore_db_instance_to_point_in_time\"><code>restore_db_instance_to_point_in_time()<\/code><\/a>"}},{"name":"restore_db_instance_to_point_in_time","modifier":"public","description":"<p>Restores a DB Instance to an arbitrary point-in-time. Users can restore to any point in time\nbefore the latestRestorableTime for up to backupRetentionPeriod days. The target database is\ncreated from the source database with the same configuration as the original database except\nthat the DB instance is created with the default DB security group.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"source_db_instance_identifier","type":"string","description":"<p>The identifier of the source DB Instance from which to restore. Constraints:<\/p>\n\n<ul><li>Must be the identifier of an existing database instance<\/li><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"req"},"name":"target_db_instance_identifier","type":"string","description":"<p>The name of the new database instance to be created. Constraints:<\/p>\n\n<ul><li>Must contain from 1 to 63 alphanumeric characters or hyphens<\/li><li>First character must be a letter<\/li><li>Cannot end with a hyphen or contain two consecutive hyphens<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>RestoreTime<\/code> - <code>string<\/code> - Optional - The date and time to restore from. Valid Values: Value must be a UTC time Constraints:<ul><li>Must be before the latest restorable time for the DB Instance<\/li><li>Cannot be specified if UseLatestRestorableTime parameter is true<\/li><\/ul>Example: <code>2009-09-07T23:45:00Z<\/code> May be passed as a number of seconds since UNIX Epoch, or any string compatible with <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a>.<\/li><li><code>UseLatestRestorableTime<\/code> - <code>boolean<\/code> - Optional - Specifies whether (<code>true<\/code>) or not (<code>false<\/code>) the DB Instance is restored from the latest backup time. Default: <code>false<\/code> Constraints: Cannot be specified if RestoreTime parameter is provided.<\/li><li><code>DBInstanceClass<\/code> - <code>string<\/code> - Optional - The compute and memory capacity of the Amazon RDS DB instance. Valid Values: <code>db.t1.micro | db.m1.small | db.m1.medium | db.m1.large | db.m1.xlarge | db.m2.2xlarge | db.m2.4xlarge<\/code> Default: The same DBInstanceClass as the original DB Instance.<\/li><li><code>Port<\/code> - <code>integer<\/code> - Optional - The port number on which the database accepts connections. Constraints: Value must be <code>1150-65535<\/code> Default: The same port as the original DB Instance.<\/li><li><code>AvailabilityZone<\/code> - <code>string<\/code> - Optional - The EC2 Availability Zone that the database instance will be created in. Default: A random, system-chosen Availability Zone. Constraint: You cannot specify the AvailabilityZone parameter if the MultiAZ parameter is set to true. Example: <code>us-east-1a<\/code><\/li><li><code>DBSubnetGroupName<\/code> - <code>string<\/code> - Optional - The DB subnet group name to use for the new instance.<\/li><li><code>MultiAZ<\/code> - <code>boolean<\/code> - Optional - Specifies if the DB Instance is a Multi-AZ deployment. Constraint: You cannot specify the AvailabilityZone parameter if the MultiAZ parameter is set to <code>true<\/code>.<\/li><li><code>PubliclyAccessible<\/code> - <code>boolean<\/code> - Optional - <\/li><li><code>AutoMinorVersionUpgrade<\/code> - <code>boolean<\/code> - Optional - Indicates that minor version upgrades will be applied automatically to the DB Instance during the maintenance window.<\/li><li><code>LicenseModel<\/code> - <code>string<\/code> - Optional - License model information for the restored DB Instance. Default: Same as source. Valid values: <code>license-included<\/code> | <code>bring-your-own-license<\/code> | <code>general-public-license<\/code><\/li><li><code>DBName<\/code> - <code>string<\/code> - Optional - The database name for the restored DB Instance. <p class=\"note\">This parameter is not used for the MySQL engine.<\/p><\/li><li><code>Engine<\/code> - <code>string<\/code> - Optional - The database engine to use for the new instance. Default: The same as source Constraint: Must be compatible with the engine of the source Example: <code>oracle-ee<\/code><\/li><li><code>Iops<\/code> - <code>integer<\/code> - Optional - The amount of Provisioned IOPS (input\/output operations per second) to be initially allocated for the DB Instance. Constraints: Must be an integer greater than 1000.<\/li><li><code>OptionGroupName<\/code> - <code>string<\/code> - Optional - <\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function restore_db_instance_to_point_in_time($source_db_instance_identifier, $target_db_instance_identifier, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['SourceDBInstanceIdentifier'] = $source_db_instance_identifier;\n    $opt['TargetDBInstanceIdentifier'] = $target_db_instance_identifier;\n    \n    \/\/ Optional DateTime\n    if (isset($opt['RestoreTime']))\n    {\n        $opt['RestoreTime'] = $this-&gt;util-&gt;convert_date_to_iso8601($opt['RestoreTime']);\n    }\n\n    return $this-&gt;authenticate('RestoreDBInstanceToPointInTime', $opt);\n}","examples":{"example":{"title":"<p>Restore a database instance to a point in time.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;restore_db_instance_to_point_in_time('myInstance', 'myInstancerestore', array(\n\t'RestoreTime' =&gt; 'yesterday',\n\t'AutoMinorVersionUpgrade' =&gt; 'true'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());"}},"related":{"method":"<a href=\"..\/amazonrds\/restore_db_instance_from_db_snapshot.html\" rel=\"..\/..\/index.html#m=amazonrds\/restore_db_instance_from_db_snapshot\"><code>restore_db_instance_from_db_snapshot()<\/code><\/a>"}},{"name":"revoke_db_security_group_ingress","modifier":"public","description":"<p>Revokes ingress from a DBSecurityGroup for previously authorized IP ranges or EC2 or VPC\nSecurity Groups. Required parameters for this API are one of CIDRIP, EC2SecurityGroupId for\nVPC, or (EC2SecurityGroupOwnerId and either EC2SecurityGroupName or EC2SecurityGroupId).<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"db_security_group_name","type":"string","description":"<p>The name of the DB Security Group to revoke ingress from.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>CIDRIP<\/code> - <code>string<\/code> - Optional - The IP range to revoke access from. Must be a valid CIDR range. If <code>CIDRIP<\/code> is specified, <code>EC2SecurityGroupName<\/code>, <code>EC2SecurityGroupId<\/code> and <code>EC2SecurityGroupOwnerId<\/code> cannot be provided.<\/li><li><code>EC2SecurityGroupName<\/code> - <code>string<\/code> - Optional - The name of the EC2 Security Group to revoke access from. For VPC DB Security Groups, <code>EC2SecurityGroupId<\/code> must be provided. Otherwise, EC2SecurityGroupOwnerId and either <code>EC2SecurityGroupName<\/code> or <code>EC2SecurityGroupId<\/code> must be provided.<\/li><li><code>EC2SecurityGroupId<\/code> - <code>string<\/code> - Optional - The id of the EC2 Security Group to revoke access from. For VPC DB Security Groups, <code>EC2SecurityGroupId<\/code> must be provided. Otherwise, EC2SecurityGroupOwnerId and either <code>EC2SecurityGroupName<\/code> or <code>EC2SecurityGroupId<\/code> must be provided.<\/li><li><code>EC2SecurityGroupOwnerId<\/code> - <code>string<\/code> - Optional - The AWS Account Number of the owner of the EC2 security group specified in the <code>EC2SecurityGroupName<\/code> parameter. The AWS Access Key ID is not an acceptable value. For VPC DB Security Groups, <code>EC2SecurityGroupId<\/code> must be provided. Otherwise, EC2SecurityGroupOwnerId and either <code>EC2SecurityGroupName<\/code> or <code>EC2SecurityGroupId<\/code> must be provided.<\/li><li><code>curlopts<\/code> - <code>array<\/code> - Optional - A set of values to pass directly into <code>curl_setopt()<\/code>, where the key is a pre-defined <code>CURLOPT_*<\/code> constant.<\/li><li><code>returnCurlHandle<\/code> - <code>boolean<\/code> - Optional - A private toggle specifying that the cURL handle be returned rather than actually completing the request. This toggle is useful for manually managed batch requests.<\/li><\/ul>"}]},"returnvalue":"<p>A <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object containing a parsed HTTP response.<\/p>","source":"public function revoke_db_security_group_ingress($db_security_group_name, $opt = null)\n{\n    if (!$opt) $opt = array();\n    $opt['DBSecurityGroupName'] = $db_security_group_name;\n    \n    return $this-&gt;authenticate('RevokeDBSecurityGroupIngress', $opt);\n}","examples":{"example":[{"title":"<p>Revoke database security group access for a given security group.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;revoke_db_security_group_ingress('mySecurityGroup', array(\n\t'EC2SecurityGroupName' =&gt; 'default',\n\t'EC2SecurityGroupOwnerId' =&gt; CFCredentials::get('@default')-&gt;account_id\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"},{"title":"<p>Revoke database security group access for a CIDR IP Range.<\/p>","code":"\/\/ Instantiate the class\n$rds = new AmazonRDS();\n\n$response = $rds-&gt;revoke_db_security_group_ingress('mySecurityGroup', array(\n\t'CIDRIP' =&gt; '0.0.0.0\/0'\n));\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}]},"related":{"method":["<a href=\"..\/amazonrds\/authorize_db_security_group_ingress.html\" rel=\"..\/..\/index.html#m=amazonrds\/authorize_db_security_group_ingress\"><code>authorize_db_security_group_ingress()<\/code><\/a>","<a href=\"..\/amazonrds\/create_db_security_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/create_db_security_group\"><code>create_db_security_group()<\/code><\/a>","<a href=\"..\/amazonrds\/delete_db_security_group.html\" rel=\"..\/..\/index.html#m=amazonrds\/delete_db_security_group\"><code>delete_db_security_group()<\/code><\/a>","<a href=\"..\/amazonrds\/describe_db_security_groups.html\" rel=\"..\/..\/index.html#m=amazonrds\/describe_db_security_groups\"><code>describe_db_security_groups()<\/code><\/a>"]}},{"name":"send","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Executes the batch request queue by sending all queued requests.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"clear_after_send","type":"boolean","description":"<p>Whether or not to clear the batch queue after sending a request. Defaults to <code>true<\/code>. Set this to <code>false<\/code> if you are caching batch responses and want to retrieve results later.<\/p>"}},"returnvalue":"<p>An array of <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> objects.<\/p>","source":"public function send($clear_after_send = true)\n{\n    if ($this-&gt;use_batch_flow)\n    {\n        \/\/ When we send the request, disable batch flow.\n        $this-&gt;use_batch_flow = false;\n\n        \/\/ If we're not caching, simply send the request.\n        if (!$this-&gt;use_cache_flow)\n        {\n            $response = $this-&gt;batch_object-&gt;send();\n            $parsed_data = array_map(array($this, 'parse_callback'), $response);\n            $parsed_data = new CFArray($parsed_data);\n\n            \/\/ Clear the queue\n            if ($clear_after_send)\n            {\n                $this-&gt;batch_object-&gt;queue = array();\n            }\n\n            return $parsed_data;\n        }\n\n        \/\/ Generate an identifier specific to this particular set of arguments.\n        $cache_id = $this-&gt;key . '_' . get_class($this) . '_' . sha1(serialize($this-&gt;batch_object));\n\n        \/\/ Instantiate the appropriate caching object.\n        $this-&gt;cache_object = new $this-&gt;cache_class($cache_id, $this-&gt;cache_location, $this-&gt;cache_expires, $this-&gt;cache_compress);\n\n        if ($this-&gt;delete_cache)\n        {\n            $this-&gt;use_cache_flow = false;\n            $this-&gt;delete_cache = false;\n            return $this-&gt;cache_object-&gt;delete();\n        }\n\n        \/\/ Invoke the cache callback function to determine whether to pull data from the cache or make a fresh request.\n        $data_set = $this-&gt;cache_object-&gt;response_manager(array($this, 'cache_callback_batch'), array($this-&gt;batch_object));\n        $parsed_data = array_map(array($this, 'parse_callback'), $data_set);\n        $parsed_data = new CFArray($parsed_data);\n\n        \/\/ Clear the queue\n        if ($clear_after_send)\n        {\n            $this-&gt;batch_object-&gt;queue = array();\n        }\n\n        \/\/ End!\n        return $parsed_data;\n    }\n\n    \/\/ Load the class\n    $null = new CFBatchRequest();\n    unset($null);\n\n    throw new CFBatchRequest_Exception('You must use $object-&gt;batch()-&gt;send()');\n}"},{"name":"set_batch_class","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_batch_class($class = 'CFBatchRequest')\n{\n    $this-&gt;batch_class = $class;\n    return $this;\n}"},{"name":"set_cache_config","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set the caching configuration to use for response caching.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"location","type":"string","description":"<p>The location to store the cache object in. This may vary by cache method.<\/p><ul><li>File - The local file system paths such as <code>.\/cache<\/code> (relative) or <code>\/tmp\/cache\/<\/code> (absolute). The location must be server-writable.<\/li><li>APC - Pass in <code>apc<\/code> to use this lightweight cache. You must have the <a href=\"http:\/\/php.net\/apc\">APC extension<\/a> installed.<\/li><li>XCache - Pass in <code>xcache<\/code> to use this lightweight cache. You must have the <a href=\"http:\/\/xcache.lighttpd.net\">XCache<\/a> extension installed.<\/li><li>Memcached - Pass in an indexed array of associative arrays. Each associative array should have a <code>host<\/code> and a <code>port<\/code> value representing a <a href=\"http:\/\/php.net\/memcached\">Memcached<\/a> server to connect to.<\/li><li>PDO - A URL-style string (e.g. <code>pdo.mysql:\/\/user:pass@localhost\/cache<\/code>) or a standard DSN-style string (e.g. <code>pdo.sqlite:\/sqlite\/cache.db<\/code>). MUST be prefixed with <code>pdo.<\/code>. See <code>CachePDO<\/code> and <a href=\"http:\/\/php.net\/pdo\">PDO<\/a> for more details.<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"gzip","type":"boolean","description":"<p>Whether or not data should be gzipped before being stored. A value of <code>true<\/code> will compress the contents before caching them. A value of <code>false<\/code> will leave the contents uncompressed. Defaults to <code>true<\/code>.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_cache_config($location, $gzip = true)\n{\n    \/\/ If location is empty, don't do anything.\n    if (empty($location))\n    {\n        return $this;\n    }\n\n    \/\/ If we have an array, we're probably passing in Memcached servers and ports.\n    if (is_array($location))\n    {\n        $this-&gt;cache_class = 'CacheMC';\n    }\n    else\n    {\n        \/\/ I would expect locations like `\/tmp\/cache`, `pdo.mysql:\/\/user:pass@hostname:port`, `pdo.sqlite:memory:`, and `apc`.\n        $type = strtolower(substr($location, 0, 3));\n        switch ($type)\n        {\n            case 'apc':\n                $this-&gt;cache_class = 'CacheAPC';\n                break;\n\n            case 'xca': \/\/ First three letters of `xcache`\n                $this-&gt;cache_class = 'CacheXCache';\n                break;\n\n            case 'pdo':\n                $this-&gt;cache_class = 'CachePDO';\n                $location = substr($location, 4);\n                break;\n\n            default:\n                $this-&gt;cache_class = 'CacheFile';\n                break;\n        }\n    }\n\n    \/\/ Set the remaining cache information.\n    $this-&gt;cache_location = $location;\n    $this-&gt;cache_compress = $gzip;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Fire a single request, and then cache the response to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Fire a single request, and then cache the response to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"}]}},{"name":"set_hostname","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set the hostname to connect to. This is useful for alternate services that are API-compatible with\nAWS, but run from a different hostname.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"hostname","type":"string","description":"<p>The alternate hostname to use in place of the default one. Useful for mock or test applications living on different hostnames.<\/p>"},{"@attributes":{"choice":"opt"},"name":"port_number","type":"integer","description":"<p>The alternate port number to use in place of the default one. Useful for mock or test applications living on different port numbers.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_hostname($hostname, $port_number = null)\n{\n    if ($this-&gt;override_hostname)\n    {\n        $this-&gt;hostname = $hostname;\n\n        if ($port_number)\n        {\n            $this-&gt;port_number = $port_number;\n            $this-&gt;hostname .= ':' . (string) $this-&gt;port_number;\n        }\n    }\n\n    return $this;\n}","examples":{"example":{"title":"<p>Configure hostname and resource prefix.<\/p>","code":"\/\/ Instantiate the class\n$ec2 = new AmazonEC2();\n\n\/\/ Point to the new endpoint\n$ec2-&gt;set_hostname('localhost', '3000');\n$ec2-&gt;set_resource_prefix('\/services\/mock');\n$ec2-&gt;allow_hostname_override(false);\n\n\/\/ Describe images\n$response = $ec2-&gt;describe_images();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"set_max_retries","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Sets the maximum number of times to retry failed requests.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"retries","type":"integer","description":"<p>The maximum number of times to retry failed requests. Defaults to <code>3<\/code>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_max_retries($retries = 3)\n{\n    $this-&gt;max_retries = $retries;\n    return $this;\n}","examples":{"example":{"title":"<p>Sets the maximum number of times to retry failed requests.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Adjust offset\n$sdb = new AmazonSDB();\n$sdb-&gt;set_max_retries(3);\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;max_retries);","result":"int(3)"}}},{"name":"set_parser_class","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfsimplexml\/index.html\" rel=\"..\/..\/index.html#i=cfsimplexml\"><code>CFSimpleXML<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_parser_class($class = 'CFSimpleXML')\n{\n    $this-&gt;parser_class = $class;\n    return $this;\n}"},{"name":"set_proxy","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set the proxy settings to use.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"proxy","type":"string","description":"<p>Accepts proxy credentials in the following format: <code>proxy:\/\/user:pass@hostname:port<\/code><\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_proxy($proxy)\n{\n    $this-&gt;proxy = $proxy;\n    return $this;\n}","examples":{"example":{"title":"<p>Set the proxy server to use for requests.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Adjust offset\n$sdb = new AmazonSDB();\n$sdb-&gt;set_proxy('proxy:\/\/user:pass@hostname:port');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;proxy);","result":"string(31) \"proxy:\/\/user:pass@hostname:port\""}}},{"name":"set_region","modifier":"public","description":"<p>This allows you to explicitly sets the region for the service to use.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"region","type":"string","description":"<p>The region to explicitly set. Available options are <a href=\"..\/amazonrds\/constants.html#region_us_e1\" rel=\"..\/..\/index.html#c=amazonrds\/region_us_e1\"><code>REGION_US_E1<\/code><\/a>, <a href=\"..\/amazonrds\/constants.html#region_us_w1\" rel=\"..\/..\/index.html#c=amazonrds\/region_us_w1\"><code>REGION_US_W1<\/code><\/a>, <a href=\"..\/amazonrds\/constants.html#region_us_w2\" rel=\"..\/..\/index.html#c=amazonrds\/region_us_w2\"><code>REGION_US_W2<\/code><\/a>, <a href=\"..\/amazonrds\/constants.html#region_eu_w1\" rel=\"..\/..\/index.html#c=amazonrds\/region_eu_w1\"><code>REGION_EU_W1<\/code><\/a>, <a href=\"..\/amazonrds\/constants.html#region_apac_se1\" rel=\"..\/..\/index.html#c=amazonrds\/region_apac_se1\"><code>REGION_APAC_SE1<\/code><\/a>, <a href=\"..\/amazonrds\/constants.html#region_apac_se2\" rel=\"..\/..\/index.html#c=amazonrds\/region_apac_se2\"><code>REGION_APAC_SE2<\/code><\/a>, <a href=\"..\/amazonrds\/constants.html#region_apac_ne1\" rel=\"..\/..\/index.html#c=amazonrds\/region_apac_ne1\"><code>REGION_APAC_NE1<\/code><\/a>, <a href=\"..\/amazonrds\/constants.html#region_sa_e1\" rel=\"..\/..\/index.html#c=amazonrds\/region_sa_e1\"><code>REGION_SA_E1<\/code><\/a>, <a href=\"..\/amazonrds\/constants.html#region_us_gov1\" rel=\"..\/..\/index.html#c=amazonrds\/region_us_gov1\"><code>REGION_US_GOV1<\/code><\/a>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_region($region)\n{\n    \/\/ @codeCoverageIgnoreStart\n    $this-&gt;set_hostname($region);\n    return $this;\n    \/\/ @codeCoverageIgnoreEnd\n}","examples":{"example":{"title":"<p>Set the region to use.<\/p>","code":"$rds = new AmazonRDS();\n\n\/\/ Set the region to US-East (Northern Virginia)\n$rds-&gt;set_region(AmazonRDS::REGION_US_E1);\n\n\/\/ Describe database engine versions\n$response = $rds-&gt;describe_db_engine_versions();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"set_request_class","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfrequest\/index.html\" rel=\"..\/..\/index.html#i=cfrequest\"><code>CFRequest<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":{},"source":"public function set_request_class($class = 'CFRequest')\n{\n    $this-&gt;request_class = $class;\n    return $this;\n}","examples":{"example":{"title":"<p>Set the request class to use in the toolkit.<\/p>","description":"<p>The new request class must extend from <code>CFRequest<\/code>.<\/p>\n\n<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Custom class to extend RequestCore\nclass TestRequestCore extends RequestCore\n{\n\tpublic function test_method()\n\t{\n\t\treturn true;\n\t}\n}\n\n\/\/ Instantiate class and set new class\n$sdb = new AmazonSDB();\n$sdb-&gt;set_request_class('TestRequestCore');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;request_class);","result":"string(15) \"TestRequestCore\""}}},{"name":"set_resource_prefix","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set the resource prefix to use. This method is useful for alternate services that are API-compatible\nwith AWS.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"prefix","type":"string","description":"<p>An alternate prefix to prepend to the resource path. Useful for mock or test applications.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_resource_prefix($prefix)\n{\n    $this-&gt;resource_prefix = $prefix;\n    return $this;\n}","examples":{"example":{"title":"<p>Configure hostname and resource prefix.<\/p>","code":"\/\/ Instantiate the class\n$ec2 = new AmazonEC2();\n\n\/\/ Point to the new endpoint\n$ec2-&gt;set_hostname('localhost', '3000');\n$ec2-&gt;set_resource_prefix('\/services\/mock');\n$ec2-&gt;allow_hostname_override(false);\n\n\/\/ Describe images\n$response = $ec2-&gt;describe_images();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"set_response_class","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_response_class($class = 'CFResponse')\n{\n    $this-&gt;response_class = $class;\n    return $this;\n}","examples":{"example":{"title":"<p>Set the response class to use in the toolkit.<\/p>","description":"<p>The new response class must extend from <code>CFResponse<\/code>.<\/p>\n\n<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Custom class to extend ResponseCore\nclass TestResponseCore extends ResponseCore\n{\n\tpublic function test_method()\n\t{\n\t\treturn true;\n\t}\n}\n\n\/\/ Instantiate class and set new class\n$sdb = new AmazonSDB();\n$sdb-&gt;set_response_class('TestResponseCore');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;response_class);","result":"string(16) \"TestResponseCore\""}}},{"name":"set_utilities_class","modifier":"public","inheritance":{"class":{"name":"CFRuntime"}},"description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfutilities\/index.html\" rel=\"..\/..\/index.html#i=cfutilities\"><code>CFUtilities<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_utilities_class($class = 'CFUtilities')\n{\n    $this-&gt;utilities_class = $class;\n    $this-&gt;util = new $this-&gt;utilities_class();\n    return $this;\n}","examples":{"example":{"title":"<p>Set the utilities class to use in the toolkit.<\/p>","description":"<p>The new utilities class must extend from <code>CFUtilities<\/code>.<\/p>\n\n<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Custom class to extend CFUtilities\nclass TestUtilities extends CFUtilities\n{\n\tpublic function test_method()\n\t{\n\t\treturn true;\n\t}\n}\n\n\/\/ Instantiate class and set new class\n$sdb = new AmazonSDB();\n$sdb-&gt;set_utilities_class('TestUtilities');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;util-&gt;test_method());","result":"bool(true)"}}}]},"constants":{"@attributes":{"count":"22"},"constant":[{"name":"BUILD","value":"20130314130000"},{"name":"DEFAULT_URL","value":"rds.us-east-1.amazonaws.com"},{"name":"NAME","value":"aws-sdk-php"},{"name":"REGION_APAC_NE1","value":"rds.ap-northeast-1.amazonaws.com"},{"name":"REGION_APAC_SE1","value":"rds.ap-southeast-1.amazonaws.com"},{"name":"REGION_APAC_SE2","value":"rds.ap-southeast-2.amazonaws.com"},{"name":"REGION_CALIFORNIA","value":"rds.us-west-1.amazonaws.com"},{"name":"REGION_EU_W1","value":"rds.eu-west-1.amazonaws.com"},{"name":"REGION_IRELAND","value":"rds.eu-west-1.amazonaws.com"},{"name":"REGION_OREGON","value":"rds.us-west-2.amazonaws.com"},{"name":"REGION_SAO_PAULO","value":"rds.sa-east-1.amazonaws.com"},{"name":"REGION_SA_E1","value":"rds.sa-east-1.amazonaws.com"},{"name":"REGION_SINGAPORE","value":"rds.ap-southeast-1.amazonaws.com"},{"name":"REGION_SYDNEY","value":"rds.ap-southeast-2.amazonaws.com"},{"name":"REGION_TOKYO","value":"rds.ap-northeast-1.amazonaws.com"},{"name":"REGION_US_E1","value":"rds.us-east-1.amazonaws.com"},{"name":"REGION_US_GOV1","value":"rds.us-gov-west-1.amazonaws.com"},{"name":"REGION_US_W1","value":"rds.us-west-1.amazonaws.com"},{"name":"REGION_US_W2","value":"rds.us-west-2.amazonaws.com"},{"name":"REGION_VIRGINIA","value":"rds.us-east-1.amazonaws.com"},{"name":"USERAGENT","value":"aws-sdk-php\/1.6.2 PHP\/5.4.9 curl\/7.15.5 openssl\/0.9.8e-fips-rhel5"},{"name":"VERSION","value":"1.6.2"}]},"properties":{"@attributes":{"count":"40"},"property":[{"name":"api_version","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The supported API version.<\/p>"},{"name":"auth_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The authentication class to use.<\/p>"},{"name":"auth_token","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The Amazon Authentication Token.<\/p>"},{"name":"batch_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFBatchRequest"}},"description":"<p>The default class to use for handling batch requests (defaults to <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a>).<\/p>"},{"name":"batch_object","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The current instantiated batch request object.<\/p>"},{"name":"cache_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The caching class to use.<\/p>"},{"name":"cache_compress","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The state of cache compression.<\/p>"},{"name":"cache_expires","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>When the cache should be considered stale.<\/p>"},{"name":"cache_location","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The caching location to use.<\/p>"},{"name":"cache_object","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The current instantiated cache object.<\/p>"},{"name":"credentials","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The credentials to use for authentication.<\/p>"},{"name":"debug_mode","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The state of the debug mode setting.<\/p>"},{"name":"delete_cache","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The state of the cache deletion setting.<\/p>"},{"name":"hostname","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The alternate hostname to use, if any.<\/p>"},{"name":"internal_batch_object","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The internally instantiated batch request object.<\/p>"},{"name":"key","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The Amazon API Key.<\/p>"},{"name":"max_retries","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"integer","value":"3"}},"description":"<p>The number of times to retry failed requests.<\/p>"},{"name":"operation","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The operation to execute.<\/p>"},{"name":"operation_prefix","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The string prefix to prepend to the operation name.<\/p>"},{"name":"override_hostname","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of the capability to override the hostname with <a href=\"..\/amazonrds\/set_hostname.html\" rel=\"..\/..\/index.html#m=amazonrds\/set_hostname\"><code>set_hostname()<\/code><\/a>.<\/p>"},{"name":"parse_the_response","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of whether the response should be parsed or not.<\/p>"},{"name":"parser_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFSimpleXML"}},"description":"<p>The default class to use for parsing XML (defaults to <a href=\"..\/cfsimplexml\/index.html\" rel=\"..\/..\/index.html#i=cfsimplexml\"><code>CFSimpleXML<\/code><\/a>).<\/p>"},{"name":"payload","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The payload to send.<\/p>"},{"name":"port_number","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The alternate port number to use, if any.<\/p>"},{"name":"proxy","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The proxy to use for connecting.<\/p>"},{"name":"redirects","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The number of times a request has been retried.<\/p>"},{"name":"registered_streaming_read_callback","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The user-defined callback function to call when a stream is read from.<\/p>"},{"name":"registered_streaming_write_callback","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The user-defined callback function to call when a stream is written to.<\/p>"},{"name":"request_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFRequest"}},"description":"<p>The default class to use for HTTP requests (defaults to <a href=\"..\/cfrequest\/index.html\" rel=\"..\/..\/index.html#i=cfrequest\"><code>CFRequest<\/code><\/a>).<\/p>"},{"name":"resource_prefix","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The alternate resource prefix to use, if any.<\/p>"},{"name":"response_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFResponse"}},"description":"<p>The default class to use for HTTP responses (defaults to <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a>).<\/p>"},{"name":"secret_key","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The Amazon API Secret Key.<\/p>"},{"name":"service","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>An identifier for the current AWS service.<\/p>"},{"name":"ssl_verification","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of SSL certificate verification.<\/p>"},{"name":"use_aws_query","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of whether auth should be handled as AWS Query.<\/p>"},{"name":"use_batch_flow","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The state of batch flow usage.<\/p>"},{"name":"use_cache_flow","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>The state of cache flow usage.<\/p>"},{"name":"use_ssl","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of SSL\/HTTPS use.<\/p>"},{"name":"util","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","description":"<p>Handle for the utility functions.<\/p>"},{"name":"utilities_class","inheritance":{"class":{"name":"CFRuntime","file":"sdk.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFUtilities"}},"description":"<p>The default class to use for utilities (defaults to <a href=\"..\/cfutilities\/index.html\" rel=\"..\/..\/index.html#i=cfutilities\"><code>CFUtilities<\/code><\/a>).<\/p>"}]}}}