{"file":{},"class":{"name":"AuthV2Query","file":"authentication\/signature_v2query.class.php","url":"..\/authv2query\/index.html","description":"<p>Implements support for Signature v2 (AWS Query).<\/p>","metadata":{"version":"2011.11.22","license":"See the included NOTICE.md file for more information.","copyright":"See the included NOTICE.md file for more information.","link":"http:\/\/aws.amazon.com\/php\/ PHP Developer Center"},"inheritance":{"class":{"name":"Signer","file":"authentication\/signer.abstract.php","url":"..\/signer\/index.html"}},"implements":{"interface":{"name":"Signable","file":"authentication\/signable.interface.php"}},"methods":{"@attributes":{"count":"2"},"method":[{"name":"__construct","modifier":"public","description":"<p>Constructs a new instance of the <a href=\"..\/authv2query\/index.html\" rel=\"..\/..\/index.html#i=authv2query\"><code>AuthV2Query<\/code><\/a> class.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"endpoint","type":"string","description":"<p>The endpoint to direct the request to.<\/p>"},{"@attributes":{"choice":"req"},"name":"operation","type":"string","description":"<p>The operation to execute as a result of this request.<\/p>"},{"@attributes":{"choice":"req"},"name":"payload","type":"array","description":"<p>The options to use as part of the payload in the request.<\/p>"},{"@attributes":{"choice":"req"},"name":"credentials","type":"CFCredential","description":"<p>The credentials to use for signing and making requests.<\/p>"}]},"returnvalue":{"@attributes":{"type":"void"}},"source":"public function __construct($endpoint, $operation, $payload, CFCredential $credentials)\n{\n    parent::__construct($endpoint, $operation, $payload, $credentials);\n}"},{"name":"authenticate","modifier":"public","description":"<p>Generates a cURL handle with all of the required authentication bits set.<\/p>","returnvalue":"<p>A cURL handle ready for executing.<\/p>","source":"public function authenticate()\n{\n    \/\/ Determine signing values\n    $current_time = time();\n    $date = gmdate(CFUtilities::DATE_FORMAT_RFC2616, $current_time);\n    $timestamp = gmdate(CFUtilities::DATE_FORMAT_ISO8601, $current_time);\n    $query = array();\n\n    \/\/ Do we have an authentication token?\n    if ($this-&gt;auth_token)\n    {\n        $headers['X-Amz-Security-Token'] = $this-&gt;auth_token;\n        $query['SecurityToken'] = $this-&gt;auth_token;\n    }\n\n    \/\/ Only add it if it exists.\n    if ($this-&gt;api_version)\n    {\n        $query['Version'] = $this-&gt;api_version;\n    }\n\n    $query['Action'] = $this-&gt;operation;\n    $query['AWSAccessKeyId'] = $this-&gt;key;\n    $query['SignatureMethod'] = 'HmacSHA256';\n    $query['SignatureVersion'] = 2;\n    $query['Timestamp'] = $timestamp;\n\n    \/\/ Merge in any options that were passed in\n    if (is_array($this-&gt;payload))\n    {\n        $query = array_merge($query, $this-&gt;payload);\n    }\n\n    \/\/ Do a case-sensitive, natural order sort on the array keys.\n    uksort($query, 'strcmp');\n\n    \/\/ Create the string that needs to be hashed.\n    $canonical_query_string = $this-&gt;util-&gt;to_signable_string($query);\n\n    \/\/ Remove the default scheme from the domain.\n    $domain = str_replace(array('http:\/\/', 'https:\/\/'), '', $this-&gt;endpoint);\n\n    \/\/ Parse our request.\n    $parsed_url = parse_url('http:\/\/' . $domain);\n\n    \/\/ Set the proper host header.\n    if (isset($parsed_url['port']) &amp;&amp; (integer) $parsed_url['port'] !== 80 &amp;&amp; (integer) $parsed_url['port'] !== 443)\n    {\n        $host_header = strtolower($parsed_url['host']) . ':' . $parsed_url['port'];\n    }\n    else\n    {\n        $host_header = strtolower($parsed_url['host']);\n    }\n\n    \/\/ Set the proper request URI.\n    $request_uri = isset($parsed_url['path']) ? $parsed_url['path'] : '\/';\n\n    \/\/ Prepare the string to sign\n    $this-&gt;string_to_sign = &quot;POST\\n$host_header\\n$request_uri\\n$canonical_query_string&quot;;\n\n    \/\/ Hash the AWS secret key and generate a signature for the request.\n    $query['Signature'] = base64_encode(hash_hmac('sha256', $this-&gt;string_to_sign, $this-&gt;secret_key, true));\n\n    \/\/ Generate the querystring from $query\n    $this-&gt;querystring = $this-&gt;util-&gt;to_query_string($query);\n\n    \/\/ Gather information to pass along to other classes.\n    $helpers = array(\n        'utilities' =&gt; $this-&gt;utilities_class,\n        'request' =&gt; $this-&gt;request_class,\n        'response' =&gt; $this-&gt;response_class,\n    );\n\n    \/\/ Compose the request.\n    $request_url = ($this-&gt;use_ssl ? 'https:\/\/' : 'http:\/\/') . $domain;\n    $request_url .= !isset($parsed_url['path']) ? '\/' : '';\n\n    \/\/ Instantiate the request class\n    $request = new $this-&gt;request_class($request_url, $this-&gt;proxy, $helpers, $this-&gt;credentials);\n    $request-&gt;set_method('POST');\n    $request-&gt;set_body($this-&gt;querystring);\n    $headers['Content-Type'] = 'application\/x-www-form-urlencoded; charset=utf-8';\n\n    \/\/ Pass along registered stream callbacks\n    if ($this-&gt;registered_streaming_read_callback)\n    {\n        $request-&gt;register_streaming_read_callback($this-&gt;registered_streaming_read_callback);\n    }\n\n    if ($this-&gt;registered_streaming_write_callback)\n    {\n        $request-&gt;register_streaming_write_callback($this-&gt;registered_streaming_write_callback);\n    }\n\n    \/\/ Sort headers\n    uksort($headers, 'strnatcasecmp');\n\n    \/\/ Add headers to request and compute the string to sign\n    foreach ($headers as $header_key =&gt; $header_value)\n    {\n        \/\/ Strip linebreaks from header values as they're illegal and can allow for security issues\n        $header_value = str_replace(array(&quot;\\r&quot;, &quot;\\n&quot;), '', $header_value);\n\n        \/\/ Add the header if it has a value\n        if ($header_value !== '')\n        {\n            $request-&gt;add_header($header_key, $header_value);\n        }\n    }\n\n    return $request;\n}"}]},"constants":{"@attributes":{"count":"0"}},"properties":{"@attributes":{"count":"4"},"property":[{"name":"credentials","inheritance":{"class":{"name":"Signer","file":"authentication\/signer.abstract.php"}},"modifier":"public","description":"<p>The credentials to use for signing and making requests.<\/p>"},{"name":"endpoint","inheritance":{"class":{"name":"Signer","file":"authentication\/signer.abstract.php"}},"modifier":"public","description":"<p>The endpoint to direct the request to.<\/p>"},{"name":"operation","inheritance":{"class":{"name":"Signer","file":"authentication\/signer.abstract.php"}},"modifier":"public","description":"<p>The operation to execute as a result of this request.<\/p>"},{"name":"payload","inheritance":{"class":{"name":"Signer","file":"authentication\/signer.abstract.php"}},"modifier":"public","description":"<p>The options to use as part of the payload in the request.<\/p>"}]}}}