{"file":{},"class":{"name":"AuthV4JSON","file":"authentication\/signature_v4json.class.php","url":"..\/authv4json\/index.html","description":"<p>Implements support for Signature v4 (Query).<\/p>","metadata":{"version":"2011.01.03","license":"See the included NOTICE.md file for more information.","copyright":"See the included NOTICE.md file for more information.","link":"http:\/\/aws.amazon.com\/php\/ PHP Developer Center"},"inheritance":{"class":{"name":"Signer","file":"authentication\/signer.abstract.php","url":"..\/signer\/index.html"}},"implements":{"interface":{"name":"Signable","file":"authentication\/signable.interface.php"}},"methods":{"@attributes":{"count":"14"},"method":[{"name":"__construct","modifier":"public","description":"<p>Constructs a new instance of the <a href=\"..\/authv4query\/index.html\" rel=\"..\/..\/index.html#i=authv4query\"><code>AuthV4Query<\/code><\/a> class.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"endpoint","type":"string","description":"<p>The endpoint to direct the request to.<\/p>"},{"@attributes":{"choice":"req"},"name":"operation","type":"string","description":"<p>The operation to execute as a result of this request.<\/p>"},{"@attributes":{"choice":"req"},"name":"payload","type":"array","description":"<p>The options to use as part of the payload in the request.<\/p>"},{"@attributes":{"choice":"req"},"name":"credentials","type":"CFCredential","description":"<p>The credentials to use for signing and making requests.<\/p>"}]},"returnvalue":{"@attributes":{"type":"void"}},"source":"public function __construct($endpoint, $operation, $payload, CFCredential $credentials)\n{\n    parent::__construct($endpoint, $operation, $payload, $credentials);\n}"},{"name":"authenticate","modifier":"public","description":"<p>Generates a cURL handle with all of the required authentication bits set.<\/p>","returnvalue":"<p>A cURL handle ready for executing.<\/p>","source":"public function authenticate()\n{\n    \/\/ Determine signing values\n    $current_time = time();\n    $timestamp = gmdate(CFUtilities::DATE_FORMAT_SIGV4, $current_time);\n\n    \/\/ Initialize\n    $x_amz_target = null;\n\n    $this-&gt;headers = array();\n    $this-&gt;signed_headers = array();\n    $this-&gt;canonical_headers = array();\n    $this-&gt;query = array();\n\n    \/\/ Prepare JSON structure\n    $this-&gt;body = json_encode($this-&gt;payload);\n    if ($this-&gt;body === '' || $this-&gt;body === '[]')\n    {\n        $this-&gt;body = '{}';\n    }\n\n    \/\/ Do we have an authentication token?\n    if ($this-&gt;auth_token)\n    {\n        $this-&gt;headers['X-Amz-Security-Token'] = $this-&gt;auth_token;\n        $this-&gt;query['SecurityToken'] = $this-&gt;auth_token;\n    }\n\n    \/\/ Manage the key-value pairs that are used in the query.\n    if (stripos($this-&gt;operation, 'x-amz-target') !== false)\n    {\n        $x_amz_target = trim(str_ireplace('x-amz-target:', '', $this-&gt;operation));\n    }\n    else\n    {\n        $this-&gt;query['Action'] = $this-&gt;operation;\n    }\n\n    \/\/ Only add it if it exists.\n    if ($this-&gt;api_version)\n    {\n        $this-&gt;query['Version'] = $this-&gt;api_version;\n    }\n\n    \/\/ Do a case-sensitive, natural order sort on the array keys.\n    uksort($this-&gt;query, 'strcmp');\n\n    \/\/ Remove the default scheme from the domain.\n    $domain = str_replace(array('http:\/\/', 'https:\/\/'), '', $this-&gt;endpoint);\n\n    \/\/ Parse our request.\n    $parsed_url = parse_url('http:\/\/' . $domain);\n\n    \/\/ Set the proper host header.\n    if (isset($parsed_url['port']) &amp;&amp; (integer) $parsed_url['port'] !== 80 &amp;&amp; (integer) $parsed_url['port'] !== 443)\n    {\n        $host_header = strtolower($parsed_url['host']) . ':' . $parsed_url['port'];\n    }\n    else\n    {\n        $host_header = strtolower($parsed_url['host']);\n    }\n\n    \/\/ Generate the querystring from $this-&gt;query\n    $this-&gt;querystring = $this-&gt;util-&gt;to_query_string($this-&gt;query);\n\n    \/\/ Gather information to pass along to other classes.\n    $helpers = array(\n        'utilities' =&gt; $this-&gt;utilities_class,\n        'request' =&gt; $this-&gt;request_class,\n        'response' =&gt; $this-&gt;response_class,\n    );\n\n    \/\/ Compose the request.\n    $request_url = ($this-&gt;use_ssl ? 'https:\/\/' : 'http:\/\/') . $domain;\n    $request_url .= !isset($parsed_url['path']) ? '\/' : '';\n\n    \/\/ Instantiate the request class\n    $request = new $this-&gt;request_class($request_url, $this-&gt;proxy, $helpers, $this-&gt;credentials);\n    $request-&gt;set_method('POST');\n    $request-&gt;set_body($this-&gt;body);\n    $this-&gt;querystring = $this-&gt;body;\n    $content_type = (stripos($this-&gt;operation, 'StorageGateway') !== false) ? 'application\/x-amz-json-1.1' : 'application\/x-amz-json-1.0';\n    $this-&gt;headers['Content-Type'] = $content_type;\n    $this-&gt;headers['X-Amz-Target'] = $x_amz_target;\n\n    \/\/ Pass along registered stream callbacks\n    if ($this-&gt;registered_streaming_read_callback)\n    {\n        $request-&gt;register_streaming_read_callback($this-&gt;registered_streaming_read_callback);\n    }\n\n    if ($this-&gt;registered_streaming_write_callback)\n    {\n        $request-&gt;register_streaming_write_callback($this-&gt;registered_streaming_write_callback);\n    }\n\n    \/\/ Add authentication headers\n    $this-&gt;headers['X-Amz-Date'] = $timestamp;\n    $this-&gt;headers['Content-Length'] = strlen($this-&gt;querystring);\n    $this-&gt;headers['Host'] = $host_header;\n\n    \/\/ Sort headers\n    uksort($this-&gt;headers, 'strnatcasecmp');\n\n    \/\/ Add headers to request and compute the string to sign\n    foreach ($this-&gt;headers as $header_key =&gt; $header_value)\n    {\n        \/\/ Strip line breaks and remove consecutive spaces. Services collapse whitespace in signature calculation\n        $header_value = preg_replace('\/\\s+\/', ' ', trim($header_value));\n\n        $request-&gt;add_header($header_key, $header_value);\n        $this-&gt;canonical_headers[] = strtolower($header_key) . ':' . $header_value;\n        $this-&gt;signed_headers[] = strtolower($header_key);\n    }\n\n    $this-&gt;headers['Authorization'] = $this-&gt;authorization($timestamp);\n    $request-&gt;add_header('Authorization', $this-&gt;headers['Authorization']);\n    $request-&gt;request_headers = $this-&gt;headers;\n\n    return $request;\n}"},{"name":"authorization","modifier":"protected","description":"<p>Generates the authorization string to use for the request.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"datetime","type":"string","description":"<p>The current timestamp.<\/p>"}},"returnvalue":"<p>The authorization string.<\/p>","source":"protected function authorization($datetime)\n{\n    $access_key_id = $this-&gt;key;\n\n    $parts = array();\n    $parts[] = &quot;AWS4-HMAC-SHA256 Credential=${access_key_id}\/&quot; . $this-&gt;credential_string($datetime);\n    $parts[] = 'SignedHeaders=' . implode(';', $this-&gt;signed_headers);\n    $parts[] = 'Signature=' . $this-&gt;hex16($this-&gt;signature($datetime));\n\n    return implode(',', $parts);\n}"},{"name":"canonical_querystring","modifier":"protected","description":"<p>The canonical query string.<\/p>","returnvalue":"<p>The canonical query string.<\/p>","source":"protected function canonical_querystring()\n{\n    if (!isset($this-&gt;canonical_querystring))\n    {\n        $this-&gt;canonical_querystring = $this-&gt;util-&gt;to_signable_string($this-&gt;query);\n    }\n\n    return $this-&gt;canonical_querystring;\n}"},{"name":"canonical_request","modifier":"protected","description":"<p>Calculate the canonical request.<\/p>","returnvalue":"<p>The canonical request.<\/p>","source":"protected function canonical_request()\n{\n    $parts = array();\n    $parts[] = 'POST';\n    $parts[] = $this-&gt;canonical_uri();\n    $parts[] = ''; \/\/ $parts[] = $this-&gt;canonical_querystring();\n    $parts[] = implode(&quot;\\n&quot;, $this-&gt;canonical_headers) . &quot;\\n&quot;;\n    $parts[] = implode(';', $this-&gt;signed_headers);\n    $parts[] = $this-&gt;hex16($this-&gt;hash($this-&gt;body));\n\n    $this-&gt;canonical_request = implode(&quot;\\n&quot;, $parts);\n\n    return $this-&gt;canonical_request;\n}"},{"name":"canonical_uri","modifier":"protected","description":"<p>The request URI path.<\/p>","returnvalue":"<p>The request URI path.<\/p>","source":"protected function canonical_uri()\n{\n    return '\/';\n}"},{"name":"credential_string","modifier":"protected","description":"<p>Generates the credential string to use for signing.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"datetime","type":"string","description":"<p>The current timestamp.<\/p>"}},"returnvalue":"<p>The credential string.<\/p>","source":"protected function credential_string($datetime)\n{\n    $parts = array();\n    $parts[] = substr($datetime, 0, 8);\n    $parts[] = $this-&gt;region();\n    $parts[] = $this-&gt;service();\n    $parts[] = 'aws4_request';\n\n    return implode('\/', $parts);\n}"},{"name":"hash","modifier":"protected","description":"<p>SHA-256 hashes the string.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"string"}},"returnvalue":"<p>Raw SHA-256 hashed string.<\/p>","source":"protected function hash($string)\n{\n    return hash('sha256', $string, true);\n}"},{"name":"hex16","modifier":"protected","description":"<p>Hex16-pack the data.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"value","type":"string","description":"<p>The data to hex16 pack.<\/p>"}},"returnvalue":"<p>The hex16-packed data.<\/p>","source":"protected function hex16($value)\n{\n    $result = unpack('H*', $value);\n    return reset($result);\n}"},{"name":"hmac","modifier":"protected","description":"<p>Applies HMAC SHA-256 encryption to the string, salted by the key.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"key"},{"@attributes":{"choice":"req"},"name":"string"}]},"returnvalue":"<p>Raw HMAC SHA-256 hashed string.<\/p>","source":"protected function hmac($key, $string)\n{\n    return hash_hmac('sha256', $string, $key, true);\n}"},{"name":"region","modifier":"protected","description":"<p>The region ID to use in the signature.<\/p>","returnvalue":"<p>The region ID.<\/p>","source":"protected function region()\n{\n    $pieces = explode('.', $this-&gt;endpoint);\n\n    \/\/ Handle cases with single\/no region (i.e., service.amazonaws.com vs. service.region.amazonaws.com)\n    if (count($pieces) &lt; 4)\n    {\n        return 'us-east-1';\n    }\n\n    return $pieces[1];\n}"},{"name":"service","modifier":"protected","description":"<p>The service ID to use in the signature.<\/p>","returnvalue":"<p>The service ID.<\/p>","source":"protected function service()\n{\n    $pieces = explode('.', $this-&gt;endpoint);\n    return $pieces[0];\n}"},{"name":"signature","modifier":"protected","description":"<p>Calculate the signature.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"datetime","type":"string","description":"<p>The current timestamp.<\/p>"}},"returnvalue":"<p>The signature.<\/p>","source":"protected function signature($datetime)\n{\n    $k_date        = $this-&gt;hmac('AWS4' . $this-&gt;secret_key, substr($datetime, 0, 8));\n    $k_region      = $this-&gt;hmac($k_date, $this-&gt;region());\n    $k_service     = $this-&gt;hmac($k_region, $this-&gt;service());\n    $k_credentials = $this-&gt;hmac($k_service, 'aws4_request');\n    $signature     = $this-&gt;hmac($k_credentials, $this-&gt;string_to_sign($datetime));\n\n    return $signature;\n}"},{"name":"string_to_sign","modifier":"protected","description":"<p>Calculate the string to sign.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"datetime","type":"string","description":"<p>The current timestamp.<\/p>"}},"returnvalue":"<p>The string to sign.<\/p>","source":"protected function string_to_sign($datetime)\n{\n    $parts = array();\n    $parts[] = 'AWS4-HMAC-SHA256';\n    $parts[] = $datetime;\n    $parts[] = $this-&gt;credential_string($datetime);\n    $parts[] = $this-&gt;hex16($this-&gt;hash($this-&gt;canonical_request()));\n\n    $this-&gt;string_to_sign = implode(&quot;\\n&quot;, $parts);\n\n    return $this-&gt;string_to_sign;\n}"}]},"constants":{"@attributes":{"count":"0"}},"properties":{"@attributes":{"count":"4"},"property":[{"name":"credentials","inheritance":{"class":{"name":"Signer","file":"authentication\/signer.abstract.php"}},"modifier":"public","description":"<p>The credentials to use for signing and making requests.<\/p>"},{"name":"endpoint","inheritance":{"class":{"name":"Signer","file":"authentication\/signer.abstract.php"}},"modifier":"public","description":"<p>The endpoint to direct the request to.<\/p>"},{"name":"operation","inheritance":{"class":{"name":"Signer","file":"authentication\/signer.abstract.php"}},"modifier":"public","description":"<p>The operation to execute as a result of this request.<\/p>"},{"name":"payload","inheritance":{"class":{"name":"Signer","file":"authentication\/signer.abstract.php"}},"modifier":"public","description":"<p>The options to use as part of the payload in the request.<\/p>"}]}}}