{"file":{},"class":{"name":"CFComplexType","file":"utilities\/complextype.class.php","url":"..\/cfcomplextype\/index.html","description":"<p>Contains utility methods used for converting array, JSON, and YAML data types into query string keys.<\/p>","metadata":{"version":"2010.11.11","license":"See the included NOTICE.md file for more information.","copyright":"See the included NOTICE.md file for more information.","link":"http:\/\/aws.amazon.com\/php\/ PHP Developer Center"},"methods":{"@attributes":{"count":"4"},"method":[{"name":"json","modifier":"public static","description":"<p>Takes a JSON object, as a string, to convert to query string keys.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"json","type":"string","description":"<p>A JSON object. The JSON string should use canonical rules (e.g., double quotes, quoted keys) as is required by PHP&#8217;s <a href=\"http:\/\/php.net\/json_encode\"><code>json_encode()<\/code><\/a> function.<\/p>"},{"@attributes":{"choice":"opt"},"name":"member","type":"string","description":"<p>The name of the &#8220;member&#8221; property that AWS uses for lists in certain services. Defaults to an empty string.<\/p>"},{"@attributes":{"choice":"opt"},"name":"default_key","type":"string","description":"<p>The default key to use when the value for <code>$data<\/code> is a string. Defaults to an empty string.<\/p>"}]},"returnvalue":"<p>The option group parameters to merge into another method&#8217;s <code>$opt<\/code> parameter.<\/p>","source":"public static function json($json, $member = '', $default_key = '')\n{\n    return self::option_group(json_decode($json, true), $member, $default_key);\n}","examples":{"example":{"title":"<p>Convert a JSON object into a parameter list.<\/p>","code":"\/\/ Test data\n$json = '{\n\t&quot;IpPermissions&quot;: [\n\t\t{\n\t\t\t&quot;IpProtocol&quot;: &quot;tcp&quot;,\n\t\t\t&quot;FromPort&quot;: 3023,\n\t\t\t&quot;ToPort&quot;: 3024,\n\t\t\t&quot;Groups&quot;: [\n\t\t\t\t{ &quot;UserId&quot;: &quot;user1&quot;, &quot;GroupName&quot;: &quot;group1&quot; },\n\t\t\t\t{ &quot;UserId&quot;: &quot;user2&quot;, &quot;GroupName&quot;: &quot;group2&quot; }\n\t\t\t],\n\t\t\t&quot;IpRanges&quot;: [\n\t\t\t\t{ &quot;CidrIp&quot;: &quot;192.168.0.1&quot; },\n\t\t\t\t{ &quot;CidrIp&quot;: &quot;192.168.0.2&quot; }\n\t\t\t]\n\t\t},\n\t\t{\n\t\t\t&quot;IpProtocol&quot;: &quot;tcp&quot;,\n\t\t\t&quot;FromPort&quot;: 3023,\n\t\t\t&quot;ToPort&quot;: 3024,\n\t\t\t&quot;Groups&quot;: [\n\t\t\t\t{ &quot;UserId&quot;: &quot;user1&quot;, &quot;GroupName&quot;: &quot;group1&quot; },\n\t\t\t\t{ &quot;UserId&quot;: &quot;user2&quot;, &quot;GroupName&quot;: &quot;group2&quot; }\n\t\t\t],\n\t\t\t&quot;IpRanges&quot;: [\n\t\t\t\t{ &quot;CidrIp&quot;: &quot;192.168.0.1&quot; },\n\t\t\t\t{ &quot;CidrIp&quot;: &quot;192.168.0.2&quot; }\n\t\t\t]\n\t\t}\n\t]\n}';\n\n$data = CFComplexType::json($json);\n\n\/\/ Success?\nvar_dump($data);","result":"array(18) {\n  [\"IpPermissions.1.IpProtocol\"]=&gt;\n  string(3) \"tcp\"\n  [\"IpPermissions.1.FromPort\"]=&gt;\n  int(3023)\n  [\"IpPermissions.1.ToPort\"]=&gt;\n  int(3024)\n  [\"IpPermissions.1.Groups.1.UserId\"]=&gt;\n  string(5) \"user1\"\n  [\"IpPermissions.1.Groups.1.GroupName\"]=&gt;\n  string(6) \"group1\"\n  [\"IpPermissions.1.Groups.2.UserId\"]=&gt;\n  string(5) \"user2\"\n  [\"IpPermissions.1.Groups.2.GroupName\"]=&gt;\n  string(6) \"group2\"\n  [\"IpPermissions.1.IpRanges.1.CidrIp\"]=&gt;\n  string(11) \"192.168.0.1\"\n  [\"IpPermissions.1.IpRanges.2.CidrIp\"]=&gt;\n  string(11) \"192.168.0.2\"\n  [\"IpPermissions.2.IpProtocol\"]=&gt;\n  string(3) \"tcp\"\n  [\"IpPermissions.2.FromPort\"]=&gt;\n  int(3023)\n  [\"IpPermissions.2.ToPort\"]=&gt;\n  int(3024)\n  [\"IpPermissions.2.Groups.1.UserId\"]=&gt;\n  string(5) \"user1\"\n  [\"IpPermissions.2.Groups.1.GroupName\"]=&gt;\n  string(6) \"group1\"\n  [\"IpPermissions.2.Groups.2.UserId\"]=&gt;\n  string(5) \"user2\"\n  [\"IpPermissions.2.Groups.2.GroupName\"]=&gt;\n  string(6) \"group2\"\n  [\"IpPermissions.2.IpRanges.1.CidrIp\"]=&gt;\n  string(11) \"192.168.0.1\"\n  [\"IpPermissions.2.IpRanges.2.CidrIp\"]=&gt;\n  string(11) \"192.168.0.2\"\n}"}}},{"name":"map","modifier":"public static","description":"<p>Takes an associative array to convert to query string keys.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"map","type":"array","description":"<p>An associative array.<\/p>"},{"@attributes":{"choice":"opt"},"name":"member","type":"string","description":"<p>The name of the &#8220;member&#8221; property that AWS uses for lists in certain services. Defaults to an empty string.<\/p>"},{"@attributes":{"choice":"opt"},"name":"default_key","type":"string","description":"<p>The default key to use when the value for <code>$data<\/code> is a string. Defaults to an empty string.<\/p>"}]},"returnvalue":"<p>The option group parameters to merge into another method&#8217;s <code>$opt<\/code> parameter.<\/p>","source":"public static function map($map, $member = '', $default_key = '')\n{\n    return self::option_group($map, $member, $default_key);\n}","examples":{"example":{"title":"<p>Convert a native PHP associative array into a parameter list.<\/p>","code":"\/\/ Test data\n$map = array(\n\t&quot;IpPermissions&quot; =&gt; array(\n\t\tarray(\n\t\t\t&quot;IpProtocol&quot; =&gt; &quot;tcp&quot;,\n\t\t\t&quot;FromPort&quot;   =&gt; 3023,\n\t\t\t&quot;ToPort&quot;     =&gt; 3024,\n\t\t\t&quot;Groups&quot;     =&gt; array(\n\t\t\t\tarray( &quot;UserId&quot; =&gt; &quot;user1&quot;, &quot;GroupName&quot; =&gt; &quot;group1&quot; ),\n\t\t\t\tarray( &quot;UserId&quot; =&gt; &quot;user2&quot;, &quot;GroupName&quot; =&gt; &quot;group2&quot; )\n\t\t\t),\n\t\t\t&quot;IpRanges&quot; =&gt; array(\n\t\t\t\tarray(&quot;CidrIp&quot; =&gt; &quot;192.168.0.1&quot;),\n\t\t\t\tarray(&quot;CidrIp&quot; =&gt; &quot;192.168.0.2&quot;)\n\t\t\t)\n\t\t),\n\t\tarray(\n\t\t\t&quot;IpProtocol&quot; =&gt; &quot;tcp&quot;,\n\t\t\t&quot;FromPort&quot;   =&gt; 3023,\n\t\t\t&quot;ToPort&quot;     =&gt; 3024,\n\t\t\t&quot;Groups&quot;     =&gt; array(\n\t\t\t\tarray( &quot;UserId&quot; =&gt; &quot;user1&quot;, &quot;GroupName&quot; =&gt; &quot;group1&quot; ),\n\t\t\t\tarray( &quot;UserId&quot; =&gt; &quot;user2&quot;, &quot;GroupName&quot; =&gt; &quot;group2&quot; )\n\t\t\t),\n\t\t\t&quot;IpRanges&quot; =&gt; array(\n\t\t\t\tarray(&quot;CidrIp&quot; =&gt; &quot;192.168.0.1&quot;),\n\t\t\t\tarray(&quot;CidrIp&quot; =&gt; &quot;192.168.0.2&quot;)\n\t\t\t)\n\t\t)\n\t)\n);\n\n$data = CFComplexType::map($map);\n\n\/\/ Success?\nvar_dump($data);","result":"array(18) {\n  [\"IpPermissions.1.IpProtocol\"]=&gt;\n  string(3) \"tcp\"\n  [\"IpPermissions.1.FromPort\"]=&gt;\n  int(3023)\n  [\"IpPermissions.1.ToPort\"]=&gt;\n  int(3024)\n  [\"IpPermissions.1.Groups.1.UserId\"]=&gt;\n  string(5) \"user1\"\n  [\"IpPermissions.1.Groups.1.GroupName\"]=&gt;\n  string(6) \"group1\"\n  [\"IpPermissions.1.Groups.2.UserId\"]=&gt;\n  string(5) \"user2\"\n  [\"IpPermissions.1.Groups.2.GroupName\"]=&gt;\n  string(6) \"group2\"\n  [\"IpPermissions.1.IpRanges.1.CidrIp\"]=&gt;\n  string(11) \"192.168.0.1\"\n  [\"IpPermissions.1.IpRanges.2.CidrIp\"]=&gt;\n  string(11) \"192.168.0.2\"\n  [\"IpPermissions.2.IpProtocol\"]=&gt;\n  string(3) \"tcp\"\n  [\"IpPermissions.2.FromPort\"]=&gt;\n  int(3023)\n  [\"IpPermissions.2.ToPort\"]=&gt;\n  int(3024)\n  [\"IpPermissions.2.Groups.1.UserId\"]=&gt;\n  string(5) \"user1\"\n  [\"IpPermissions.2.Groups.1.GroupName\"]=&gt;\n  string(6) \"group1\"\n  [\"IpPermissions.2.Groups.2.UserId\"]=&gt;\n  string(5) \"user2\"\n  [\"IpPermissions.2.Groups.2.GroupName\"]=&gt;\n  string(6) \"group2\"\n  [\"IpPermissions.2.IpRanges.1.CidrIp\"]=&gt;\n  string(11) \"192.168.0.1\"\n  [\"IpPermissions.2.IpRanges.2.CidrIp\"]=&gt;\n  string(11) \"192.168.0.2\"\n}"}}},{"name":"option_group","modifier":"public static","description":"<p>A protected method that is used by <a href=\"..\/cfcomplextype\/json.html\" rel=\"..\/..\/index.html#m=cfcomplextype\/json\"><code>json()<\/code><\/a>, <a href=\"..\/cfcomplextype\/yaml.html\" rel=\"..\/..\/index.html#m=cfcomplextype\/yaml\"><code>yaml()<\/code><\/a> and <a href=\"..\/cfcomplextype\/map.html\" rel=\"..\/..\/index.html#m=cfcomplextype\/map\"><code>map()<\/code><\/a>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"data","type":"string|array","description":"<p>The data to iterate over.<\/p>"},{"@attributes":{"choice":"opt"},"name":"member","type":"string","description":"<p>The name of the &#8220;member&#8221; property that AWS uses for lists in certain services. Defaults to an empty string.<\/p>"},{"@attributes":{"choice":"opt"},"name":"key","type":"string","description":"<p>The default key to use when the value for <code>$data<\/code> is a string. Defaults to an empty string.<\/p>"},{"@attributes":{"choice":"opt"},"name":"out","type":"array","description":"<p>INTERNAL ONLY. The array that contains the calculated values up to this point.<\/p>"}]},"returnvalue":"<p>The option group parameters to merge into another method&#8217;s <code>$opt<\/code> parameter.<\/p>","source":"public static function option_group($data, $member = '', $key = '', &amp;$out = array())\n{\n    $reset = $key;\n\n    if (is_array($data))\n    {\n        foreach ($data as $k =&gt; $v)\n        {\n            \/\/ Avoid 0-based indexes.\n            if (is_int($k))\n            {\n                $k = $k + 1;\n\n                if ($member !== '')\n                {\n                    $key .= '.' . $member;\n                }\n            }\n\n            $key .= ($key === '' ? $k : '.' . $k);\n\n            if (is_array($v))\n            {\n                self::option_group($v, $member, $key, $out);\n            }\n            elseif ($v instanceof CFStepConfig)\n            {\n                self::option_group($v-&gt;get_config(), $member, $key, $out);\n            }\n            else\n            {\n                $out[$key] = $v;\n            }\n\n            $key = $reset;\n        }\n    }\n    else\n    {\n        $out[$key] = $data;\n    }\n\n    return $out;\n}"},{"name":"yaml","modifier":"public static","description":"<p>Takes a YAML object, as a string, to convert to query string keys.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"yaml","type":"string","description":"<p>A YAML object.<\/p>"},{"@attributes":{"choice":"opt"},"name":"member","type":"string","description":"<p>The name of the &#8220;member&#8221; property that AWS uses for lists in certain services. Defaults to an empty string.<\/p>"},{"@attributes":{"choice":"opt"},"name":"default_key","type":"string","description":"<p>The default key to use when the value for <code>$data<\/code> is a string. Defaults to an empty string.<\/p>"}]},"returnvalue":"<p>The option group parameters to merge into another method&#8217;s <code>$opt<\/code> parameter.<\/p>","source":"public static function yaml($yaml, $member = '', $default_key = '')\n{\n    return self::option_group(sfYaml::load($yaml), $member, $default_key);\n}","examples":{"example":[{"title":"<p>Convert a short-form YAML object into a parameter list.<\/p>","code":"\/\/ Test data\n$yaml = '\nIpPermissions:\n  - { IpProtocol: tcp, FromPort: 3023, ToPort: 3024, Groups: [{ UserId: user1, GroupName: group1 }, { UserId: user2, GroupName: group2 }], IpRanges: [{ CidrIp: 192.168.0.1 }, { CidrIp: 192.168.0.2 }] }\n  - { IpProtocol: tcp, FromPort: 3023, ToPort: 3024, Groups: [{ UserId: user1, GroupName: group1 }, { UserId: user2, GroupName: group2 }], IpRanges: [{ CidrIp: 192.168.0.1 }, { CidrIp: 192.168.0.2 }] }\n';\n\n$data = CFComplexType::yaml($yaml);\n\n\/\/ Success?\nvar_dump($data);","result":"array(18) {\n  [\"IpPermissions.1.IpProtocol\"]=&gt;\n  string(3) \"tcp\"\n  [\"IpPermissions.1.FromPort\"]=&gt;\n  int(3023)\n  [\"IpPermissions.1.ToPort\"]=&gt;\n  int(3024)\n  [\"IpPermissions.1.Groups.1.UserId\"]=&gt;\n  string(5) \"user1\"\n  [\"IpPermissions.1.Groups.1.GroupName\"]=&gt;\n  string(6) \"group1\"\n  [\"IpPermissions.1.Groups.2.UserId\"]=&gt;\n  string(5) \"user2\"\n  [\"IpPermissions.1.Groups.2.GroupName\"]=&gt;\n  string(6) \"group2\"\n  [\"IpPermissions.1.IpRanges.1.CidrIp\"]=&gt;\n  string(11) \"192.168.0.1\"\n  [\"IpPermissions.1.IpRanges.2.CidrIp\"]=&gt;\n  string(11) \"192.168.0.2\"\n  [\"IpPermissions.2.IpProtocol\"]=&gt;\n  string(3) \"tcp\"\n  [\"IpPermissions.2.FromPort\"]=&gt;\n  int(3023)\n  [\"IpPermissions.2.ToPort\"]=&gt;\n  int(3024)\n  [\"IpPermissions.2.Groups.1.UserId\"]=&gt;\n  string(5) \"user1\"\n  [\"IpPermissions.2.Groups.1.GroupName\"]=&gt;\n  string(6) \"group1\"\n  [\"IpPermissions.2.Groups.2.UserId\"]=&gt;\n  string(5) \"user2\"\n  [\"IpPermissions.2.Groups.2.GroupName\"]=&gt;\n  string(6) \"group2\"\n  [\"IpPermissions.2.IpRanges.1.CidrIp\"]=&gt;\n  string(11) \"192.168.0.1\"\n  [\"IpPermissions.2.IpRanges.2.CidrIp\"]=&gt;\n  string(11) \"192.168.0.2\"\n}"},{"title":"<p>Convert a long-form YAML object into a parameter list.<\/p>","code":"\/\/ Test data\n$yaml = '\nIpPermissions:\n  0:\n    IpProtocol: tcp\n    FromPort: 3023\n    ToPort: 3024\n    Groups:\n      0:\n        UserId: user1\n        GroupName: group1\n      1:\n        UserId: user2\n        GroupName: group2\n    IpRanges:\n      0:\n        CidrIp: 192.168.0.1\n      1:\n        CidrIp: 192.168.0.2\n  1:\n    IpProtocol: tcp\n    FromPort: 3023\n    ToPort: 3024\n    Groups:\n      - { UserId: user1, GroupName: group1 }\n      - { UserId: user2, GroupName: group2 }\n    IpRanges:\n      - { CidrIp: 192.168.0.1 }\n      - { CidrIp: 192.168.0.2 }\n';\n\n$data = CFComplexType::yaml($yaml);\n\n\/\/ Success?\nvar_dump($data);","result":"array(18) {\n  [\"IpPermissions.1.IpProtocol\"]=&gt;\n  string(3) \"tcp\"\n  [\"IpPermissions.1.FromPort\"]=&gt;\n  int(3023)\n  [\"IpPermissions.1.ToPort\"]=&gt;\n  int(3024)\n  [\"IpPermissions.1.Groups.1.UserId\"]=&gt;\n  string(5) \"user1\"\n  [\"IpPermissions.1.Groups.1.GroupName\"]=&gt;\n  string(6) \"group1\"\n  [\"IpPermissions.1.Groups.2.UserId\"]=&gt;\n  string(5) \"user2\"\n  [\"IpPermissions.1.Groups.2.GroupName\"]=&gt;\n  string(6) \"group2\"\n  [\"IpPermissions.1.IpRanges.1.CidrIp\"]=&gt;\n  string(11) \"192.168.0.1\"\n  [\"IpPermissions.1.IpRanges.2.CidrIp\"]=&gt;\n  string(11) \"192.168.0.2\"\n  [\"IpPermissions.2.IpProtocol\"]=&gt;\n  string(3) \"tcp\"\n  [\"IpPermissions.2.FromPort\"]=&gt;\n  int(3023)\n  [\"IpPermissions.2.ToPort\"]=&gt;\n  int(3024)\n  [\"IpPermissions.2.Groups.1.UserId\"]=&gt;\n  string(5) \"user1\"\n  [\"IpPermissions.2.Groups.1.GroupName\"]=&gt;\n  string(6) \"group1\"\n  [\"IpPermissions.2.Groups.2.UserId\"]=&gt;\n  string(5) \"user2\"\n  [\"IpPermissions.2.Groups.2.GroupName\"]=&gt;\n  string(6) \"group2\"\n  [\"IpPermissions.2.IpRanges.1.CidrIp\"]=&gt;\n  string(11) \"192.168.0.1\"\n  [\"IpPermissions.2.IpRanges.2.CidrIp\"]=&gt;\n  string(11) \"192.168.0.2\"\n}"}]}}]},"constants":{"@attributes":{"count":"0"}},"properties":{"@attributes":{"count":"0"}}}}