{"file":{},"class":{"name":"CFCredentials","file":"utilities\/credentials.class.php","url":"..\/cfcredentials\/index.html","description":"<p>The <a href=\"..\/cfcredentials\/index.html\" rel=\"..\/..\/index.html#i=cfcredentials\"><code>CFCredentials<\/code><\/a> class enables developers to easily switch between multiple sets of credentials.<\/p>","metadata":{"version":"2012.07.13","license":"See the included NOTICE.md file for more information.","copyright":"See the included NOTICE.md file for more information.","link":"http:\/\/aws.amazon.com\/php\/ PHP Developer Center"},"methods":{"@attributes":{"count":"4"},"method":[{"name":"__construct","modifier":"final private","description":"<p>Prevents this class from being constructed<\/p>","returnvalue":{},"source":"final private function __construct() {}"},{"name":"get","modifier":"public static","description":"<p>Retrieves the requested credentials from the internal credential store.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"credential_name"}},"returnvalue":"<p>A stdClass object where the properties represent the keys that were provided.<\/p>","source":"public static function get($credential_name = self::DEFAULT_KEY)\n{\n    \/\/ Make sure the credential set exists\n    if (!isset(self::$credentials[$credential_name]))\n    {\n        throw new CFCredentials_Exception('The credential set, &quot;' . $credential_name . '&quot;, does not exist and cannot be retrieved.');\n    }\n\n    \/\/ Return the credential set as an object\n    return new CFCredential(self::$credentials[$credential_name]);\n}","examples":{"example":{"title":"<p>Set the credential sets that should be available to the application, and read back the values.<\/p>","code":"\/\/ Dependencies\nrequire_once dirname(__FILE__) . '\/..\/..\/sdk.class.php';\nrequire_once dirname(__FILE__) . '\/..\/testutil.inc.php';\n\n\/**\n * Create a group of credential sets. This would normally go inside the config.inc.php file.\n *\n * Each credential set is assigned a name. Any name that PHP allows as an array key may be used.\n * `@inherit` is a keyword that will tell the class to use the inherited credential set as a base,\n * then apply changes on top of it. The &quot;default&quot; credential set can be its own credential set,\n * or can reference another credential set by name.\n *\/\nCFCredentials::set(array(\n\n\t\/\/ Credentials for the production environment\n\t'production' =&gt; array(\n\t\t'key' =&gt; 'production-key',\n\t\t'secret' =&gt; 'production-secret',\n\t\t'cache' =&gt; 'apc',\n\t\t'something_else' =&gt; 'abc123',\n\t\t'aardvark' =&gt; 'applesauce'\n\t),\n\n\t\/\/ Credentials for the staging environment. Inherits from &quot;production&quot;.\n\t'staging' =&gt; array(\n\t\t'@inherit' =&gt; 'production',\n\t\t'key' =&gt; 'staging-key',\n\t\t'secret' =&gt; 'staging-secret',\n\t),\n\n\t\/\/ Credentials for the development environment. Inherits from &quot;staging&quot;, and in-turn, &quot;production&quot;.\n\t'development' =&gt; array(\n\t\t'@inherit' =&gt; 'staging',\n\t\t'key' =&gt; 'development-key',\n\t\t'secret' =&gt; 'development-secret',\n\t),\n\n\t\/\/ Any name can be used for a credential set.\n\t'my-own-personal-credential-set' =&gt; array(\n\t\t'@inherit' =&gt; 'staging'\n\t),\n\n\t\/\/ Set a default credential set for ambiguous cases.\n\t'@default' =&gt; 'development'\n));\n\n\/**\n * Requesting data back from the credential sets.\n *\/\necho CFCredentials::get('production')-&gt;key . PHP_EOL;\necho CFCredentials::get('production')-&gt;aardvark . PHP_EOL;\n\necho PHP_EOL;\n\necho CFCredentials::get('staging')-&gt;key . PHP_EOL;\necho CFCredentials::get('staging')-&gt;aardvark . PHP_EOL;\n\necho PHP_EOL;\n\necho CFCredentials::get('development')-&gt;key . PHP_EOL;\necho CFCredentials::get('development')-&gt;aardvark . PHP_EOL;\n\necho PHP_EOL;\n\necho CFCredentials::get('my-own-personal-credential-set')-&gt;key . PHP_EOL;\n\necho PHP_EOL;\n\necho CFCredentials::get()-&gt;key . PHP_EOL;","result":"production-key\napplesauce\n\nstaging-key\napplesauce\n\ndevelopment-key\napplesauce\n\nstaging-key\n\ndevelopment-key"}}},{"name":"list_sets","modifier":"public static","description":"<p>Retrieves a list of all available credential set names.<\/p>","returnvalue":"<p>A list of all available credential set names.<\/p>","source":"public static function list_sets()\n{\n    return new CFArray(array_keys(self::$credentials));\n}"},{"name":"set","modifier":"public static","description":"<p>Stores the credentials for re-use.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"credential_sets","type":"array","description":"<p>The named credential sets that should be made available to the application.<\/p>"}},"returnvalue":{"@attributes":{"type":"void"}},"source":"public static function set(array $credential_sets)\n{\n    \/\/ Make sure a default credential set is specified or can be inferred\n    if (count($credential_sets) === 1)\n    {\n        $credential_sets[self::DEFAULT_KEY] = reset($credential_sets);\n    }\n\n    \/\/ Resolve any @inherit tags\n    foreach ($credential_sets as $credential_name =&gt; &amp;$credential_set)\n    {\n        if (is_array($credential_set))\n        {\n            foreach ($credential_set as $credential_key =&gt; &amp;$credential_value)\n            {\n                if ($credential_key === self::INHERIT_KEY)\n                {\n                    if (!isset($credential_sets[$credential_value]))\n                    {\n                        throw new CFCredentials_Exception('The credential set, &quot;' . $credential_value . '&quot;, does not exist and cannot be inherited.');\n                    }\n\n                    $credential_set = array_merge($credential_sets[$credential_value], $credential_set);\n                    unset($credential_set[self::INHERIT_KEY]);\n                }\n            }\n        }\n    }\n\n    \/\/ Normalize the value of the @default credential set\n    if (isset($credential_sets[self::DEFAULT_KEY]))\n    {\n        $default = $credential_sets[self::DEFAULT_KEY];\n        if (is_string($default))\n        {\n            if (!isset($credential_sets[$default]))\n            {\n                throw new CFCredentials_Exception('The credential set, &quot;' . $default . '&quot;, does not exist and cannot be used as the default credential set.');\n            }\n\n            $credential_sets[self::DEFAULT_KEY] = $credential_sets[$default];\n        }\n    }\n\n    \/\/ Store the credentials\n    self::$credentials = $credential_sets;\n}","examples":{"example":{"title":"<p>Set the credential sets that should be available to the application, and read back the values.<\/p>","code":"\/\/ Dependencies\nrequire_once dirname(__FILE__) . '\/..\/..\/sdk.class.php';\nrequire_once dirname(__FILE__) . '\/..\/testutil.inc.php';\n\n\/**\n * Create a group of credential sets. This would normally go inside the config.inc.php file.\n *\n * Each credential set is assigned a name. Any name that PHP allows as an array key may be used.\n * `@inherit` is a keyword that will tell the class to use the inherited credential set as a base,\n * then apply changes on top of it. The &quot;default&quot; credential set can be its own credential set,\n * or can reference another credential set by name.\n *\/\nCFCredentials::set(array(\n\n\t\/\/ Credentials for the production environment\n\t'production' =&gt; array(\n\t\t'key' =&gt; 'production-key',\n\t\t'secret' =&gt; 'production-secret',\n\t\t'cache' =&gt; 'apc',\n\t\t'something_else' =&gt; 'abc123',\n\t\t'aardvark' =&gt; 'applesauce'\n\t),\n\n\t\/\/ Credentials for the staging environment. Inherits from &quot;production&quot;.\n\t'staging' =&gt; array(\n\t\t'@inherit' =&gt; 'production',\n\t\t'key' =&gt; 'staging-key',\n\t\t'secret' =&gt; 'staging-secret',\n\t),\n\n\t\/\/ Credentials for the development environment. Inherits from &quot;staging&quot;, and in-turn, &quot;production&quot;.\n\t'development' =&gt; array(\n\t\t'@inherit' =&gt; 'staging',\n\t\t'key' =&gt; 'development-key',\n\t\t'secret' =&gt; 'development-secret',\n\t),\n\n\t\/\/ Any name can be used for a credential set.\n\t'my-own-personal-credential-set' =&gt; array(\n\t\t'@inherit' =&gt; 'staging'\n\t),\n\n\t\/\/ Set a default credential set for ambiguous cases.\n\t'@default' =&gt; 'development'\n));\n\n\/**\n * Requesting data back from the credential sets.\n *\/\necho CFCredentials::get('production')-&gt;key . PHP_EOL;\necho CFCredentials::get('production')-&gt;aardvark . PHP_EOL;\n\necho PHP_EOL;\n\necho CFCredentials::get('staging')-&gt;key . PHP_EOL;\necho CFCredentials::get('staging')-&gt;aardvark . PHP_EOL;\n\necho PHP_EOL;\n\necho CFCredentials::get('development')-&gt;key . PHP_EOL;\necho CFCredentials::get('development')-&gt;aardvark . PHP_EOL;\n\necho PHP_EOL;\n\necho CFCredentials::get('my-own-personal-credential-set')-&gt;key . PHP_EOL;\n\necho PHP_EOL;\n\necho CFCredentials::get()-&gt;key . PHP_EOL;","result":"production-key\napplesauce\n\nstaging-key\napplesauce\n\ndevelopment-key\napplesauce\n\nstaging-key\n\ndevelopment-key"}}}]},"constants":{"@attributes":{"count":"2"},"constant":[{"name":"DEFAULT_KEY","value":"@default"},{"name":"INHERIT_KEY","value":"@inherit"}]},"properties":{"@attributes":{"count":"1"},"property":{"name":"credentials","modifier":"protected static","description":"<p>Stores the credentials<\/p>"}}}}