{"file":{},"class":{"name":"CFPolicy","file":"utilities\/policy.class.php","url":"..\/cfpolicy\/index.html","description":"<p>Simplifies the process of signing JSON policy documents.<\/p>","metadata":{"version":"2011.04.25","license":"See the included NOTICE.md file for more information.","copyright":"See the included NOTICE.md file for more information.","link":"http:\/\/aws.amazon.com\/php\/ PHP Developer Center"},"methods":{"@attributes":{"count":"7"},"method":[{"name":"__construct","modifier":"public","description":"<p>Constructs a new instance of this class.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auth","type":"CFRuntime","description":"<p>An instance of any authenticated AWS object that is an instance of <a href=\"..\/cfruntime\/index.html\" rel=\"..\/..\/index.html#i=cfruntime\"><code>CFRuntime<\/code><\/a> (e.g. <a href=\"..\/amazonec2\/index.html\" rel=\"..\/..\/index.html#i=amazonec2\"><code>AmazonEC2<\/code><\/a>, <a href=\"..\/amazons3\/index.html\" rel=\"..\/..\/index.html#i=amazons3\"><code>AmazonS3<\/code><\/a>).<\/p>"},{"@attributes":{"choice":"req"},"name":"policy","type":"string|array","description":"<p>The associative array representing the S3 policy to use, or a string of JSON content.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function __construct($auth, $policy)\n{\n    $this-&gt;auth = $auth;\n\n    if (is_array($policy)) \/\/ We received an associative array...\n    {\n        $this-&gt;json_policy = json_encode($policy);\n    }\n    else \/\/ We received a valid, parseable JSON string...\n    {\n        $this-&gt;json_policy = json_encode(json_decode($policy, true));\n    }\n\n    return $this;\n}","seealso":{"link":["<a href=\"http:\/\/docs.amazonwebservices.com\/AmazonS3\/2006-03-01\/dev\/index.html?HTTPPOSTForms.html\">S3 Policies<\/a>","<a href=\"http:\/\/docs.amazonwebservices.com\/AmazonS3\/latest\/dev\/index.html?AccessPolicyLanguage.html\">Access Policy Language<\/a>"]}},{"name":"decode_policy","modifier":"public static","description":"<p>Decode a policy that was returned from the service.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"response","type":"string","description":"<p>The policy returned by AWS that you want to decode into an object.<\/p>"}},"returnvalue":"<p>The Base64-encoded, signed JSON string.<\/p>","source":"public static function decode_policy($response)\n{\n    return json_decode(urldecode($response), true);\n}"},{"name":"get_json","modifier":"public","description":"<p>Gets the JSON string with the whitespace removed.<\/p>","returnvalue":"<p>The JSON string without extraneous whitespace.<\/p>","source":"public function get_json()\n{\n    return $this-&gt;json_policy;\n}","examples":{"example":{"title":"<p>Create a policy and get the JSON policy document.<\/p>","code":"\/\/ Instantiate S3\n$s3 = new AmazonS3();\n\n\/\/ Create a new S3 policy using an associative array\n$policy = new CFPolicy($s3, array(\n\t'Version' =&gt; '2008-10-17',\n\t'Statement' =&gt; array(\n\t\tarray( \/\/ Statement #1\n\t\t\t'Sid' =&gt; 'AddPerm',\n\t\t\t'Effect' =&gt; 'Allow',\n\t\t\t'Principal' =&gt; array(\n\t\t\t\t'AWS' =&gt; '*'\n\t\t\t),\n\t\t\t'Action' =&gt; array('s3:ReadObject'),\n\t\t\t'Resource' =&gt; array('arn:aws:s3:::my-bucket\/*')\n\t\t)\n\t)\n));\n\n\/\/ Success?\nvar_dump($policy-&gt;get_json());","result":"string(172) \"{\"Version\":\"2008-10-17\",\"Statement\":[{\"Sid\":\"AddPerm\",\"Effect\":\"Allow\",\"Principal\":{\"AWS\":\"*\"},\"Action\":[\"s3:ReadObject\"],\"Resource\":[\"arn:aws:s3:::aws-php-sdk-test\\\/*\"]}]}\""}},"related":{"method":["<AmazonIAM::decode_policy()>","<a href=\"..\/amazoniam\/delete_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/delete_group_policy\"><code>AmazonIAM::delete_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/delete_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/delete_user_policy\"><code>AmazonIAM::delete_user_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/get_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/get_group_policy\"><code>AmazonIAM::get_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/get_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/get_user_policy\"><code>AmazonIAM::get_user_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/list_group_policies.html\" rel=\"..\/..\/index.html#m=amazoniam\/list_group_policies\"><code>AmazonIAM::list_group_policies()<\/code><\/a>","<a href=\"..\/amazoniam\/list_user_policies.html\" rel=\"..\/..\/index.html#m=amazoniam\/list_user_policies\"><code>AmazonIAM::list_user_policies()<\/code><\/a>","<a href=\"..\/amazoniam\/put_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/put_group_policy\"><code>AmazonIAM::put_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/put_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/put_user_policy\"><code>AmazonIAM::put_user_policy()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_key.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_key\"><code>get_key()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_policy.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_policy\"><code>get_policy()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_policy_signature.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_policy_signature\"><code>get_policy_signature()<\/code><\/a>"]}},{"name":"get_key","modifier":"public","description":"<p>Get the key from the authenticated instance.<\/p>","returnvalue":"<p>The key from the authenticated instance.<\/p>","source":"public function get_key()\n{\n    return $this-&gt;auth-&gt;key;\n}","examples":{"example":{"title":"<p>Create a policy and get the account key.<\/p>","code":"\/\/ Instantiate S3\n$s3 = new AmazonS3();\n\n\/\/ Create a new S3 policy using an associative array\n$policy = new s3Policy($s3, array(\n\t'Version' =&gt; '2008-10-17',\n\t'Statement' =&gt; array(\n\t\tarray( \/\/ Statement #1\n\t\t\t'Sid' =&gt; 'AddPerm',\n\t\t\t'Effect' =&gt; 'Allow',\n\t\t\t'Principal' =&gt; array(\n\t\t\t\t'AWS' =&gt; '*'\n\t\t\t),\n\t\t\t'Action' =&gt; array('s3:ReadObject'),\n\t\t\t'Resource' =&gt; array('arn:aws:s3:::my-bucket\/*')\n\t\t)\n\t)\n));\n\n\/\/ Success?\nvar_dump($policy-&gt;get_key());","result":"string(20) \"AKIAIFUPOHJBYISR6QBA\""}},"related":{"method":["<AmazonIAM::decode_policy()>","<a href=\"..\/amazoniam\/delete_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/delete_group_policy\"><code>AmazonIAM::delete_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/delete_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/delete_user_policy\"><code>AmazonIAM::delete_user_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/get_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/get_group_policy\"><code>AmazonIAM::get_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/get_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/get_user_policy\"><code>AmazonIAM::get_user_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/list_group_policies.html\" rel=\"..\/..\/index.html#m=amazoniam\/list_group_policies\"><code>AmazonIAM::list_group_policies()<\/code><\/a>","<a href=\"..\/amazoniam\/list_user_policies.html\" rel=\"..\/..\/index.html#m=amazoniam\/list_user_policies\"><code>AmazonIAM::list_user_policies()<\/code><\/a>","<a href=\"..\/amazoniam\/put_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/put_group_policy\"><code>AmazonIAM::put_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/put_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/put_user_policy\"><code>AmazonIAM::put_user_policy()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_json.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_json\"><code>get_json()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_policy.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_policy\"><code>get_policy()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_policy_signature.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_policy_signature\"><code>get_policy_signature()<\/code><\/a>"]}},{"name":"get_policy","modifier":"public","description":"<p>Base64-encodes the JSON string.<\/p>","returnvalue":"<p>The Base64-encoded version of the JSON string.<\/p>","source":"public function get_policy()\n{\n    return base64_encode($this-&gt;json_policy);\n}","examples":{"example":{"title":"<p>Create a policy and get the Base64-encoded policy document.<\/p>","code":"\/\/ Instantiate S3\n$s3 = new AmazonS3();\n\n\/\/ Create a new S3 policy using an associative array\n$policy = new s3Policy($s3, array(\n\t'Version' =&gt; '2008-10-17',\n\t'Statement' =&gt; array(\n\t\tarray( \/\/ Statement #1\n\t\t\t'Sid' =&gt; 'AddPerm',\n\t\t\t'Effect' =&gt; 'Allow',\n\t\t\t'Principal' =&gt; array(\n\t\t\t\t'AWS' =&gt; '*'\n\t\t\t),\n\t\t\t'Action' =&gt; array('s3:ReadObject'),\n\t\t\t'Resource' =&gt; array('arn:aws:s3:::my-bucket\/*')\n\t\t)\n\t)\n));\n\n\/\/ Success?\nvar_dump($policy-&gt;get_policy());","result":"string(232) \"eyJWZXJzaW9uIjoiMjAwOC0xMC0xNyIsIlN0YXRlbWVudCI6W3siU2lkIjoiQWRkUGVybSIsIkVmZmVjdCI6IkFsbG93IiwiUHJpbmNpcGFsIjp7IkFXUyI6IioifSwiQWN0aW9uIjpbInMzOlJlYWRPYmplY3QiXSwiUmVzb3VyY2UiOlsiYXJuOmF3czpzMzo6OmF3cy1waHAtc2RrLXRlc3RcLyoiXX1dfQ==\""}},"related":{"method":["<AmazonIAM::decode_policy()>","<a href=\"..\/amazoniam\/delete_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/delete_group_policy\"><code>AmazonIAM::delete_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/delete_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/delete_user_policy\"><code>AmazonIAM::delete_user_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/get_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/get_group_policy\"><code>AmazonIAM::get_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/get_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/get_user_policy\"><code>AmazonIAM::get_user_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/list_group_policies.html\" rel=\"..\/..\/index.html#m=amazoniam\/list_group_policies\"><code>AmazonIAM::list_group_policies()<\/code><\/a>","<a href=\"..\/amazoniam\/list_user_policies.html\" rel=\"..\/..\/index.html#m=amazoniam\/list_user_policies\"><code>AmazonIAM::list_user_policies()<\/code><\/a>","<a href=\"..\/amazoniam\/put_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/put_group_policy\"><code>AmazonIAM::put_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/put_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/put_user_policy\"><code>AmazonIAM::put_user_policy()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_json.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_json\"><code>get_json()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_key.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_key\"><code>get_key()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_policy_signature.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_policy_signature\"><code>get_policy_signature()<\/code><\/a>"]}},{"name":"get_policy_signature","modifier":"public","description":"<p>Gets the JSON string with the whitespace removed.<\/p>","returnvalue":"<p>The Base64-encoded, signed JSON string.<\/p>","source":"public function get_policy_signature()\n{\n    return base64_encode(hash_hmac('sha1', $this-&gt;get_policy(), $this-&gt;auth-&gt;secret_key));\n}","examples":{"example":{"title":"<p>Create a policy and get the policy signature.<\/p>","code":"\/\/ Instantiate S3\n$s3 = new AmazonS3();\n\n\/\/ Create a new S3 policy using an associative array\n$policy = new s3Policy($s3, array(\n\t'Version' =&gt; '2008-10-17',\n\t'Statement' =&gt; array(\n\t\tarray( \/\/ Statement #1\n\t\t\t'Sid' =&gt; 'AddPerm',\n\t\t\t'Effect' =&gt; 'Allow',\n\t\t\t'Principal' =&gt; array(\n\t\t\t\t'AWS' =&gt; '*'\n\t\t\t),\n\t\t\t'Action' =&gt; array('s3:ReadObject'),\n\t\t\t'Resource' =&gt; array('arn:aws:s3:::my-bucket\/*')\n\t\t)\n\t)\n));\n\n\/\/ Success?\nvar_dump($policy-&gt;get_policy_signature());","result":"string(56) \"OWU3NzU4MGMxZmY5NDM0YmI1Mjk0Y2FkMWIzZWFjMDFiYzdkMjdlMg==\""}},"related":{"method":["<AmazonIAM::decode_policy()>","<a href=\"..\/amazoniam\/delete_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/delete_group_policy\"><code>AmazonIAM::delete_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/delete_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/delete_user_policy\"><code>AmazonIAM::delete_user_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/get_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/get_group_policy\"><code>AmazonIAM::get_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/get_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/get_user_policy\"><code>AmazonIAM::get_user_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/list_group_policies.html\" rel=\"..\/..\/index.html#m=amazoniam\/list_group_policies\"><code>AmazonIAM::list_group_policies()<\/code><\/a>","<a href=\"..\/amazoniam\/list_user_policies.html\" rel=\"..\/..\/index.html#m=amazoniam\/list_user_policies\"><code>AmazonIAM::list_user_policies()<\/code><\/a>","<a href=\"..\/amazoniam\/put_group_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/put_group_policy\"><code>AmazonIAM::put_group_policy()<\/code><\/a>","<a href=\"..\/amazoniam\/put_user_policy.html\" rel=\"..\/..\/index.html#m=amazoniam\/put_user_policy\"><code>AmazonIAM::put_user_policy()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_json.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_json\"><code>get_json()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_key.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_key\"><code>get_key()<\/code><\/a>","<a href=\"..\/cfpolicy\/get_policy.html\" rel=\"..\/..\/index.html#m=cfpolicy\/get_policy\"><code>get_policy()<\/code><\/a>"]}},{"name":"init","modifier":"public static","description":"<p>Alternate approach to constructing a new instance. Supports chaining.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"auth","type":"CFRuntime","description":"<p>An instance of any authenticated AWS object that is an instance of <a href=\"..\/cfruntime\/index.html\" rel=\"..\/..\/index.html#i=cfruntime\"><code>CFRuntime<\/code><\/a> (e.g. <a href=\"..\/amazonec2\/index.html\" rel=\"..\/..\/index.html#i=amazonec2\"><code>AmazonEC2<\/code><\/a>, <a href=\"..\/amazons3\/index.html\" rel=\"..\/..\/index.html#i=amazons3\"><code>AmazonS3<\/code><\/a>).<\/p>"},{"@attributes":{"choice":"req"},"name":"policy","type":"string|array","description":"<p>The associative array representing the S3 policy to use, or a string of JSON content.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public static function init($auth, $policy)\n{\n    if (version_compare(PHP_VERSION, '5.3.0', '&lt;'))\n    {\n        throw new Exception('PHP 5.3 or newer is required to instantiate a new class with CLASS::init().');\n    }\n\n    $self = get_called_class();\n    return new $self($auth, $policy);\n}"}]},"constants":{"@attributes":{"count":"0"}},"properties":{"@attributes":{"count":"2"},"property":[{"name":"auth","modifier":"public","description":"<p>Stores the object that contains the authentication credentials.<\/p>"},{"name":"json_policy","modifier":"public","description":"<p>Stores the policy object that we&#8217;re working with.<\/p>"}]}}}