{"file":{},"class":{"name":"CFRequest","file":"utilities\/request.class.php","url":"..\/cfrequest\/index.html","description":"<p>Wraps the underlying <code>RequestCore<\/code> class with some AWS-specific customizations.<\/p>","metadata":{"version":"2011.12.02","license":"See the included NOTICE.md file for more information.","copyright":"See the included NOTICE.md file for more information.","link":"http:\/\/aws.amazon.com\/php\/ PHP Developer Center"},"inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"methods":{"@attributes":{"count":"28"},"method":[{"name":"__construct","modifier":"public","description":"<p>Constructs a new instance of this class.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"opt"},"name":"url","type":"string","description":"<p>The URL to request or service endpoint to query.<\/p>"},{"@attributes":{"choice":"opt"},"name":"proxy","type":"string","description":"<p>The faux-url to use for proxy settings. Takes the following format: <code>proxy:\/\/user:pass@hostname:port<\/code><\/p>"},{"@attributes":{"choice":"opt"},"name":"helpers","type":"array","description":"<p>An associative array of classnames to use for request, and response functionality. Gets passed in automatically by the calling class.<\/p>"},{"@attributes":{"choice":"opt"},"name":"credentials","type":"CFCredential","description":"<p>The credentials to use for signing and making requests.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function __construct($url = null, $proxy = null, $helpers = null, CFCredential $credentials = null)\n{\n    parent::__construct($url, $proxy, $helpers);\n\n    \/\/ Standard settings for all requests\n    $this-&gt;set_useragent(CFRUNTIME_USERAGENT);\n    $this-&gt;credentials = $credentials;\n    $this-&gt;cacert_location = ($this-&gt;credentials['certificate_authority'] ? $this-&gt;credentials['certificate_authority'] : false);\n\n    if (strpos(parse_url($url, PHP_URL_HOST), 'dynamodb') === 0)\n        {\n            $this-&gt;use_gzip_enconding = false;\n        }\n\n    return $this;\n}"},{"name":"__destruct","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Destructs the instance. Closes opened file handles.<\/p>","returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"249","end":"262","lines":"14"}}},{"name":"add_header","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Adds a custom HTTP header to the cURL request.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"key","type":"string","description":"<p>The custom HTTP header to set.<\/p>"},{"@attributes":{"choice":"req"},"name":"value","type":"mixed","description":"<p>The value to assign to the custom HTTP header.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"289","end":"293","lines":"5"}},"examples":{"example":{"title":"<p>Add a custom header to the request (without firing it).<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;prep_request();\n$http-&gt;add_header('x-requestcore-header', 'value');\n\nvar_dump($http-&gt;request_headers);","result":"array(3) {\n  [\"Expect\"]=&gt;\n  string(12) \"100-continue\"\n  [\"Connection\"]=&gt;\n  string(5) \"close\"\n  [\"x-requestcore-header\"]=&gt;\n  string(5) \"value\"\n}"}}},{"name":"get_response_body","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Get the HTTP response body from the request.<\/p>","returnvalue":"<p>The response body.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"982","end":"985","lines":"4"}},"examples":{"example":{"title":"<p>Display the response body of the request.<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;send_request();\n\nvar_dump($http-&gt;get_response_body());","result":"string(48) \"abcdefghijklmnopqrstuvwxyz\n0123456789\n!@#$%^&amp;*()\""}}},{"name":"get_response_code","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Get the HTTP response code from the request.<\/p>","returnvalue":"<p>The HTTP response code.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"992","end":"995","lines":"4"}},"examples":{"example":{"title":"<p>Display the response code for the request.<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;send_request();\n\nvar_dump($http-&gt;get_response_code());","result":"int(200)"}}},{"name":"get_response_header","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Get the HTTP response headers from the request.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"header","type":"string","description":"<p>A specific header value to return. Defaults to all headers.<\/p>"}},"returnvalue":"<p>All or selected header values.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"968","end":"975","lines":"8"}},"examples":{"example":[{"title":"<p>Display the Content-Type header of the response.<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;send_request();\n\nvar_dump($http-&gt;get_response_header('content-type'));","result":"string(25) \"text\/plain; charset=utf-8\""},{"title":"<p>Display the URL that was requested.<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;send_request();\n$info = $http-&gt;get_response_header('_info');\n\nvar_dump($info['url']);","result":"string(71) \"http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt\""}]}},{"name":"prep_request","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Prepares and adds the details of the cURL request. This can be passed along to a <a href=\"http:\/\/php.net\/curl_multi_exec\"><code>curl_multi_exec()<\/code><\/a>\nfunction.<\/p>","returnvalue":"<p>The handle for the cURL object.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"618","end":"771","lines":"154"}}},{"name":"process_response","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Take the post-processed cURL data and break it down into useful header\/body\/info chunks. Uses the\ndata stored in the <code>curl_handle<\/code> and <code>response<\/code> properties unless replacement data is passed in via\nparameters.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"opt"},"name":"curl_handle","type":"resource","description":"<p>The reference to the already executed cURL request.<\/p>"},{"@attributes":{"choice":"opt"},"name":"response","type":"string","description":"<p>The actual response content itself that needs to be parsed.<\/p>"}]},"returnvalue":"<p>A <ResponseCore> object containing a parsed HTTP response.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"782","end":"828","lines":"47"}}},{"name":"register_streaming_read_callback","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Register a callback function to execute whenever a data stream is read from using\n<a href=\"..\/cfrequest\/streaming_read_callback.html\" rel=\"..\/..\/index.html#m=cfrequest\/streaming_read_callback\"><code>CFRequest::streaming_read_callback()<\/code><\/a>.<\/p>\n\n<p>The user-defined callback function should accept three arguments:<\/p>\n\n<ul><li><code>$curl_handle<\/code> - <code>resource<\/code> - Required - The cURL handle resource that represents the in-progress transfer.<\/li><li><code>$file_handle<\/code> - <code>resource<\/code> - Required - The file handle resource that represents the file on the local file system.<\/li><li><code>$length<\/code> - <code>integer<\/code> - Required - The length in kilobytes of the data chunk that was transferred.<\/li><\/ul>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"callback","type":"string|array|function","description":"<p>The callback function is called by <a href=\"http:\/\/php.net\/call_user_func\"><code>call_user_func()<\/code><\/a>, so you can pass the following values:<\/p>\n\n<ul><li>The name of a global function to execute, passed as a string.<\/li><li>A method to execute, passed as <code>array('ClassName', 'MethodName')<\/code>.<\/li><li>An anonymous function (PHP 5.3+).<\/li><\/ul>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"503","end":"508","lines":"6"}}},{"name":"register_streaming_write_callback","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Register a callback function to execute whenever a data stream is written to using\n<a href=\"..\/cfrequest\/streaming_write_callback.html\" rel=\"..\/..\/index.html#m=cfrequest\/streaming_write_callback\"><code>CFRequest::streaming_write_callback()<\/code><\/a>.<\/p>\n\n<p>The user-defined callback function should accept two arguments:<\/p>\n\n<ul><li><code>$curl_handle<\/code> - <code>resource<\/code> - Required - The cURL handle resource that represents the in-progress transfer.<\/li><li><code>$length<\/code> - <code>integer<\/code> - Required - The length in kilobytes of the data chunk that was transferred.<\/li><\/ul>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"callback","type":"string|array|function","description":"<p>The callback function is called by <a href=\"http:\/\/php.net\/call_user_func\"><code>call_user_func()<\/code><\/a>, so you can pass the following values:<\/p>\n\n<ul><li>The name of a global function to execute, passed as a string.<\/li><li>A method to execute, passed as <code>array('ClassName', 'MethodName')<\/code>.<\/li><li>An anonymous function (PHP 5.3+).<\/li><\/ul>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"527","end":"532","lines":"6"}}},{"name":"remove_header","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Removes an HTTP header from the cURL request.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"key","type":"string","description":"<p>The custom HTTP header to set.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"301","end":"308","lines":"8"}},"examples":{"example":{"title":"<p>Add a custom header to the request, and then remove it (without firing it).<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;prep_request();\n\n$http-&gt;add_header('x-requestcore-header', 'value');\n$http-&gt;remove_header('x-requestcore-header');\n\nvar_dump($http-&gt;request_headers);","result":"array(2) {\n  [\"Expect\"]=&gt;\n  string(12) \"100-continue\"\n  [\"Connection\"]=&gt;\n  string(5) \"close\"\n}"}}},{"name":"send_multi_request","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Sends the request using <a href=\"http:\/\/php.net\/curl_multi_exec\"><code>curl_multi_exec()<\/code><\/a>, enabling parallel requests. Uses the &#8220;rolling&#8221; method.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"handles","type":"array","description":"<p>An indexed array of cURL handles to process simultaneously.<\/p>"},{"@attributes":{"choice":"opt"},"name":"opt","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>callback<\/code> - <code>string|array<\/code> - Optional - The string name of a function to pass the response data to. If this is a method, pass an array where the <code>[0]<\/code> index is the class and the <code>[1]<\/code> index is the method name.<\/li><li><code>limit<\/code> - <code>integer<\/code> - Optional - The number of simultaneous requests to make. This can be useful for scaling around slow server responses. Defaults to trusting cURLs judgement as to how many to use.<\/li><\/ul>"}]},"returnvalue":"<p>Post-processed cURL responses.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"872","end":"956","lines":"85"}},"examples":{"example":{"title":"<p>Collect cURL handles for two requests, fire them, and then display the response bodies.<\/p>","code":"$http = new RequestCore();\n\n$responses = $http-&gt;send_multi_request(array(\n\t$http-&gt;set_request_url('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt')-&gt;prep_request(),\n\t$http-&gt;set_request_url('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request2.txt')-&gt;prep_request()\n));\n\n$bodies = array(\n\t$responses[0]-&gt;body,\n\t$responses[1]-&gt;body,\n);\n\nvar_dump($bodies);","result":"array(2) {\n  [0]=&gt;\n  string(48) \"abcdefghijklmnopqrstuvwxyz\n0123456789\n!@#$%^&amp;*()\"\n  [1]=&gt;\n  string(48) \")(*&amp;^%$#@!\n9876543210\nzyxwvutsrqponmljkihgfedcba\"\n}"}}},{"name":"send_request","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Sends the request, calling necessary utility functions to update built-in properties.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"parse","type":"boolean","description":"<p>Whether to parse the response with ResponseCore or not.<\/p>"}},"returnvalue":"<p>The resulting unparsed data from the request.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"836","end":"861","lines":"26"}},"examples":{"example":[{"title":"<p>Send the request, parse it with ResponseCore, and display only the body.<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$response = $http-&gt;send_request(true);\n\nvar_dump($response-&gt;body);","result":"string(48) \"abcdefghijklmnopqrstuvwxyz\n0123456789\n!@#$%^&amp;*()\""},{"title":"<p>Set the URL via set_request_url() instead of the constructor, send the request, parse it with ResponseCore, and display only the body.<\/p>","code":"$http = new RequestCore();\n\n$http-&gt;set_request_url('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$response = $http-&gt;send_request(true);\n\nvar_dump($response-&gt;body);","result":"string(48) \"abcdefghijklmnopqrstuvwxyz\n0123456789\n!@#$%^&amp;*()\""},{"title":"<p>Send a HEAD request instead of a GET request, and display the response's Content-Type and non-body (since this is a HEAD request).<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;set_method($http::HTTP_HEAD);\n$response = $http-&gt;send_request(true);\n\nvar_dump($response-&gt;header['content-type']);\nvar_dump($response-&gt;body);","result":"string(25) \"text\/plain; charset=utf-8\"\nbool(false)"}]}},{"name":"set_body","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Set the body to send in the request.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"body","type":"string","description":"<p>The textual content to send along in the body of the request.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"340","end":"344","lines":"5"}}},{"name":"set_credentials","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Sets the credentials to use for authentication.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"user","type":"string","description":"<p>The username to authenticate with.<\/p>"},{"@attributes":{"choice":"req"},"name":"pass","type":"string","description":"<p>The password to authenticate with.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"275","end":"280","lines":"6"}},"examples":{"example":[{"title":"<p>Pass credentials to set_credentials(), and display the username to use.<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;set_credentials('user', 'pass');\n$http-&gt;prep_request();\n\nvar_dump($http-&gt;username);","result":"string(4) \"user\""},{"title":"<p>Pass credentials to set_credentials(), and display the password to use.<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;set_credentials('user', 'pass');\n$http-&gt;prep_request();\n\nvar_dump($http-&gt;password);","result":"string(4) \"pass\""}]}},{"name":"set_curlopts","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Set additional CURLOPT settings. These will merge with the default settings, and override if\nthere is a duplicate.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"curlopts","type":"array","description":"<p>A set of key-value pairs that set <code>CURLOPT<\/code> options. These will merge with the existing CURLOPTs, and ones passed here will override the defaults. Keys should be the <code>CURLOPT_*<\/code> constants, not strings.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"365","end":"369","lines":"5"}}},{"name":"set_method","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Set the method type for the request.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"method","type":"string","description":"<p>One of the following constants: <a href=\"..\/cfrequest\/constants.html#http_get\" rel=\"..\/..\/index.html#c=cfrequest\/http_get\"><code>HTTP_GET<\/code><\/a>, <a href=\"..\/cfrequest\/constants.html#http_post\" rel=\"..\/..\/index.html#c=cfrequest\/http_post\"><code>HTTP_POST<\/code><\/a>, <a href=\"..\/cfrequest\/constants.html#http_put\" rel=\"..\/..\/index.html#c=cfrequest\/http_put\"><code>HTTP_PUT<\/code><\/a>, <a href=\"..\/cfrequest\/constants.html#http_head\" rel=\"..\/..\/index.html#c=cfrequest\/http_head\"><code>HTTP_HEAD<\/code><\/a>, <a href=\"..\/cfrequest\/constants.html#http_delete\" rel=\"..\/..\/index.html#c=cfrequest\/http_delete\"><code>HTTP_DELETE<\/code><\/a>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"316","end":"320","lines":"5"}}},{"name":"set_proxy","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Set the proxy to use for making requests.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"proxy","type":"string","description":"<p>The faux-url to use for proxy settings. Takes the following format: <code>proxy:\/\/user:pass@hostname:port<\/code><\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"462","end":"470","lines":"9"}},"examples":{"example":{"title":"<p>Pass in a proxy DSN string, and process it with prep_request() (without firing the request).<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;set_proxy('proxy:\/\/user:pass@hostname:80');\n$http-&gt;prep_request();\n\nvar_dump($http-&gt;proxy);","result":"array(5) {\n  [\"scheme\"]=&gt;\n  string(5) \"proxy\"\n  [\"host\"]=&gt;\n  string(8) \"hostname\"\n  [\"port\"]=&gt;\n  int(80)\n  [\"user\"]=&gt;\n  string(4) \"user\"\n  [\"pass\"]=&gt;\n  string(4) \"pass\"\n}"}}},{"name":"set_read_file","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Sets the file to read from while streaming up.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"location","type":"string","description":"<p>The readable location to read from.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"421","end":"427","lines":"7"}}},{"name":"set_read_stream","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Sets the resource to read from while streaming up. Reads the stream from its current position until\nEOF or <code>$size<\/code> bytes have been read. If <code>$size<\/code> is not given it will be determined by <a href=\"http:\/\/php.net\/fstat\"><code>fstat()<\/code><\/a> and\n<a href=\"http:\/\/php.net\/ftell\"><code>ftell()<\/code><\/a>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"resource","type":"resource","description":"<p>The readable resource to read from.<\/p>"},{"@attributes":{"choice":"opt"},"name":"size","type":"integer","description":"<p>The size of the stream to read.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"393","end":"413","lines":"21"}}},{"name":"set_read_stream_size","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Sets the length in bytes to read from the stream while streaming up.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"size","type":"integer","description":"<p>The length in bytes to read from the stream.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"377","end":"382","lines":"6"}}},{"name":"set_request_url","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Set the URL to make the request to.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"url","type":"string","description":"<p>The URL to make the request to.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"352","end":"356","lines":"5"}},"examples":{"example":{"title":"<p>Send a request and display the URL we requested.<\/p>","code":"$http = new RequestCore();\n$http-&gt;set_request_url('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;send_request();\n\nvar_dump($http-&gt;response_info['url']);","result":"string(71) \"http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt\""}}},{"name":"set_seek_position","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Set the intended starting seek position.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"position","type":"integer","description":"<p>The byte-position of the stream to begin reading from.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"478","end":"483","lines":"6"}}},{"name":"set_useragent","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Sets a custom useragent string for the class.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"ua","type":"string","description":"<p>The useragent string to use.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"328","end":"332","lines":"5"}},"examples":{"example":{"title":"<p>Set a useragent string to use for HTTP requests.<\/p>","code":"$http = new RequestCore('http:\/\/github.com\/skyzyx\/requestcore\/raw\/master\/_tests\/test_request.txt');\n$http-&gt;set_useragent('SampleUserAgentString');\n\nvar_dump($http-&gt;useragent);","result":"string(21) \"SampleUserAgentString\""}}},{"name":"set_write_file","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Sets the file to write to while streaming down.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"location","type":"string","description":"<p>The writeable location to write to.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"448","end":"454","lines":"7"}}},{"name":"set_write_stream","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>Sets the resource to write to while streaming down.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"resource","type":"resource","description":"<p>The writeable resource to write to.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"435","end":"440","lines":"6"}}},{"name":"streaming_read_callback","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>A callback function that is invoked by cURL for streaming up.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"curl_handle","type":"resource","description":"<p>The cURL handle for the request.<\/p>"},{"@attributes":{"choice":"req"},"name":"file_handle","type":"resource","description":"<p>The open file handle resource.<\/p>"},{"@attributes":{"choice":"req"},"name":"length","type":"integer","description":"<p>The maximum number of bytes to read.<\/p>"}]},"returnvalue":"<p>Binary data from a stream.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"546","end":"576","lines":"31"}}},{"name":"streaming_write_callback","modifier":"public","inheritance":{"class":{"name":"RequestCore"}},"description":"<p>A callback function that is invoked by cURL for streaming down.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"curl_handle","type":"resource","description":"<p>The cURL handle for the request.<\/p>"},{"@attributes":{"choice":"req"},"name":"data","type":"binary","description":"<p>The data to write.<\/p>"}]},"returnvalue":"<p>The number of bytes written.<\/p>","source":{"@attributes":{"file":"lib\/requestcore\/requestcore.class.php","start":"585","end":"610","lines":"26"}}}]},"constants":{"@attributes":{"count":"5"},"constant":[{"name":"HTTP_DELETE","value":"DELETE"},{"name":"HTTP_GET","value":"GET"},{"name":"HTTP_HEAD","value":"HEAD"},{"name":"HTTP_POST","value":"POST"},{"name":"HTTP_PUT","value":"PUT"}]},"properties":{"@attributes":{"count":"32"},"property":[{"name":"allow_set_time_limit","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>Whether or not the set_time_limit function should be called.<\/p>"},{"name":"cacert_location","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The location of the cacert.pem file to use.<\/p>"},{"name":"credentials","modifier":"public","description":"<p>The active credential set.<\/p>"},{"name":"curl_handle","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The handle for the cURL object.<\/p>"},{"name":"curlopts","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>Custom CURLOPT settings.<\/p>"},{"name":"debug_mode","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The state of debug mode.<\/p>"},{"name":"method","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The method by which the request is being made.<\/p>"},{"name":"password","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The password to use for the request.<\/p>"},{"name":"proxy","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>Stores the proxy settings to use for the request.<\/p>"},{"name":"read_file","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>File to read from while streaming up.<\/p>"},{"name":"read_stream","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The resource to read from while streaming up.<\/p>"},{"name":"read_stream_read","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The length already read from the stream.<\/p>"},{"name":"read_stream_size","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The size of the stream to read from.<\/p>"},{"name":"registered_streaming_read_callback","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The user-defined callback function to call when a stream is read from.<\/p>"},{"name":"registered_streaming_write_callback","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The user-defined callback function to call when a stream is written to.<\/p>"},{"name":"request_body","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The body being sent in the request.<\/p>"},{"name":"request_class","modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFRequest"}},"description":"<p>The default class to use for HTTP Requests (defaults to <a href=\"..\/cfrequest\/index.html\" rel=\"..\/..\/index.html#i=cfrequest\"><code>CFRequest<\/code><\/a>).<\/p>"},{"name":"request_headers","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The headers being sent in the request.<\/p>"},{"name":"request_url","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The URL being requested.<\/p>"},{"name":"response","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The response returned by the request.<\/p>"},{"name":"response_body","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The body returned by the request.<\/p>"},{"name":"response_class","modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFResponse"}},"description":"<p>The default class to use for HTTP Responses (defaults to <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a>).<\/p>"},{"name":"response_code","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The HTTP status code returned by the request.<\/p>"},{"name":"response_headers","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The headers returned by the request.<\/p>"},{"name":"response_info","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>Additional response data.<\/p>"},{"name":"seek_position","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>Stores the intended starting seek position.<\/p>"},{"name":"ssl_verification","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of SSL certificate verification.<\/p>"},{"name":"use_gzip_enconding","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>Whether or not to use gzip encoding via CURLOPT_ENCODING<\/p>"},{"name":"useragent","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","initializer":{"@attributes":{"type":"string","value":"RequestCore\/1.4.4"}},"description":"<p>Default useragent string to use.<\/p>"},{"name":"username","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The username to use for the request.<\/p>"},{"name":"write_file","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>File to write to while streaming down.<\/p>"},{"name":"write_stream","inheritance":{"class":{"name":"RequestCore","file":"lib\/requestcore\/requestcore.class.php"}},"modifier":"public","description":"<p>The resource to write to while streaming down.<\/p>"}]}}}