{"file":{},"class":{"name":"CFRuntime","file":"sdk.class.php","url":"..\/cfruntime\/index.html","description":"<p>Core functionality and default settings shared across all SDK classes. All methods and properties in this\nclass are inherited by the service-specific classes.<\/p>","metadata":{"version":"2012.05.25","license":"See the included NOTICE.md file for more information.","copyright":"See the included NOTICE.md file for more information.","link":"http:\/\/aws.amazon.com\/php\/ PHP Developer Center"},"methods":{"@attributes":{"count":"29"},"method":[{"name":"__construct","modifier":"public","description":"<p>The constructor. This class should not be instantiated directly. Rather, a service-specific class\nshould be instantiated.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"options","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>certificate_authority<\/code> - <code>boolean<\/code> - Optional - Determines which Cerificate Authority file to use. A value of boolean <code>false<\/code> will use the Certificate Authority file available on the system. A value of boolean <code>true<\/code> will use the Certificate Authority provided by the SDK. Passing a file system path to a Certificate Authority file (chmodded to <code>0755<\/code>) will use that. Leave this set to <code>false<\/code> if you&#8217;re not sure.<\/li><li><code>credentials<\/code> - <code>string<\/code> - Optional - The name of the credential set to use for authentication.<\/li><li><code>default_cache_config<\/code> - <code>string<\/code> - Optional - This option allows a preferred storage type to be configured for long-term caching. This can be changed later using the <a href=\"..\/cfruntime\/set_cache_config.html\" rel=\"..\/..\/index.html#m=cfruntime\/set_cache_config\"><code>set_cache_config()<\/code><\/a> method. Valid values are: <code>apc<\/code>, <code>xcache<\/code>, or a file system path such as <code>.\/cache<\/code> or <code>\/tmp\/cache\/<\/code>.<\/li><li><code>key<\/code> - <code>string<\/code> - Optional - Your AWS key, or a session key. If blank, the default credential set will be used.<\/li><li><code>instance_profile_timeout<\/code> - <code>integer<\/code> - Optional - When retrieving IAM instance profile credentials, there is a hard connection timeout that defaults to 2 seconds to prevent unnecessary on non-EC2 systems. This setting allows you to change that timeout if needed.<\/li><li><code>secret<\/code> - <code>string<\/code> - Optional - Your AWS secret key, or a session secret key. If blank, the default credential set will be used.<\/li><li><code>token<\/code> - <code>string<\/code> - Optional - An AWS session token.<\/li><li><code>use_instance_profile_credentials<\/code> - <code>boolean<\/code> - Optional - Forces the use of IAM Instance Profile credentials, even when regular credentials are provided.<\/li><\/ul>"}},"returnvalue":{"@attributes":{"type":"void"}},"source":"public function __construct(array $options = array())\n{\n    \/\/ Instantiate the utilities class.\n    $this-&gt;util = new $this-&gt;utilities_class();\n\n    \/\/ Determine the current service.\n    $this-&gt;service = get_class($this);\n\n    \/\/ Create credentials based on the options\n    $runtime_credentials = new CFCredential($options);\n    $credentials_provided = false;\n\n    \/\/ Retrieve a credential set from config.inc.php if it exists\n    if (isset($options['credentials']))\n    {\n        \/\/ Use a specific credential set and merge with the runtime credentials\n        $this-&gt;credentials = CFCredentials::get($options['credentials'])\n            -&gt;merge($runtime_credentials);\n    }\n    else\n    {\n        try\n        {\n            \/\/ Use the default credential set and merge with the runtime credentials\n            $this-&gt;credentials = CFCredentials::get(CFCredentials::DEFAULT_KEY)\n                -&gt;merge($runtime_credentials);\n        }\n        catch (CFCredentials_Exception $e)\n        {\n            \/\/ Only the runtime credentials were provided\n            $this-&gt;credentials = $runtime_credentials;\n        }\n    }\n\n    \/\/ Check if keys were actually provided\n    if (isset($this-&gt;credentials['key']) &amp;&amp; isset($this-&gt;credentials['secret']))\n    {\n        $credentials_provided = true;\n    }\n\n    \/\/ Check for an instance profile credentials override\n    if (isset($this-&gt;credentials['use_instance_profile_credentials']) &amp;&amp; $this-&gt;credentials['use_instance_profile_credentials'])\n    {\n        $credentials_provided = false;\n    }\n\n    \/\/ Automatically enable whichever caching mechanism is set to default.\n    $this-&gt;set_cache_config($this-&gt;credentials-&gt;default_cache_config);\n\n    \/\/ If no credentials were provided, try to get them from the EC2 instance profile\n    if (!$credentials_provided)\n    {\n        \/\/ Default caching mechanism is required\n        if (!$this-&gt;credentials-&gt;default_cache_config)\n        {\n            \/\/ @codeCoverageIgnoreStart\n            throw new CFCredentials_Exception('No credentials were provided. The SDK attempts to retrieve Instance '\n                . 'Profile credentials from the EC2 Instance Metadata Service, but doing this requires the '\n                . '&quot;default_cache_config&quot; option to be set in the config.inc.php file or constructor. In order to '\n                . 'cache the retrieved credentials.');\n            \/\/ @codeCoverageIgnoreEnd\n        }\n\n        \/\/ Instantiate and invoke the cache for instance profile credentials\n        $cache = new $this-&gt;cache_class('instance_profile_credentials', $this-&gt;cache_location, 0, $this-&gt;cache_compress);\n        if ($data = $cache-&gt;read())\n        {\n            $cache-&gt;expire_in((strtotime($data['expires']) - time()) * 0.85);\n        }\n        $instance_profile_credentials = $cache-&gt;response_manager(array($this, 'cache_instance_profile_credentials'), array($cache, $options));\n\n        $this-&gt;credentials-&gt;key    = $instance_profile_credentials['key'];\n        $this-&gt;credentials-&gt;secret = $instance_profile_credentials['secret'];\n        $this-&gt;credentials-&gt;token  = $instance_profile_credentials['token'];\n    }\n\n    \/\/ Set internal credentials after they are resolved\n    $this-&gt;key = $this-&gt;credentials-&gt;key;\n    $this-&gt;secret_key = $this-&gt;credentials-&gt;secret;\n    $this-&gt;auth_token = $this-&gt;credentials-&gt;token;\n}"},{"name":"allow_hostname_override","modifier":"public","description":"<p>Disables any subsequent use of the <a href=\"..\/cfruntime\/set_hostname.html\" rel=\"..\/..\/index.html#m=cfruntime\/set_hostname\"><code>set_hostname()<\/code><\/a> method.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"override","type":"boolean","description":"<p>Whether or not subsequent calls to <a href=\"..\/cfruntime\/set_hostname.html\" rel=\"..\/..\/index.html#m=cfruntime\/set_hostname\"><code>set_hostname()<\/code><\/a> should be obeyed. A <code>false<\/code> value disables the further effectiveness of <a href=\"..\/cfruntime\/set_hostname.html\" rel=\"..\/..\/index.html#m=cfruntime\/set_hostname\"><code>set_hostname()<\/code><\/a>. Defaults to <code>true<\/code>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function allow_hostname_override($override = true)\n{\n    $this-&gt;override_hostname = $override;\n    return $this;\n}","examples":{"example":{"title":"<p>Configure hostname and resource prefix.<\/p>","code":"\/\/ Instantiate the class\n$ec2 = new AmazonEC2();\n\n\/\/ Point to the new endpoint\n$ec2-&gt;set_hostname('localhost', '3000');\n$ec2-&gt;set_resource_prefix('\/services\/mock');\n$ec2-&gt;allow_hostname_override(false);\n\n\/\/ Describe images\n$response = $ec2-&gt;describe_images();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"authenticate","modifier":"public","description":"<p>Default, shared method for authenticating a connection to AWS.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"operation","type":"string","description":"<p>Indicates the operation to perform.<\/p>"},{"@attributes":{"choice":"req"},"name":"payload","type":"array","description":"<p>An associative array of parameters for authenticating. See the individual methods for allowed keys.<\/p>"}]},"returnvalue":"<p>Object containing a parsed HTTP response.<\/p>","source":"public function authenticate($operation, $payload)\n{\n    $original_payload = $payload;\n    $method_arguments = func_get_args();\n    $curlopts = array();\n    $return_curl_handle = false;\n\n    if (substr($operation, 0, strlen($this-&gt;operation_prefix)) !== $this-&gt;operation_prefix)\n    {\n        $operation = $this-&gt;operation_prefix . $operation;\n    }\n\n    \/\/ Extract the custom CURLOPT settings from the payload\n    if (is_array($payload) &amp;&amp; isset($payload['curlopts']))\n    {\n        $curlopts = $payload['curlopts'];\n        unset($payload['curlopts']);\n    }\n\n    \/\/ Determine whether the response or curl handle should be returned\n    if (is_array($payload) &amp;&amp; isset($payload['returnCurlHandle']))\n    {\n        $return_curl_handle = isset($payload['returnCurlHandle']) ? $payload['returnCurlHandle'] : false;\n        unset($payload['returnCurlHandle']);\n    }\n\n    \/\/ Use the caching flow to determine if we need to do a round-trip to the server.\n    if ($this-&gt;use_cache_flow)\n    {\n        \/\/ Generate an identifier specific to this particular set of arguments.\n        $cache_id = $this-&gt;key . '_' . get_class($this) . '_' . $operation . '_' . sha1(serialize($method_arguments));\n\n        \/\/ Instantiate the appropriate caching object.\n        $this-&gt;cache_object = new $this-&gt;cache_class($cache_id, $this-&gt;cache_location, $this-&gt;cache_expires, $this-&gt;cache_compress);\n\n        if ($this-&gt;delete_cache)\n        {\n            $this-&gt;use_cache_flow = false;\n            $this-&gt;delete_cache = false;\n            return $this-&gt;cache_object-&gt;delete();\n        }\n\n        \/\/ Invoke the cache callback function to determine whether to pull data from the cache or make a fresh request.\n        $data = $this-&gt;cache_object-&gt;response_manager(array($this, 'cache_callback'), $method_arguments);\n\n        \/\/ Parse the XML body\n        $data = $this-&gt;parse_callback($data);\n\n        \/\/ End!\n        return $data;\n    }\n\n    \/*%******************************************************************************************%*\/\n\n    \/\/ Signer\n    $signer = new $this-&gt;auth_class($this-&gt;hostname, $operation, $payload, $this-&gt;credentials);\n    $signer-&gt;key = $this-&gt;key;\n    $signer-&gt;secret_key = $this-&gt;secret_key;\n    $signer-&gt;auth_token = $this-&gt;auth_token;\n    $signer-&gt;api_version = $this-&gt;api_version;\n    $signer-&gt;utilities_class = $this-&gt;utilities_class;\n    $signer-&gt;request_class = $this-&gt;request_class;\n    $signer-&gt;response_class = $this-&gt;response_class;\n    $signer-&gt;use_ssl = $this-&gt;use_ssl;\n    $signer-&gt;proxy = $this-&gt;proxy;\n    $signer-&gt;util = $this-&gt;util;\n    $signer-&gt;registered_streaming_read_callback = $this-&gt;registered_streaming_read_callback;\n    $signer-&gt;registered_streaming_write_callback = $this-&gt;registered_streaming_write_callback;\n    $request = $signer-&gt;authenticate();\n\n    \/\/ Update RequestCore settings\n    $request-&gt;request_class = $this-&gt;request_class;\n    $request-&gt;response_class = $this-&gt;response_class;\n    $request-&gt;ssl_verification = $this-&gt;ssl_verification;\n\n    \/*%******************************************************************************************%*\/\n\n    \/\/ Debug mode\n    if ($this-&gt;debug_mode)\n    {\n        $request-&gt;debug_mode = $this-&gt;debug_mode;\n    }\n\n    \/\/ Set custom CURLOPT settings\n    if (count($curlopts))\n    {\n        $request-&gt;set_curlopts($curlopts);\n    }\n\n    \/\/ Manage the (newer) batch request API or the (older) returnCurlHandle setting.\n    if ($this-&gt;use_batch_flow)\n    {\n        $handle = $request-&gt;prep_request();\n        $this-&gt;batch_object-&gt;add($handle);\n        $this-&gt;use_batch_flow = false;\n\n        return $handle;\n    }\n    elseif ($return_curl_handle)\n    {\n        return $request-&gt;prep_request();\n    }\n\n    \/\/ Send!\n    $request-&gt;send_request();\n\n    \/\/ Prepare the response.\n    $headers = $request-&gt;get_response_header();\n    $headers['x-aws-stringtosign'] = $signer-&gt;string_to_sign;\n\n    if (isset($signer-&gt;canonical_request))\n    {\n        $headers['x-aws-canonicalrequest'] = $signer-&gt;canonical_request;\n    }\n\n    $headers['x-aws-request-headers'] = $request-&gt;request_headers;\n    $headers['x-aws-body'] = $signer-&gt;querystring;\n\n    $data = new $this-&gt;response_class($headers, ($this-&gt;parse_the_response === true) ? $this-&gt;parse_callback($request-&gt;get_response_body()) : $request-&gt;get_response_body(), $request-&gt;get_response_code());\n\n    $response_body = (string) $request-&gt;get_response_body();\n\n    \/\/ Was it Amazon's fault the request failed? Retry the request until we reach $max_retries.\n    if (\n        (integer) $request-&gt;get_response_code() === 500 || \/\/ Internal Error (presumably transient)\n        (integer) $request-&gt;get_response_code() === 503)   \/\/ Service Unavailable (presumably transient)\n    {\n        if ($this-&gt;redirects &lt;= $this-&gt;max_retries)\n        {\n            \/\/ Exponential backoff\n            $delay = (integer) (pow(4, $this-&gt;redirects) * 100000);\n            usleep($delay);\n            $this-&gt;redirects++;\n            $data = $this-&gt;authenticate($operation, $original_payload);\n        }\n    }\n\n    \/\/ DynamoDB has additional, custom logic for retrying requests\n    else\n    {\n        \/\/ If the request to DynamoDB was throttled, we need to retry\n        $need_to_retry_dynamodb_request = (\n            (integer) $request-&gt;get_response_code() === 400 &amp;&amp;\n            stripos($response_body, 'com.amazonaws.dynamodb.') !== false &amp;&amp;\n            stripos($response_body, 'ProvisionedThroughputExceededException') !== false\n        );\n\n        \/\/ If the CRC32 of the response does not match the expected value, we need to retry\n        $response_headers = $request-&gt;get_response_header();\n        if (!$need_to_retry_dynamodb_request &amp;&amp; isset($response_headers['x-amz-crc32']))\n        {\n            $crc32_expected = $response_headers['x-amz-crc32'];\n            $crc32_actual = hexdec(hash('crc32b', $response_body));\n            $need_to_retry_dynamodb_request = ($crc32_expected != $crc32_actual);\n        }\n\n        \/\/ Perform retry if necessary using a more aggressive exponential backoff\n        if ($need_to_retry_dynamodb_request)\n        {\n            if ($this-&gt;redirects === 0)\n            {\n                $this-&gt;redirects++;\n                $data = $this-&gt;authenticate($operation, $original_payload);\n            }\n            elseif ($this-&gt;redirects &lt;= max($this-&gt;max_retries, 10))\n            {\n                \/\/ Exponential backoff\n                $delay = (integer) (pow(2, ($this-&gt;redirects - 1)) * 50000);\n                usleep($delay);\n                $this-&gt;redirects++;\n                $data = $this-&gt;authenticate($operation, $original_payload);\n            }\n        }\n    }\n\n    $this-&gt;redirects = 0;\n    return $data;\n}"},{"name":"batch","modifier":"public","description":"<p>Specifies that the intended request should be queued for a later batch request.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"queue","type":"CFBatchRequest","description":"<p>The <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a> instance to use for managing batch requests. If not available, it generates a new instance of <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function batch(CFBatchRequest &amp;$queue = null)\n{\n    if ($queue)\n    {\n        $this-&gt;batch_object = $queue;\n    }\n    elseif ($this-&gt;internal_batch_object)\n    {\n        $this-&gt;batch_object = &amp;$this-&gt;internal_batch_object;\n    }\n    else\n    {\n        $this-&gt;internal_batch_object = new $this-&gt;batch_class();\n        $this-&gt;batch_object = &amp;$this-&gt;internal_batch_object;\n    }\n\n    $this-&gt;use_batch_flow = true;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Batch several requests together, and then cache the responses to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"},{"title":"<p>Batch several requests together, and then cache the responses to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"}]}},{"name":"cache","modifier":"public","description":"<p>Specifies that the resulting <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object should be cached according to the settings from\n<a href=\"..\/cfruntime\/set_cache_config.html\" rel=\"..\/..\/index.html#m=cfruntime\/set_cache_config\"><code>set_cache_config()<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"expires","type":"string|integer","description":"<p>The time the cache is to expire. Accepts a number of seconds as an integer, or an amount of time, as a string, that is understood by <a href=\"http:\/\/php.net\/strtotime\"><code>strtotime()<\/code><\/a> (e.g. &#8220;1 hour&#8221;).<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function cache($expires)\n{\n    \/\/ Die if they haven't used set_cache_config().\n    if (!$this-&gt;cache_class)\n    {\n        throw new CFRuntime_Exception('Must call set_cache_config() before using cache()');\n    }\n\n    if (is_string($expires))\n    {\n        $expires = strtotime($expires);\n        $this-&gt;cache_expires = $expires - time();\n    }\n    elseif (is_int($expires))\n    {\n        $this-&gt;cache_expires = $expires;\n    }\n\n    $this-&gt;use_cache_flow = true;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Fire a single request, and then cache the response to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Fire a single request, and then cache the response to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Fire a single request, and then cache the response to Memcache.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config(array(\n\tarray('host' =&gt; '127.0.0.1')\n));\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Batch several requests together, and then cache the responses to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"},{"title":"<p>Batch several requests together, and then cache the responses to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ Prepare for parallel requests\n$sdb-&gt;batch()-&gt;list_domains();\n$sdb-&gt;batch()-&gt;list_domains();\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;batch()-&gt;cache('1 minute')-&gt;send(false);\nvar_dump($response[0]-&gt;isOK());\nvar_dump($response[1]-&gt;isOK());","result":"bool(true)\nbool(true)\nbool(true)\nbool(true)"}]}},{"name":"cache_instance_profile_credentials","modifier":"public","description":"<p>Fetches and caches EC2 instance profile credentials. This is meant to be used by the constructor, and is not to\nbe manually invoked.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"cache","type":"CacheCore","description":"<p>The a reference to the cache object that is being used to handle the caching.<\/p>"},{"@attributes":{"choice":"req"},"name":"options","type":"array","description":"<p>The options that were passed into the constructor.<\/p>"}]},"returnvalue":"<p>The data to be cached, or NULL.<\/p>","source":"public function cache_instance_profile_credentials($cache, $options)\n{\n    $instance_profile_url = 'http:\/\/169.254.169.254\/latest\/meta-data\/iam\/security-credentials\/';\n    $connect_timeout = isset($options['instance_profile_timeout']) ? $options['instance_profile_timeout'] : 2;\n\n    try\n    {\n        \/\/ Make a call to the EC2 Metadata Service to find the available instance profile\n        $request = new RequestCore($instance_profile_url);\n        $request-&gt;set_curlopts(array(CURLOPT_CONNECTTIMEOUT =&gt; $connect_timeout));\n        $response = $request-&gt;send_request(true);\n\n        if ($response-&gt;isOK())\n        {\n            \/\/ Get the instance profile name\n            $profile = (string) $response-&gt;body;\n\n            \/\/ Make a call to the EC2 Metadata Service to get the instance profile credentials\n            $request = new RequestCore($instance_profile_url . $profile);\n            $request-&gt;set_curlopts(array(CURLOPT_CONNECTTIMEOUT =&gt; $connect_timeout));\n            $response = $request-&gt;send_request(true);\n\n            if ($response-&gt;isOK())\n            {\n                \/\/ Get the credentials\n                $credentials = json_decode($response-&gt;body, true);\n\n                if ($credentials['Code'] === 'Success')\n                {\n                    \/\/ Determine the expiration time\n                    $expiration_time = strtotime((string) $credentials['Expiration']);\n                    $expiration_duration = round(($expiration_time - time()) * 0.85);\n                    $cache-&gt;expire_in($expiration_duration);\n\n                    \/\/ Return the credential information\n                    return array(\n                        'key'     =&gt; $credentials['AccessKeyId'],\n                        'secret'  =&gt; $credentials['SecretAccessKey'],\n                        'token'   =&gt; $credentials['Token'],\n                        'expires' =&gt; $credentials['Expiration'],\n                    );\n                }\n            }\n        }\n    }\n    catch (cURL_Exception $e)\n    {\n        \/\/ The EC2 Metadata Service does not exist or had timed out.\n        \/\/ An exception will be thrown on the next line.\n    }\n\n    \/\/ @codeCoverageIgnoreStart\n    throw new CFCredentials_Exception('No credentials were provided. The SDK attempted to retrieve Instance '\n        . 'Profile credentials from the EC2 Instance Metadata Service, but failed to do so. Instance profile '\n        . 'credentials are only accessible on EC2 instances configured with a specific IAM role.');\n    \/\/ @codeCoverageIgnoreEnd\n}"},{"name":"cache_sts_credentials","modifier":"public","description":"<p>Fetches and caches STS credentials. This is meant to be used by the constructor, and is not to be\nmanually invoked.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"cache","type":"CacheCore","description":"<p>The a reference to the cache object that is being used to handle the caching.<\/p>"},{"@attributes":{"choice":"req"},"name":"options","type":"array","description":"<p>The options that were passed into the constructor.<\/p>"}]},"returnvalue":"<p>The data to be cached, or NULL.<\/p>","source":"public function cache_sts_credentials($cache, $options)\n{\n    $token = new AmazonSTS($options);\n    $response = $token-&gt;get_session_token();\n\n    if ($response-&gt;isOK())\n    {\n        \/\/ Update the expiration\n        $expiration_time = strtotime((string) $response-&gt;body-&gt;GetSessionTokenResult-&gt;Credentials-&gt;Expiration);\n        $expiration_duration = round(($expiration_time - time()) * 0.85);\n        $cache-&gt;expire_in($expiration_duration);\n\n        \/\/ Return the important data\n        $credentials = $response-&gt;body-&gt;GetSessionTokenResult-&gt;Credentials;\n\n        return array(\n            'key'     =&gt; (string) $credentials-&gt;AccessKeyId,\n            'secret'  =&gt; (string) $credentials-&gt;SecretAccessKey,\n            'token'   =&gt; (string) $credentials-&gt;SessionToken,\n            'expires' =&gt; (string) $credentials-&gt;Expiration,\n        );\n    }\n\n    \/\/ @codeCoverageIgnoreStart\n    throw new STS_Exception('Temporary credentials from the AWS Security '\n        . 'Token Service could not be retrieved using the provided long '\n        . 'term credentials. It\\'s possible that the provided long term '\n        . 'credentials were invalid.');\n    \/\/ @codeCoverageIgnoreEnd\n}"},{"name":"delete_cache","modifier":"public","description":"<p>Deletes a cached <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> object using the specified cache storage type.<\/p>","returnvalue":"<p>A value of <code>true<\/code> if cached object exists and is successfully deleted, otherwise <code>false<\/code>.<\/p>","source":"public function delete_cache()\n{\n    $this-&gt;use_cache_flow = true;\n    $this-&gt;delete_cache = true;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Delete a cached response from APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ Delete the data\n$response = $sdb-&gt;delete_cache()-&gt;list_domains();\nvar_dump($response);","result":"bool(false)"},{"title":"<p>Delete a cached response from the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ Delete the data\n$response = $sdb-&gt;delete_cache()-&gt;list_domains();\nvar_dump($response);","result":"bool(true)"},{"title":"<p>Delete a cached response from Memcache.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config(array(\n\tarray('host' =&gt; '127.0.0.1')\n));\n\n\/\/ Delete the data\n$response = $sdb-&gt;delete_cache()-&gt;list_domains();\nvar_dump($response);","result":"bool(true)"}]}},{"name":"disable_ssl","modifier":"public","description":"<p>Disables SSL\/HTTPS connections for hosts that don&#8217;t support them. Some services, however, still\nrequire SSL support.<\/p>\n\n<p>This method will throw a user warning when invoked, which can be hidden by changing your\n<a href=\"http:\/\/php.net\/error_reporting\"><code>error_reporting()<\/code><\/a> settings.<\/p>","returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function disable_ssl()\n{\n    trigger_error('Disabling SSL connections is potentially unsafe and highly discouraged.', E_USER_WARNING);\n    $this-&gt;use_ssl = false;\n    return $this;\n}","examples":{"example":{"title":"<p>Disable SSL for the service.<\/p>","description":"<ul>\n<li><strong>Note:<\/strong> It is not recommended to disable SSL.<\/li>\n<li><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/li>\n<\/ul>","code":"\/\/ Disable SSL (Using @ to suppress the warning from the output)\n$sdb = new AmazonSDB();\n@$sdb-&gt;disable_ssl(true);\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;use_ssl);","result":"bool(false)"}},"changelog":{"change":"<p>This method now throws a user warning when invoked.<\/p>"}},{"name":"disable_ssl_verification","modifier":"public","description":"<p>Disables the verification of the SSL Certificate Authority. Doing so can enable an attacker to carry\nout a man-in-the-middle attack.<\/p>\n\n<p>https:\/\/secure.wikimedia.org\/wikipedia\/en\/wiki\/Man-in-the-middle_attack<\/p>\n\n<p>This method will throw a user warning when invoked, which can be hidden by changing your\n<a href=\"http:\/\/php.net\/error_reporting\"><code>error_reporting()<\/code><\/a> settings.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"ssl_verification"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function disable_ssl_verification($ssl_verification = false)\n{\n    trigger_error('Disabling the verification of SSL certificates can lead to man-in-the-middle attacks. It is potentially unsafe and highly discouraged.', E_USER_WARNING);\n    $this-&gt;ssl_verification = $ssl_verification;\n    return $this;\n}"},{"name":"enable_debug_mode","modifier":"public","description":"<p>Enables HTTP request\/response header logging to <code>STDERR<\/code>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"enabled","type":"boolean","description":"<p>Whether or not to enable debug mode. Defaults to <code>true<\/code>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function enable_debug_mode($enabled = true)\n{\n    $this-&gt;debug_mode = $enabled;\n    return $this;\n}","examples":{"example":{"title":"<p>Enable the logging of HTTP request\/response headers to <code>STDERR<\/code>.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Adjust offset\n$sdb = new AmazonSDB();\n$sdb-&gt;enable_debug_mode();\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;debug_mode);","result":"bool(true)"}}},{"name":"factory","modifier":"public static","description":"<p>Alternate approach to constructing a new instance. Supports chaining.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"options","type":"array","description":"<p>An associative array of parameters that can have the following keys:<\/p>\n\n<ul><li><code>certificate_authority<\/code> - <code>boolean<\/code> - Optional - Determines which Cerificate Authority file to use. A value of boolean <code>false<\/code> will use the Certificate Authority file available on the system. A value of boolean <code>true<\/code> will use the Certificate Authority provided by the SDK. Passing a file system path to a Certificate Authority file (chmodded to <code>0755<\/code>) will use that. Leave this set to <code>false<\/code> if you&#8217;re not sure.<\/li><li><code>credentials<\/code> - <code>string<\/code> - Optional - The name of the credential set to use for authentication.<\/li><li><code>default_cache_config<\/code> - <code>string<\/code> - Optional - This option allows a preferred storage type to be configured for long-term caching. This can be changed later using the <a href=\"..\/cfruntime\/set_cache_config.html\" rel=\"..\/..\/index.html#m=cfruntime\/set_cache_config\"><code>set_cache_config()<\/code><\/a> method. Valid values are: <code>apc<\/code>, <code>xcache<\/code>, or a file system path such as <code>.\/cache<\/code> or <code>\/tmp\/cache\/<\/code>.<\/li><li><code>key<\/code> - <code>string<\/code> - Optional - Your AWS key, or a session key. If blank, the default credential set will be used.<\/li><li><code>secret<\/code> - <code>string<\/code> - Optional - Your AWS secret key, or a session secret key. If blank, the default credential set will be used.<\/li><li><code>token<\/code> - <code>string<\/code> - Optional - An AWS session token.<\/li><\/ul>"}},"returnvalue":{"@attributes":{"type":"void"}},"source":"public static function factory(array $options = array())\n{\n    if (version_compare(PHP_VERSION, '5.3.0', '&lt;'))\n    {\n        throw new Exception('PHP 5.3 or newer is required to instantiate a new class with CLASS::factory().');\n    }\n\n    $self = get_called_class();\n    return new $self($options);\n}"},{"name":"register_streaming_read_callback","modifier":"public","description":"<p>Register a callback function to execute whenever a data stream is read from using\n<a href=\"..\/cfrequest\/streaming_read_callback.html\" rel=\"..\/..\/index.html#m=cfrequest\/streaming_read_callback\"><code>CFRequest::streaming_read_callback()<\/code><\/a>.<\/p>\n\n<p>The user-defined callback function should accept three arguments:<\/p>\n\n<ul><li><code>$curl_handle<\/code> - <code>resource<\/code> - Required - The cURL handle resource that represents the in-progress transfer.<\/li><li><code>$file_handle<\/code> - <code>resource<\/code> - Required - The file handle resource that represents the file on the local file system.<\/li><li><code>$length<\/code> - <code>integer<\/code> - Required - The length in kilobytes of the data chunk that was transferred.<\/li><\/ul>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"callback","type":"string|array|function","description":"<p>The callback function is called by <a href=\"http:\/\/php.net\/call_user_func\"><code>call_user_func()<\/code><\/a>, so you can pass the following values:<\/p>\n\n<ul><li>The name of a global function to execute, passed as a string.<\/li><li>A method to execute, passed as <code>array('ClassName', 'MethodName')<\/code>.<\/li><li>An anonymous function (PHP 5.3+).<\/li><\/ul>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function register_streaming_read_callback($callback)\n{\n    $this-&gt;registered_streaming_read_callback = $callback;\n    return $this;\n}","examples":{"example":{"title":"<p>Register a callback function to execute when data is read from an open stream.<\/p>","code":"$s3 = new AmazonS3();\n\n\/\/ Define callback function\nfunction callback($curl_handle, $file_handle, $length)\n{\n\t$total = curl_getinfo($curl_handle, CURLINFO_CONTENT_LENGTH_DOWNLOAD);\n\t$transferred = curl_getinfo($curl_handle, CURLINFO_SIZE_DOWNLOAD);\n\t$percentage = floor(($transferred \/ $total) * 100);\n\n\techo 'Transferred ' . $transferred . ' KB out of ' . $total . ' KB (' . $percentage . '%)' . PHP_EOL;\n}\n\n\/\/ Register a callback function to execute when a stream is read locally.\n$s3-&gt;register_streaming_read_callback('callback');\n\n\/\/ Download a public object.\n$response = $s3-&gt;create_object('my-bucket', 'my-large-file.mp4', array(\n\t'fileUpload' =&gt; '.\/uploads\/my-large-file.mp4'\n));","result":"Transferred 7200 bytes out of 9000000 (0%)\n...\nTransferred 9000000 bytes out of 9000000 (100%)"}}},{"name":"register_streaming_write_callback","modifier":"public","description":"<p>Register a callback function to execute whenever a data stream is written to using\n<a href=\"..\/cfrequest\/streaming_write_callback.html\" rel=\"..\/..\/index.html#m=cfrequest\/streaming_write_callback\"><code>CFRequest::streaming_write_callback()<\/code><\/a>.<\/p>\n\n<p>The user-defined callback function should accept two arguments:<\/p>\n\n<ul><li><code>$curl_handle<\/code> - <code>resource<\/code> - Required - The cURL handle resource that represents the in-progress transfer.<\/li><li><code>$length<\/code> - <code>integer<\/code> - Required - The length in kilobytes of the data chunk that was transferred.<\/li><\/ul>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"callback","type":"string|array|function","description":"<p>The callback function is called by <a href=\"http:\/\/php.net\/call_user_func\"><code>call_user_func()<\/code><\/a>, so you can pass the following values:<\/p>\n\n<ul><li>The name of a global function to execute, passed as a string.<\/li><li>A method to execute, passed as <code>array('ClassName', 'MethodName')<\/code>.<\/li><li>An anonymous function (PHP 5.3+).<\/li><\/ul>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function register_streaming_write_callback($callback)\n{\n    $this-&gt;registered_streaming_write_callback = $callback;\n    return $this;\n}","examples":{"example":{"title":"<p>Register a callback function to execute when data is written to an open stream.<\/p>","code":"$s3 = new AmazonS3();\n\n\/\/ Define callback function\nfunction callback($curl_handle, $length)\n{\n\t$total = curl_getinfo($curl_handle, CURLINFO_CONTENT_LENGTH_DOWNLOAD);\n\t$transferred = curl_getinfo($curl_handle, CURLINFO_SIZE_DOWNLOAD);\n\t$percentage = floor(($transferred \/ $total) * 100);\n\n\techo 'Transferred ' . $transferred . ' KB out of ' . $total . ' KB (' . $percentage . '%)' . PHP_EOL;\n}\n\n\/\/ Register a callback function to execute when a stream is written locally.\n$s3-&gt;register_streaming_write_callback('callback');\n\n\/\/ Download a public object.\n$response = $s3-&gt;get_object('my-bucket', 'my-large-file.mp4', array(\n\t'fileDownload' =&gt; '.\/downloads\/my-large-file.mp4'\n));","result":"Transferred 7200 bytes out of 9000000 (0%)\n...\nTransferred 9000000 bytes out of 9000000 (100%)"}}},{"name":"send","modifier":"public","description":"<p>Executes the batch request queue by sending all queued requests.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"clear_after_send","type":"boolean","description":"<p>Whether or not to clear the batch queue after sending a request. Defaults to <code>true<\/code>. Set this to <code>false<\/code> if you are caching batch responses and want to retrieve results later.<\/p>"}},"returnvalue":"<p>An array of <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a> objects.<\/p>","source":"public function send($clear_after_send = true)\n{\n    if ($this-&gt;use_batch_flow)\n    {\n        \/\/ When we send the request, disable batch flow.\n        $this-&gt;use_batch_flow = false;\n\n        \/\/ If we're not caching, simply send the request.\n        if (!$this-&gt;use_cache_flow)\n        {\n            $response = $this-&gt;batch_object-&gt;send();\n            $parsed_data = array_map(array($this, 'parse_callback'), $response);\n            $parsed_data = new CFArray($parsed_data);\n\n            \/\/ Clear the queue\n            if ($clear_after_send)\n            {\n                $this-&gt;batch_object-&gt;queue = array();\n            }\n\n            return $parsed_data;\n        }\n\n        \/\/ Generate an identifier specific to this particular set of arguments.\n        $cache_id = $this-&gt;key . '_' . get_class($this) . '_' . sha1(serialize($this-&gt;batch_object));\n\n        \/\/ Instantiate the appropriate caching object.\n        $this-&gt;cache_object = new $this-&gt;cache_class($cache_id, $this-&gt;cache_location, $this-&gt;cache_expires, $this-&gt;cache_compress);\n\n        if ($this-&gt;delete_cache)\n        {\n            $this-&gt;use_cache_flow = false;\n            $this-&gt;delete_cache = false;\n            return $this-&gt;cache_object-&gt;delete();\n        }\n\n        \/\/ Invoke the cache callback function to determine whether to pull data from the cache or make a fresh request.\n        $data_set = $this-&gt;cache_object-&gt;response_manager(array($this, 'cache_callback_batch'), array($this-&gt;batch_object));\n        $parsed_data = array_map(array($this, 'parse_callback'), $data_set);\n        $parsed_data = new CFArray($parsed_data);\n\n        \/\/ Clear the queue\n        if ($clear_after_send)\n        {\n            $this-&gt;batch_object-&gt;queue = array();\n        }\n\n        \/\/ End!\n        return $parsed_data;\n    }\n\n    \/\/ Load the class\n    $null = new CFBatchRequest();\n    unset($null);\n\n    throw new CFBatchRequest_Exception('You must use $object-&gt;batch()-&gt;send()');\n}"},{"name":"set_batch_class","modifier":"public","description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_batch_class($class = 'CFBatchRequest')\n{\n    $this-&gt;batch_class = $class;\n    return $this;\n}"},{"name":"set_cache_config","modifier":"public","description":"<p>Set the caching configuration to use for response caching.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"location","type":"string","description":"<p>The location to store the cache object in. This may vary by cache method.<\/p><ul><li>File - The local file system paths such as <code>.\/cache<\/code> (relative) or <code>\/tmp\/cache\/<\/code> (absolute). The location must be server-writable.<\/li><li>APC - Pass in <code>apc<\/code> to use this lightweight cache. You must have the <a href=\"http:\/\/php.net\/apc\">APC extension<\/a> installed.<\/li><li>XCache - Pass in <code>xcache<\/code> to use this lightweight cache. You must have the <a href=\"http:\/\/xcache.lighttpd.net\">XCache<\/a> extension installed.<\/li><li>Memcached - Pass in an indexed array of associative arrays. Each associative array should have a <code>host<\/code> and a <code>port<\/code> value representing a <a href=\"http:\/\/php.net\/memcached\">Memcached<\/a> server to connect to.<\/li><li>PDO - A URL-style string (e.g. <code>pdo.mysql:\/\/user:pass@localhost\/cache<\/code>) or a standard DSN-style string (e.g. <code>pdo.sqlite:\/sqlite\/cache.db<\/code>). MUST be prefixed with <code>pdo.<\/code>. See <code>CachePDO<\/code> and <a href=\"http:\/\/php.net\/pdo\">PDO<\/a> for more details.<\/li><\/ul>"},{"@attributes":{"choice":"opt"},"name":"gzip","type":"boolean","description":"<p>Whether or not data should be gzipped before being stored. A value of <code>true<\/code> will compress the contents before caching them. A value of <code>false<\/code> will leave the contents uncompressed. Defaults to <code>true<\/code>.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_cache_config($location, $gzip = true)\n{\n    \/\/ If location is empty, don't do anything.\n    if (empty($location))\n    {\n        return $this;\n    }\n\n    \/\/ If we have an array, we're probably passing in Memcached servers and ports.\n    if (is_array($location))\n    {\n        $this-&gt;cache_class = 'CacheMC';\n    }\n    else\n    {\n        \/\/ I would expect locations like `\/tmp\/cache`, `pdo.mysql:\/\/user:pass@hostname:port`, `pdo.sqlite:memory:`, and `apc`.\n        $type = strtolower(substr($location, 0, 3));\n        switch ($type)\n        {\n            case 'apc':\n                $this-&gt;cache_class = 'CacheAPC';\n                break;\n\n            case 'xca': \/\/ First three letters of `xcache`\n                $this-&gt;cache_class = 'CacheXCache';\n                break;\n\n            case 'pdo':\n                $this-&gt;cache_class = 'CachePDO';\n                $location = substr($location, 4);\n                break;\n\n            default:\n                $this-&gt;cache_class = 'CacheFile';\n                break;\n        }\n    }\n\n    \/\/ Set the remaining cache information.\n    $this-&gt;cache_location = $location;\n    $this-&gt;cache_compress = $gzip;\n\n    return $this;\n}","examples":{"example":[{"title":"<p>Fire a single request, and then cache the response to APC.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('apc');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"},{"title":"<p>Fire a single request, and then cache the response to the file system.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Instantiate\n$sdb = new AmazonSDB();\n$sdb-&gt;set_cache_config('.\/cache');\n\n\/\/ First time pulls live data\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());\n\n\/\/ Second time pulls from cache\n$response = $sdb-&gt;cache(10)-&gt;list_domains();\nvar_dump($response-&gt;isOK());","result":"bool(true)\nbool(true)"}]}},{"name":"set_hostname","modifier":"public","description":"<p>Set the hostname to connect to. This is useful for alternate services that are API-compatible with\nAWS, but run from a different hostname.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"hostname","type":"string","description":"<p>The alternate hostname to use in place of the default one. Useful for mock or test applications living on different hostnames.<\/p>"},{"@attributes":{"choice":"opt"},"name":"port_number","type":"integer","description":"<p>The alternate port number to use in place of the default one. Useful for mock or test applications living on different port numbers.<\/p>"}]},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_hostname($hostname, $port_number = null)\n{\n    if ($this-&gt;override_hostname)\n    {\n        $this-&gt;hostname = $hostname;\n\n        if ($port_number)\n        {\n            $this-&gt;port_number = $port_number;\n            $this-&gt;hostname .= ':' . (string) $this-&gt;port_number;\n        }\n    }\n\n    return $this;\n}","examples":{"example":{"title":"<p>Configure hostname and resource prefix.<\/p>","code":"\/\/ Instantiate the class\n$ec2 = new AmazonEC2();\n\n\/\/ Point to the new endpoint\n$ec2-&gt;set_hostname('localhost', '3000');\n$ec2-&gt;set_resource_prefix('\/services\/mock');\n$ec2-&gt;allow_hostname_override(false);\n\n\/\/ Describe images\n$response = $ec2-&gt;describe_images();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"set_max_retries","modifier":"public","description":"<p>Sets the maximum number of times to retry failed requests.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"retries","type":"integer","description":"<p>The maximum number of times to retry failed requests. Defaults to <code>3<\/code>.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_max_retries($retries = 3)\n{\n    $this-&gt;max_retries = $retries;\n    return $this;\n}","examples":{"example":{"title":"<p>Sets the maximum number of times to retry failed requests.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Adjust offset\n$sdb = new AmazonSDB();\n$sdb-&gt;set_max_retries(3);\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;max_retries);","result":"int(3)"}}},{"name":"set_parser_class","modifier":"public","description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfsimplexml\/index.html\" rel=\"..\/..\/index.html#i=cfsimplexml\"><code>CFSimpleXML<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_parser_class($class = 'CFSimpleXML')\n{\n    $this-&gt;parser_class = $class;\n    return $this;\n}"},{"name":"set_proxy","modifier":"public","description":"<p>Set the proxy settings to use.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"proxy","type":"string","description":"<p>Accepts proxy credentials in the following format: <code>proxy:\/\/user:pass@hostname:port<\/code><\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_proxy($proxy)\n{\n    $this-&gt;proxy = $proxy;\n    return $this;\n}","examples":{"example":{"title":"<p>Set the proxy server to use for requests.<\/p>","description":"<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Adjust offset\n$sdb = new AmazonSDB();\n$sdb-&gt;set_proxy('proxy:\/\/user:pass@hostname:port');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;proxy);","result":"string(31) \"proxy:\/\/user:pass@hostname:port\""}}},{"name":"set_request_class","modifier":"public","description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfrequest\/index.html\" rel=\"..\/..\/index.html#i=cfrequest\"><code>CFRequest<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":{},"source":"public function set_request_class($class = 'CFRequest')\n{\n    $this-&gt;request_class = $class;\n    return $this;\n}","examples":{"example":{"title":"<p>Set the request class to use in the toolkit.<\/p>","description":"<p>The new request class must extend from <code>CFRequest<\/code>.<\/p>\n\n<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Custom class to extend RequestCore\nclass TestRequestCore extends RequestCore\n{\n\tpublic function test_method()\n\t{\n\t\treturn true;\n\t}\n}\n\n\/\/ Instantiate class and set new class\n$sdb = new AmazonSDB();\n$sdb-&gt;set_request_class('TestRequestCore');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;request_class);","result":"string(15) \"TestRequestCore\""}}},{"name":"set_resource_prefix","modifier":"public","description":"<p>Set the resource prefix to use. This method is useful for alternate services that are API-compatible\nwith AWS.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"prefix","type":"string","description":"<p>An alternate prefix to prepend to the resource path. Useful for mock or test applications.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_resource_prefix($prefix)\n{\n    $this-&gt;resource_prefix = $prefix;\n    return $this;\n}","examples":{"example":{"title":"<p>Configure hostname and resource prefix.<\/p>","code":"\/\/ Instantiate the class\n$ec2 = new AmazonEC2();\n\n\/\/ Point to the new endpoint\n$ec2-&gt;set_hostname('localhost', '3000');\n$ec2-&gt;set_resource_prefix('\/services\/mock');\n$ec2-&gt;allow_hostname_override(false);\n\n\/\/ Describe images\n$response = $ec2-&gt;describe_images();\n\n\/\/ Success?\nvar_dump($response-&gt;isOK());","result":"bool(true)"}}},{"name":"set_response_class","modifier":"public","description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_response_class($class = 'CFResponse')\n{\n    $this-&gt;response_class = $class;\n    return $this;\n}","examples":{"example":{"title":"<p>Set the response class to use in the toolkit.<\/p>","description":"<p>The new response class must extend from <code>CFResponse<\/code>.<\/p>\n\n<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Custom class to extend ResponseCore\nclass TestResponseCore extends ResponseCore\n{\n\tpublic function test_method()\n\t{\n\t\treturn true;\n\t}\n}\n\n\/\/ Instantiate class and set new class\n$sdb = new AmazonSDB();\n$sdb-&gt;set_response_class('TestResponseCore');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;response_class);","result":"string(16) \"TestResponseCore\""}}},{"name":"set_utilities_class","modifier":"public","description":"<p>Set a custom class for this functionality. Use this method when extending\/overriding existing classes\nwith new functionality.<\/p>\n\n<p>The replacement class must extend from <a href=\"..\/cfutilities\/index.html\" rel=\"..\/..\/index.html#i=cfutilities\"><code>CFUtilities<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"class","type":"string","description":"<p>The name of the new class to use for this functionality.<\/p>"}},"returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function set_utilities_class($class = 'CFUtilities')\n{\n    $this-&gt;utilities_class = $class;\n    $this-&gt;util = new $this-&gt;utilities_class();\n    return $this;\n}","examples":{"example":{"title":"<p>Set the utilities class to use in the toolkit.<\/p>","description":"<p>The new utilities class must extend from <code>CFUtilities<\/code>.<\/p>\n\n<p><strong>Note:<\/strong> This method is inherited by all service-specific classes.<\/p>","code":"\/\/ Custom class to extend CFUtilities\nclass TestUtilities extends CFUtilities\n{\n\tpublic function test_method()\n\t{\n\t\treturn true;\n\t}\n}\n\n\/\/ Instantiate class and set new class\n$sdb = new AmazonSDB();\n$sdb-&gt;set_utilities_class('TestUtilities');\n\n\/\/ Test if the value was set\nvar_dump($sdb-&gt;util-&gt;test_method());","result":"bool(true)"}}}]},"constants":{"@attributes":{"count":"4"},"constant":[{"name":"BUILD","value":"20130314130000"},{"name":"NAME","value":"aws-sdk-php"},{"name":"USERAGENT","value":"aws-sdk-php\/1.6.2 PHP\/5.4.9 curl\/7.15.5 openssl\/0.9.8e-fips-rhel5"},{"name":"VERSION","value":"1.6.2"}]},"properties":{"@attributes":{"count":"40"},"property":[{"name":"api_version","modifier":"public","description":"<p>The supported API version.<\/p>"},{"name":"auth_class","modifier":"public","description":"<p>The authentication class to use.<\/p>"},{"name":"auth_token","modifier":"public","description":"<p>The Amazon Authentication Token.<\/p>"},{"name":"batch_class","modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFBatchRequest"}},"description":"<p>The default class to use for handling batch requests (defaults to <a href=\"..\/cfbatchrequest\/index.html\" rel=\"..\/..\/index.html#i=cfbatchrequest\"><code>CFBatchRequest<\/code><\/a>).<\/p>"},{"name":"batch_object","modifier":"public","description":"<p>The current instantiated batch request object.<\/p>"},{"name":"cache_class","modifier":"public","description":"<p>The caching class to use.<\/p>"},{"name":"cache_compress","modifier":"public","description":"<p>The state of cache compression.<\/p>"},{"name":"cache_expires","modifier":"public","description":"<p>When the cache should be considered stale.<\/p>"},{"name":"cache_location","modifier":"public","description":"<p>The caching location to use.<\/p>"},{"name":"cache_object","modifier":"public","description":"<p>The current instantiated cache object.<\/p>"},{"name":"credentials","modifier":"public","description":"<p>The credentials to use for authentication.<\/p>"},{"name":"debug_mode","modifier":"public","description":"<p>The state of the debug mode setting.<\/p>"},{"name":"delete_cache","modifier":"public","description":"<p>The state of the cache deletion setting.<\/p>"},{"name":"hostname","modifier":"public","description":"<p>The alternate hostname to use, if any.<\/p>"},{"name":"internal_batch_object","modifier":"public","description":"<p>The internally instantiated batch request object.<\/p>"},{"name":"key","modifier":"public","description":"<p>The Amazon API Key.<\/p>"},{"name":"max_retries","modifier":"public","initializer":{"@attributes":{"type":"integer","value":"3"}},"description":"<p>The number of times to retry failed requests.<\/p>"},{"name":"operation","modifier":"public","description":"<p>The operation to execute.<\/p>"},{"name":"operation_prefix","modifier":"public","description":"<p>The string prefix to prepend to the operation name.<\/p>"},{"name":"override_hostname","modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of the capability to override the hostname with <a href=\"..\/cfruntime\/set_hostname.html\" rel=\"..\/..\/index.html#m=cfruntime\/set_hostname\"><code>set_hostname()<\/code><\/a>.<\/p>"},{"name":"parse_the_response","modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of whether the response should be parsed or not.<\/p>"},{"name":"parser_class","modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFSimpleXML"}},"description":"<p>The default class to use for parsing XML (defaults to <a href=\"..\/cfsimplexml\/index.html\" rel=\"..\/..\/index.html#i=cfsimplexml\"><code>CFSimpleXML<\/code><\/a>).<\/p>"},{"name":"payload","modifier":"public","description":"<p>The payload to send.<\/p>"},{"name":"port_number","modifier":"public","description":"<p>The alternate port number to use, if any.<\/p>"},{"name":"proxy","modifier":"public","description":"<p>The proxy to use for connecting.<\/p>"},{"name":"redirects","modifier":"public","description":"<p>The number of times a request has been retried.<\/p>"},{"name":"registered_streaming_read_callback","modifier":"public","description":"<p>The user-defined callback function to call when a stream is read from.<\/p>"},{"name":"registered_streaming_write_callback","modifier":"public","description":"<p>The user-defined callback function to call when a stream is written to.<\/p>"},{"name":"request_class","modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFRequest"}},"description":"<p>The default class to use for HTTP requests (defaults to <a href=\"..\/cfrequest\/index.html\" rel=\"..\/..\/index.html#i=cfrequest\"><code>CFRequest<\/code><\/a>).<\/p>"},{"name":"resource_prefix","modifier":"public","description":"<p>The alternate resource prefix to use, if any.<\/p>"},{"name":"response_class","modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFResponse"}},"description":"<p>The default class to use for HTTP responses (defaults to <a href=\"..\/cfresponse\/index.html\" rel=\"..\/..\/index.html#i=cfresponse\"><code>CFResponse<\/code><\/a>).<\/p>"},{"name":"secret_key","modifier":"public","description":"<p>The Amazon API Secret Key.<\/p>"},{"name":"service","modifier":"public","description":"<p>An identifier for the current AWS service.<\/p>"},{"name":"ssl_verification","modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of SSL certificate verification.<\/p>"},{"name":"use_aws_query","modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of whether auth should be handled as AWS Query.<\/p>"},{"name":"use_batch_flow","modifier":"public","description":"<p>The state of batch flow usage.<\/p>"},{"name":"use_cache_flow","modifier":"public","description":"<p>The state of cache flow usage.<\/p>"},{"name":"use_ssl","modifier":"public","initializer":{"@attributes":{"type":"boolean","value":"true"}},"description":"<p>The state of SSL\/HTTPS use.<\/p>"},{"name":"util","modifier":"public","description":"<p>Handle for the utility functions.<\/p>"},{"name":"utilities_class","modifier":"public","initializer":{"@attributes":{"type":"string","value":"CFUtilities"}},"description":"<p>The default class to use for utilities (defaults to <a href=\"..\/cfutilities\/index.html\" rel=\"..\/..\/index.html#i=cfutilities\"><code>CFUtilities<\/code><\/a>).<\/p>"}]}}}