{"file":{},"class":{"name":"CFUtilities","file":"utilities\/utilities.class.php","url":"..\/cfutilities\/index.html","description":"<p>Contains a set of utility methods for connecting to, and working with, AWS.<\/p>","metadata":{"version":"2010.09.30","license":"See the included NOTICE.md file for more information.","copyright":"See the included NOTICE.md file for more information.","link":"http:\/\/aws.amazon.com\/php\/ PHP Developer Center"},"methods":{"@attributes":{"count":"17"},"method":[{"name":"__construct","modifier":"public","description":"<p>Constructs a new instance of this class.<\/p>","returnvalue":"<p>A reference to the current instance.<\/p>","source":"public function __construct()\n{\n    return $this;\n}"},{"name":"convert_date_to_iso8601","modifier":"public","description":"<p>Checks to see if a date stamp is ISO-8601 formatted, and if not, makes it so.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"datestamp","type":"string","description":"<p>A date stamp, or a string that can be parsed into a date stamp.<\/p>"}},"returnvalue":"<p>An ISO-8601 formatted date stamp.<\/p>","source":"public function convert_date_to_iso8601($datestamp)\n{\n    if (!preg_match('\/\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2}((\\+|-)\\d{2}:\\d{2}|Z)\/m', $datestamp))\n    {\n        return gmdate(self::DATE_FORMAT_ISO8601, strtotime($datestamp));\n    }\n\n    return $datestamp;\n}","examples":{"example":{"title":"<p>Convert a date into ISO-8601 format.<\/p>","code":"\/\/ Instantiate the object\n$s3 = new AmazonS3();\n\n\/\/ Test for ISO-8601, and convert if not.\n$date_correct = '2009-08-01T07:00:00Z';\n$date_incorrect = '1 August 2009, 7am GMT';\n\n\/\/ Success?\nvar_dump($s3-&gt;util-&gt;convert_date_to_iso8601($date_correct));\nvar_dump($s3-&gt;util-&gt;convert_date_to_iso8601($date_incorrect));","result":"string(20) \"2009-08-01T07:00:00Z\"\nstring(20) \"2009-08-01T07:00:00Z\""}}},{"name":"convert_response_to_array","modifier":"public","description":"<p>Converts a SimpleXML response to an array structure.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"response","type":"ResponseCore","description":"<p>A response value.<\/p>"}},"returnvalue":"<p>The response value as a standard, multi-dimensional array.<\/p>","source":"public function convert_response_to_array(ResponseCore $response)\n{\n    return json_decode(json_encode($response), true);\n}","examples":{"example":{"title":"<p>Converts a SimpleXML response into a standard associative array.<\/p>","code":"\/\/ Get a ResponseCore response\n$pas = new AmazonPAS();\n$data = $pas-&gt;item_search('skillet awake');\n\n\/\/ Convert SimpleXMLElement to Array.\n$data = $pas-&gt;util-&gt;convert_response_to_array($data);\n\n\/\/ Success?\nvar_dump($data['body']['Items']['Request']['ItemSearchRequest']['Keywords']);","result":"string(13) \"skillet awake\""}}},{"name":"decode_uhex","modifier":"public","description":"<p>Decodes <code>\\uXXXX<\/code> entities into their real unicode character equivalents.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"s","type":"string","description":"<p>The string to decode.<\/p>"}},"returnvalue":"<p>The decoded string.<\/p>","source":"public function decode_uhex($s)\n{\n    preg_match_all('\/\\\\\\u([0-9a-f]{4})\/i', $s, $matches);\n    $matches = $matches[count($matches) - 1];\n    $map = array();\n\n    foreach ($matches as $match)\n    {\n        if (!isset($map[$match]))\n        {\n            $map['\\u' . $match] = html_entity_decode('&amp;#' . hexdec($match) . ';', ENT_NOQUOTES, 'UTF-8');\n        }\n    }\n\n    return str_replace(array_keys($map), $map, $s);\n}"},{"name":"encode_signature2","modifier":"public","description":"<p>Encode the value according to RFC 3986.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"string","type":"string","description":"<p>String to convert.<\/p>"}},"returnvalue":"<p>URL-friendly sign-able string.<\/p>","source":"public function encode_signature2($string)\n{\n    $string = rawurlencode($string);\n    return str_replace('%7E', '~', $string);\n}"},{"name":"generate_guid","modifier":"public","description":"<p>Generates a random GUID.<\/p>","metadata":{"author":"<a href=\"http:\/\/www.php.net\/manual\/en\/function.com-create-guid.php#99425\">Alix Axel<\/a>","license":"http:\/\/www.php.net\/license\/3_01.txt PHP License"},"returnvalue":"<p>A random GUID.<\/p>","source":"public function generate_guid()\n{\n    return sprintf(\n        '%04X%04X-%04X-%04X-%04X-%04X%04X%04X',\n        mt_rand(0, 65535),\n        mt_rand(0, 65535),\n        mt_rand(0, 65535),\n        mt_rand(16384, 20479),\n        mt_rand(32768, 49151),\n        mt_rand(0, 65535),\n        mt_rand(0, 65535),\n        mt_rand(0, 65535)\n    );\n}"},{"name":"hex_to_base64","modifier":"public","description":"<p>Convert a HEX value to Base64.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"str","type":"string","description":"<p>Value to convert.<\/p>"}},"returnvalue":"<p>Base64-encoded string.<\/p>","source":"public function hex_to_base64($str)\n{\n    $raw = '';\n\n    for ($i = 0; $i &lt; strlen($str); $i += 2)\n    {\n        $raw .= chr(hexdec(substr($str, $i, 2)));\n    }\n\n    return base64_encode($raw);\n}"},{"name":"is_base64","modifier":"public","description":"<p>Determines whether the data is Base64 encoded or not.<\/p>","metadata":{"license":"http:\/\/us.php.net\/manual\/en\/function.base64-decode.php#81425 PHP License"},"parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"s","type":"string","description":"<p>The string to test.<\/p>"}},"returnvalue":"<p>Whether the string is Base64 encoded or not.<\/p>","source":"public function is_base64($s)\n{\n    return (bool) preg_match('\/^[a-zA-Z0-9\\\/\\r\\n+]*={0,2}$\/', $s);\n}"},{"name":"is_json","modifier":"public","description":"<p>Determines whether the data is a JSON string or not.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"s","type":"string","description":"<p>The string to test.<\/p>"}},"returnvalue":"<p>Whether the string is a valid JSON object or not.<\/p>","source":"public function is_json($s)\n{\n    return !!(json_decode($s) instanceof stdClass);\n}"},{"name":"json_encode","modifier":"public","description":"<p>Can be removed once all calls are updated.<\/p>","metadata":{"deprecated":"Use <a href=\"http:\/\/php.net\/json_encode\"><code>json_encode()<\/code><\/a> instead."},"parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"obj","type":"mixed","description":"<p>The PHP object to convert into a JSON string.<\/p>"}},"returnvalue":"<p>A JSON string.<\/p>","source":"public function json_encode($obj)\n{\n    return json_encode($obj);\n}"},{"name":"konst","modifier":"public","description":"<p>Retrieves the value of a class constant, while avoiding the <code>T_PAAMAYIM_NEKUDOTAYIM<\/code> error. Misspelled because <code>const<\/code> is a reserved word.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"class","type":"object","description":"<p>An instance of the class containing the constant.<\/p>"},{"@attributes":{"choice":"req"},"name":"const","type":"string","description":"<p>The name of the constant to retrieve.<\/p>"}]},"returnvalue":"<p>The value of the class constant.<\/p>","source":"public function konst($class, $const)\n{\n    if (is_string($class))\n    {\n        $ref = new ReflectionClass($class);\n    }\n    else\n    {\n        $ref = new ReflectionObject($class);\n    }\n\n    return $ref-&gt;getConstant($const);\n}","examples":{"example":{"title":"<p>Very simple example of konst().<\/p>","code":"$s3 = new AmazonS3();\n$rfc2616 = $s3-&gt;util-&gt;konst($s3-&gt;util, 'DATE_FORMAT_RFC2616');\n$date = gmdate($rfc2616, 946684800);\n\nvar_dump($date);","result":"string(29) \"Sat, 01 Jan 2000 00:00:00 GMT\""}}},{"name":"query_to_array","modifier":"public","description":"<p>Convert a query string into an associative array. Multiple, identical keys will become an indexed array.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"qs","type":"string","description":"<p>Query string to convert.<\/p>"}},"returnvalue":"<p>Associative array of keys and values.<\/p>","source":"public function query_to_array($qs)\n{\n    $query = explode('&amp;', $qs);\n    $data = array();\n\n    foreach ($query as $q)\n    {\n        $q = explode('=', $q);\n\n        if (isset($data[$q[0]]) &amp;&amp; is_array($data[$q[0]]))\n        {\n            $data[$q[0]][] = urldecode($q[1]);\n        }\n        else if (isset($data[$q[0]]) &amp;&amp; !is_array($data[$q[0]]))\n        {\n            $data[$q[0]] = array($data[$q[0]]);\n            $data[$q[0]][] = urldecode($q[1]);\n        }\n        else\n        {\n            $data[urldecode($q[0])] = urldecode($q[1]);\n        }\n    }\n    return $data;\n}","examples":{"example":[{"title":"<p>Decodes a query string into an associative array.<\/p>","code":"\/\/ Instantiate\n$s3 = new AmazonS3();\n\n\/\/ Test data\n$qs = $s3-&gt;util-&gt;to_query_string(array(\n\t'a' =&gt; 1,\n\t'b' =&gt; 2,\n\t'c' =&gt; 3,\n));\n\n\/\/ Success?\nvar_dump($qs);\nvar_dump($util-&gt;query_to_array($qs));","result":"string(11) \"a=1&amp;b=2&amp;c=3\"\narray(3) {\n  [\"a\"]=&gt;\n  string(1) \"1\"\n  [\"b\"]=&gt;\n  string(1) \"2\"\n  [\"c\"]=&gt;\n  string(1) \"3\"\n}"},{"title":"<p>Decodes a query string into an associative array. Supports the same key more than once.<\/p>","code":"\/\/ Instantiate\n$s3 = new AmazonS3();\n\n\/\/ Test data\nvar_dump($s3-&gt;util-&gt;query_to_array('a=1&amp;a=2&amp;a=3&amp;b=2&amp;c=3'));","result":"array(3) {\n  [\"a\"]=&gt;\n  array(3) {\n    [0]=&gt;\n    string(1) \"1\"\n    [1]=&gt;\n    string(1) \"2\"\n    [2]=&gt;\n    string(1) \"3\"\n  }\n  [\"b\"]=&gt;\n  string(1) \"2\"\n  [\"c\"]=&gt;\n  string(1) \"3\"\n}"}]}},{"name":"size_readable","modifier":"public","description":"<p>Return human readable file sizes.<\/p>","metadata":{"author":["<a href=\"mailto:aidan@php.net\">Aidan Lister<\/a>","<a href=\"mailto:ryan@getcloudfusion.com\">Ryan Parman<\/a>"],"license":"http:\/\/www.php.net\/license\/3_01.txt PHP License"},"parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"size","type":"integer","description":"<p>Filesize in bytes.<\/p>"},{"@attributes":{"choice":"opt"},"name":"unit","type":"string","description":"<p>The maximum unit to use. Defaults to the largest appropriate unit.<\/p>"},{"@attributes":{"choice":"opt"},"name":"default","type":"string","description":"<p>The format for the return string. Defaults to <code>%01.2f %s<\/code>.<\/p>"}]},"returnvalue":"<p>The human-readable file size.<\/p>","source":"public function size_readable($size, $unit = null, $default = null)\n{\n    \/\/ Units\n    $sizes = array('B', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB');\n    $mod = 1024;\n    $ii = count($sizes) - 1;\n\n    \/\/ Max unit\n    $unit = array_search((string) $unit, $sizes);\n    if ($unit === null || $unit === false)\n    {\n        $unit = $ii;\n    }\n\n    \/\/ Return string\n    if ($default === null)\n    {\n        $default = '%01.2f %s';\n    }\n\n    \/\/ Loop\n    $i = 0;\n    while ($unit != $i &amp;&amp; $size &gt;= 1024 &amp;&amp; $i &lt; $ii)\n    {\n        $size \/= $mod;\n        $i++;\n    }\n\n    return sprintf($default, $size, $sizes[$i]);\n}","examples":{"example":{"title":"<p>Convert a number of bytes into a human-readable format.<\/p>","code":"\/\/ Instantiate\n$s3 = new AmazonS3();\n\n\/\/ Test data\nvar_dump($s3-&gt;util-&gt;size_readable(9876543210));","result":"string(7) \"9.20 GB\""}},"seealso":{"link":"<a href=\"http:\/\/aidanlister.com\/repos\/v\/function.size_readable.php\">Original Function<\/a>"}},{"name":"time_hms","modifier":"public","description":"<p>Convert a number of seconds into Hours:Minutes:Seconds.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"seconds","type":"integer","description":"<p>The number of seconds to convert.<\/p>"}},"returnvalue":"<p>The formatted time.<\/p>","source":"public function time_hms($seconds)\n{\n    $time = '';\n\n    \/\/ First pass\n    $hours = (int) ($seconds \/ 3600);\n    $seconds = $seconds % 3600;\n    $minutes = (int) ($seconds \/ 60);\n    $seconds = $seconds % 60;\n\n    \/\/ Cleanup\n    $time .= ($hours) ? $hours . ':' : '';\n    $time .= ($minutes &lt; 10 &amp;&amp; $hours &gt; 0) ? '0' . $minutes : $minutes;\n    $time .= ':';\n    $time .= ($seconds &lt; 10) ? '0' . $seconds : $seconds;\n\n    return $time;\n}","examples":{"example":{"title":"<p>Converts a number of seconds into HH:MM:SS format.<\/p>","code":"\/\/ Instantiate\n$s3 = new AmazonS3();\n\n\/\/ Test data\nvar_dump($s3-&gt;util-&gt;time_hms(98765));","result":"string(8) \"27:26:05\""}}},{"name":"to_query_string","modifier":"public","description":"<p>Convert an associative array into a query string.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"array","type":"array","description":"<p>Array to convert.<\/p>"}},"returnvalue":"<p>URL-friendly query string.<\/p>","source":"public function to_query_string($array)\n{\n    $temp = array();\n\n    foreach ($array as $key =&gt; $value)\n    {\n        if (is_string($key) &amp;&amp; !is_array($value))\n        {\n            $temp[] = rawurlencode($key) . '=' . rawurlencode($value);\n        }\n    }\n\n    return implode('&amp;', $temp);\n}","examples":{"example":{"title":"<p>Convert an associative array into a query string.<\/p>","code":"\/\/ Instantiate\n$s3 = new AmazonS3();\n\n\/\/ Test data\nvar_dump($s3-&gt;util-&gt;to_query_string(array(\n\t'a' =&gt; 1,\n\t'b' =&gt; 2,\n\t'c' =&gt; 3,\n)));","result":"string(11) \"a=1&amp;b=2&amp;c=3\""}}},{"name":"to_signable_string","modifier":"public","description":"<p>Convert an associative array into a sign-able string.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"array","type":"array","description":"<p>Array to convert.<\/p>"}},"returnvalue":"<p>URL-friendly sign-able string.<\/p>","source":"public function to_signable_string($array)\n{\n    $t = array();\n\n    foreach ($array as $k =&gt; $v)\n    {\n        $t[] = $this-&gt;encode_signature2($k) . '=' . $this-&gt;encode_signature2($v);\n    }\n\n    return implode('&amp;', $t);\n}"},{"name":"try_these","modifier":"public","description":"<p>Returns the first value that is set. Based on <a href=\"http:\/\/api.prototypejs.org\/language\/Try\/these\/\">Try.these()<\/a> from <a href=\"http:\/\/prototypejs.org\">Prototype<\/a>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"attrs","type":"array","description":"<p>The attributes to test, as strings. Intended for testing properties of the $base object, but also works with variables if you place an @ symbol at the beginning of the command.<\/p>"},{"@attributes":{"choice":"opt"},"name":"base","type":"object","description":"<p>The base object to use, if any.<\/p>"},{"@attributes":{"choice":"opt"},"name":"default","type":"mixed","description":"<p>What to return if there are no matches. Defaults to <code>null<\/code>.<\/p>"}]},"returnvalue":"<p>Either a matching property of a given object, boolean <code>false<\/code>, or any other data type you might choose.<\/p>","source":"public function try_these($attrs, $base = null, $default = null)\n{\n    if ($base)\n    {\n        foreach ($attrs as $attr)\n        {\n            if (isset($base-&gt;$attr))\n            {\n                return $base-&gt;$attr;\n            }\n        }\n    }\n    else\n    {\n        foreach ($attrs as $attr)\n        {\n            if (isset($attr))\n            {\n                return $attr;\n            }\n        }\n    }\n\n    return $default;\n}","examples":{"example":{"title":"<p>Return the first non-falsey value.<\/p>","code":"\/\/ Test data\n$obj = new StdClass;\n$obj-&gt;a = null;\n$obj-&gt;b = null;\n$obj-&gt;c = 1;\n$obj-&gt;d = null;\n\n\/\/ Instantiate\n$s3 = new AmazonS3();\n\n\/\/ Success?\nvar_dump($s3-&gt;util-&gt;try_these(array('a', 'b', 'c', 'd'), $obj, true));","result":"int(1)"}}}]},"constants":{"@attributes":{"count":"4"},"constant":[{"name":"DATE_FORMAT_ISO8601","value":"Y-m-d\\TH:i:s\\Z"},{"name":"DATE_FORMAT_MYSQL","value":"Y-m-d H:i:s"},{"name":"DATE_FORMAT_RFC2616","value":"D, d M Y H:i:s \\G\\M\\T"},{"name":"DATE_FORMAT_SIGV4","value":"Ymd\\THis\\Z"}]},"properties":{"@attributes":{"count":"0"}}}}