{"file":{},"class":{"name":"CachePDO","file":"lib\/cachecore\/cachepdo.class.php","url":"..\/cachepdo\/index.html","description":"<p>Container for all PDO-based cache methods. Inherits additional methods from <CacheCore>. Adheres\nto the ICacheCore interface.<\/p>","metadata":{"version":"2012.04.17","copyright":["2006-2012 Ryan Parman","2006-2010 Foleeo, Inc.","2012 Amazon.com, Inc. or its affiliates.","2008-2010 Contributors"],"license":"http:\/\/opensource.org\/licenses\/bsd-license.php Simplified BSD License","link":["http:\/\/github.com\/skyzyx\/cachecore CacheCore","http:\/\/getcloudfusion.com CloudFusion","http:\/\/php.net\/pdo PDO"]},"inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"implements":{"interface":{"name":"ICacheCore","file":"lib\/cachecore\/icachecore.interface.php"}},"methods":{"@attributes":{"count":"13"},"method":[{"name":"__construct","modifier":"public","description":"<p>Constructs a new instance of this class.<\/p>\n\n<p>Tested with <a href=\"http:\/\/mysql.com\">MySQL 5.0.x<\/a>, <a href=\"http:\/\/postgresql.com\">PostgreSQL<\/a>, and\n<a href=\"http:\/\/sqlite.org\">SQLite 3.x<\/a>. SQLite 2.x is assumed to work. No other PDO-supported databases have\nbeen tested (e.g. Oracle, Microsoft SQL Server, IBM DB2, ODBC, Sybase, Firebird). Feel free to send\npatches for additional database support.<\/p>\n\n<p>See <a href=\"http:\/\/php.net\/pdo\">http:\/\/php.net\/pdo<\/a> for more information.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"name","type":"string","description":"<p>A name to uniquely identify the cache object.<\/p>"},{"@attributes":{"choice":"opt"},"name":"location","type":"string","description":"<p>The location to store the cache object in. This may vary by cache method. The default value is NULL.<\/p>"},{"@attributes":{"choice":"opt"},"name":"expires","type":"integer","description":"<p>The number of seconds until a cache object is considered stale. The default value is 0.<\/p>"},{"@attributes":{"choice":"opt"},"name":"gzip","type":"boolean","description":"<p>Whether data should be gzipped before being stored. The default value is true.<\/p>"}]},"returnvalue":"<p>Reference to the cache object.<\/p>","source":"public function __construct($name, $location = null, $expires = 0, $gzip = true)\n{\n    \/\/ Make sure the name is no longer than 40 characters.\n    $name = sha1($name);\n\n    \/\/ Call parent constructor and set id.\n    parent::__construct($name, $location, $expires, $gzip);\n    $this-&gt;id = $this-&gt;name;\n    $options = array();\n\n    \/\/ Check if the location contains :\/\/ (e.g. mysql:\/\/user:pass@hostname:port\/table)\n    if (stripos($location, ':\/\/') === false)\n    {\n        \/\/ No? Just pass it through.\n        $this-&gt;dsn = parse_url($location);\n        $this-&gt;dsn_string = $location;\n    }\n    else\n    {\n        \/\/ Yes? Parse and set the DSN\n        $this-&gt;dsn = parse_url($location);\n        $this-&gt;dsn_string = $this-&gt;dsn['scheme'] . ':host=' . $this-&gt;dsn['host'] . ((isset($this-&gt;dsn['port'])) ? ';port=' . $this-&gt;dsn['port'] : '') . ';dbname=' . substr($this-&gt;dsn['path'], 1);\n    }\n\n    \/\/ Make sure that user\/pass are defined.\n    $user = isset($this-&gt;dsn['user']) ? $this-&gt;dsn['user'] : null;\n    $pass = isset($this-&gt;dsn['pass']) ? $this-&gt;dsn['pass'] : null;\n\n    \/\/ Set persistence for databases that support it.\n    switch ($this-&gt;dsn['scheme'])\n    {\n        case 'mysql': \/\/ MySQL\n        case 'pgsql': \/\/ PostgreSQL\n            $options[PDO::ATTR_PERSISTENT] = true;\n            break;\n    }\n\n    \/\/ Instantiate a new PDO object with a persistent connection.\n    $this-&gt;pdo = new PDO($this-&gt;dsn_string, $user, $pass, $options);\n\n    \/\/ Define prepared statements for improved performance.\n    $this-&gt;create = $this-&gt;pdo-&gt;prepare(&quot;INSERT INTO cache (id, expires, data) VALUES (:id, :expires, :data)&quot;);\n    $this-&gt;read = $this-&gt;pdo-&gt;prepare(&quot;SELECT id, expires, data FROM cache WHERE id = :id&quot;);\n    $this-&gt;reset = $this-&gt;pdo-&gt;prepare(&quot;UPDATE cache SET expires = :expires WHERE id = :id&quot;);\n    $this-&gt;delete = $this-&gt;pdo-&gt;prepare(&quot;DELETE FROM cache WHERE id = :id&quot;);\n}"},{"name":"create","modifier":"public","description":"<p>Creates a new cache.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"data","type":"mixed","description":"<p>The data to cache.<\/p>"}},"returnvalue":"<p>Whether the operation was successful.<\/p>","source":"public function create($data)\n{\n    $data = serialize($data);\n    $data = $this-&gt;gzip ? gzcompress($data) : $data;\n\n    $this-&gt;create-&gt;bindParam(':id', $this-&gt;id);\n    $this-&gt;create-&gt;bindParam(':data', $data);\n    $this-&gt;create-&gt;bindParam(':expires', $this-&gt;generate_timestamp());\n\n    return (bool) $this-&gt;create-&gt;execute();\n}"},{"name":"delete","modifier":"public","description":"<p>Deletes a cache.<\/p>","returnvalue":"<p>Whether the operation was successful.<\/p>","source":"public function delete()\n{\n    $this-&gt;delete-&gt;bindParam(':id', $this-&gt;id);\n    return $this-&gt;delete-&gt;execute();\n}"},{"name":"expire_in","modifier":"public","inheritance":{"class":{"name":"CacheCore"}},"description":"<p>Set the number of seconds until a cache expires.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"seconds"}},"returnvalue":{"@attributes":{"type":"$this"}},"source":{"@attributes":{"file":"lib\/cachecore\/cachecore.class.php","start":"101","end":"105","lines":"5"}}},{"name":"generate_timestamp","modifier":"protected","description":"<p>Returns a timestamp value apropriate to the current database type.<\/p>","returnvalue":"<p>Timestamp for MySQL and PostgreSQL, integer value for SQLite.<\/p>","source":"protected function generate_timestamp()\n{\n    \/\/ Define 'expires' settings differently.\n    switch ($this-&gt;dsn['scheme'])\n    {\n        \/\/ These support timestamps.\n        case 'mysql': \/\/ MySQL\n        case 'pgsql': \/\/ PostgreSQL\n            $expires = date(DATE_FORMAT_MYSQL, time());\n            break;\n\n        \/\/ These support integers.\n        case 'sqlite': \/\/ SQLite 3\n        case 'sqlite2': \/\/ SQLite 2\n            $expires = time();\n            break;\n    }\n\n    return $expires;\n}"},{"name":"get_drivers","modifier":"public","description":"<p>Returns a list of supported PDO database drivers. Identical to &lt;PDO::getAvailableDrivers()>.<\/p>","returnvalue":"<p>The list of supported database drivers.<\/p>","source":"public function get_drivers()\n{\n    return PDO::getAvailableDrivers();\n}","seealso":{"link":"<a href=\"http:\/\/php.net\/pdo.getavailabledrivers\">PHP Method<\/a>"}},{"name":"init","modifier":"public static","inheritance":{"class":{"name":"CacheCore"}},"description":"<p>Allows for chaining from the constructor. Requires PHP 5.3 or newer.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"name","type":"string","description":"<p>A name to uniquely identify the cache object.<\/p>"},{"@attributes":{"choice":"opt"},"name":"location","type":"string","description":"<p>The location to store the cache object in. This may vary by cache method. The default value is NULL.<\/p>"},{"@attributes":{"choice":"opt"},"name":"expires","type":"integer","description":"<p>The number of seconds until a cache object is considered stale. The default value is 0.<\/p>"},{"@attributes":{"choice":"opt"},"name":"gzip","type":"boolean","description":"<p>Whether data should be gzipped before being stored. The default value is true.<\/p>"}]},"returnvalue":"<p>Reference to the cache object.<\/p>","source":{"@attributes":{"file":"lib\/cachecore\/cachecore.class.php","start":"84","end":"93","lines":"10"}}},{"name":"is_expired","modifier":"public","description":"<p>Checks whether the cache object is expired or not.<\/p>","returnvalue":"<p>Whether the cache is expired or not.<\/p>","source":"public function is_expired()\n{\n    if ($this-&gt;timestamp() + $this-&gt;expires &lt; time())\n    {\n        return true;\n    }\n\n    return false;\n}"},{"name":"read","modifier":"public","description":"<p>Reads a cache.<\/p>","returnvalue":"<p>Either the content of the cache object, or boolean <code>false<\/code>.<\/p>","source":"public function read()\n{\n    if (!$this-&gt;store_read)\n    {\n        $this-&gt;read-&gt;bindParam(':id', $this-&gt;id);\n        $this-&gt;read-&gt;execute();\n        $this-&gt;store_read = $this-&gt;read-&gt;fetch(PDO::FETCH_ASSOC);\n    }\n\n    if ($this-&gt;store_read)\n    {\n        $data = $this-&gt;store_read['data'];\n        $data = $this-&gt;gzip ? gzuncompress($data) : $data;\n\n        return unserialize($data);\n    }\n\n    return false;\n}"},{"name":"reset","modifier":"public","description":"<p>Resets the freshness of the cache.<\/p>","returnvalue":"<p>Whether the operation was successful.<\/p>","source":"public function reset()\n{\n    $this-&gt;reset-&gt;bindParam(':id', $this-&gt;id);\n    $this-&gt;reset-&gt;bindParam(':expires', $this-&gt;generate_timestamp());\n    return (bool) $this-&gt;reset-&gt;execute();\n}"},{"name":"response_manager","modifier":"public","inheritance":{"class":{"name":"CacheCore"}},"description":"<p>Provides a simple, straightforward cache-logic mechanism. Useful for non-complex response caches.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"callback","type":"string|function","description":"<p>The name of the function to fire when we need to fetch new data to cache.<\/p>"},{"@attributes":{"choice":"opt"},"name":"params","type":"array","description":"<p>Parameters to pass into the callback function, as an array.<\/p>"}]},"returnvalue":"<p>The cached data being requested.<\/p>","source":{"@attributes":{"file":"lib\/cachecore\/cachecore.class.php","start":"114","end":"143","lines":"30"}}},{"name":"timestamp","modifier":"public","description":"<p>Retrieves the timestamp of the cache.<\/p>","returnvalue":"<p>Either the Unix time stamp of the cache creation, or boolean <code>false<\/code>.<\/p>","source":"public function timestamp()\n{\n    if (!$this-&gt;store_read)\n    {\n        $this-&gt;read-&gt;bindParam(':id', $this-&gt;id);\n        $this-&gt;read-&gt;execute();\n        $this-&gt;store_read = $this-&gt;read-&gt;fetch(PDO::FETCH_ASSOC);\n    }\n\n    if ($this-&gt;store_read)\n    {\n        $value = $this-&gt;store_read['expires'];\n\n        \/\/ If 'expires' isn't yet an integer, convert it into one.\n        if (!is_numeric($value))\n        {\n            $value = strtotime($value);\n        }\n\n        $this-&gt;timestamp = date('U', $value);\n        return $this-&gt;timestamp;\n    }\n\n    return false;\n}"},{"name":"update","modifier":"public","description":"<p>Updates an existing cache.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"data","type":"mixed","description":"<p>The data to cache.<\/p>"}},"returnvalue":"<p>Whether the operation was successful.<\/p>","source":"public function update($data)\n{\n    $this-&gt;delete();\n    return $this-&gt;create($data);\n}"}]},"constants":{"@attributes":{"count":"0"}},"properties":{"@attributes":{"count":"15"},"property":[{"name":"create","modifier":"public","description":"<p>Holds the prepared statement for creating an entry.<\/p>"},{"name":"delete","modifier":"public","description":"<p>Holds the prepared statement for deleting an entry.<\/p>"},{"name":"dsn","modifier":"public","description":"<p>Holds the parsed URL components.<\/p>"},{"name":"dsn_string","modifier":"public","description":"<p>Holds the PDO-friendly version of the connection string.<\/p>"},{"name":"expires","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>The number of seconds before a cache object is considered stale.<\/p>"},{"name":"gzip","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>Stores whether or not the content should be gzipped when stored<\/p>"},{"name":"id","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>Used internally to uniquely identify the location + name of the cache object.<\/p>"},{"name":"location","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>Where to store the cache.<\/p>"},{"name":"name","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>A name to uniquely identify the cache object by.<\/p>"},{"name":"pdo","modifier":"public","description":"<p>Reference to the PDO connection object.<\/p>"},{"name":"read","modifier":"public","description":"<p>Holds the prepared statement for reading an entry.<\/p>"},{"name":"reset","modifier":"public","description":"<p>Holds the prepared statement for resetting the expiry of an entry.<\/p>"},{"name":"store_read","modifier":"public","description":"<p>Holds the response of the read so we only need to fetch it once instead of doing\nmultiple queries.<\/p>"},{"name":"timestamp","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>Stores the time when the cache object was created.<\/p>"},{"name":"update","modifier":"public","description":"<p>Holds the prepared statement for updating an entry.<\/p>"}]}}}