{"file":{},"class":{"name":"CacheXCache","file":"lib\/cachecore\/cachexcache.class.php","url":"..\/cachexcache\/index.html","description":"<p>Container for all XCache-based cache methods. Inherits additional methods from <CacheCore>. Adheres\nto the ICacheCore interface.<\/p>","metadata":{"version":"2012.04.17","copyright":["2006-2012 Ryan Parman","2006-2010 Foleeo, Inc.","2012 Amazon.com, Inc. or its affiliates.","2008-2010 Contributors"],"license":"http:\/\/opensource.org\/licenses\/bsd-license.php Simplified BSD License","link":["http:\/\/github.com\/skyzyx\/cachecore CacheCore","http:\/\/getcloudfusion.com CloudFusion","http:\/\/xcache.lighttpd.net XCache"]},"inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"implements":{"interface":{"name":"ICacheCore","file":"lib\/cachecore\/icachecore.interface.php"}},"methods":{"@attributes":{"count":"11"},"method":[{"name":"__construct","modifier":"public","description":"<p>Constructs a new instance of this class.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"name","type":"string","description":"<p>A name to uniquely identify the cache object.<\/p>"},{"@attributes":{"choice":"opt"},"name":"location","type":"string","description":"<p>The location to store the cache object in. This may vary by cache method. The default value is NULL.<\/p>"},{"@attributes":{"choice":"opt"},"name":"expires","type":"integer","description":"<p>The number of seconds until a cache object is considered stale. The default value is 0.<\/p>"},{"@attributes":{"choice":"opt"},"name":"gzip","type":"boolean","description":"<p>Whether data should be gzipped before being stored. The default value is true.<\/p>"}]},"returnvalue":"<p>Reference to the cache object.<\/p>","source":"public function __construct($name, $location = null, $expires = 0, $gzip = true)\n{\n    parent::__construct($name, null, $expires, $gzip);\n    $this-&gt;id = $this-&gt;name;\n}"},{"name":"create","modifier":"public","description":"<p>Creates a new cache.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"data","type":"mixed","description":"<p>The data to cache.<\/p>"}},"returnvalue":"<p>Whether the operation was successful.<\/p>","source":"public function create($data)\n{\n    $data = serialize($data);\n    $data = $this-&gt;gzip ? gzcompress($data) : $data;\n\n    return xcache_set($this-&gt;id, $data, $this-&gt;expires);\n}"},{"name":"delete","modifier":"public","description":"<p>Deletes a cache.<\/p>","returnvalue":"<p>Whether the operation was successful.<\/p>","source":"public function delete()\n{\n    return xcache_unset($this-&gt;id);\n}"},{"name":"expire_in","modifier":"public","inheritance":{"class":{"name":"CacheCore"}},"description":"<p>Set the number of seconds until a cache expires.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"seconds"}},"returnvalue":{"@attributes":{"type":"$this"}},"source":{"@attributes":{"file":"lib\/cachecore\/cachecore.class.php","start":"101","end":"105","lines":"5"}}},{"name":"init","modifier":"public static","inheritance":{"class":{"name":"CacheCore"}},"description":"<p>Allows for chaining from the constructor. Requires PHP 5.3 or newer.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"name","type":"string","description":"<p>A name to uniquely identify the cache object.<\/p>"},{"@attributes":{"choice":"opt"},"name":"location","type":"string","description":"<p>The location to store the cache object in. This may vary by cache method. The default value is NULL.<\/p>"},{"@attributes":{"choice":"opt"},"name":"expires","type":"integer","description":"<p>The number of seconds until a cache object is considered stale. The default value is 0.<\/p>"},{"@attributes":{"choice":"opt"},"name":"gzip","type":"boolean","description":"<p>Whether data should be gzipped before being stored. The default value is true.<\/p>"}]},"returnvalue":"<p>Reference to the cache object.<\/p>","source":{"@attributes":{"file":"lib\/cachecore\/cachecore.class.php","start":"84","end":"93","lines":"10"}}},{"name":"is_expired","modifier":"public","description":"<p>Defined here, but always returns false. XCache manages it&#8217;s own expirations. It&#8217;s worth\nmentioning that if the server is configured for a long xcache.var_gc_interval then it IS\npossible for expired data to remain in the var cache, though it is not possible to access\nit.<\/p>","returnvalue":"<p>Whether the cache is expired or not.<\/p>","source":"public function is_expired()\n{\n    return false;\n}"},{"name":"read","modifier":"public","description":"<p>Reads a cache.<\/p>","returnvalue":"<p>Either the content of the cache object, or boolean <code>false<\/code>.<\/p>","source":"public function read()\n{\n    if ($data = xcache_get($this-&gt;id))\n    {\n        $data = $this-&gt;gzip ? gzuncompress($data) : $data;\n        return unserialize($data);\n    }\n\n    return false;\n}"},{"name":"reset","modifier":"public","description":"<p>Implemented here, but always returns <code>false<\/code>. XCache manages its own expirations.<\/p>","returnvalue":"<p>Whether the operation was successful.<\/p>","source":"public function reset()\n{\n    return false;\n}"},{"name":"response_manager","modifier":"public","inheritance":{"class":{"name":"CacheCore"}},"description":"<p>Provides a simple, straightforward cache-logic mechanism. Useful for non-complex response caches.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"callback","type":"string|function","description":"<p>The name of the function to fire when we need to fetch new data to cache.<\/p>"},{"@attributes":{"choice":"opt"},"name":"params","type":"array","description":"<p>Parameters to pass into the callback function, as an array.<\/p>"}]},"returnvalue":"<p>The cached data being requested.<\/p>","source":{"@attributes":{"file":"lib\/cachecore\/cachecore.class.php","start":"114","end":"143","lines":"30"}}},{"name":"timestamp","modifier":"public","description":"<p>Implemented here, but always returns <code>false<\/code>. XCache manages its own expirations.<\/p>","returnvalue":"<p>Either the Unix time stamp of the cache creation, or boolean <code>false<\/code>.<\/p>","source":"public function timestamp()\n{\n    return false;\n}"},{"name":"update","modifier":"public","description":"<p>Updates an existing cache.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"data","type":"mixed","description":"<p>The data to cache.<\/p>"}},"returnvalue":"<p>Whether the operation was successful.<\/p>","source":"public function update($data)\n{\n    $data = serialize($data);\n    $data = $this-&gt;gzip ? gzcompress($data) : $data;\n\n    return xcache_set($this-&gt;id, $data, $this-&gt;expires);\n}"}]},"constants":{"@attributes":{"count":"0"}},"properties":{"@attributes":{"count":"6"},"property":[{"name":"expires","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>The number of seconds before a cache object is considered stale.<\/p>"},{"name":"gzip","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>Stores whether or not the content should be gzipped when stored<\/p>"},{"name":"id","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>Used internally to uniquely identify the location + name of the cache object.<\/p>"},{"name":"location","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>Where to store the cache.<\/p>"},{"name":"name","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>A name to uniquely identify the cache object by.<\/p>"},{"name":"timestamp","inheritance":{"class":{"name":"CacheCore","file":"lib\/cachecore\/cachecore.class.php"}},"modifier":"public","description":"<p>Stores the time when the cache object was created.<\/p>"}]}}}