{"file":{},"class":{"name":"DynamoDBSessionHandler","file":"extensions\/dynamodbsessionhandler.class.php","url":"..\/dynamodbsessionhandler\/index.html","description":"<p>Provides an interface for using Amazon DynamoDB as a session store by hooking into\nPHP&#8217;s session handler hooks. This class is not auto-loaded, and must be included\nmanually or via the <code>AmazonDynamoDB::register_session_hander()<\/code> method.<\/p>\n\n<p>Once registered, You may use the native <code>$_SESSION<\/code> superglobal and\nsession functions, and the sessions will be stored automatically in DynamoDB.<\/p>","methods":{"@attributes":{"count":"17"},"method":[{"name":"__construct","modifier":"public","description":"<p>Initializes the session handler and prepares the configuration options.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"dynamodb","type":"AmazonDynamoDB","description":"<p>An instance of the DynamoDB client.<\/p>"},{"@attributes":{"choice":"opt"},"name":"options","type":"array","description":"<p>Configuration options.<\/p>"}]},"returnvalue":{},"source":"public function __construct(AmazonDynamoDB $dynamodb, array $options = array())\n{\n    \/\/ Store the AmazonDynamoDB client for use in the session handler\n    $this-&gt;_dynamodb = $dynamodb;\n\n    \/\/ Do type conversions on options and store the values\n    foreach ($options as $key =&gt; $value)\n    {\n        if (isset(self::$_option_types[$key]))\n        {\n            settype($value, self::$_option_types[$key]);\n            $this-&gt;{'_' . $key} = $value;\n        }\n    }\n\n    \/\/ Make sure the lifetime is positive. Use the gc_maxlifetime otherwise\n    if ($this-&gt;_session_lifetime &lt;= 0)\n    {\n        $this-&gt;_session_lifetime = (integer) ini_get('session.gc_maxlifetime');\n    }\n}"},{"name":"__destruct","modifier":"public","description":"<p>Destruct the session handler and make sure the session gets written.<\/p>\n\n<p>NOTE: It is usually better practice to call <code>session_write_close()<\/code>\nmanually in your application as soon as session modifications are complete. This\nis especially true if session locking is enabled (which it is by default).<\/p>","returnvalue":{},"source":"public function __destruct()\n{\n    session_write_close();\n}","related":{"method":"<http:\/\/php.net\/manual\/en\/function.session-set-save-handler.php#refsect1-function.session-set-save-handler-notes>"}},{"name":"_id","modifier":"protected","description":"<p>Prefix the session ID with the session name and prepare for DynamoDB usage<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"id","type":"string","description":"<p>The session ID.<\/p>"}},"returnvalue":"<p>The HashKeyElement value formatted as an array.<\/p>","source":"protected function _id($id)\n{\n    return trim($this-&gt;_session_name . '_' . $id, '_');\n}"},{"name":"_lock_and_read","modifier":"protected","description":"<p>Acquires a lock on a session in DynamoDB using conditional updates.<\/p>\n\n<p>WARNING: There is a <code>while(true);<\/code> in here.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"id","type":"string","description":"<p>The session ID.<\/p>"}},"returnvalue":"<p>The response from DynamoDB.<\/p>","source":"protected function _lock_and_read($id)\n{\n    $now = time();\n    $timeout = $now + $this-&gt;_max_lock_wait_time;\n\n    do\n    {\n        \/\/ Acquire the lock\n        $response = $this-&gt;_dynamodb-&gt;update_item(array(\n            'TableName'        =&gt; $this-&gt;_table_name,\n            'Key'              =&gt; array('HashKeyElement' =&gt; $this-&gt;_dynamodb-&gt;attribute($this-&gt;_id($id))),\n            'AttributeUpdates' =&gt; array('lock' =&gt; $this-&gt;_dynamodb-&gt;attribute(1, 'update')),\n            'Expected'         =&gt; array('lock' =&gt; array('Exists' =&gt; false)),\n            'ReturnValues'     =&gt; 'ALL_NEW',\n        ));\n\n        \/\/ If lock succeeds (or times out), exit the loop, otherwise sleep and try again\n        if ($response-&gt;isOK() || $now &gt;= $timeout)\n        {\n            return $response;\n        }\n        elseif (stripos((string) $response-&gt;body-&gt;asXML(), 'ConditionalCheckFailedException') !== false)\n        {\n            usleep(rand($this-&gt;_min_lock_retry_utime, $this-&gt;_max_lock_retry_utime));\n\n            $now = time();\n        }\n        else\n        {\n            return null;\n        }\n    }\n    while (true);\n}"},{"name":"close","modifier":"public","description":"<p>Close a session from writing<\/p>\n\n<p>Part of the standard PHP session handler interface<\/p>","returnvalue":"<p>Success<\/p>","source":"public function close()\n{\n    if (!$this-&gt;_session_written)\n    {\n        \/\/ Ensure that the session is unlocked even if the write did not happen\n        $id = $this-&gt;_open_session;\n        $response = $this-&gt;_dynamodb-&gt;update_item(array(\n            'TableName'        =&gt; $this-&gt;_table_name,\n            'Key'              =&gt; array('HashKeyElement' =&gt; $this-&gt;_dynamodb-&gt;attribute($this-&gt;_id($id))),\n            'AttributeUpdates' =&gt; array(\n                'expires'  =&gt; $this-&gt;_dynamodb-&gt;attribute(time() + $this-&gt;_session_lifetime, 'update'),\n                'lock'     =&gt; array('Action' =&gt; AmazonDynamoDB::ACTION_DELETE)\n            ),\n        ));\n\n        $this-&gt;_session_written = $response-&gt;isOK();\n    }\n\n    $this-&gt;_open_session = null;\n\n    return $this-&gt;_session_written;\n}"},{"name":"close_session","modifier":"public","description":"<p>Delegates to <code>session_commit()<\/code><\/p>","returnvalue":"<p>Chainable.<\/p>","source":"public function close_session()\n{\n    session_commit();\n\n    return $this;\n}"},{"name":"create_sessions_table","modifier":"public","description":"<p>Creates a table in DynamoDB for session storage according to provided configuration options.<\/p>\n\n<p>Note: Table creation may also be done via the AWS Console, which might make the most sense for this use case.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"opt"},"name":"read_capacity_units","type":"integer","description":"<p>Read capacity units for table throughput.<\/p>"},{"@attributes":{"choice":"opt"},"name":"write_capacity_units","type":"integer","description":"<p>Write capacity units for table throughput.<\/p>"}]},"returnvalue":"<p>Returns <code>true<\/code> on success.<\/p>","source":"public function create_sessions_table($read_capacity_units = 10, $write_capacity_units = 5)\n{\n    $response = $this-&gt;_dynamodb-&gt;create_table(array(\n        'TableName' =&gt; $this-&gt;_table_name,\n        'KeySchema' =&gt; array(\n            'HashKeyElement' =&gt; array(\n                'AttributeName' =&gt; $this-&gt;_hash_key,\n                'AttributeType' =&gt; AmazonDynamoDB::TYPE_STRING,\n            )\n        ),\n        'ProvisionedThroughput' =&gt; array(\n            'ReadCapacityUnits' =&gt; $read_capacity_units,\n            'WriteCapacityUnits' =&gt; $write_capacity_units,\n        )\n    ));\n\n    return $response-&gt;isOK();\n}"},{"name":"delete_sessions_table","modifier":"public","description":"<p>Deletes the session storage table from DynamoDB.<\/p>\n\n<p>Note: Table deletion may also be done via the AWS Console, which might make the most sense for this use case.<\/p>","returnvalue":"<p>Returns <code>true<\/code> on success.<\/p>","source":"public function delete_sessions_table()\n{\n    $response = $this-&gt;_dynamodb-&gt;delete_table(array('TableName' =&gt; $this-&gt;_table_name));\n\n    return $response-&gt;isOK();\n}"},{"name":"destroy","modifier":"public","description":"<p>Delete a session stored in DynamoDB.<\/p>\n\n<p>Part of the standard PHP session handler interface.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"id","type":"string","description":"<p>The session ID.<\/p>"},{"@attributes":{"choice":"opt"},"name":"garbage_collect_mode","type":"boolean","description":"<p>Whether or not the handler is doing garbage collection.<\/p>"}]},"returnvalue":"<p>Whether or not the operation succeeded.<\/p>","source":"public function destroy($id, $garbage_collect_mode = false)\n{\n    \/\/ Make sure we don't prefix the ID a second time\n    if (!$garbage_collect_mode)\n    {\n        $id = $this-&gt;_id($id);\n    }\n\n    $delete_options = array(\n        'TableName' =&gt; $this-&gt;_table_name,\n        'Key'       =&gt; array('HashKeyElement' =&gt; $this-&gt;_dynamodb-&gt;attribute($id)),\n    );\n\n    \/\/ Make sure not to garbage collect locked sessions\n    if ($garbage_collect_mode &amp;&amp; $this-&gt;_session_locking)\n    {\n        $delete_options['Expected'] = array('lock' =&gt; array('Exists' =&gt; false));\n    }\n\n    \/\/ Send the delete request to DynamoDB\n    $response = $this-&gt;_dynamodb-&gt;delete_item($delete_options);\n\n    $this-&gt;_session_written = $response-&gt;isOK();\n\n    return $this-&gt;_session_written;\n}"},{"name":"destroy_session","modifier":"public","description":"<p>Delegates to <code>session_destroy()<\/code><\/p>","returnvalue":"<p>Chainable.<\/p>","source":"public function destroy_session()\n{\n    session_destroy();\n\n    return $this;\n}"},{"name":"garbage_collect","modifier":"public","description":"<p>Performs garbage collection on the sessions stored in the DynamoDB table.<\/p>\n\n<p>Part of the standard PHP session handler interface.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"maxlifetime","type":"integer","description":"<p>The value of <code>session.gc_maxlifetime<\/code>. Ignored.<\/p>"}},"returnvalue":"<p>Whether or not the operation succeeded.<\/p>","source":"public function garbage_collect($maxlifetime = null)\n{\n    \/\/ Send a search request to DynamoDB looking for expired sessions\n    $response = $this-&gt;_dynamodb-&gt;scan(array(\n        'TableName'  =&gt; $this-&gt;_table_name,\n        'ScanFilter' =&gt; array('expires' =&gt; array(\n            'AttributeValueList' =&gt; array($this-&gt;_dynamodb-&gt;attribute(time())),\n            'ComparisonOperator' =&gt; AmazonDynamoDB::CONDITION_LESS_THAN,\n        )),\n    ));\n\n    \/\/ Delete the expired sessions\n    if ($response-&gt;isOK())\n    {\n        $deleted = array();\n\n        \/\/ Get the ID of and delete each session that is expired\n        foreach ($response-&gt;body-&gt;Items as $item)\n        {\n            $id = (string) $item-&gt;{$this-&gt;_hash_key}-&gt;{AmazonDynamoDB::TYPE_STRING};\n            $deleted[$id] = $this-&gt;destroy($id, true);\n        }\n\n        \/\/ Return true if all of the expired sessions were deleted\n        return (array_sum($deleted) === count($deleted));\n    }\n\n    return false;\n}"},{"name":"is_session_open","modifier":"public","description":"<p>Checks if the session is open and writable.<\/p>","returnvalue":"<p>Whether or not the session is still open for writing.<\/p>","source":"public function is_session_open()\n{\n    return (boolean) $this-&gt;_open_session;\n}"},{"name":"open","modifier":"public","description":"<p>Open a session for writing. Triggered by <a href=\"http:\/\/php.net\/session_start\"><code>session_start()<\/code><\/a>.<\/p>\n\n<p>Part of the standard PHP session handler interface.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"save_path","type":"string","description":"<p>The session save path (see <a href=\"http:\/\/php.net\/session_save_path\"><code>session_save_path()<\/code><\/a>).<\/p>"},{"@attributes":{"choice":"req"},"name":"session_name","type":"string","description":"<p>The session name (see <a href=\"http:\/\/php.net\/session_name\"><code>session_name()<\/code><\/a>).<\/p>"}]},"returnvalue":"<p>Whether or not the operation succeeded.<\/p>","source":"public function open($save_path, $session_name)\n{\n    $this-&gt;_save_path    = $save_path;\n    $this-&gt;_session_name = $session_name;\n    $this-&gt;_open_session = session_id();\n\n    return true;\n}"},{"name":"open_session","modifier":"public","description":"<p>Delegates to <code>session_start()<\/code><\/p>","returnvalue":"<p>Chainable.<\/p>","source":"public function open_session()\n{\n    session_start();\n\n    return $this;\n}"},{"name":"read","modifier":"public","description":"<p>Read a session stored in DynamoDB<\/p>\n\n<p>Part of the standard PHP session handler interface<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"id","type":"string","description":"<p>The session ID.<\/p>"}},"returnvalue":"<p>The session data.<\/p>","source":"public function read($id)\n{\n    $result = '';\n\n    \/\/ Get the session data from DynamoDB (acquire a lock if locking is enabled)\n    if ($this-&gt;_session_locking)\n    {\n        $response = $this-&gt;_lock_and_read($id);\n        $node_name = 'Attributes';\n    }\n    else\n    {\n        $response = $this-&gt;_dynamodb-&gt;get_item(array(\n            'TableName'      =&gt; $this-&gt;_table_name,\n            'Key'            =&gt; array('HashKeyElement' =&gt; $this-&gt;_dynamodb-&gt;attribute($this-&gt;_id($id))),\n            'ConsistentRead' =&gt; $this-&gt;_consistent_reads,\n        ));\n        $node_name = 'Item';\n    }\n\n    if ($response &amp;&amp; $response-&gt;isOK())\n    {\n        $item = array();\n\n        \/\/ Get the data from the DynamoDB response\n        if ($response-&gt;body-&gt;{$node_name})\n        {\n            foreach ($response-&gt;body-&gt;{$node_name}-&gt;children() as $key =&gt; $value)\n            {\n                $type = $value-&gt;children()-&gt;getName();\n                $item[$key] = $value-&gt;{$type}-&gt;to_string();\n            }\n        }\n\n        if (isset($item['expires']) &amp;&amp; isset($item['data']))\n        {\n            \/\/ Check the expiration date before using\n            if ($item['expires'] &gt; time())\n            {\n                $result = $item['data'];\n            }\n            else\n            {\n                $this-&gt;destroy($id);\n            }\n        }\n    }\n\n    return $result;\n}"},{"name":"register","modifier":"public","description":"<p>Register DynamoDB as a session handler.<\/p>\n\n<p>Uses the PHP-provided method to register this class as a session handler.<\/p>","returnvalue":"<p>Chainable.<\/p>","source":"public function register()\n{\n    session_set_save_handler(\n        array($this, 'open'),\n        array($this, 'close'),\n        array($this, 'read'),\n        array($this, 'write'),\n        array($this, 'destroy'),\n        array($this, 'garbage_collect')\n    );\n\n    return $this;\n}"},{"name":"write","modifier":"public","description":"<p>Write a session to DynamoDB.<\/p>\n\n<p>Part of the standard PHP session handler interface.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"id","type":"string","description":"<p>The session ID.<\/p>"},{"@attributes":{"choice":"req"},"name":"data","type":"string","description":"<p>The session data.<\/p>"}]},"returnvalue":"<p>Whether or not the operation succeeded.<\/p>","source":"public function write($id, $data)\n{\n    \/\/ Write the session data to DynamoDB\n    $response = $this-&gt;_dynamodb-&gt;put_item(array(\n        'TableName' =&gt; $this-&gt;_table_name,\n        'Item'      =&gt; $this-&gt;_dynamodb-&gt;attributes(array(\n            $this-&gt;_hash_key =&gt; $this-&gt;_id($id),\n            'expires'        =&gt; time() + $this-&gt;_session_lifetime,\n            'data'           =&gt; $data,\n        )),\n    ));\n\n    $this-&gt;_session_written = $response-&gt;isOK();\n\n    return $this-&gt;_session_written;\n}"}]},"constants":{"@attributes":{"count":"0"}},"properties":{"@attributes":{"count":"14"},"property":[{"name":"_consistent_reads","modifier":"protected","initializer":{"@attributes":{"type":"boolean","value":"true"}},"metadata":{"var":"boolean Whether or not the session handler should do consistent reads from DynamoDB."}},{"name":"_dynamodb","modifier":"protected","metadata":{"var":"AmazonDynamoDB The DyanamoDB client."}},{"name":"_hash_key","modifier":"protected","initializer":{"@attributes":{"type":"string","value":"id"}},"metadata":{"var":"string The name of the hash key in the DynamoDB sessions table."}},{"name":"_max_lock_retry_utime","modifier":"protected","initializer":{"@attributes":{"type":"integer","value":"100000"}},"metadata":{"var":"integer Maximum time, in microseconds, that the session handler should wait to retry acquiring a lock."}},{"name":"_max_lock_wait_time","modifier":"protected","initializer":{"@attributes":{"type":"integer","value":"30"}},"metadata":{"var":"integer Maximum time, in seconds, that the session handler should take to acquire a lock."}},{"name":"_min_lock_retry_utime","modifier":"protected","initializer":{"@attributes":{"type":"integer","value":"10000"}},"metadata":{"var":"integer Minimum time, in microseconds, that the session handler should wait to retry acquiring a lock."}},{"name":"_open_session","modifier":"protected","metadata":{"var":"boolean Keeps track of if the session is open."}},{"name":"_option_types","modifier":"protected static","initializer":{"@attributes":{"type":"array","value":"array( \"table_name\" => \"string\", \"hash_key\" => \"string\", \"session_lifetime\" => \"integer\", \"consistent_reads\" => \"boolean\", \"session_locking\" => \"boolean\", \"max_lock_wait_time\" => \"integer\", \"min_lock_retry_utime\" => \"integer\", \"max_lock_retry_utime\" => \"integer\" )"}},"metadata":{"var":"array Type for casting the configuration options."}},{"name":"_save_path","modifier":"protected","metadata":{"var":"string The session save path (see <a href=\"http:\/\/php.net\/session_save_path\"><code>session_save_path()<\/code><\/a>)."}},{"name":"_session_lifetime","modifier":"protected","metadata":{"var":"integer The lifetime of an inactive session before it should be garbage collected."}},{"name":"_session_locking","modifier":"protected","initializer":{"@attributes":{"type":"boolean","value":"true"}},"metadata":{"var":"boolean Whether or not the session handler should do session locking."}},{"name":"_session_name","modifier":"protected","metadata":{"var":"string The session name (see <a href=\"http:\/\/php.net\/session_name\"><code>session_name()<\/code><\/a>)."}},{"name":"_session_written","modifier":"protected","metadata":{"var":"boolean Keeps track of whether the session is open."}},{"name":"_table_name","modifier":"protected","initializer":{"@attributes":{"type":"string","value":"sessions"}},"metadata":{"var":"string The name of the DynamoDB table in which to store sessions."}}]}}}