{"file":{},"class":{"name":"S3StreamWrapper","file":"extensions\/s3streamwrapper.class.php","url":"..\/s3streamwrapper\/index.html","description":"<p>Provides an interface for accessing Amazon S3 using PHP&#8217;s native file management functions.<\/p>\n\n<p>Amazon S3 file patterns take the following form: <code>s3:\/\/bucket\/object<\/code>.<\/p>","methods":{"@attributes":{"count":"22"},"method":[{"name":"client","modifier":"public","description":"<p>Fetches the client for the protocol being used.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"opt"},"name":"protocol","type":"string","description":"<p>The protocol associated with this stream wrapper.<\/p>"}},"returnvalue":"<p>The S3 client associated with this stream wrapper.<\/p>","source":"public function client($protocol = null)\n{\n    if ($protocol == null)\n    {\n        if ($parsed = parse_url($this-&gt;path))\n        {\n            $protocol = $parsed['scheme'];\n        }\n        else\n        {\n            trigger_error(__CLASS__ . ' could not determine the protocol of the stream wrapper in use.');\n        }\n    }\n\n    return self::$_clients[$protocol];\n}"},{"name":"dir_closedir","modifier":"public","description":"<p>Close directory handle. This method is called in response to <a href=\"http:\/\/php.net\/closedir\"><code>closedir()<\/code><\/a>.<\/p>\n\n<p>Since Amazon S3 doesn&#8217;t have real directories, always return <code>true<\/code>.<\/p>","returnvalue":{"@attributes":{"type":"boolean"}},"source":"public function dir_closedir()\n{\n    $this-&gt;position = 0;\n    $this-&gt;path = null;\n    $this-&gt;file_list = null;\n    $this-&gt;open_file = null;\n    $this-&gt;seek_position = 0;\n    $this-&gt;eof = false;\n    $this-&gt;buffer = null;\n    $this-&gt;object_size = 0;\n\n    return true;\n}"},{"name":"dir_opendir","modifier":"public","description":"<p>Open directory handle. This method is called in response to <a href=\"http:\/\/php.net\/opendir\"><code>opendir()<\/code><\/a>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"path","type":"string","description":"<p>Specifies the URL that was passed to <a href=\"http:\/\/php.net\/opendir\"><code>opendir()<\/code><\/a>.<\/p>"},{"@attributes":{"choice":"req"},"name":"options","type":"integer","description":"<p>Not used. Passed in by <a href=\"http:\/\/php.net\/opendir\"><code>opendir()<\/code><\/a>.<\/p>"}]},"returnvalue":"<p>Returns <code>true<\/code> on success or <code>false<\/code> on failure.<\/p>","source":"public function dir_opendir($path, $options)\n{\n    $this-&gt;path = $path;\n    list($protocol, $bucket, $object_name) = $this-&gt;parse_path($path);\n\n    $pattern = '\/^' . self::regex_token($object_name) . '(.*)[^\\\/$]\/';\n\n    $this-&gt;file_list = $this-&gt;client($protocol)-&gt;get_object_list($bucket, array(\n        'pcre' =&gt; $pattern\n    ));\n\n    return (count($this-&gt;file_list)) ? true : false;\n}"},{"name":"dir_readdir","modifier":"public","description":"<p>This method is called in response to <a href=\"http:\/\/php.net\/readdir\"><code>readdir()<\/code><\/a>.<\/p>","returnvalue":"<p>Should return a string representing the next filename, or <code>false<\/code> if there is no next file.<\/p>","source":"public function dir_readdir()\n{\n    if (isset($this-&gt;file_list[$this-&gt;position]))\n    {\n        $out = $this-&gt;file_list[$this-&gt;position];\n        $this-&gt;position++;\n    }\n    else\n    {\n        $out = false;\n    }\n\n    return $out;\n}"},{"name":"dir_rewinddir","modifier":"public","description":"<p>This method is called in response to <a href=\"http:\/\/php.net\/rewinddir\"><code>rewinddir()<\/code><\/a>.<\/p>\n\n<p>Should reset the output generated by <a href=\"http:\/\/php.net\/streamwrapper.dir_readdir\"><code>streamWrapper::dir_readdir()<\/code><\/a>. i.e.: The next call to\n<a href=\"http:\/\/php.net\/streamwrapper.dir_readdir\"><code>streamWrapper::dir_readdir()<\/code><\/a> should return the first entry in the location returned by\n<a href=\"http:\/\/php.net\/streamwrapper.dir_opendir\"><code>streamWrapper::dir_opendir()<\/code><\/a>.<\/p>","returnvalue":"<p>Returns <code>true<\/code> on success or <code>false<\/code> on failure.<\/p>","source":"public function dir_rewinddir()\n{\n    $this-&gt;position = 0;\n\n    return true;\n}"},{"name":"mkdir","modifier":"public","description":"<p>Create a new bucket. This method is called in response to <a href=\"http:\/\/php.net\/mkdir\"><code>mkdir()<\/code><\/a>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"path","type":"string","description":"<p>The bucket name to create.<\/p>"},{"@attributes":{"choice":"req"},"name":"mode","type":"integer","description":"<p>Permissions. 700-range permissions map to ACL_PUBLIC. 600-range permissions map to ACL_AUTH_READ. All other permissions map to ACL_PRIVATE. Expects octal form.<\/p>"},{"@attributes":{"choice":"req"},"name":"options","type":"integer","description":"<p>Ignored.<\/p>"}]},"returnvalue":"<p>Whether the bucket was created successfully or not.<\/p>","source":"public function mkdir($path, $mode, $options)\n{\n    \/\/ Get the value that was *actually* passed in as mode, and default to 0\n    $trace_slice = array_slice(debug_backtrace(), -1);\n    $mode = isset($trace_slice[0]['args'][1]) ? decoct($trace_slice[0]['args'][1]) : 0;\n\n    $this-&gt;path = $path;\n    list($protocol, $bucket, $object_name) = $this-&gt;parse_path($path);\n\n    if (in_array($mode, range(700, 799)))\n    {\n        $acl = AmazonS3::ACL_PUBLIC;\n    }\n    elseif (in_array($mode, range(600, 699)))\n    {\n        $acl = AmazonS3::ACL_AUTH_READ;\n    }\n    else\n    {\n        $acl = AmazonS3::ACL_PRIVATE;\n    }\n\n    $client = $this-&gt;client($protocol);\n    $region = $client-&gt;hostname;\n    $response = $client-&gt;create_bucket($bucket, $region, $acl);\n\n    return $response-&gt;isOK();\n}"},{"name":"parse_path","modifier":"public","description":"<p>Parses an S3 URL into the parts needed by the stream wrapper.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"path","type":"string","description":"<p>The path to parse.<\/p>"}},"returnvalue":"<p>An array of 3 items: protocol, bucket, and object name ready for <code>list()<\/code>.<\/p>","source":"public function parse_path($path)\n{\n    $url = parse_url($path);\n\n    return array(\n        $url['scheme'],                                       \/\/ Protocol\n        $url['host'],                                         \/\/ Bucket\n        (isset($url['path']) ? substr($url['path'], 1) : ''), \/\/ Object\n    );\n}"},{"name":"regex_token","modifier":"public static","description":"<p>Makes the given token PCRE-compatible.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"token","type":"string","description":"<p>The token<\/p>"}},"returnvalue":"<p>The PCRE-compatible version of the token<\/p>","source":"public static function regex_token($token)\n{\n    $token = str_replace('\/', '\\\/', $token);\n    $token = quotemeta($token);\n    return str_replace('\\\\\\\\', '\\\\', $token);\n}"},{"name":"register","modifier":"public static","description":"<p>Registers the S3StreamWrapper class as a stream wrapper.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"opt"},"name":"s3","type":"AmazonS3","description":"<p>An instance of the AmazonS3 client.<\/p>"},{"@attributes":{"choice":"opt"},"name":"protocol","type":"string","description":"<p>The name of the protocol to register.<\/p>"}]},"returnvalue":"<p>Whether or not the registration succeeded.<\/p>","source":"public static function register(AmazonS3 $s3 = null, $protocol = 's3')\n{\n    S3StreamWrapper::$_clients[$protocol] = $s3 ? $s3 : new AmazonS3();\n\n    return stream_wrapper_register($protocol, 'S3StreamWrapper');\n}"},{"name":"rename","modifier":"public","description":"<p>Renames a file or directory. This method is called in response to <a href=\"http:\/\/php.net\/rename\"><code>rename()<\/code><\/a>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"path_from","type":"string","description":"<p>The URL to the current file.<\/p>"},{"@attributes":{"choice":"req"},"name":"path_to","type":"string","description":"<p>The URL which the <code>$path_from<\/code> should be renamed to.<\/p>"}]},"returnvalue":"<p>Returns <code>true<\/code> on success or <code>false<\/code> on failure.<\/p>","source":"public function rename($path_from, $path_to)\n{\n    list($protocol, $from_bucket_name, $from_object_name) = $this-&gt;parse_path($path_from);\n    list($protocol, $to_bucket_name, $to_object_name) = $this-&gt;parse_path($path_to);\n\n    $copy_response = $this-&gt;client($protocol)-&gt;copy_object(\n        array('bucket' =&gt; $from_bucket_name, 'filename' =&gt; $from_object_name),\n        array('bucket' =&gt; $to_bucket_name,   'filename' =&gt; $to_object_name  )\n    );\n\n    if ($copy_response-&gt;isOK())\n    {\n        $delete_response = $this-&gt;client($protocol)-&gt;delete_object($from_bucket_name, $from_object_name);\n\n        if ($delete_response-&gt;isOK())\n        {\n            return true;\n        }\n    }\n\n    return false;\n}"},{"name":"rmdir","modifier":"public","description":"<p>This method is called in response to <a href=\"http:\/\/php.net\/rmdir\"><code>rmdir()<\/code><\/a>.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"path","type":"string","description":"<p>The bucket name to create.<\/p>"},{"@attributes":{"choice":"req"},"name":"context","type":"boolean","description":"<p>Ignored.<\/p>"}]},"returnvalue":"<p>Whether the bucket was deleted successfully or not.<\/p>","source":"public function rmdir($path, $context)\n{\n    $this-&gt;path = $path;\n    list($protocol, $bucket, $object_name) = $this-&gt;parse_path($path);\n\n    $response = $this-&gt;client($protocol)-&gt;delete_bucket($bucket);\n\n    return $response-&gt;isOK();\n}"},{"name":"stream_close","modifier":"public","description":"<p>Close a resource. This method is called in response to <a href=\"http:\/\/php.net\/fclose\"><code>fclose()<\/code><\/a>.<\/p>\n\n<p>All resources that were locked, or allocated, by the wrapper should be released.<\/p>","returnvalue":{"@attributes":{"type":"void"}},"source":"public function stream_close()\n{\n    $this-&gt;position = 0;\n    $this-&gt;path = null;\n    $this-&gt;file_list = null;\n    $this-&gt;open_file = null;\n    $this-&gt;seek_position = 0;\n    $this-&gt;eof = false;\n    $this-&gt;buffer = null;\n    $this-&gt;object_size = 0;\n}"},{"name":"stream_eof","modifier":"public","description":"<p>Tests for end-of-file on a file pointer. This method is called in response to <a href=\"http:\/\/php.net\/feof\"><code>feof()<\/code><\/a>.<\/p>","returnvalue":{"@attributes":{"type":"boolean"}},"source":"public function stream_eof()\n{\n    return $this-&gt;eof;\n}"},{"name":"stream_flush","modifier":"public","description":"<p>Flushes the output. This method is called in response to <a href=\"http:\/\/php.net\/fflush\"><code>fflush()<\/code><\/a>. If you have cached data in\nyour stream but not yet stored it into the underlying storage, you should do so now.<\/p>\n\n<p>Since this implementation doesn&#8217;t buffer streams, simply return <code>true<\/code>.<\/p>","returnvalue":"<p>Whether or not flushing succeeded<\/p>","source":"public function stream_flush()\n{\n    if ($this-&gt;buffer === null)\n    {\n        return false;\n    }\n\n    list($protocol, $bucket, $object_name) = $this-&gt;parse_path($this-&gt;path);\n\n    $response = $this-&gt;client($protocol)-&gt;create_object($bucket, $object_name, array(\n        'body' =&gt; $this-&gt;buffer,\n    ));\n\n    $this-&gt;seek_position = 0;\n    $this-&gt;buffer = null;\n    $this-&gt;eof = true;\n\n    return $response-&gt;isOK();\n}"},{"name":"stream_open","modifier":"public","description":"<p>Opens file or URL. This method is called immediately after the wrapper is initialized\n(e.g., by <a href=\"http:\/\/php.net\/fopen\"><code>fopen()<\/code><\/a> and <a href=\"http:\/\/php.net\/file_get_contents\"><code>file_get_contents()<\/code><\/a>).<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"path","type":"string","description":"<p>Specifies the URL that was passed to the original function.<\/p>"},{"@attributes":{"choice":"req"},"name":"mode","type":"string","description":"<p>Ignored.<\/p>"},{"@attributes":{"choice":"req"},"name":"options","type":"integer","description":"<p>Ignored.<\/p>"},{"@attributes":{"choice":"req"},"name":"opened_path"}]},"returnvalue":"<p>Returns <code>true<\/code> on success or <code>false<\/code> on failure.<\/p>","source":"public function stream_open($path, $mode, $options, &amp;$opened_path)\n{\n    $opened_path = $path;\n    $this-&gt;open_file = $path;\n    $this-&gt;path = $path;\n    $this-&gt;seek_position = 0;\n    $this-&gt;object_size = 0;\n\n    return true;\n}"},{"name":"stream_read","modifier":"public","description":"<p>Read from stream. This method is called in response to <a href=\"http:\/\/php.net\/fread\"><code>fread()<\/code><\/a> and <a href=\"http:\/\/php.net\/fgets\"><code>fgets()<\/code><\/a>.<\/p>\n\n<p>It is important to avoid reading files that are near to or larger than the amount of memory\nallocated to PHP, otherwise &#8220;out of memory&#8221; errors will occur.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"count","type":"integer","description":"<p>Always equal to 8192. PHP is fun, isn&#8217;t it?<\/p>"}},"returnvalue":"<p>The contents of the Amazon S3 object.<\/p>","source":"public function stream_read($count)\n{\n    if ($this-&gt;eof)\n    {\n        return false;\n    }\n\n    list($protocol, $bucket, $object_name) = $this-&gt;parse_path($this-&gt;path);\n\n    if ($this-&gt;seek_position &gt; 0 &amp;&amp; $this-&gt;object_size)\n    {\n        if ($count + $this-&gt;seek_position &gt; $this-&gt;object_size)\n        {\n            $count = $this-&gt;object_size - $this-&gt;seek_position;\n        }\n\n        $start = $this-&gt;seek_position;\n        $end = $this-&gt;seek_position + $count;\n\n        $response = $this-&gt;client($protocol)-&gt;get_object($bucket, $object_name, array(\n            'range' =&gt; $start . '-' . $end\n        ));\n    }\n    else\n    {\n        $response = $this-&gt;client($protocol)-&gt;get_object($bucket, $object_name);\n        $this-&gt;object_size = isset($response-&gt;header['content-length']) ? $response-&gt;header['content-length'] : 0;\n    }\n\n    if (!$response-&gt;isOK())\n    {\n        return false;\n    }\n\n    $data = substr($response-&gt;body, 0, min($count, $this-&gt;object_size));\n    $this-&gt;seek_position += strlen($data);\n\n\n    if ($this-&gt;seek_position &gt;= $this-&gt;object_size)\n    {\n        $this-&gt;eof = true;\n        $this-&gt;seek_position = 0;\n        $this-&gt;object_size = 0;\n    }\n\n    return $data;\n}"},{"name":"stream_seek","modifier":"public","description":"<p>Seeks to specific location in a stream. This method is called in response to <a href=\"http:\/\/php.net\/fseek\"><code>fseek()<\/code><\/a>. The read\/write\nposition of the stream should be updated according to the <code>$offset<\/code> and <code>$whence<\/code>\nparameters.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"offset","type":"integer","description":"<p>The number of bytes to offset from the start of the file.<\/p>"},{"@attributes":{"choice":"req"},"name":"whence","type":"integer","description":"<p>Ignored. Always uses <code>SEEK_SET<\/code>.<\/p>"}]},"returnvalue":"<p>Whether or not the seek was successful.<\/p>","source":"public function stream_seek($offset, $whence)\n{\n    $this-&gt;seek_position = $offset;\n\n    return true;\n}"},{"name":"stream_stat","modifier":"public","description":"<p>Retrieve information about a file resource.<\/p>","returnvalue":"<p>Returns the same data as a call to <a href=\"http:\/\/php.net\/stat\"><code>stat()<\/code><\/a>.<\/p>","source":"public function stream_stat()\n{\n    return $this-&gt;url_stat($this-&gt;path, null);\n}"},{"name":"stream_tell","modifier":"public","description":"<p>Retrieve the current position of a stream. This method is called in response to <a href=\"http:\/\/php.net\/ftell\"><code>ftell()<\/code><\/a>.<\/p>","returnvalue":"<p>Returns the current position of the stream.<\/p>","source":"public function stream_tell()\n{\n    return $this-&gt;seek_position;\n}"},{"name":"stream_write","modifier":"public","description":"<p>Write to stream. This method is called in response to <a href=\"http:\/\/php.net\/fwrite\"><code>fwrite()<\/code><\/a>.<\/p>\n\n<p>It is important to avoid reading files that are larger than the amount of memory allocated to PHP,\notherwise &#8220;out of memory&#8221; errors will occur.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"data","type":"string","description":"<p>The data to write to the stream.<\/p>"}},"returnvalue":"<p>The number of bytes that were written to the stream.<\/p>","source":"public function stream_write($data)\n{\n    $size = strlen($data);\n\n    $this-&gt;seek_position = $size;\n    $this-&gt;buffer .= $data;\n\n    return $this-&gt;seek_position;\n}"},{"name":"unlink","modifier":"public","description":"<p>Delete a file. This method is called in response to <a href=\"http:\/\/php.net\/unlink\"><code>unlink()<\/code><\/a>.<\/p>","parameters":{"parameter":{"@attributes":{"choice":"req"},"name":"path","type":"string","description":"<p>The file URL which should be deleted.<\/p>"}},"returnvalue":"<p>Returns <code>true<\/code> on success or <code>false<\/code> on failure.<\/p>","source":"public function unlink($path)\n{\n    $this-&gt;path = $path;\n    list($protocol, $bucket, $object_name) = $this-&gt;parse_path($path);\n\n    $response = $this-&gt;client($protocol)-&gt;delete_object($bucket, $object_name);\n\n    return $response-&gt;isOK();\n}"},{"name":"url_stat","modifier":"public","description":"<p>This method is called in response to all <a href=\"http:\/\/php.net\/stat\"><code>stat()<\/code><\/a> related functions.<\/p>","parameters":{"parameter":[{"@attributes":{"choice":"req"},"name":"path","type":"string","description":"<p>The file path or URL to stat. Note that in the case of a URL, it must be a <code>:\/\/<\/code> delimited URL. Other URL forms are not supported.<\/p>"},{"@attributes":{"choice":"req"},"name":"flags","type":"integer","description":"<p>Holds additional flags set by the streams API. This implementation ignores all defined flags.<\/p>"}]},"returnvalue":"<p>Should return as many elements as <a href=\"http:\/\/php.net\/stat\"><code>stat()<\/code><\/a> does. Unknown or unavailable values should be set to a rational value (usually <code>0<\/code>).<\/p>","source":"public function url_stat($path, $flags)\n{\n    \/\/ Defaults\n    $out = array();\n    $out[0] = $out['dev'] = 0;\n    $out[1] = $out['ino'] = 0;\n    $out[2] = $out['mode'] = 0;\n    $out[3] = $out['nlink'] = 0;\n    $out[4] = $out['uid'] = 0;\n    $out[5] = $out['gid'] = 0;\n    $out[6] = $out['rdev'] = 0;\n    $out[7] = $out['size'] = 0;\n    $out[8] = $out['atime'] = 0;\n    $out[9] = $out['mtime'] = 0;\n    $out[10] = $out['ctime'] = 0;\n    $out[11] = $out['blksize'] = 0;\n    $out[12] = $out['blocks'] = 0;\n\n    $this-&gt;path = $path;\n    list($protocol, $bucket, $object_name) = $this-&gt;parse_path($this-&gt;path);\n\n    $file = null;\n    $mode = 0;\n\n    if ($object_name)\n    {\n        $response = $this-&gt;client($protocol)-&gt;list_objects($bucket, array(\n            'prefix' =&gt; $object_name\n        ));\n\n        if (!$response-&gt;isOK())\n        {\n            return $out;\n        }\n\n        \/\/ Ummm... yeah...\n        if (is_object($response-&gt;body))\n        {\n            $file = $response-&gt;body-&gt;Contents[0];\n        }\n        else\n        {\n            $body = simplexml_load_string($response-&gt;body);\n            $file = $body-&gt;Contents[0];\n        }\n    }\n    else\n    {\n        $response = $this-&gt;client($protocol)-&gt;list_objects($bucket);\n\n        if (!$response-&gt;isOK())\n        {\n            return $out;\n        }\n    }\n\n    \/*\n    Type &amp; Permission bitwise values (only those that pertain to S3).\n    Simulate the concept of a &quot;directory&quot;. Nothing has an executable bit because there's no executing on S3.\n    Reference: http:\/\/docstore.mik.ua\/orelly\/webprog\/pcook\/ch19_13.htm\n\n    0100000 =&gt; type:   regular file\n    0040000 =&gt; type:   directory\n    0000400 =&gt; owner:  read permission\n    0000200 =&gt; owner:  write permission\n    0000040 =&gt; group:  read permission\n    0000020 =&gt; group:  write permission\n    0000004 =&gt; others: read permission\n    0000002 =&gt; others: write permission\n    *\/\n\n    \/\/ File or directory?\n    \/\/ @todo: Add more detailed support for permissions. Currently only takes OWNER into account.\n    if (!$object_name) \/\/ Root of the bucket\n    {\n        $mode = octdec('0040777');\n    }\n    elseif ($file)\n    {\n        $mode = (str_replace('\/\/', '\/', $object_name . '\/') === (string) $file-&gt;Key) ? octdec('0040777') : octdec('0100777'); \/\/ Directory, Owner R\/W : Regular File, Owner R\/W\n    }\n    else\n    {\n        $mode = octdec('0100777');\n    }\n\n    \/\/ Update stat output\n    $out[2] = $out['mode'] = $mode;\n    $out[4] = $out['uid'] = (isset($file) ? (string) $file-&gt;Owner-&gt;ID : 0);\n    $out[7] = $out['size'] = (isset($file) ? (string) $file-&gt;Size : 0);\n    $out[8] = $out['atime'] = (isset($file) ? date('U', strtotime((string) $file-&gt;LastModified)) : 0);\n    $out[9] = $out['mtime'] = (isset($file) ? date('U', strtotime((string) $file-&gt;LastModified)) : 0);\n    $out[10] = $out['ctime'] = (isset($file) ? date('U', strtotime((string) $file-&gt;LastModified)) : 0);\n\n    return $out;\n}"}]},"constants":{"@attributes":{"count":"0"}},"properties":{"@attributes":{"count":"9"},"property":[{"name":"_clients","modifier":"protected static","metadata":{"var":"array An array of AmazonS3 clients registered as stream wrappers."}},{"name":"buffer","modifier":"public"},{"name":"eof","modifier":"public"},{"name":"file_list","modifier":"public"},{"name":"object_size","modifier":"public"},{"name":"open_file","modifier":"public"},{"name":"path","modifier":"public"},{"name":"position","modifier":"public"},{"name":"seek_position","modifier":"public"}]}}}